BeginPackage["DocumentationBuild`Build`HTML`"]

(* Needed for now so that AntLog resolves correctly. *)
AppendTo[$ContextPath, "Global`"]

Begin["`Private`"]

Needs["DocumentationBuild`"]; (* For DocumentationBuildHTML *)
Needs["DocumentationBuild`Utils`"];
Needs["DocumentationBuild`Info`"];
Needs["DocumentationBuild`Export`"];
Needs["DocumentationBuild`Build`Utils`"];
Needs["DocumentationBuild`Common`"]; (* For DocsError *)
Needs["DocumentationBuild`Internal`"]; (* For CheckShouldUseInternalHTMLAssets *)
Needs["DocumentationBuild`Settings`"];

DocumentationBuildHTML::warning = "``";
DocumentationBuildHTML::error = "``";

Options[DocumentationBuildHTML] = {
	(* FIXME: Should be this a required parameter, not just an option? Maybe in the typical
	          case this is the paclet name? *)
	"LinkBase" -> Automatic,
	"Debug" -> False,
	(* Whether generated HTML files should be suffixed with a two letter (ISO 639-1-like)
	   language abbreviation code. *)
	"UseLanguageExtension" -> False,
	(* TODO:
		Determine exactly and then document what this option controls. I *think* it controls
		whether assets from DocumentationAssets are copied into the output directory.
	*)
	"CompleteHTMLQ" -> False,
	Language -> "English",
	(* Function used to provide citation content for a symbol reference page. *)
	(* Note:
		Defaulting this option to the current $CitationCallback value is important for
		preserving LoadInternalSettings[]'s assignment of $CitationCallback in internal
		builds. *)
	"CitationFunction" :> DocumentationBuild`Settings`$CitationCallback,
	(* TODO: Remove this option? "Full" is the only valid value. *)
	"BuildType" -> "Full",
	(* This option is only used if "Debug" is True. *)
	"IntermediateNBOutputDirectory" -> None,
	(* What percentage of notebooks need to build to HTML successfully in order for the
	   overall build to be considered "successful". During the development cycle, setting
	   this to a non-100% value is useful for lessening the fragility of the docs builds. *)
	"SuccessThreshold" -> 1.0,
	(* Debugging option. If this it set, it overrides the set of files which would
	   otherwise be built, and instead forces just the single specified file to be built. *)
	"SingleFile" -> None
};

DocumentationBuildHTML[
	inputDir0_?StringQ,
	outputDir0_?StringQ,
	(* This can be a predicate function or a String/StringExpression. This is applied to
	   the *absolute* file name (including the `inputDir` prefix). *)
	filter : Except[_ -> _] : (True &),
	OptionsPattern[]
] := Module[{
	(* Expand any '~' or relative file paths. *)
	inputDir = ExpandFileName[inputDir0],
	outputDir = ExpandFileName[outputDir0],
	requiredSuccessRate = OptionValue["SuccessThreshold"],
	useLanguageExtension = TrueQ @ OptionValue["UseLanguageExtension"],
	(* Note:
		Replace "" with Automatic to handle the empty AntProperty["app.name"] case that
		html.xml might pass. *)
	linkBase = Replace[OptionValue["LinkBase"], "" -> Automatic],
	wsmInProductQ,
	htmlOutputDir,
	lang = OptionValue[Language],
	guidesDir,
	feMemory,
	singleFile,
	files,
	result,
	successfulFilesCount = 0,
	skippedFilesCount = 0,
	failedFilesCount = 0
},
Block[{
	DocumentationBuild`Settings`$DocumentationBuildLanguage = lang,

	(* Clear core information cache for this build. This prevents accidentally using a
	   a cached value from a previous build. *)
	DocumentationBuild`Info`$CoreInformation = <||>,

	(* Validate the user-provided citation function value. *)
	DocumentationBuild`Settings`$CitationCallback = Replace[OptionValue["CitationFunction"], {
		value:(_Function | None ) :> value,
		value_ :> DocsError["Invalid value for \"CitationFunction\" option: ``. CitationFunction must be a Function or None.", value]
	}]
},
	Assert[NumberQ[requiredSuccessRate]];

	(* Note: I think increasing the recursion limit is necessary because XMLTransforms can
	         recurse quite a bit. I'm unsure though. *)
	$RecursionLimit = 1000;

	(*--------------------------------------------*)
	(* Determine output directory for built html. *)
	(*--------------------------------------------*)

	wsmInProductQ = And[
		StringContainsQ[inputDir, "MathModelica", IgnoreCase -> True],
		TrueQ @ OptionValue["CompleteHTMLQ"]
	];

	(*
		There are three different supported configurations for the layout of the code
		built into the output diretory. The third configuration is used only for the
		Wolfram internal Wolfram System Modeler In-Product documentation builds.

		# [Option 1]: CompleteHTMLQ === False:

			Just built HTML, without the standard bundle of CSS/JS

			OutDir/
				ref/*.html
				guide/*.html
				tutorial/*.html
				...

		# [Option 2]: CompleteHTMLQ === True:

			Built HTML placed in an html/ directory, with standard bundle copied
			alongside. This is intended for HTML documentation which is intended to be
			deployed stand-alone, typically for self-hosted documentation.

			OutDir/
				html/
					ref/*.html
					guide/*.html
					tutorial/*.html
					...
				css/
				js/
				fonts/
				...

		# [Option 3]: System Modeler's In-Product build

			OutDir/
				ref/*.html
				guide/*.html
				tutorial/*.html
				(webm|lwg|wsm)-assets/
	*)

	(* User variables *)
	(* if creating local html, add 'html' dir to output path *)
	htmlOutputDir = Which[
		(* If this is an internal build with custom assets, don't place built HTML into
		   an html/ subdirectory. *)
		TrueQ @ OptionValue["CompleteHTMLQ"] && wsmInProductQ,
			outputDir,
		TrueQ @ OptionValue["CompleteHTMLQ"],
			AppendHTMLDirectory[outputDir],
		True,
			outputDir
	];

	AntLog["DocumentationBuildHTML: inputDir  = " <> ToString[inputDir, InputForm]];
	AntLog["DocumentationBuildHTML: htmlOutputDir = " <> ToString[htmlOutputDir, InputForm]];

	AntLog["DocumentationBuildHTML: useLanguageExtension = " <> ToString[useLanguageExtension, InputForm]];
	AntLog["DocumentationBuildHTML: CompleteHTMLQ = " <> ToString[OptionValue["CompleteHTMLQ"], InputForm]];


	guidesDir = DocumentationBuild`Utils`GetGuidesDirectory[inputDir, lang];

	DocumentationBuild`Transform`debug = False;
	DocumentationBuild`Export`$DocumentationExportDebug = TrueQ @ OptionValue["Debug"];

	If[TrueQ @ OptionValue["Debug"],
		feMemory = InputForm @ UsingFrontEnd[MemoryInUse[$FrontEnd]];
		AntLog[StringJoin["BuildHTML: Current FE memory: ", ToString @ feMemory, "\n"]];
		AntLog[StringJoin["BuildHTML: DocumentationBuild Directory: ", DocumentationBuild`Common`$DocumentationBuildDirectory]];
		AntLog["BuildHTML: Guides directory: " <> guidesDir];
	];

	Replace[OptionValue["BuildType"], {
		valid:"Full" :> valid,
		unknown_ :> Throw[StringForm["Unknown HTML build type: ``", unknown] ]
	}];

	singleFile = OptionValue["SingleFile"];
	AntLog["BuildHTML: singleFile = "<>ToString[singleFile, InputForm]];

	If[singleFile =!= None && FileExistsQ[singleFile],
		files = {singleFile}
	,
		files = GetDocumentationNotebooks[inputDir, lang, filter]
	];

	AntLog["BuildHTML: First[files] = "<>ToString[First[files], InputForm]];
	AntLog["BuildHTML: Last[files] = "<>ToString[Last[files], InputForm]];

	SetCombinedSourceInformation[linkBase, inputDir, lang];

	SetSystemEnglishSourceInformationIfNeeded[linkBase, lang];

	DocumentationBuild`Navigation`$Pages = DocumentationBuild`Navigation`ComputeLinkTrails[guidesDir];
	If[TrueQ @ OptionValue["Debug"],
		AntLog["Number of pages: "<>ToString@Length@DocumentationBuild`Navigation`$Pages]
	];

	Assert[MatchQ[files, {___?StringQ}]];

	result = BuildWithProgress[
		Function[{filename},
			TimeConstrained[
				buildFileToHTML[
					inputDir,
					htmlOutputDir,
					filename,
					"Debug" -> OptionValue["Debug"],
					"CompleteHTMLQ" -> OptionValue["CompleteHTMLQ"],
					Language -> OptionValue[Language],
					"IntermediateNBOutputDirectory" -> OptionValue["IntermediateNBOutputDirectory"],
					"UseLanguageExtension" -> useLanguageExtension
				]
				,
				(* Limit the build time of any individual file to 1 hour. *)
				3600,
				Failure["Timeout", <||>]
			]
		],
		files,
		requiredSuccessRate,
		"Building HTML"
	];

	Assert[FailureQ[result] || MatchQ[result, _Success]];

	If[TrueQ[OptionValue["CompleteHTMLQ"]],
		TimeConstrained[
			copyCompleteWebAssets[
				outputDir,
				CheckShouldUseInternalHTMLAssets[inputDir]
			],
			2400
		]
	];

	Developer`UninstallFrontEnd[];

	(* Don't leave the stylesheet cache notebooks visible in Notebooks[]. *)
	DocumentationBuild`Common`EmptyStylesheetCache[];

	result
]] ~Catch~ $DocsErrorTag

(**************************************)
(* buildFileToHTML[]                  *)
(**************************************)

Options[buildFileToHTML] = {
	"Debug" -> False,
	"CompleteHTMLQ" -> False,
	Language -> "English",
	"IntermediateNBOutputDirectory" -> None,
	(* Whether the extension on the HTML file should be `.html` or `.html.<ext>` where `ext`
	   is one of en/zh/es/ja/ko/ru/pt-br. *)
	"UseLanguageExtension" -> True
}

buildFileToHTML[
	inputDir_?StringQ,
	htmlOutputDir_?StringQ,
	inFile_?StringQ,
	OptionsPattern[]
] := Module[{
	language = OptionValue[Language],
	completeHTMLQ = TrueQ @ OptionValue["CompleteHTMLQ"],
	info,
	flag,
	langext,
	uri,
	pacletName,
	webmQ,
	lwgManagerQ,
	wsmlinkQ, wsmlibQ,
	wsmQ,
	webAssetsPath,
	urlBaseRemote,
	htmlTemplateIncludes,
	time,
	res,
	outFile
},
	DocumentationBuild`Transform`debug =
		DocumentationBuild`Export`$DocumentationExportDebug =
			TrueQ @ OptionValue["Debug"];

	(*-------------------------------------------------------------------------*)
	(* Get notebook information, check whether this notebook should be skipped *)
	(*-------------------------------------------------------------------------*)

	info = GetNotebookInformation[Get @ inFile];

	If[FailureQ[info],
		(* TODO: Make this an error, not a skip? This might happen when non-authoring type
		         notebooks are in the Documentation directory (which might be common, e.g.
				 example data?). *)
		(* Notebook to convert doesn't contain proper EntityType *)
		AntLog["            File's Entity Type not found -- file not included in build."];
		Return[None];
	];

	If[MissingQ @ Lookup[info, "EntityType"],
		(* If GetNotebookInformation returns a non-failure, then it must have an
		   "EntityType" value. *)
		DocsError["Notebook information is missing EntityType"];
	];

	Which[
		(* ignore files and CopyFile to output dir without converting *)
		!DocumentationBuild`Utils`ConvertFileQ[ inFile],
			AntLog["            File marked to ignore -- file not included in build."];
			Return[None];
	,
		(* pass over flagged nbs which shouldn't be included within on-web distribution *)
		MatchQ[ flag = DocumentFlag[ inFile] , "ExcisedFlag" | "FutureFlag" | "InternalFlag" | "TemporaryFlag"],
			AntLog["            " <> flag <> " -- file not included in build."];
			Return[None]
	,
		(* Convert everything else *)
		True, Null
	];

	(*-----------------------------------------------------------*)
	(* Set conditionals for special handling of Wolfram products *)
	(*-----------------------------------------------------------*)

	uri = Lookup[info, "URI", DocsError["Documentation file is missing URI"]];
	pacletName = Lookup[info, "PacletName"];

	webmQ = pacletName === "webMathematica";
	lwgManagerQ = pacletName === "LightweightGridManager";
	(* WSM *)
	wsmlinkQ = Or[
		StringMatchQ[uri, "*WSMLink/*"],
		(* Formatting note: e956a0f0:
			Round two of adapting to the fact that SystemModeler is now
			System Modeler with space.
		*)
		StringMatchQ[pacletName, "*SystemModeler*"|"*System Modeler*"]
	];
	wsmlibQ = StringMatchQ[pacletName, "Wolfram SystemModeler Library"];
	wsmQ = Or[wsmlinkQ, wsmlibQ];

	(*-------------------------------------*)
	(* Determine the output HTML file name *)
	(*-------------------------------------*)

	(* TODO Fix hack for Preview pages *)
	(* The following shoudl work but doesn't for Preview pages *)

	outFile = StringReplace[inFile, {".nb" -> "", ToFileName[{inputDir, language}] -> $PathnameSeparator}];
	outFile = StringReplace[outFile, {".nb" -> "", ToFileName[{inputDir}] -> $PathnameSeparator}];

	(* .en, .zh, .es, .ja, etc. *)
	langext = "." <> LanguageAbbreviation[language];

	outFile = DocumentationBuild`Common`FilePathToDocURI[
		StringJoin[
			htmlOutputDir,
			StringReplace[outFile, {".nb" -> ""}],
			If[TrueQ @ OptionValue["UseLanguageExtension"],
				".html" <> langext,
				".html"
			]
		],
		$PathnameSeparator
	];

	(* TODO clean up this path name section *)
	outFile = FixedPoint[
		StringReplace[#, {
			"System\\" -> "",
			"Packages\\" -> "",
			"Documentation\\" <> language <> "\\" -> "",
			language <> "\\" -> "",
			"\\\\" -> "\\",
			"//" -> "/"
		}] &,
		outFile
	];

	(* Formatting note: 91dc5304d5f99475cd9e4e59bca3bc47384f3159:
		Updates for LightweightGridManager docs

		https://jira.wolfram.com/jira/browse/WLDOCS-3229
		https://bugs.wolfram.com/show?number=395458

		Like webMathematica, LightweightGridManager's documentation is intended
		for viewing in-product in a tomcat installation rather than on the web
		at ref.w.c.

		Standalone docs like webM and lwgM are usually presented only in
		English, which makes the filename language extension ".en" superfluous.
		Also, it seems tomcat doesn't handle language negotiation based on
		filename extension.

		- Rethinking filename language extensions.  They've been determined in
		html2.m (DocumentationBuildHTML) based solely on the presence of a string in
		the input file's pathname, but I'd like to have better evidence of a file's
		type before assigning a filename.  The definitive way to do that is in
		Export.m when we check 'pacletName'.  So I've moved the language extension
		code to the earliest part of ExportWebPage for use throughout.
			connorg update:
				I've lifted this back into DocumentationBuildHTML, with the ultimate
				aim of moving this logic out of the core DocumentationBuild logic
				entirely. Ideally the LightweightGrid(Manager) docs build
				configuration would just set "UseLanguageExtension" -> False, rather
				than relying on DocumentationBuild to handle this case specially.

		- Then if we're building lightweightgrid docs, remove any language
		extension.  If this works well, do the same for webM docs and remove
		extension code from html2.m.
	*)
	If[lwgManagerQ,
		outFile = StringReplace[outFile, langext ~~ EndOfString -> ""]
	];

	(*-------------------------------------------------------*)
	(* Determine the relative path to the static web assets. *)
	(*-------------------------------------------------------*)

	(* Determine the location of CSS/JS assets relative to the html page. *)
	(* This path is embedded in the built HTML page. *)
	webAssetsPath = Which[
		TrueQ[webmQ],
			"/webMathematica/Resources/webm-assets/"
		,
		TrueQ[lwgManagerQ],
			"/WolframLightweightGrid/Documentation/lwg-assets/"
		,
		completeHTMLQ && TrueQ[wsmQ],
			"../Resources/wsm-assets/"
		,
		TrueQ[completeHTMLQ],
			(* this changes depending on page (e.g ref/Sin v. ref/messages/func/tag) *)
			(* TODO:
				This logic assums that `uri` is of the form "PacletName/ref/SomeSymbol".
				This will be true for the vast majority of paclets, but not System
				documentation. Generate a warning or an error if `uri` is not a valid
				paclet documentation URI; add an undocumented option to disable that
				behavior if the documentation being built is the system documentation.

				Note:
					The above needs to take into account internal paclets which provide
					System documentation, but aren't part of the main Notebooks builds.
			*)
			StringRepeat["../", StringCount[uri, "/"]]
		,
		True,
			StringRepeat["../", StringCount[uri, "/"]-1]
	];

	BuildLog["webAssetsPath: " <> webAssetsPath];

	(*------------------------------------------------------------------*)
	(* Determine URL base for documentation outside the current paclet. *)
	(*------------------------------------------------------------------*)

	(* Formatting note: a410487c:
		Updates for in-product WSM documentation

		- new variable wsmQ: True when we're processing any WSM build - web or in-product, to more easily
		make the distinction betweewn WSM for web (completeHTMLQ is False) and WSM for in-product docs
		(completeHTMLQ is True) since the two use different assets.
		- new layout file WSMTutorialsStandalone.html for in-product wsm docs.  "Standalone" and "complete
		html" have long been synonyms that refer to the in-product docs that use their own set of
		includes without outside resources.
	*)
	(* Formatting note: 8d264e32:
		https://jira.wolfram.com/jira/browse/SYSTEMMODELER-1236

		simplify the handling of WSM web paclet:// links:
		- do nothing to in-product links
		- redirect the links to /system-modeler/ without munging the value of
		  urlBaseRemote, which could have consequences elsewhere
	*)
	(* Formatting note: 3fac0ee9:
		https://jira.wolfram.com/jira/browse/SYSTEMMODELER-2176

		Adjust the value of urlBaseRemote: if we're processing a SystemModeler notebook,
		set the base url to /system--modeler/ for use in all link transforms.  Otherwise,
		accept whatever has been provided and default to /language/.

		Formatting note: 91dc5304:
			- For standalone docs, urlBaseRemote should be a string that produces
			working links in-product.  Locations and methods differ per product.
	*)
	(* Formatting note: 90634f12:
		Handle webMathematica links and filenames
		ExportWebPage: test whether we're building a webm file and if so:

		1. use a special url base that provides correct addresses within
			webm's installation in tomcat/apache.

		2. trust that ExportWebPage has been sent the correct filename ending
			in .html and don't append .en to it.  webm's tomcat/apache doesn't
			handle language negotiation out of the box and cannot find
			.html.en files unless the entire filename is given explicitly.
	*)
	(* This path is used as the base for ButtonBox/TemplateBox hyperlinks to
	   documentation items which are considered 'outside' the documentation package
	   currently being built. *)
	urlBaseRemote = Which[
		TrueQ[wsmQ],
			If[completeHTMLQ,
				If[TrueQ[wsmlibQ],
					"wsm:///sma/",
					"wsm:///ref/"
				],
				(* WSM web docs, not in-product docs *)
				"/system-modeler/"
			],
		TrueQ[webmQ],
			"/webMathematica/Resources/Documentation/English/",
		TrueQ[lwgManagerQ],
			"/WolframLightweightGrid/Documentation/",
		TrueQ[completeHTMLQ],
			(* TODO: Use https:// here. Update StringMatchQ checks for "http://". Grep:
				`String.*Q.*"http` *)
			"http://reference.wolfram.com/language/",
		True,
			"/language/"
	];

	(*-------------------------------------------------------------------------*)
	(* Choose the template values to substitute to link to the bundled assets. *)
	(*-------------------------------------------------------------------------*)

	htmlTemplateIncludes = Which[
		(TrueQ[wsmQ] && TrueQ[completeHTMLQ])
		|| TrueQ[webmQ]
		|| TrueQ[lwgManagerQ],
			$InternalBundleHTMLTemplateIncludes
		,
		completeHTMLQ,
			$StandardBundleHTMLTemplateIncludes
		,
		True,
			(* Leave the default `<!--#include virtual=".."-->` directives untouched. *)
			<||>
	];

	DocsAssert[MatchQ[htmlTemplateIncludes, Association[(_?StringQ -> (_?StringQ | _Function)) ...]]];

	htmlTemplateIncludes = ResolveHTMLTemplateIncludes[
		htmlTemplateIncludes,
		<|
			(* la will always be a language's 2-letter ISO 639-1 code as used in our documentation *)
			"la" -> LanguageAbbreviation[language],
			"webAssetsPath" -> webAssetsPath,
			"completeHTMLQ" -> completeHTMLQ,
			"wsmQ" -> wsmQ
		|>
	];

	DocsAssert[MatchQ[htmlTemplateIncludes, Association[(_?StringQ -> _?StringQ) ...]]];

	(*----------------------------------*)
	(* Perform the conversion into HTML *)
	(*----------------------------------*)

	{time, res} = AbsoluteTiming @ (
		(* Convert notebook to html *)
		DocumentationBuild`Common`PubsEvaluateWithFE @ ExportDocumentation[
			outFile,
			inFile,
			"Debug" -> TrueQ @ OptionValue["Debug"],
			"ExportFormats" -> If[TrueQ @ OptionValue["Debug"] && OptionValue["IntermediateNBOutputDirectory"] =!= None,
				{
					"Notebook" -> {"OutputDirectory" -> OptionValue["IntermediateNBOutputDirectory"]},
					"HTML"
				},
				"HTML"
			],
			"Language"->language,
			"WebAssetsPath" -> webAssetsPath,
			"URLBaseRemote" -> urlBaseRemote,
			"HTMLTemplateIncludes" -> htmlTemplateIncludes,
			"CompleteHTMLQ" -> TrueQ @ OptionValue["CompleteHTMLQ"]
		]
	);

	(* TODO:
		Assert that this new `outFile` is the same as the `outFile` passed to ExportDocumentation?
	*)
	outFile = Replace[res, {
		_?FailureQ :> Return[res],
		{"HTML" -> out_?FailureQ} :> Return[out],
		{"HTML" -> out_?StringQ} :> out,
		_ :> DocsError[
			"Unrecognized return value while processing ``: ``",
			InputForm[inFile],
			InputForm[res]
		]
	}];

	(* Output file test *)
	If[FileType[outFile] === None || FileByteCount[outFile] < 1,
		DocsError["Output file does not exist after completion of ExportDocumentation: ``", outFile];
	];

	AntLog["outFile = "<>outFile];

	Success["buildFileToHTML", <|
		"HTML" -> outFile,
		"Time" -> Quantity[time, "Seconds"]
	|>]
]

(**************************************)
(* copyCompleteWebAssets[]            *)
(**************************************)

(* Copy the CSS/JS/image assets necessary to view the built HTML in a standalone
   environment. *)
copyCompleteWebAssets[
	outputDir_?StringQ,
	assets : (Automatic | {_?StringQ, _?StringQ}) : Automatic
] := Module[{
	assetSourceDir,
	assetOutputDir,
	assetOutputPrefix,
	htmlTemplateFiles,
	parent,
	outFile
},
	(* TODO:
		Re-scrape the standard reference.wolfram.com assets and include
		them in DocumentationBuild/Assets/, so that they can be bundled with
		built HTML in this case. The "html/2014/standard" directory contains
		JS/CSS which are no longer compatible with the HTML produced by
		ExportWebPage (presumably since the large redesign in 2017 which is
		mentioned in the git history).

		At the moment "StandardBundle" consists of the "wsm-assets", copied from
		DocumentationAssets back into DocumentationBuild. At the moment, the
		"wsm-assets" seem to be the most recently updated / complete which are easily
		available.
	*)

	{assetSourceDir, assetOutputPrefix} = Replace[assets, {
		custom:{_?StringQ, _?StringQ} :> custom,
		Automatic :> {DocsAsset["WebData", "StandardBundle"], ""},
		_ :> DocsAssert[False, "Unreachable: unexpected 'assets' parameter specification: ``", assets]
	}];

	DocsAssert[DirectoryQ[assetSourceDir], "bad asset source directory: ``", assetSourceDir];

	assetOutputDir = FileNameJoin[{outputDir, assetOutputPrefix}];

	(* Copy ExampleData files/dir ignoring 'bad' files *)
	htmlTemplateFiles = FileNames["*", assetSourceDir, Infinity];
	htmlTemplateFiles = Select[
		htmlTemplateFiles,
		(
			!MemberQ[FileNameSplit[#], "CVS"]
			&& FileType[#] === File
		)&
	];
	htmlTemplateFiles = DeleteCases[htmlTemplateFiles, s_ /; StringContainsQ[s, "mathematica.htaccess"]];

	Map[
		Function[{file},
			outFile = FileNameJoin[{
				assetOutputDir,
				(* Get the relative path for the asset *)
				FileNameDrop[file, Length @ FileNameSplit @ assetSourceDir]
			}];

			(* Create the directory for `outFile` and copy it over. *)
			parent = DirectoryName[outFile];
			If[!DirectoryQ[parent],
				result = CreateDirectory[parent];
				If[FailureQ[result],
					DocsError["Error creating HTML asset directory: ``", result];
				];
			];

			CopyFile[file, outFile, OverwriteTarget -> True];
		],
		htmlTemplateFiles
	];
]


End[]

EndPackage[]