BeginPackage["DocumentationBuild`Build`Notebooks`"]

Needs["DocumentationBuild`"]; (* For DocumentationBuildNotebooks *)
Needs["DocumentationBuild`Common`"]; (* For DocsError[] *)
Needs["DocumentationBuild`Utils`"];
Needs["DocumentationBuild`Build`Utils`"];

Begin["`Private`"]

DocumentationBuildNotebooks::error = "``";
DocumentationBuildNotebooks::warning = "``";

(*--------------------------------------------------------*)

Options[DocumentationBuildNotebooks] = {
	(* TODO: Use `:> $Language` here. Check the other functions with a Language parameter. *)
	Language -> "English",
	"Debug" -> False,
	(* FIXME: Should be this a required parameter, not just an option? Maybe in the typical
	          case this is the paclet name? *)
	"LinkBase" -> Automatic,
	(* TODO: Determine and document when and why a user might want to change this option.
	         Maybe it shouldn't even be an option, and a search index should just always
			 be built? *)
	(* NOTE: This used to be two options, because there are actually two different indexes
	         associated with a paclet: The SpellIndex and the SearchIndex. These were two
			 separate options in the old notebook.xml Ant script, but I combined them into
			 one for this WL interface.
			 Update: Actually there are three indexes: Index, SpellIndex, and SearchIndex.
	*)
	"BuildSearchIndex" -> True,
	"ProgressDescription" -> Automatic,
	(* Function used to provide citation content for a symbol reference page. *)
	(* Note:
		Defaulting this option to the current $CitationCallback value is important for
		preserving LoadInternalSettings[]'s assignment of $CitationCallback in internal
		builds. *)
	"CitationFunction" :> DocumentationBuild`Settings`$CitationCallback,
	(* What percentage of notebooks need to build successfully in order for the overall
	   build to be considered "successful". During the development cycle, setting this to
	   a non-100% value is useful for lessening the fragility of the docs builds. *)
	"SuccessThreshold" -> 1.0
};

DocumentationBuildNotebooks[
	inputDir0_?StringQ,
	outputDir0_?StringQ,
	(* This can be a predicate function or a String/StringExpression. This is applied to
	   the *relative* file name (without the `inputDir` prefix). *)
	filter : Except[_ -> _] : (True &),
	OptionsPattern[]
] := Module[{
    (* Expand '~` in the path, which are not handled correctly by DocumentationSearch`. *)
	inputDir = ExpandFileName[inputDir0],
    outputDir = ExpandFileName[outputDir0],
	successRate = OptionValue["SuccessThreshold"],
	language = OptionValue[Language],
	debug = TrueQ[OptionValue["Debug"]],
	buildIndex = TrueQ[OptionValue["BuildSearchIndex"]],
	(* TODO: Is "" a valid link base? Maybe this should be a required parameter? *)
	linkBase = Replace[OptionValue["LinkBase"], "" -> Automatic],
	sourceInfo,
	files,
	searchMetaData = <||>,
	res,
	searchIndexResult,
	result
},
Block[{
	DocumentationBuild`Settings`$DocumentationBuildLanguage = language,

	(* Clear core information cache for this build. This prevents accidentally using a
	   a cached value from a previous build. *)
	DocumentationBuild`Info`$CoreInformation = <||>,

	(* Validate the user-provided citation function value. *)
	DocumentationBuild`Settings`$CitationCallback = Replace[OptionValue["CitationFunction"], {
		value:(_Function | None ) :> value,
		value_ :> DocsError["Invalid value for \"CitationFunction\" option: ``. CitationFunction must be a Function or None.", value]
	}]
},
	Assert[NumberQ[successRate]];

	(* kill if reference.nb is 0k or missing *)
	If[
		!FileExistsQ[FileNameJoin[{$InstallationDirectory, "SystemFiles", "FrontEnd", "StyleSheets", "Wolfram","Reference.nb"}] ] ||
		!FileExistsQ[FileNameJoin[{$InstallationDirectory, "SystemFiles", "FrontEnd", "StyleSheets", "Wolfram","Reference-ja.nb"}] ]
	,
		Message[DocumentationBuildNotebooks::error, "Reference.nb or Reference-ja.nb is corrupted or missing."];
		Return[$Failed];
	];

	guidesDir = DocumentationBuild`Utils`GetGuidesDirectory[inputDir, language];

	DocumentationBuild`Transform`debug = debug;
	DocumentationBuild`Export`$DocumentationExportDebug = debug;

	files = GetDocumentationNotebooks[inputDir, language, filter];

	(* Create Index directory and NotebookIndexer Java Object *)
	If[ buildIndex, (
		(* Create Index directory *)
		indexDir = If[Position[FileNameSplit[outputDir], language] === {},
			ToFileName[{outputDir, language, "Index"}]
			, (* else *)
			ToFileName[{
				FileNameTake[
					outputDir,
					{1, Position[FileNameSplit[outputDir], language][[1, 1]]}
				],
				"Index"
			}]
		];
		If[FileType[indexDir] === Directory,
			DeleteDirectory[indexDir, DeleteContents -> True]
		];
		CreateDirectory[indexDir];
		If[FileType[indexDir] =!= Directory,
			Message[DocumentationBuildNotebooks::error, "Creation of Index directory failed."];
			Return[$Failed];
		];

		(* Create Speeling Index directory *)
		indexSpellDir = ToFileName[{DirectoryName[indexDir]}, "SpellIndex"];
		If[FileType[indexSpellDir] === Directory,
			DeleteDirectory[indexSpellDir, DeleteContents -> True]
		];
		CreateDirectory[indexSpellDir];
	)];

	SetCombinedSourceInformation[linkBase, inputDir, language];

	DocumentationBuild`Info`SetSystemEnglishSourceInformationIfNeeded[linkBase, language];

	(* Computing guide link trails *)
	DocumentationBuild`Navigation`$Pages = DocumentationBuild`Navigation`ComputeLinkTrails[guidesDir];
	If[debug,
		Print["Number of pages: " <> ToString@Length@DocumentationBuild`Navigation`$Pages]
	];

	(*-------------------------------------------*)
	(* Loop over every input file to build them. *)
	(*-------------------------------------------*)

	If[!MatchQ[files, {___?StringQ}],
		Throw["`files` is not a list of strings: " <> ToString[InputForm[files]]];
	];

	result = BuildWithProgress[
		Function[file,
			Module[{outFile, nbResult},
				outFile = FileNameJoin[{outputDir, StringReplace[file, inputDir -> ""]}];

				nbResult = buildDocumentationNotebook[
					file,
					outFile,
					(*
						Options
					*)
					Language -> language
				];

				nbResult = Replace[nbResult, {success:Success[___], metaData_} :> (
					If[buildIndex,
						AssociateTo[searchMetaData, outFile -> metaData];
					];
					success
				)];

				nbResult
			]
		],
		files,
		successRate,
		Replace[OptionValue["ProgressDescription"], {
			desc_?StringQ :> desc,
			Automatic -> "Building documentation",
			bad_ :> (
				Message[DocumentationBuild::warning, StringForm[
					"Ignoring non-String \"ProgressDescription\" value: ``",
					InputForm[bad]
				]];
				"Building documentation"
			)
		}]
	];

	(* Don't leave the stylesheet cache notebooks visible in Notebooks[]. *)
	DocumentationBuild`Common`EmptyStylesheetCache[];

	If[FailureQ[result],
		Return[result];
	];

	(*----------------------------------------------------------*)
	(* Create Index, SearchIndex, and SpellIndex search indices *)
	(*----------------------------------------------------------*)

	Assert[MatchQ[searchMetaData, <| RepeatedNull[_?FileExistsQ -> _?ListQ] |>]];

	If[buildIndex,
		Needs["DocumentationSearch`"];

		(* NOTE:
			We need to generate the three different search indexes, using a fresh JVM for
			each one to accommodate these constraints:

				* DocumentationSearch` and TextSearch` indexing/searching is done using
				  Lucene, a Java library.
				* Java code is not permitted to run on iOS, so instead a library called
				  CLucene is used; this is how documentation searching is done on iOS.
				* CLucene is essentially abandonware, and does not support reading with
				  index files created by Lucene versions newer than ~2.1.
				* Modern versions of Lucene cannot generate index files in a format
				  compatible with CLucene.

			Ergo:

				* In order to support iOS, we need to generate an index file which can be
				  read by CLucene.
				* In order to do _that_, we need to keep around an old (2.1) version of
				  Lucene.

			However:

				* Modern functions like TextSearch use a much newer version of Lucene (8.3
				  in WL 12.2 at time of writing).
				* Multiple copies of Lucene cannot be on the classpath for a single JVM;
				  they conflict and result in Java exceptions being thrown.

			> To verify the above, you can run `$ find . | grep lucene-core` from the
			> Mathematica.app directory, and you'll see that DocumentationSearch` and
			> TextSearch` appear to use two different versions of Lucene.

			So therefore we need to isolate calls to different generations of the
			DocumentationSearch` API, accomplished using the `withFreshJVM[]` function
			provided by Todd Gayley.

			Separating these tasks was originally accomplished in the notebook.xml Ant
			script of yore by having a separate `<mathematica fresh="true">..</mathematica>`
			target just for performing this indexing task. `withFreshJVM[]` performs that
			function now.

			See original commit for context: 903ba8e
		*)

		(*---------------------*)
		(* Create legacy index *)
		(*---------------------*)

		res = withFreshJVM[
			createDocumentationSearchLegacyIndex[searchMetaData]
		];

		If[FailureQ[res] || !MatchQ[res, Success[___]],
			(* Assume a message was already issued. *)
			Return[$Failed];
		];

		(*-----------------------*)
		(* Create spelling index *)
		(*-----------------------*)

		res = withFreshJVM[
			DocumentationSearch`CreateSpellIndex[indexDir, indexSpellDir]
		];

		If[res =!= Null,
			Message[
				DocumentationBuildNotebooks::error,
				StringForm["error creating documentation spelling index: ``", res]
			];
			Return[$Failed];
		];

		(*---------------------*)
		(* Create modern index *)
		(*---------------------*)

		searchIndexResult = withFreshJVM[
			createDocumentationSearchIndex[
				inputDir,
				outputDir,
				files,
				Language -> language
			]
		];

		If[FailureQ[searchIndexResult],
			Return[searchIndexResult];
		];
	];

	result
]] ~Catch~ $DocsErrorTag

(*--------------------------------------------------------*)

DocumentationBuildNotebooksIncremental[
	inputDir0_?StringQ,
	outputDir0_?StringQ,
	resultsLocation_?StringQ,
	opts:OptionsPattern[DocumentationBuildNotebooks]
] := Module[{
    (* Expand '~` in the path, which are not handled correctly by DocumentationSearch`. *)
	inputDir = ExpandFileName[inputDir0],
	previousProcessedFiles,
	processedFiles,
	result
},
	previousProcessedFiles = If[FileExistsQ[resultsLocation],
		Get[resultsLocation]
		,
		(* This must be the first build. Default to empty metadata. *)
		<||>
	];

	DocsAssert[
		AssociationQ[previousProcessedFiles],
		"Incremental results data does not have the expected form. It may have been
		 corrupted. Suggestion: delete the incremental results metadata file, and rebuild."
	];

	result = DocumentationBuildNotebooks[
		inputDir,
		outputDir0,
		(* Build only files from `inputDir` that have changed or been added since the
		   last build. *)
		ChangedFilesFilter[previousProcessedFiles, inputDir, outputDir0],
		opts
	];

	processedFiles = result["ProcessedFiles"];

	DocsAssert[
		AssociationQ[processedFiles],
		"ERROR: \"ProcessedFiles\" key was missing or had unexpected value. Full build result was: ``",
		InputForm[result]
	];

	processedFiles = Association[KeyValueMap[
		Function[{file, result},
			FileNameDrop[file, Length @ FileNameSplit[inputDir]]
				-> <| "Result" -> result, "FileHash" -> FileHash[file, $DocsFileHashType, $DocsFileHashEncoding] |>
		],
		processedFiles
	]];

	(* Join previously processed files and newly processed files. *)
	Put[
		Join[previousProcessedFiles, processedFiles],
		resultsLocation
	];

	result
] ~Catch~ $DocsErrorTag

(**************************************)
(* buildDocumentationNotebook[]       *)
(**************************************)

Options[buildDocumentationNotebook] = {
	(* TODO: Use `:> $Language` here. Check the other functions with a Language parameter. *)
	Language -> "English"
};

buildDocumentationNotebook[
	inFile_?StringQ,
	outFile_?StringQ,
	OptionsPattern[]
] := Module[{
	language = OptionValue[Language],
	inFileNBExpr,
	entityType,
	feprocessid,
	result,
	time,
	metaData
},
	(* billw 20161220: testing for full error messages in webm build *)
	Off[General::stop];

	(* User variables *)
	inFileNBExpr = DocumentationBuild`Utils`GetQuietly[inFile];

	(* Entity Type / Page Type *)
	entityType = DocumentationBuild`Info`GetNotebookCategorization[inFileNBExpr, "Entity Type"];

	Which[
		(* ignore files and CopyFile to output dir without converting *)
		!DocumentationBuild`Utils`ConvertFileQ[ inFileNBExpr], (
			(* TODO: What does this case represent? ConvertFileQ[] checks for the presense of
					the flag "UndeployInDocsBuild", but I can't find *any* instances of that
					string in stash.wolfram.com other than in the definition of ConvertFileQ[]
					itself. I think this may be an ancient legacy flag. *)
			(* File marked to ignore -- including in build by copying to output dir. *)
			CreateDirectory[ DirectoryName[outFile] ];
			CopyFile[ inFile, outFile];
			Return[None];
		),
		(* Notebook to convert doesn't contain proper EntityType/PageType *)
		entityType === $Failed, (
			(* File's Entity Type not found -- file not included in build. *)
			Return[None];
		),
		(* pass over flagged nbs which shouldn't be included within on-web distribution *)
		MatchQ[ DocumentationBuild`Utils`DocumentFlag[ inFileNBExpr] , "ExcisedFlag" | "FutureFlag" | "InternalFlag" | "TemporaryFlag"], (
			(* Flagged files are not included in build. *)
			Return[None];
		)
	];

	feprocessid = UsingFrontEnd[SystemInformation["FrontEnd", "ProcessID"]];
	UsingFrontEnd[SetOptions[$FrontEnd, RenderingOptions -> {"HardwareAntialiasingQuality" -> 1}]];

	(* Convert everything else *)
	time = AbsoluteTiming[
		{result, metaData} = DocumentationBuild`Make`MakeNotebook[
			inFile,
			"outfile" -> outFile,
			"Language" -> language
		];
	][[1]];

	(* Output file test *)
	If[FileType[outFile] =!= File || FileByteCount[outFile] < 1,

		(* If the file wasn't written because the FE crashed, issue a message. *)
		If[feprocessid =!= UsingFrontEnd[SystemInformation["FrontEnd", "ProcessID"]],
			Message[DocumentationBuildNotebooks::error, StringForm["FrontEnd crashed while processing '``'", inFile]]
			Return[
				Failure["DocumentationBuild", <|
					"MessageTemplate" -> "FrontEnd crashed while processing '``'",
					"MessageParameters" -> {InputForm[inFile]}
				|>],
				Module
			];
		];

		Return[$Failed, Module];
	];

	(* TODO: Remove this statement after it's been in the master builds for a few weeks
	         without being hit. This renaming logic I think it just an artifact of
			 copy-pasted code; there shouldn't be any .html files in a Notebooks builds. *)
	If[StringContainsQ[outFile, ".html"],
		DocsError["Unexpected '.html' in Notebooks build file name: ``", outFile];
	];
	(* outFile = StringReplace[outFile, ".html" -> ".html." <> LanguageAbbreviation[language]]; *)

	If[FailureQ[result],
		Return[result, Module];
	];

	Assert[MatchQ[time, _?NumberQ]];

	{
		Success["BuildDocumentationNotebook", <|
			"Time" -> Quantity[time, "Seconds"]
		|>],
		metaData
	}
]

(******************************************)
(* createDocumentationSearchLegacyIndex[] *)
(******************************************)

createDocumentationSearchLegacyIndex[searchMetaData_?AssociationQ] := Module[{
	indexer,
	res
},
	indexer = Quiet[
		DocumentationSearch`NewDocumentationNotebookIndexer[indexDir],
		{Java::excptn, LoadJavaClass::fail, KeepJavaObject::obj}
	];

	If[
		Or[
			indexer === DocumentationSearch`DocumentationNotebookIndexer[$Failed],
			Head[indexer] =!= DocumentationSearch`DocumentationNotebookIndexer,
			FailureQ[indexer]
		]
		,
		Message[DocumentationBuildNotebooks::error,
			"Setup of NotebookIndexer failed due to internal Java error."
		];
		Return[$Failed];
	];

	(*----------------------*)
	(* Perform the indexing *)
	(*----------------------*)

	Scan[
		Replace[(outFile_?FileExistsQ -> metaData_?ListQ) :> Module[{
			plainText
		},
			(* gather plain text for search engine *)
			plainText = Import[outFile, {"NB", "Plaintext"}];

			(* add text to index *)
			If[StringQ[plainText],
				DocumentationSearch`AddDocumentationNotebook[
					indexer,
					plainText,
					metaData
				];
			];
		]],
		Normal[searchMetaData]
	];

	(*-------------------*)
	(* Close the indexer *)
	(*-------------------*)

	res = DocumentationSearch`CloseDocumentationNotebookIndexer[indexer];
	If[res =!= Null,
		Message[
			DocumentationBuildNotebooks::error,
			StringForm["error closing documentation notebook indexer: ``", res]
		];
		Return[$Failed];
	];

	Success["LegacySearchIndex", <||>]
]

(**************************************)
(* createDocumentationSearchIndex[]   *)
(**************************************)

Options[createDocumentationSearchIndex] = {
	Language -> "English",
	(* TODO: Is there any reason for this to ever be something *other* than SearchIndex?
	         Maybe this shouldn't be an option? *)
	"SearchIndexName" -> "SearchIndex"
};

createDocumentationSearchIndex[
	inputDir_?StringQ,
	outputDir_?StringQ,
	files0:{___?StringQ},
	OptionsPattern[]
] := Module[{
	files = files0,
	language,
	textSearchIndexName,
	textSearchIndexVersion,
	result,
	index
},
	(* User variables *)
	language = OptionValue[Language];
	textSearchIndexName = OptionValue["SearchIndexName"];

	(* NEW: Create a TextSearchIndex - only works with M11.2+ *)
	If[$VersionNumber < 11.2,
		DocsError["Building documentation search index requires version 11.2 or newer."];
	];

	If[language =!= "English",
		(* Note: Non-English language documentation is still searchable via the older
		//       CLucene-based indexing. *)
		DocsWarning[StringJoin[
			"Building modern documentation search index currently only supports English. ",
			"Skipping generation of modern search index."
		]];
		Return[Null, Module];
	];

	(* This initializes TextSearch`PackageScope`$CurrentVersion. *)
	textSearchIndexVersion = StringTrim[ToString[TextSearch`PackageScope`$CurrentVersion], "."];

	textSearchIndexDir = If[
		Position[FileNameSplit[outputDir], language] === {}
		, (* then *)
		FileNameJoin[{outputDir, language, textSearchIndexName, textSearchIndexVersion}]
		, (* else *)
		FileNameJoin[{
			FileNameTake[outputDir, {1, Position[FileNameSplit[outputDir], language][[1, 1]]}],
			textSearchIndexName,
			textSearchIndexVersion
		}]
	];
	If[DirectoryQ[textSearchIndexDir],
		DeleteDirectory[textSearchIndexDir, DeleteContents -> True]
	];
	If[!DirectoryQ[FileNameDrop[textSearchIndexDir]],
		CreateDirectory[FileNameDrop[textSearchIndexDir]]
	];
	(* Create the index dir even if it will stay empty, per bug 339703. *)
	(* Only in M11.3+ because TextSearch`PackageScope`addSubIndex chokes on empty indices in earlier versions. *)
	If[$VersionNumber >= 11.3,
		CreateDirectory[textSearchIndexDir]
	];

	(* TODO: get rid of this hack in favor of a smoother solution. Maybe when the old index will go away. *)
	(* NOTE: This is a hacky way of avoiding indexing input files which weren't included
	         in the output (typically due the presence of the FutureFlag/ExcisedFlag. *)
	files = Select[
		FileNameJoin[{outputDir, StringReplace[#, inputDir -> ""]}]& /@ files,
		FileExistsQ
	];

	If[files === {},
		(* No files to index. Empty TextSearch index created. *)
		Return[];
	];

	result = Block[{
		TextSearch`PackageScope`$ShowJavaExceptions = True
	},
		(* FIXME: Determine the origin of this message and fix it in DocumentationSearch. *)
		Quiet[
			DocumentationSearch`CreateDocumentationIndex[
				files,
				FileNameDrop[textSearchIndexDir, -1],
				textSearchIndexVersion,
				OverwriteTarget -> True
			],
			{CreateSearchIndex::uf}
		]
	];

	index = MatchQ[result, _SearchIndexObject]
		&& MatchQ[TextSearch[result, "function"], _SearchResultObject];

	If[!TrueQ[index],
		Message[DocumentationBuildNotebooks::error, "Creation of TextSearch index failed"];
		Return[$Failed];
	];

]

(**************************************)
(* withFreshJVM[]                     *)
(**************************************)

Attributes[withFreshJVM] = {HoldFirst}

withFreshJVM[expr_] := Module[{
	newJVM
},
	(* This is just a wrapper that ensures a fresh JVM is used for the build. This is done
	   purely as a workaround for bug 407121. *)
	Internal`WithLocalSettings[
		newJVM = InstallJava[ForceLaunch -> True, Default -> False];
		(* So that when DocumentationSearch calls AddToClassPath for its Lucene libs, they
		   go at the front. *)
		SetOptions[AddToClassPath, Prepend -> True]
		,
		UseJVM[
			GetJVM[newJVM],
			expr
		]
		,
		(* Restore the option value *)
		SetOptions[AddToClassPath, Prepend -> False];
		UninstallJava[newJVM]
	]
]


End[]

EndPackage[]