BeginPackage["DocumentationBuild`Build`SystemNotebooks`"]

Begin["`Private`"]

Needs["XML`"];
Needs["Transmogrify`"];
Needs["JLink`"];

Needs["DocumentationBuild`"];
Needs["DocumentationBuild`Make`"];
Needs["DocumentationBuild`Common`"];       (* DocsAssert *)
Needs["DocumentationBuild`Build`Utils`"];

(* Ensure that the Ant* utilities are still visible. Overtime, refactoring should make
   this unnecessary. *)
PrependTo[$ContextPath, "Global`"];

Global`AntLog;

Assert[Context[AntLog] === "Global`"];

DocumentationBuildSystemNotebooks::error = "``";

Options[DocumentationBuildSystemNotebooks] = {
	Language -> "English",
	"Debug" -> False,
	(* Whether to generate a manifest.txt file in the output directory.
	   Defaults to False because this takes several seconds, which is a high overhead
	   when trying to build a small number of files quickly (e.g. for testing purposes). *)
	"GenerateManifest" -> False,
	(* Whether the output directory should contain built notebooks with both high
	   resolution and low resolution images. *)
	"HighAndLowResolution" -> False,
	(* The version which is being targeted by this build. See comment on
	   Options[MakeNotebook]. *)
	"ReleaseVersion" -> OptionValue[MakeNotebook, "ReleaseVersion"],
	(* TODO: Determine and document how the files in this directory are intended to be
	         built. *)
	"CoreInformationDirectory" -> Automatic,
	(* The percentage of files which must be successfully processed (or skipped) for the
	   build to count as a success. Release builds should have this set at 1.0, but
	   development builds will typically set it lower. *)
	"SuccessThreshold" -> 1.0,
	"SingleFile" -> None
};

DocumentationBuildSystemNotebooks[
	inputDir0_?StringQ,
	outputDir0_?StringQ,
	buildGroup : (_?IntegerQ | All),
	filters0 : Except[_?OptionQ] : (True &),
	OptionsPattern[]
] := Module[{
	(* Expand '~' components before passing these paths down. *)
	inputDir = ExpandFileName[inputDir0],
	outputDir = ExpandFileName[outputDir0],
	debug = TrueQ[OptionValue["Debug"]],
	language = OptionValue[Language],
	filters = Replace[filters0, filter:Except[_?ListQ] :> {filter}],
	highAndLowResolution = TrueQ[OptionValue["HighAndLowResolution"]],
	releaseVersion = OptionValue["ReleaseVersion"],
	successRate = OptionValue["SuccessThreshold"],
	singleFile = OptionValue["SingleFile"],
	generateManifest = TrueQ[OptionValue["GenerateManifest"]],
	coreInfoFile,
	notebooks,
	result
},
Block[{
	DocumentationBuild`Settings`$DocumentationBuildLanguage = language
},
	DocumentationBuild`Transform`debug = debug;
	DocumentationBuild`Export`$DocumentationExportDebug = debug;

	DocsAssert[ListQ[filters]];

	(* Initialize various $-variables in DocumentationBuild`Settings`. *)
	result = DocumentationBuild`Internal`LoadInternalSettings[];
	If[FailureQ[result],
		Return[result, Module];
	];

	coreInfoFile = Replace[OptionValue["CoreInformationDirectory"], {
		Automatic -> Automatic,
		dir_?DirectoryQ :> FileNameJoin[{
			dir,
			StringJoin[
				"SourceInformation.",
				Switch[language, "Japanese", "ja", "ChineseSimplified", "zh", "Spanish", "es", _,"en"],
				".m"
			]
		}],
		dir_ :> DocsError[
			"Error building System notebooks: core information directory '``' does not exist",
			dir
		]
	}];

	DocumentationBuild`Info`SetCoreInformation[language, coreInfoFile, {"System"}];

	DocumentationBuild`Info`SetSystemEnglishSourceInformationIfNeeded["Mathematica", language];

	notebooks = If[FileExistsQ[singleFile],
		{singleFile}
		,
		If[
			FailureQ @ setupMakeMathematicaNotebooks[
				inputDir, outputDir, highAndLowResolution, language, buildGroup,
				generateManifest
			],
			Return[$Failed]
		];

		AntLog["building sub-group "<>ToString[buildGroup]];

		getSystemDocumentationNotebooks[inputDir, language, buildGroup, filters]
	];

	AntLog["Notebooks to process: "<>ToString@Length[notebooks]];

	If[FailureQ[notebooks],
		Return[notebooks];
	];

	If[!MatchQ[notebooks, {___?StringQ}],
		Message[DocumentationBuildSystemNotebooks::error, StringForm[
			"Result is not a List of notebook files: ", notebooks
		]];
		Return[$Failed];
	];

	result = BuildWithProgress[
		Function[file,
			makeMathematicaNotebook[
				inputDir,
				outputDir,
				highAndLowResolution,
				language,
				file,
				releaseVersion
			]
		],
		notebooks,
		successRate,
		"Building Wolfram System documentation"
	];

	(* Don't leave the stylesheet cache notebooks visible in Notebooks[]. *)
	DocumentationBuild`Common`EmptyStylesheetCache[];

	(* What is the file size of all the documents *)
	(* TODO: Re-enable this as a field on the Success[..] object which should be returned.
	         This should probably be time constrained.
	If[highAndLowResolution,
		AntLog["Total filesize1: " <> ToString @ Total @ Map[FileByteCount, FileNames["*.nb", outputDir1, Infinity]] <> " bytes"];
		AntLog["Total filesize2: " <> ToString @ Total @ Map[FileByteCount, FileNames["*.nb", outputDir2, Infinity]] <> " bytes"]
		,
		AntLog["Total filesize: " <> ToString @ Total @ Map[FileByteCount, FileNames["*.nb", outputDir, Infinity]] <> " bytes"]
	];
	*)

	result
]] ~Catch~ $DocsErrorTag


DocumentationBuildSystemNotebooksIncremental[
	inputDir0_?StringQ,
	outputDir0_?StringQ,
	resultsLocation_?StringQ,
	opts:OptionsPattern[DocumentationBuildSystemNotebooks]
] := Module[{
	(* Expand '~' components before passing these paths down. *)
	inputDir = ExpandFileName[inputDir0],
	previousProcessedFiles,
	processedFiles,
	result
},
	previousProcessedFiles = If[FileExistsQ[resultsLocation],
		Get[resultsLocation]
		,
		DocsWarning[StringJoin[
			"Unable to locate previous build results for incremental rebuild.",
			" All documentation files will be rebuilt."
		]];

		<||>
	];

	DocsAssert[AssociationQ[previousProcessedFiles]];

	result = DocumentationBuildSystemNotebooks[
		inputDir,
		outputDir0,
		All,
		ChangedFilesFilter[previousProcessedFiles, inputDir, outputDir0],
		opts
	];

	processedFiles = result["ProcessedFiles"];

	DocsAssert[
		AssociationQ[processedFiles],
		"ERROR: \"ProcessedFiles\" key was missing or had unexpected value. Full build result was: ``",
		InputForm[result]
	];

	processedFiles = Association[KeyValueMap[
		Function[{file, result},
			FileNameDrop[file, Length @ FileNameSplit[inputDir]]
				-> <| "Result" -> result, "FileHash" -> FileHash[file, $DocsFileHashType, $DocsFileHashEncoding] |>
		],
		processedFiles
	]];

	(* Join previously processed files and newly processed files. *)
	Put[Join[previousProcessedFiles, processedFiles], resultsLocation];

	result
] ~Catch~ $DocsErrorTag


setupMakeMathematicaNotebooks[
	inputDir_?StringQ,
	outputDir_?StringQ,
	highAndLowResolution_?BooleanQ,
	language_?StringQ,
	buildGroup : (_?IntegerQ | All),
	generateManifest_?BooleanQ
] := Module[{
	exampleDataCheckoutDir
},

	AntLog @ StringJoin[
		"M- Version: ", ToString[SystemInformation["Small"], InputForm], "\n",
		"Installation Directory: ", ToString@$InstallationDirectory, "\n"
	];

	(* kill if reference.nb is 0k or missing *)
	If[
		FileByteCount[ToFileName[{$InstallationDirectory,"SystemFiles","FrontEnd","StyleSheets", "Wolfram"},"Reference.nb"] ] < 1 ||
		FileByteCount[ToFileName[{$InstallationDirectory,"SystemFiles","FrontEnd","StyleSheets", "Wolfram"},"Reference.nb"] ]===$Failed||
		FileByteCount[ToFileName[{$InstallationDirectory,"SystemFiles","FrontEnd","StyleSheets", "Wolfram"},"Reference-ja.nb"] ] < 1 ||
		FileByteCount[ToFileName[{$InstallationDirectory,"SystemFiles","FrontEnd","StyleSheets", "Wolfram"},"Reference-ja.nb"] ] === $Failed
		,
		AntLog["Reference.nb or Reference-ja.nb is corrupted or missing. Terminating. "];
		(* Failed build *)
		AntLog["Build failed"];
		(* Quit with a non-zero exit code *)
		Quit[1]
	];

	(* User variables *)

	If[highAndLowResolution,
		outputDir1 = FileNameJoin[{outputDir, "LowResolution"}];
		outputDir2 = FileNameJoin[{outputDir, "HighResolution"}];
	];

	Quiet[
		If[highAndLowResolution,
			CreateDirectory[outputDir1];
			CreateDirectory[outputDir2]
			,
			CreateDirectory[outputDir]
		],
		{CreateDirectory::filex}
	];

	If[generateManifest, Module[{entries},
		entries = Monitor[
			FileNames["Entries", FileNameJoin[FileNameSplit[inputDir]], Infinity],
			"Collecting CVS 'Entries' files to generate manifest.txt ..."
		];

		Monitor[
			If[highAndLowResolution,
				exportToTSV[ FileNameJoin[Flatten[{FileNameSplit[outputDir1], "manifest.txt"}]], entries ];
				exportToTSV[ FileNameJoin[Flatten[{FileNameSplit[outputDir2], "manifest.txt"}]], entries ];
				,
				exportToTSV[ FileNameJoin[Flatten[{FileNameSplit[outputDir ], "manifest.txt"}]], entries ]
			];
			,
			"Reading 'Entries' files and writing manifest.txt ..."
		]
	]];

	AntLog["Creating ExampleData folder"];
	(* This copyDirectoryNoCVS utility function copies the contents of a directory to another location, with some caveats:
		- it copies all files that are in the top level of the source dir.
		- it only copies subdirectories if they are named in the subdirsToInclude argument.
		- it excludes CVS directories at any level, always.
		The idea is to copy all files from ExampleData, ignoring all subdirectories except for specifically-requested ones.
	*)
	copyDirectoryNoCVS[srcDir_String, destDir_String, subdirsToInclude:{___String}:{}] :=
		Module[{subdirsPatt},
			(* This is the pattern we will use below to delete subdirs that do not match the subdirsToInclude. *)
			subdirsPatt = Alternatives @@ (FileNameJoin[{destDir, #}] ~~ ($PathnameSeparator | EndOfString) & /@ subdirsToInclude);
			Quiet[
				CreateDirectory[FileNameDrop[destDir]];

				(* The function works by  using CopyDirectory to copy the entire directory tree, and then it goes back and deletes
					any CVS dirs or other subdirectories not requested to be included.
				*)
				CopyDirectory[srcDir, destDir];
				,
				{CreateDirectory::filex, CopyDirectory::filex}
			];
			FileSystemScan[
				If[DirectoryQ[#] && (StringContainsQ[#, $PathnameSeparator ~~ "CVS" ~~ ($PathnameSeparator | EndOfString)] || ! StringContainsQ[#, subdirsPatt]),
					DeleteDirectory[#, DeleteContents -> True]
				]&,
				destDir, Infinity, IncludeDirectories -> True
			]
		];

	(* The place from which to copy the ExampleData files. *)
	exampleDataCheckoutDir = FileNameJoin[{inputDir, "Documentation", language, "System", "ExampleData"}];

    (* Copy ExampleData folder.
       ExampleData is only checked out in the Notebooks-Part01 build *)
	If[buildGroup === 1 || buildGroup === All,
		If[!FileExistsQ[exampleDataCheckoutDir],
			DocsError[
				"Input directory does not have the expected form: ExampleData folder does not exist: ``",
				exampleDataCheckoutDir
			];
		];

		(* Add any directories in ExampleData that we want to include in layouts: *)
		exampleDataSubdirsToInclude = {"SamplePaclet"};
		If[highAndLowResolution,
			copyDirectoryNoCVS[exampleDataCheckoutDir, FileNameJoin[{outputDir1, "Documentation", language, "System", "ExampleData"}], exampleDataSubdirsToInclude];
			copyDirectoryNoCVS[exampleDataCheckoutDir, FileNameJoin[{outputDir2, "Documentation", language, "System", "ExampleData"}], exampleDataSubdirsToInclude],
		(* else *)
			copyDirectoryNoCVS[exampleDataCheckoutDir, FileNameJoin[{outputDir, "Documentation", language, "System", "ExampleData"}], exampleDataSubdirsToInclude]
		];
		AntLog["Copied over ExampleData files"]
		,
		AntLog["Skipping ExampleData copy"]
	];

	AntLog @ StringJoin[
		"Current FE memory: ", ToString[UsingFrontEnd[MemoryInUse[$FrontEnd]]], "\n",
		"DocumentationBuild Directory: ", DocumentationBuild`Common`$DocumentationBuildDirectory, "\n"
	];
]

getSystemDocumentationNotebooks[
	inputDir_?StringQ,
	language_?StringQ,
	buildGroup : (_?IntegerQ | All),
	filters0_?ListQ
] := Module[{
	files,
	filegroups,
	filter,
	filters = filters0
},
	(* Gather files to process *)
	filegroups = {
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","Guides"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","HowTos"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","ReferencePages"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","Tutorials"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","ExamplePages"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","Workflows"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Documentation",language,"System","WorkflowGuides"}]],
		FileNameJoin[Flatten[{FileNameSplit[inputDir],"Packages"}]]
	};

	files = Flatten[Map[DocumentationBuild`Common`notebookFileNames, filegroups]];

	DocsAssert[MatchQ[files, {___?StringQ}]];

	filter = Replace[buildGroup, {
		All -> (True &),
		group_?IntegerQ :> buildGroupFilter[buildGroup, language]
	}];

	filters = Prepend[filters, filter];

	files = FilterFiles[inputDir, files, filters];

	DocsAssert[MatchQ[files, {___?StringQ}]];

	(* remove files which aren't under 'language' *)
	files = Select[files, StringMatchQ[#, "*" ~~ ($PathnameSeparator | "/") ~~ language ~~ "*"]& ];
	(* remove files which aren't notebooks *)
	files = Select[files, StringMatchQ[#, "*.nb"]& ];
	(* remove files which are excluded *)
	excluded = Map[
		($PathnameSeparator ~~ #) &,
		{
			(* XX Root guide rename change XX *)
			"Guides" ~~ $PathnameSeparator ~~ "Mathematica.nb",
			(* excluding this until it can be redone *)
			"FirstFiveMinutesWithMathematica.nb",
			"VirtualBookContents.nb", "QuickOverview.nb", "HandsOn", "WorkflowAgenda.nb",
			"WorkflowRoot.nb"
		}
	];
	files = (DeleteCases[files, f_ /; ! StringFreeQ[f, Alternatives @@ #, IgnoreCase -> True]] &[excluded]);

	files
]

buildGroupFilter[
	buildGroup_?IntegerQ,
	language_?StringQ
] := (
	(* distributing notebooks into groups to parallelize build -- largely to split out workflows and split up ref pages *)
	(* Groups: {{"A", "B"}, {"C", "D"}, {"E", "F", "H"}, {"G"}, {"I", "J", "K"}, {"L"}, {"M", "N", "O", "P"}, {"Q", "R", "T"}, {"S"},
	{"U", "V", "W", "X", "Y", "Z", "$"}}; *)

	Switch[buildGroup,
		1, (* Sets of smaller pages *)
			Function[file,
				MatchQ[
					FileNameSplit[file],
					Alternatives[
						{"Documentation", language, "System", "Guides", ___},
						{"Documentation", language, "System", "HowTos", ___},
						{"Documentation", language, "System", "Tutorials", ___},
						{"Documentation", language, "System", "ExamplePages", ___},
						{"Packages", ___}
					]
				]
			]
		,
		2, (* workflows and workflow guides *)
			Function[file,
				MatchQ[
					FileNameSplit[file],
					Alternatives[
						{"Documentation", language, "System", "Workflows", ___},
						{"Documentation", language, "System", "WorkflowGuides", ___}
					]
				]
			]
		,
		3, (* non-Symbol System` ref pages *)
			(
				!StringFreeQ[#, "System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator]
				&& StringFreeQ[#, $PathnameSeparator~~"Symbols"~~$PathnameSeparator]
			)&,
		4, (* start of System symbols *)
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("A"|"B")~~"*.nb"]&,
		5,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("C"|"D")~~"*.nb"]&,
		6,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("E"|"F"|"H")~~"*.nb"]&,
		7,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~"G*.nb"]&,
		8,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("I"|"J"|"K")~~"*.nb"]&,
		9,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~"L*.nb"]&,
		10,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("M"|"N"|"O"|"P")~~"*.nb"]&,
		11,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("Q"|"R"|"T")~~"*.nb"]&,
		12,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~"S*.nb"]&,
		13,
			StringMatchQ[#, "*System"~~$PathnameSeparator~~"ReferencePages"~~$PathnameSeparator~~"Symbols"~~$PathnameSeparator~~("U"|"V"|"W"|"X"|"Y"|"Z"|"$")~~"*.nb"]&,
		_,
			DocsError["Unhandled buildgroup value: ``", buildGroup];
	]
)

(**************************************)
(* makeMathematicaNotebook[]          *)
(**************************************)

makeMathematicaNotebook[
	inputDir_?StringQ,
	outputDir_?StringQ,
	highAndLowResolution_?BooleanQ,
	language_?StringQ,
	filename_?StringQ,
	releaseVersion_?StringQ
] := (
	result = TimeConstrained[
		doMakeMathematicaNotebook[
			inputDir,
			outputDir,
			highAndLowResolution,
			language,
			filename,
			releaseVersion
		],
		(* Number of seconds in the original MakeMathematicaNotebooks.xml
			script. ~26 minutes. *)
		1600,
		"TimeoutFailure"
	];

	If[result =!= "TimeoutFailure",
		Return[result];
	];

	(* FIXME:
		Below we just pretend that this file was skipped, when really it failed due to a
		timeout. I'm leaving this as-is for now because I want to know which files still
		fail due to a timeout (but without aborting the whole build just because a handful
		timed out).
	*)
	AntLog[StringJoin[
		"\nERROR:\nERROR:\nERROR: Timeout failure while processing: ",
		filename,
		"\nERROR:\nERROR:"
	]];

	Return[None];
)

doMakeMathematicaNotebook[
	inputDir_?StringQ,
	outputDir_?StringQ,
	highAndLowResolution_?BooleanQ,
	language_?StringQ,
	inFile_?StringQ,
	releaseVersion_?StringQ
] := With[{
	trustNotebooks = True
},
Module[{
	(* NOTE:
		This is always set to True, because historically all the the system notebooks
		builds included this debug logging. Setting this to false / removing this is
		dependent on a deeper overhaul of how timing metrics are collected deep down in
		MakeNotebook[] and the rest of DocumentationBuild`.

		TODO: Once the above is done, make this with the "Debug" option, orthogonal to
		      collection of timing data.
	*)
	timingDebug = True,
	feMemory,
	res
},
	(* Log time *)
	AntLog@DateString[{"Hour", ":", "Minute", ":", "Second", " ", "MonthName", " ", "Day", ", ", "Year"}];

	(* $DocumentationBuildDebug=True;
	Global`$AntBuildQ = True; *)

	(* User variables *)
	If[highAndLowResolution,
		outputDir1 = FileNameJoin[{outputDir, "LowResolution"}];
		outputDir2 = FileNameJoin[{outputDir, "HighResolution"}];
	];


	inFileNBExpr = DocumentationBuild`Utils`GetQuietly[inFile];
	If[highAndLowResolution,
		outFile1 = StringJoin[outputDir1, StringReplace[inFile, {inputDir->""}] ];
		outFile2 = StringJoin[outputDir2, StringReplace[inFile, {inputDir->""}] ]
		,
		outFile = StringJoin[outputDir, StringReplace[inFile, {inputDir->""}] ]
	];

	(* Entity Type / Page Type *)
	entityType = DocumentationBuild`Info`GetNotebookCategorization[inFileNBExpr, "Entity Type"];

	ignorefiles = Flatten[{
		StartOfString|$PathnameSeparator~~".", "FirstFiveMinutesWithMathematica.nb", "VirtualBookContents.nb", "FullNotationPalette.nb",
		"NotationPalette.nb", "HandsOnFirstLook01.nb", "HandsOnFirstLook02.nb", "HandsOnFirstLook03.nb", "HandsOnFirstLook04.nb",
		"HandsOnFirstLook05.nb", "HandsOnFirstLook06.nb", "HandsOnFirstLook07.nb", "HandsOnFirstLook08.nb", "HandsOnFirstLook09.nb",
		"HandsOnFirstLook10.nb", "HandsOnFirstLook11.nb", "HandsOnFirstLook12.nb", "HandsOnFirstLook13.nb", "HandsOnFirstLook14.nb",
		"HandsOnFirstLook15.nb", "HandsOnFirstLook16.nb", "HandsOnFirstLook17.nb", "HandsOnFirstLook18.nb", "HandsOnFirstLook19.nb",
		"HandsOnFirstLook20.nb", "HandsOnFirstLook21.nb", "HandsOnFirstLook22.nb","HandsOnTour01.nb", "HandsOnTour02.nb", "HandsOnTour03.nb",
		"HandsOnTour04.nb", "HandsOnTour05.nb", "HandsOnTour06.nb", "HandsOnTour07.nb", "HandsOnTour08.nb", "HandsOnTour09.nb",
		"HandsOnTour10.nb", "HandsOnTour11.nb", "HandsOnTour12.nb", "HandsOnTour13.nb", "HandsOnTour14.nb", "HandsOnTour15.nb",
		"HandsOnTour16.nb", "HandsOnTour17.nb", "QuickOverview.nb"
	}];

	feprocessid = getFEProcessID[];

	(* FIXME:
		Setting this on $FrontEnd instead of $FrontEndSession is a potential
		security vulnerabilty. (And even just setting it on $FrontEndSession is
		risky.) Though I'll note that at least as of WL v13.0 (and possibly earlier), this
		command doesn't seem to do anything, because `NotebookSecurityOptions` is not a
		System` symbols, and ends up resolving in this packages private context. SetOptions
		only recognizes the value if it is set using the string "NotebookSecurityOptions",
		or a symbol whose Context is equal to the current $Context. *)
	DocumentationBuild`Common`PubsEvaluateWithFE[SetOptions[
		$FrontEnd,
		If[trustNotebooks, NotebookSecurityOptions->{"TrustByDefault"->True}, Sequence[]],
		RenderingOptions -> {"HardwareAntialiasingQuality" -> 1}
	]];

	AntLog["FrontEnd Memory: " <> ToString[UsingFrontEnd[MemoryInUse[$FrontEnd]]]];
	AntLog["Kernel Memory: " <>ToString@MemoryInUse[]];
	AntLog["FrontEnd ProcessID: " <> ToString[feprocessid]];
	AntLog["Kernel ProcessID: " <>ToString[$ProcessID]];
	AntLog["Quality Level: " <>ToString[InputForm @ DocumentationBuild`Common`PubsEvaluateWithFE[CurrentValue[$FrontEnd, {RenderingOptions, "HardwareAntialiasingQuality"}]]]];
	DocumentationBuild`Common`PubsEvaluateWithFE[SetOptions[$FrontEnd, RenderingOptions -> {"HardwareAntialiasingQuality" -> 1}]];
	AntLog["Quality Level Second Time: " <>ToString[InputForm @ DocumentationBuild`Common`PubsEvaluateWithFE[CurrentValue[$FrontEnd, {RenderingOptions, "HardwareAntialiasingQuality"}]]]];
	DocumentationBuild`Common`PubsEvaluateWithFE[SetOptions[$FrontEndSession, RenderingOptions -> {"HardwareAntialiasingQuality" -> 1}]];
	AntLog["Quality Level Session: " <>ToString[InputForm @ DocumentationBuild`Common`PubsEvaluateWithFE[CurrentValue[$FrontEndSession, {RenderingOptions, "HardwareAntialiasingQuality"}]]]];


	AntLog["Processing: " <> FileNameDrop[inFile, Length @ FileNameSplit[inputDir]]];

	Which[
		(* ignore files and CopyFile to output dir without converting *)
		DocumentationBuild`Utils`ConvertFileQ[ inFile, ignorefiles],
			AntLog["          File marked to ignore -- including in build by copying to output dir."];
			If[highAndLowResolution,
				CreateDirectory[ DirectoryName[outFile1] ];
				CopyFile[ inFile, outFile1];
				CreateDirectory[ DirectoryName[outFile2] ];
				CopyFile[ inFile, outFile2]
				,
				CreateDirectory[ DirectoryName[outFile] ];
				CopyFile[ inFile, outFile];
			];
			Return[None];
		,
		(* Commenting this out for now; might need it again
		(* Alternative Mathematica Root Guide Page *)
		(FileBaseName[inFile]=="Mathematica") && (FileExistsQ[FileNames["MathematicaAlternate.nb", appPath, Infinity][[1]]]),
			AntLog["            Alternative Mathematica Root Guide Page  -- including in build by copying local copy to output dir."];
			If[highAndLowResolution,
				CreateDirectory[ DirectoryName[outFile1] ];
				alternativeNB=First[FileNames["MathematicaAlternate.nb", appPath, Infinity]];
				CopyFile[ alternativeNB, outFile1];
				CreateDirectory[ DirectoryName[outFile2] ];
				alternativeNB=First[FileNames["MathematicaAlternate.nb", appPath, Infinity]];
				CopyFile[ alternativeNB, outFile2]
				,
				CreateDirectory[ DirectoryName[outFile] ];
				alternativeNB=First[FileNames["MathematicaAlternate.nb", appPath, Infinity]];
				CopyFile[ alternativeNB, outFile]
			];
			Return[];
		,
		*)
		(* Notebook to convert doesn't contain proper EntityType/PageType *)
		entityType === $Failed,
			AntLog["            File's Entity Type not found -- file not included in build."];
			Return[None];
		,
		(* pass over flagged nbs which shouldn't be included within distribution *)
		TrueQ[
			(DocumentationBuild`Make`Private`CheckVersion[
				("New" /. DocumentationBuild`Info`GetNotebookHistoryData[inFileNBExpr]),
				"ReleaseVersion" -> releaseVersion,
				"ThresholdVersion" -> Automatic
			] > 1) ||
			(DocumentationBuild`Make`Private`CheckVersion[
				("Modified" /. DocumentationBuild`Info`GetNotebookHistoryData[inFileNBExpr]),
				"ReleaseVersion" -> releaseVersion,
				"ThresholdVersion" -> Automatic
			] > 1)
		],
			AntLog["             -- files not included in build due to history data greater than " <> releaseVersion];
			Return[None];
		,
		(* pass over flagged nbs which shouldn't be included within distribution *)
		MatchQ[ flag = DocumentationBuild`Utils`DocumentFlag[ inFileNBExpr] , "ExcisedFlag" | "FutureFlag" | "InternalFlag" | "TemporaryFlag"],
			AntLog["            " <> flag <> " -- files not included in build."];
			Return[None];
		,
		(* Convert everything else *)
		True, Null
	];


	If[highAndLowResolution,
		time = AbsoluteTiming[
			{res, metaData} = DocumentationBuild`Make`MakeNotebook[
				inFile,
				"outfile"->outFile1,
				"outfile2"->outFile2,
				"Language" -> language,
				"Debug" -> timingDebug
			];

			If[(fileBuildFailedQ[outFile1] || fileBuildFailedQ[outFile2]) && feprocessid =!= getFEProcessID[],
				AntLog["Reprocessing: Due to Frontend restart"];

				{res, metaData} = DocumentationBuild`Make`MakeNotebook[
					inFile,
					"outfile"->outFile1,
					"outfile2"->outFile2,
					"Language" -> language,
					"Debug" -> timingDebug
				]
			]
		][[1]];

		(* Output file test *)
		If[(fileBuildFailedQ[outFile1] || fileBuildFailedQ[outFile2] ||
			(ValueQ[DocumentationBuild`Private`rastertime] && SameQ[Chop[DocumentationBuild`Private`rastertime, 10^-5], 0])),
			AntLog["*** Output file test failed *** "];
			res = $Failed
		]
		,
		time = AbsoluteTiming[
			{res, metaData} = DocumentationBuild`Make`MakeNotebook[
				inFile,
				"outfile"->outFile,
				"Language" -> language,
				"Debug" -> timingDebug
			];

			If[fileBuildFailedQ[outFile] && feprocessid =!= getFEProcessID[],
				AntLog["Reprocessing: Due to Frontend restart"];

				{res, metaData} = DocumentationBuild`Make`MakeNotebook[
					inFile,
					"outfile"->outFile,
					"Language" -> language,
					"Debug" -> timingDebug
				]
			]
		][[1]];

		(* Output file test *)
		If[(fileBuildFailedQ[outFile] ||
			(ValueQ[DocumentationBuild`Private`rastertime] && SameQ[Chop[DocumentationBuild`Private`rastertime, 10^-5], 0])),
			AntLog["*** Output file test failed *** "];
			res = $Failed
		];
	];


	If[FailureQ[res],
		Return[res];
	];

	If[highAndLowResolution,
		AntLog["outFile1: "<>ToString[outFile1]];
		AntLog[StringJoin[ "File Size: ", ToString[If[FileExistsQ[outFile1], FileByteCount[outFile1],"unknown"]], " bytes" ]];
		AntLog["outFile2: "<>ToString[outFile2]];
		AntLog[StringJoin[ "File Size: ", ToString[If[FileExistsQ[outFile2], FileByteCount[outFile2],"unknown"]], " bytes" ]];
		,
		AntLog["outFile: "<>ToString[outFile]];
		AntLog[StringJoin[ "File Size: ", ToString[If[FileExistsQ[outFile], FileByteCount[outFile],"unknown"]], " bytes" ]];
	];
	AntLog[StringJoin[ "Success!    ", ToString[time], " secs" ]];


	(* Force quit the kernel after n requests *)
	feMemory = UsingFrontEnd[MemoryInUse[$FrontEnd]];
	If[ feMemory > 1*^9, (
		AntLog[StringJoin[
			"Noticed the FE memory was ",
			ToString[InputForm[feMemory]],
			" which is greater than 1*^9, will run UninstallFrontEnd to alleviate this issue..."
		]];
		Developer`UninstallFrontEnd[];
		AntLog["Finished quitting FE"];
	)];

	Success["makeMathematicaNotebook", <|
		"Time" -> Quantity[time, "Seconds"]
	|>]
]]

(**************************************)
(* Utilities                          *)
(**************************************)

repository[entry_] := StringReplace[entry, "Entries" -> "Repository"];

exportToTSV[outfile_?StringQ, entries_?ListQ] := Export[
	outfile,
	Cases[
		Table[
			Map[
				StringReplace[#, {"//\n/" -> "/"}] &,
				Cases[
					Cases[
						Map[
							StringSplit[#, "/"] &,
							ReadList[entries[[i]], "String"]
						],
						{z_, w_, __} :> {
							ReadList[repository[entries[[i]]], "String"][[1]] <> "/" <> z,
							w
						},
						Infinity
					],
					{a_String, b_String},
					Infinity
				]
			],
			{i, 1, Length[entries]}
		],
		{a_String, b_String},
		2
	],
	"TSV"
]

getFEProcessID[] := DocumentationBuild`Common`PubsEvaluateWithFE[
	SystemInformation["FrontEnd", "ProcessID"]
]

fileBuildFailedQ[file_?StringQ] := With[{
	minFileSize = 1300
},
	FileType[file] === None || FileByteCount[file] < minFileSize
]



End[]

EndPackage[]