BeginPackage["DocumentationBuild`Build`Utils`"]

GetDocumentationNotebooks
FilterFiles
ChangedFilesFilter

SetCombinedSourceInformation
CreateCombinedSourceInformation

BuildWithProgress

$DocsFileHashType = "SHA256"
$DocsFileHashEncoding = "Base64Encoding"

Begin["`Private`"]

Needs["DocumentationBuild`"];
Needs["DocumentationBuild`Common`"];

(* Load GeneralUtilities without putting it on the context path. We only want to explicilty
   use symbols. *)
Block[{$ContextPath = $ContextPath},
    Needs["GeneralUtilities`"]
]

(**************************************)
(* GetDocumentationNotebooks          *)
(**************************************)

GetDocumentationNotebooks[
	inputDir_?StringQ,
	language_?StringQ,
	(* This can be a predicate function or a String/StringExpression. This is applied to
	   the *relative* file name (without the `inputDir` prefix). *)
	filter0_
] := Module[{
	files,
	(* Canonicalize string pattern filter to a predicate function filter. *)
	filter = Replace[filter0, _StringExpression | _String :> StringMatchQ[filter0]],
	excluded
},
	(* Gather files to process *)
	files = DocumentationBuild`Common`notebookFileNames[inputDir];

	(* Filter the files (with relative paths) using the `filter` provided by the user. *)
	files = FilterFiles[inputDir, files, {filter}];

	(* remove files which aren't under 'language' *)
	files = Select[
		files,
		StringMatchQ[#, "*" ~~ ($PathnameSeparator | "/") ~~ language ~~ "*"]&
	];

	(* remove files which aren't notebooks *)
	files = Select[files, StringMatchQ[#, "*.nb"]& ];

	(* remove files which are excluded *)
	excluded = Join[
		Map[($PathnameSeparator ~~ #) &, {"UndefinedStyles.nb", "UnusedMaterial.nb", "ReferencePagesReport.nb"}],
		Map[($PathnameSeparator ~~ # ~~ $PathnameSeparator) &, {"OriginalReferencePages"}]
	];
	files = (
		DeleteCases[files, f_ /; ! StringFreeQ[f, Alternatives @@ #, IgnoreCase -> True]] &[excluded]
	);

	files
]

(**************************************)
(* BuildWithProgress                  *)
(**************************************)

BuildWithProgress[
	processFile_Function,
	notebooks_?ListQ,
	requiredSuccessRate_?NumberQ,
	description_?StringQ
] := Module[{
	processedFiles = <||>,
	processedFileCount = 0,
	successfulFileCount = 0,
	skippedFileCount = 0,
	failedFilesCount = 0,
	max = {Quantity[0.0, "Seconds"], None},
	totalTime,
	aborted = False,
	remaining,
	metadata
},
	(*----------------------------------------------------*)
	(* Perform the main loop over each notebook file      *)
	(*----------------------------------------------------*)

	(* Use CheckAbort[] so that partial results can be returned if/when the user aborts
	   part way through *)
	totalTime = AbsoluteTiming @ CheckAbort[
		GeneralUtilities`ComputeWithProgress[
			Function[{callback},
				Scan[
					Function[file, Module[{
						time, res
					},
						Assert[StringQ[file]];
						Assert[FileExistsQ[file]];

						(* Note: Report progress *before* actually processing any files.
						   This causes the progress indicator cell to display as early as
						   possible, which is decidedly better UX when it's possible for
						   the first file to take e.g. more than 10 seconds to process. *)
						callback[<|
							"progress" -> processedFileCount / Length[notebooks],
							"processed" -> processedFileCount,
							"totalNotebooks" -> Length[notebooks],
							"successful" -> successfulFileCount,
							"failed" -> failedFilesCount,
							"skipped" -> skippedFileCount,
							"filename" -> FileNameTake[file, -1]
						|>];

						(*---------------------------------------------------------*)
						(* Call `processFile` and handle DocsError[..] exceptions. *)
						(*-------------------------------------------------------- *)

						(* `res should be Success[..], None, or FailureQ. *)
						(* Catch any DocsError[..] exceptions which processFile didn't
						   choose to handle. *)
						res = Catch[
							processFile[file],
							$DocsErrorTag,
							Function[{failure, tag}, "CaughtException"[failure]]
						];

						res = Replace[res, {
							(* Include the `file` name in the thrown Failure message. *)
							"CaughtException"[Failure[
								tag:"DocumentationError",
								data:KeyValuePattern[{"MessageTemplate" -> template_?StringQ}]
							]] :> Failure[tag, Append[
								data,
								"MessageTemplate" -> StringJoin[
									"Error processing '" <> file <> "': ",
									template
								]
							]],
							"CaughtException"[inner_] :> (
								DocsWarning["Unexpected DocsError exception format: ``", InputForm[inner]];
								inner
							)
						}];

						(*-------------------*)
						(* Record the result *)
						(*-------------------*)

						Replace[res, {
							_?FailureQ :> (
								failedFilesCount += 1;
							),
							None :> (
								skippedFileCount += 1;
							),
							success_Success :> (
								successfulFileCount += 1;
							),
							_ :> (
								Message[DocumentationBuild::error, StringForm[
									"Unrecognized result type returned from processFile: ``",
									res
								]];
							)
						}];

						processedFileCount += 1;
						AssociateTo[processedFiles, file -> res];
					]],
					notebooks
				];
			],
			<|
				"Static" -> description,
				"Dynamic" -> StringJoin[
					description,
					"\n",
					"Processing `filename`",
					"\n",
					"`processed` of `totalNotebooks`:  `successful` successful, `skipped` skipped, `failed` failed"
				]
			|>,
			"MinimumProgressTime" -> 0
		]
		,
		remaining = Length[notebooks] - processedFileCount;
		Message[DocumentationBuild::error, StringForm[
			"Abort occurred while in main loop. Returning partial results early. `` `` not processed.",
			remaining,
			Pluralize[{"file was", "files were"}, remaining]
		]];
		aborted = True;
	];

	totalTime = totalTime[[1]];

	Assert[NumberQ[totalTime]];

	(*----------------------------------------------------*)
	(* Package up the results into a Success or Failure   *)
	(*----------------------------------------------------*)

	If[!aborted,
		Assert[Length[processedFiles] === Length[notebooks]];
	];

	max = MaximalBy[
		Map[
			#["Time"]&,
			Select[processedFiles, MatchQ[_Success]]
		],
		Identity
	];

	If[max =!= <||>,
		(* Transform `"file" -> time` into {time, "file"}. *)
		max = Reverse[List @@ First[Normal[max]]];

		Assert[
			MatchQ[max, {_?QuantityQ, None | _?StringQ}],
			"max has unexpected value: ``",
			InputForm[max]
		];
	];

	(* This information is included both the Success and Failure case result object. *)
	metadata = <|
		(* Note: It's significant that this is the first item in this Association; this
		   causes it to appear in the un-expanded form of the Success/Failure object
		   summary box. *)
		"TotalTime" -> Quantity[totalTime, "Seconds"],
		(* Which file took the longest to process? *)
		"MaximumFileTime" -> max,
		(* TODO: Make these actual lists of files, not just counts. Figure out how to make
		         this look good in the console output. *)
		"ProcessedFiles" -> processedFiles,
		"FailedFilesCount" -> failedFilesCount
	|>;

	constructBuildResult[
		requiredSuccessRate,
		processedFileCount,
		successfulFileCount,
		skippedFileCount,
		metadata,
		aborted
	]
]

(**************************************)
(* constructBuildResult               *)
(**************************************)

(* Construct a Success or Failure object based on the outcome of a documentation build. *)
constructBuildResult[
	requiredSuccessRate_?NumberQ,
	(* The total number of files checked in the main build loop:
	      converted ("successful") + skipped + failed.
	   If the build was aborted, this number will be less than the total number of input
	   files.
	*)
	processedFileCount_?IntegerQ,
	successfulFileCount_?IntegerQ,
	skippedFileCount_?IntegerQ,
	callerMetadata_?AssociationQ,
	aborted_?BooleanQ
] := Module[{metadata, successRate, error},
	(* Assert[processedFileCount === successfulFileCount + numFailedFiles]; *)

	successRate = If[processedFileCount === 0,
		Indeterminate,
		(successfulFileCount + skippedFileCount) / processedFileCount
	];

	metadata = Join[
		If[aborted,
			<| "Aborted" -> True |>,
			<||>
		],
		callerMetadata,
		<|
			"PercentSucceeded" -> Quantity[100.0 * successRate, "Percent"],
			"ProcessedFilesCount" -> processedFileCount,
			"SuccessfulFilesCount" -> successfulFileCount,
			"SkippedFilesCount" -> skippedFileCount
		|>
	];

	Which[
		aborted
	,
		error = ToString[StringForm @@ {
			"``% (``) of processed files were successful.",
			100.0 * successRate,
			ToString[successfulFileCount + skippedFileCount] <> "/" <> ToString[processedFileCount]
		}];

		Failure["DocumentationBuild", Join[
			metadata,
			<|
				"MessageTemplate" -> "Documentation build aborted. " <> error
			|>
		]]
	,
		successRate === Indeterminate || successRate >= requiredSuccessRate
	,
		(* Successful build *)
		Success["DocumentationBuild", Join[
			metadata,
			<|
				"MessageTemplate" -> "Documentation build completed successfully."
			|>
		]]
	,
		True
	,
		(* Failed build *)
		error = ToString[StringForm @@ {
			"``% (``) of files were processed successfully. A minimum success rate of ``% is required.",
			100.0 * successRate,
			ToString[successfulFileCount + skippedFileCount] <> "/" <> ToString[processedFileCount],
			100.0 * requiredSuccessRate
		}];

		Assert[StringQ[error]];

		Message[DocumentationBuild::error, error];

		Failure["DocumentationBuild", Join[
			metadata,
			<|
				"MessageTemplate" -> "Documentation build failed. " <> error
			|>
		]]
	]
]

(**********************************************************)

SetCombinedSourceInformation[
	linkBase_,
	inputDir_?StringQ,
	language_?StringQ
] := Module[{sourceInfo},
	If[!MatchQ[linkBase, _?StringQ | Automatic],
		DocsError[
			"invalid \"LinkBase\" setting: ``. Expected a string or Automatic.",
			InputForm[linkBase]
		];
	];

	sourceInfo = If[
		MemberQ[
			{Automatic, "Mathematica", "webMathematica", "webMathematicaDocumentation"},
			linkBase
		],
		Automatic
		,
		CreateCombinedSourceInformation[linkBase, inputDir, language]
	];

	DocsAssert[
		MatchQ[sourceInfo, Automatic | _?AssociationQ],
		"Expected sourceInfo to be Automatic or Association, got: ``",
		Short[sourceInfo]
	];

	DocumentationBuild`Info`SetCoreInformation[
		language,
		sourceInfo,
		If[sourceInfo === Automatic,
			{"System"},
			{"System", linkBase}
		]
	];
]

SetCombinedSourceInformation[args___] :=
	DocsError["SetCombinedSourceInformation: bad arguments: ``", {args}]

(* Create combined source information data for the documentation directory at `inputDir`
   and the System documentation. *)
CreateCombinedSourceInformation[
	linkBase_?StringQ,
	inputDir_?StringQ,
	language_?StringQ
] := Module[{
	systemSourceInfo,
	docsSourceInfo
},
	(* Get the source information for the System documentation. *)
	systemSourceInfo = DocumentationBuild`Info`LoadCoreInformation[Automatic, language];

	(* Compute the source information for this documentation input directory. *)
	docsSourceInfo = DocumentationBuild`Info`CreateSourceInformation[inputDir, language, linkBase];

	DocsAssert[MatchQ[systemSourceInfo, <| (_?StringQ -> {(_?StringQ -> _) ...}) ...|>]];
	DocsAssert[MatchQ[docsSourceInfo  , {(_?StringQ -> {(_?StringQ -> _) ...}) ...}]];

	docsSourceInfo = Association[docsSourceInfo];

	DocsAssert[AssociationQ[systemSourceInfo] && AssociationQ[docsSourceInfo]];

	(* Verify that the URIs in the system source info and the URIs in the local
		docs do not clash/overlap. *)
	Module[{intersection = Intersection[Keys[systemSourceInfo], Keys[docsSourceInfo]]},
		If[intersection =!= {},
			DocsWarning[
				"Intersection between system documentation and provided documentation"
				<> " source information: ``",
				InputForm[Take[intersection, UpTo[5]]]
			];
		];
	];

	(* Return the combined System info + local docs info. *)
	Join[systemSourceInfo, docsSourceInfo]
]

(**************************************)
(* Incremental build utilities        *)
(**************************************)

(*--------------------------------------------------------*)

(* processedFileHashes should be an association of the form <| "filePath" -> "fileHash" |>,
   representing the identity of a previously processed file as it's location and content
   hash. *)
ChangedFilesFilter[
	previousResults_?AssociationQ,
	inputDir_?StringQ,
	outputDir_?StringQ
] := (
	(* This could fail if the `previousResults` were loaded from a file generated by a
	   previous version of the incremental build system that had a different format. *)
	DocsAssert[
		MatchQ[
			previousResults,
			<|
				RepeatedNull[
					_?StringQ -> <|
						"Result" -> (_Success | _?FailureQ | None),
						"FileHash" -> _?StringQ
					|>
				]
			|>
		],
		"Incremental results data does not have the expected form. It may have been corrupted,
		 or the format may have changed. Suggestion: delete the incremental build results
		 directory, and rebuild."
	];

	(* DocsAssert[MatchQ[
		processedFileHashes,
		<| (_?StringQ -> (_?StringQ | $Failed)) ... |>
	]]; *)

	Function[relFile, Module[{
		entry,
		inputFile,
		outputFile,
		previousHash,
		currentHash
	},
		entry = Lookup[previousResults, relFile];

		If[MissingQ[entry],
			(* If a file hasn't ever been processed before, build it. *)
			Return[True, Module];
		];

		If[FailureQ[entry["Result"]],
			(* If a file previously failed to process, re-process it (even if it's hash
			   hasn't changed). *)
			Return[True, Module];
		];

		inputFile = FileNameJoin[{inputDir, relFile}];
		outputFile = FileNameJoin[{outputDir, relFile}];

		previousHash = entry["FileHash"];
		currentHash = FileHash[inputFile, $DocsFileHashType, $DocsFileHashEncoding];

		(* If the file hash of the input file has changed (i.e. it's contents have
		   changed), re-process it. *)
		If[currentHash =!= previousHash,
			Return[True, Module];
		];

		(* If the output file doesn't exist in the output directory, build this file. This
		   can happen e.g. if the built paclet contents directory was deleted, but the
		   'incremental' folder containing the build metadata was not deleted. *)
		If[!FileExistsQ[outputFile],
			Return[True, Module];
		];

		(* If the file has been processed before, it's input file hash hasn't changed, and
		   the corresponding output file is present in the output directory, don't rebuild
		   it. *)
		(* FIXME: Also store and check the file hash of the output file, and rebuild if
		          the output file has changed. *)
		False
	]]
)

(*--------------------------------------------------------*)

FilterFiles[
	baseDir_?StringQ,
	files0:{___?StringQ},
	(* Predicate function or list of predicate functions. A file must pass each filter
	   to be present in the return list. The file path passed to the predicate functions
	   is relative to `baseDir`. *)
	filters0_
] := Module[{
	files = files0,
	filters = Replace[filters0, filter:Except[_?ListQ] :> {filter}]
},
	DocsAssert[ListQ[filters]];

	(* Make the file paths relative to `baseDir` for filtering. *)
	files = Map[
		FileNameDrop[#, Length[FileNameSplit[baseDir]]] &,
		files
	];

	files = SelectAll[files, filters];
	DocsAssert[MatchQ[files, {___?StringQ}]];

	(* Make the file paths absolute (starting with `baseDir`) again. *)
	files = Map[FileNameJoin[{baseDir, #}] &, files];

	files
]

(*--------------------------------------------------------*)

(* SelectAll selects elements of `list` which satisfy every predicate in `predicates`. *)
SelectAll[list_?ListQ, predicates_?ListQ] :=
	Select[list, Function[elem, AllTrue[predicates, #[elem] &]]]




End[]

EndPackage[]