BeginPackage["DocumentationBuild`Citations`"]

(*
	This package contains logic for building citation text.

	Formats are based on:

	Text:
		https://mail-archive.wolfram.com/archive/t-newdocs/2020/Mar00/0082.html

	CMS:
		https://www.chicagomanualofstyle.org/tools_citationguide/citation-guide-1.html#cg-website

	BibTeX:

		https://www.bibtex.com/e/entry-types/#misc
		https://www.bibtex.com/f/title-field/
		https://www.bibtex.com/f/howpublished-field/

		https://tex.stackexchange.com/questions/408530/what-characters-are-allowed-to-use-as-delimiters-for-bibtex-keys

	BibLaTeX:
		Section 2.1.1 entry for @online (pg. 11) in:
			http://linorg.usp.br/CTAN/macros/latex/contrib/biblatex/doc/biblatex.pdf

*)

CitationText::usage = "CitationText[info, style] produces the textual citation content \
needed to cite $info in the citation $style in a bibliography."

$CitationStyleTemplates::usage = "$CitationStyleTemplates contains string templates used \
by CitationText to construct a citation of the specified style."


Begin["`Private`"]


Needs["DocumentationBuild`Common`"] (* For DocsAssert, DocsError *)

(**********************************************************)

$CitationStyleTemplates = <|
	"Text" -> "`Creator` (`CreationYear`), `Title`, `Description`, `RetrievedFromURL`",
	(* TODO: `ContainedIn` should be italicized. *)
	"CMS" -> <|
		"WithModifiedYear" -> "`Creator`. \"`Title`\". `ContainedIn`, `CreationYear`. Last modified `LastModifiedYear`. `RetrievedFromURL`.",
		"WithoutModifiedYear" -> "`Creator`. \"`Title`\". `ContainedIn`, `CreationYear`. `RetrievedFromURL`."
	|>,
	"APA" -> "`Creator`. (`CreationYear`). `Title`. `ContainedIn`. Retrieved from `RetrievedFromURL`",
	(* TODO: Add note=\"[version `PacletVersion`]\" *)
	"BibTeX" :> "@misc{`TeXCitationKey`, author=\"{`Creator`}\", title=\"{`Title`}\", year=\"`CreationOrModifiedYear`\", howpublished=\"\\url{`RetrievedFromURL`}\"}",
	(* TODO: Add note=\"[version `PacletVersion`]\" *)
	"BibLaTeX" :> "@online{`TeXCitationKey`, author={`Creator`}, title={`Title`}, year={`CreationOrModifiedYear`}, url={`RetrievedFromURL`}}"
|>;

(**********************************************************)

CitationText[citationInfo0_?AssociationQ, style_?StringQ] := Module[{
	citationInfo = citationInfo0,
	containedIn,
	modifiedYear,
	template,
	result
},
	(*-----------------------------*)
	(* Validate the `citationInfo` *)
	(*-----------------------------*)

	DocsAssert[
		MatchQ[
			citationInfo,
			KeyValuePattern[{
				"Title" -> _?StringQ,
				"Creator" -> _?StringQ,
				(* TODO: Accept DateObject? *)
				"CreationYear" -> _?StringQ,
				"RetrievedFromURL" -> _?StringQ,
				(* Brief description of what this resource type is. E.g.
				   "Wolfram Language function".

				   This is used by the "Text" style, and is (presumably) intended to give
				   some extra context to citations for resources that have very short
				   titles, like e.g. "Plot". *)
				"Description" -> _?StringQ
			}]
		],
		"CitationText info does not contain all the necessary fields, or the type of " <>
		"some fields is invalid: ``",
		InputForm[citationInfo]
	];

	modifiedYear = Lookup[citationInfo, "LastModifiedYear", None];
	containedIn = Lookup[citationInfo, "ContainedIn", None];

	DocsAssert[
		MatchQ[modifiedYear, _?StringQ | None],
		"CitationText \"LastModifiedYear\" must be a string or None. Got: ``",
		InputForm[modifiedYear]
	];

	DocsAssert[
		MatchQ[containedIn, _?StringQ | None],
		"CitationText \"ContainedIn\" must be a string or None. Got: ``",
		InputForm[containedIn]
	];

	(* Set the "CreationOrModifiedYear" derived info field, which is the most recent
	   date associated with this resource. *)
	citationInfo["CreationOrModifiedYear"] =
		Replace[modifiedYear, None -> citationInfo["CreationYear"]];

	(*-------------------------------*)
	(* Determine the template to use *)
	(*-------------------------------*)

	template = Lookup[
		$CitationStyleTemplates,
		style,
		DocsError["CitationText: unknown citation style: ``", InputForm[style]]
	];

	template = Replace[template, {
		_?StringQ -> template,
		(* Some templates have subvariants depending on whether a modified date is
		   available. *)
		<|
			"WithModifiedYear" -> withModifiedYear_?StringQ,
			"WithoutModifiedYear" -> withoutModifiedYear_?StringQ
		|> :> If[modifiedYear =!= None,
			withModifiedYear,
			withoutModifiedYear
		],
		other_ :> DocsError[
			"CitationText: $CitationStyleTemplates entry for `` has an unexpected form: ``",
			InputForm[style],
			InputForm[other]
		]
	}];

	(*---------------------------------*)
	(* Apply the style string template *)
	(*---------------------------------*)

	template = StringTemplate[
		template,
		InsertionFunction -> Function[term,
			If[MissingQ[term],
				DocsError[
					"CitationText: citation style `` requires missing citation info field: ``",
					InputForm[style],
					InputForm[term]
				]
			];
			term
		]
	];

	result = TemplateApply[template, citationInfo];

	DocsAssert[MatchQ[result, _?StringQ]];

	result
]




End[]

EndPackage[]