(* ::Package:: *)

(* :Title: DocumentationBuild.m *)

(* :Authors:
	Andrew Hunt, Shaun McCance
	andy@wolfram.com, shaunm@wolfram.com
*)

(* :Package Version: 0.50 *)

(* :Mathematica Version: 6.0 *)
                     
(* :Copyright: (c) 2005, Andrew Hunt, Shaun McCance
               All rights reserved. 
*)

(* :Requirements: *)
(* :Discussion:
	Documentation conversion from :
		* authored nbs -> in-product notebooks
		* authored nbs -> web pages
		* authored nbs -> ___
*)

BeginPackage["DocumentationBuild`Common`", {
  "JLink`",
  "Transmogrify`",
  "XML`"
}]
(* Exported symbols added here with SymbolName::usage *)  


(* Very general messages. The number of possible errors which can occur during a
   documentation build is vast. Having a unique message name for each one isn't really
   worth the curration effort. *)
DocumentationBuild`DocumentationBuild::error = "``";
DocumentationBuild`DocumentationBuild::warning = "warning: ``";
DocumentationBuild`DocumentationBuild::assertfail = "``";
(* Note: PacletTools will Quiet this message when loading DocumentationBuild. *)
DocumentationBuild`DocumentationBuild::nodocsassets = "``";

DocsWarning::usage = "DocsWarning[formatStr, args___] generates a message indicating a non-fatal warning encountered during the documentation build process";
DocsError::usage = "DocsError[formatStr, args___] throws a Failure object indicating an error encountered during during the documentation build process";
$DocsErrorTag
DocsAssert::usage = "DocsAssert[cond, formatStr, args___] throws a Failure object indicating a failed assertion encountered during the documentation build process";
DocsRemoved::usage = "DocsRemoved[name] generates an error indicating that the function name has been removed from DocumentationBuild."

DocsAsset::usage = "DocsAsset[kind, name] returns the file path to a documentation asset or throws if it cannot be located";

$DocumentationBuildDirectory::usage = "Path to the DocumentationBuild base directory.";
$DocumentationBuildTransformDirectory::usage = "Path to the DocumentationBuild XMLTransform directory.";

$DocumentationBuildDebug::usage = "";

ServerLog;
$ServerBuildLogQ;
separator;

PubsEvaluateWithFE::usage = "PubsEvaluateWithFE[expr] is a wrapper for Developer`UseFrontEnd[...] which also sets various options useful for building documentation.";

`$ImageCount = 0;

FilePathToDocURI::usage = "FilePathToDocURI[path] returns the documentation url form of path.";

ToPacletURI::usage = "ToPacletURI[uri] canonicalizes uri to start with the \"paclet:\" prefix.";

CacheStylesheet::usage = "CacheStylesheet[nb] opens the stylesheet used by nb in an invisible notebook, to force the FrontEnd to cache it."
EmptyStylesheetCache::usage = "";

ExcludeFilesFromList::usage = "ExcludeFilesFromList[list, excludedFiles] removes files within list which match patterns defined in excludedFiles";
ExcludePathsFromList::usage = "ExcludePathsFromList[list, excludedPaths] removes paths within list which match patterns defined in excludedPaths";
ConvertGraphicsToBitmaps::usage = "";
TagBitmapConditionalizedGraphics::usage = "";


FilepathToURI::usage = "FilepathToURI[path] returns the documentation url form of path.";

notebookFileNames::usage = "";
Options[notebookFileNames] = {
  ExcludeDirectories->{"CVS", "Internal", "RawMaterial", "ExampleData", "Future", "RawGuides", "Future", "PackagesMirror", (*"Messages", "Compatibility",*)
  (* exclude per lbyrge *)
  "AppleScript", "XMLCapabilities",
  "BarAndPieCharts"~~$PathnameSeparator,
  "ClusterAnalysis"~~$PathnameSeparator,
  "ComputationalGeometry2D"~~$PathnameSeparator,
  "ComputationalGeometry3D"~~$PathnameSeparator,
  "DeveloperUtilities"~~$PathnameSeparator,
  "ErrorBarPlot"~~$PathnameSeparator,
  "ExperimentalFunctions"~~$PathnameSeparator,
  "FourierTrig"~~$PathnameSeparator,
  "StatisticsCommon"~~$PathnameSeparator,
  "SIUnits"~~$PathnameSeparator
  (* Remove documentation within Pubs cvs tree which are being provided by developers within their applications  *)
(*    "Packages"~~$PathnameSeparator~~"DatabaseLink",
    "Packages"~~$PathnameSeparator~~"GUIKit",
    "Packages"~~$PathnameSeparator~~"JLink",
    "Packages"~~$PathnameSeparator~~"NETLink",
    "Packages"~~$PathnameSeparator~~"WebServices",*)
    }, 
  ExcludeFiles -> Flatten[{StartOfString|$PathnameSeparator~~".",   "VirtualBookContents.nb" }],
  HoursAgo->Infinity,
  FileList->All
};

DocumentationBuild`$FunctionPaclet = False;
DocumentationBuild`$ffmpeg = "";

UseDocumentationBuildXMLTransforms::usage = "Add DocumentationBuild's XMLTransforms directory to $XMLTransformPath if it isn't already present"

(***************************************
  Begin Package
***************************************)

Begin["`Private`"]

(* Print full messages *)
$MessagePrePrint = .;

(* Set base directory *)
$DocumentationBuildDirectory = FileNameDrop[DirectoryName[ System`Private`$InputFileName ], -1];
$DocumentationBuildTransformDirectory = ToFileName[{$DocumentationBuildDirectory, "XMLTransforms"}];

(* FindFile["DocumentationAssets`"] could return any of:
 - DocumentationAssets/DocumentationAssets.mx
 - DocumentationAssets/DocumentationAssets.m
 - DocumentationAssets/Kernel/init.m
 - $Failed
For initial implementation, assume FindFile finds DocumentationAssets/Kernel/init.m or fails.  *)

(* Set Debug to False *)
$DocumentationBuildDebug = False;

(* log messages on server *)
ServerLog[mes_]:= If[ $ServerBuildLogQ === True, Print[mes]];
(*  *)
separator = StringJoin @@ Table["*", {50}];


(* Add DocumentationBuild\XMLTransforms if it isn't already on path *)
If[ !MemberQ[$DocumentationBuildTransformDirectory, Transmogrify`$XMLTransformPath], 
	PrependTo[Transmogrify`$XMLTransformPath, $DocumentationBuildTransformDirectory ];
	(* force transforms to be a part of path TODO: this shouldn't be necessary *)
	If[ !MemberQ[#, Transmogrify`$XMLTransforms], 
		PrependTo[Transmogrify`$XMLTransforms, #]
  ] & /@ Select[
    FileNames["*.m", $DocumentationBuildTransformDirectory, Infinity],
    Transmogrify`XMLTransformQ[#]&
  ]
];

(************************  DocsWarning[]  ***********************)

DocsWarning[formatStr_?StringQ, args___] := Message[
  DocumentationBuild`DocumentationBuild::warning,
  ToString[StringForm[formatStr, args]]
]

(************************  DocsError[]  *************************)

$DocsErrorTag

(* Generate a message and an exception. *)
DocsError[formatStr_?StringQ, args___] := (
	Message[
		DocumentationBuild`DocumentationBuild::error,
		(* Note: Use '@@' to avoid behavior described in bug #240412. *)
		ToString[StringForm @@ {formatStr, args}]
	];

	Throw[
		Failure["DocumentationError", <|
			"MessageTemplate" -> formatStr,
			"MessageParameters" -> {args}
		|>],
		$DocsErrorTag
	]
)

DocsError[args___] := Throw[
  Failure["DocumentationError", <|
    "MessageTemplate" -> ToString[StringForm[
      "Unknown documentation build error occurred: ``",
      StringJoin[Map[ToString, {args}]]
    ]]
  |>],
  $DocsErrorTag
]

(************************  DocsAssert[]  *************************)

Attributes[DocsAssert] = {HoldFirst}

DocsAssert[
	cond_,
	formatStr : _?StringQ,
	args___
] := If[!TrueQ[cond],
	Message[
		DocumentationBuild`DocumentationBuild::assertfail,
		(* Note: Use '@@' to avoid behavior described in bug #240412. *)
		ToString[StringForm @@ {formatStr, args}]
	];

	Throw[
		Failure["DocumentationError", <|
			"MessageTemplate" -> "DocsAssert[..] failed: " <> formatStr,
			"MessageParameters" -> {args}
		|>],
		$DocsErrorTag
	]
]

DocsAssert[cond_] :=
	DocsAssert[
		cond,
		"DocsAssert[..] of expression failed: ``",
		(* HoldForm so that the error shows the unevaluated asserted expression. *)
		HoldForm @ InputForm @ cond
	]

DocsAssert[args___] := Throw[
	Failure["DocumentationError", <|
		"MessageTemplate" -> ToString[StringForm[
			"Malformed DocsAssert[..] call: ``",
			StringJoin[Map[ToString, {args}]]
		]]
	|>],
	$DocsErrorTag
]

(************************  DocsRemoved[]  *************************)

(*
	Generate an error indicating that a function has been removed.

	Calling DocsRemoved is a safer alternative to removing a function/symbol completely.

	Due to the flexible nature of Wolfram Language, if we simply removed the function
	definition entirely, there is no guarantee that an obvious error would result in
	external code that had been relying on the removed function.

	Situations where this error is actually generated should be very rare, as functions
	should only be removed if it's already believed they are dead code. Calling
	DocsRemoved is a hedge against the possiblity that there may be uses of
	DocumentationBuild functions that are not known or easily findable through a code
	search.

	A function that has been replaced with DocsRemoved is a candidate for complete removal
	at some point in the future where a high level of confidence has been established that
	the function is, in fact, completely and utterly unused.
*)
DocsRemoved[name_?StringQ] :=
	DocsError["``: this function has been removed from DocumentationBuild."]

DocsRemoved[args___] :=
	DocsError["DocsRemoved: wrong arguments: ``", {args}]


(************************  DocsAsset[]  *************************)

DocsAsset["HTMLTransform",        name_?StringQ] := lookupDocsAsset["Standard", {"XMLTransforms", "HTML"}, name]
DocsAsset["HTMLTransformLayout",  name_?StringQ] := lookupDocsAsset["Standard", {"XMLTransforms", "HTML", "Layouts"}, name]
DocsAsset["TextCharacterReplacementRules"      ] := lookupDocsAsset["Standard", {"Assets", "Data"}, "TextCharacterReplacementRules.m"]
DocsAsset["LocalizationData"                   ] := lookupDocsAsset["Standard", {"Assets", "Data"}, "Localization.m"]
DocsAsset["CloudPrerequisiteData.m"            ] := lookupDocsAsset["Standard", {"Assets", "Data"}, "CloudPrerequisiteData.m"]
DocsAsset["Data:WebCharacterReplacements"      ] := lookupDocsAsset["Standard", {"Assets", "Data"}, "WebCharacterReplacements.m"]
DocsAsset["Data:SystemGuidePageWebVisitsRanks" ] := lookupDocsAsset["Standard", {"Assets", "Data"}, "SystemGuidePageWebVisitsRanks.m"]
DocsAsset["Data:SystemSourceInfo",name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Data", "SystemSourceInfo"}, name]
(* NOTE: The colon ':' character is not significant. This is just an identifier. *)
DocsAsset["Graphics:WorkflowTab", name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Graphics", "WorkflowTab"}, name]
DocsAsset["Graphics:Opener",      name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Graphics", "Opener"}, name]
DocsAsset["Graphics:Anchor",      name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Graphics", "Anchor"}, name]
DocsAsset["Graphics",             name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Graphics"}, name]
DocsAsset["Images",               name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Images"}, name]
DocsAsset["WorkflowTemplate",     name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "WorkflowTemplates"}, name]

DocsAsset["WebData",              name_?StringQ] := lookupDocsAsset["Standard", {"Assets", "Web"}, name]

DocsAsset["InternalData",         name_?StringQ] := lookupDocsAsset["Internal", {"Internal", "data"}, name]
DocsAsset["InternalOnlineImage",  name_?StringQ] := lookupDocsAsset["Internal", {"Internal", "online", "Images"}, name]
DocsAsset["Internal:Web",         name_?StringQ] := lookupDocsAsset["Internal", {"Internal", "web"}, name]

DocsAsset[args___] := DocsError[
	"Unknown documentation asset: ``",
	InputForm[{args}]
]

lookupDocsAsset[type_?StringQ, assetDir0:{__?StringQ}, name_?StringQ] := Module[{
	root,
	assetDir,
	path
},
	root = Replace[type, {
		"Standard" -> $DocumentationBuildDirectory,
		"Internal" :> Replace[
			DocumentationBuild`Internal`InternalAssetsDirectory[],
			None :> DocsError[
				"Unable to lookup internal documentation asset: ``. DocumentationAssets is unavailable.",
				name
			]
		],
		_ :> DocsError["Unknown documentation origin: ``", InputForm[type]]
	}];

	assetDir = FileNameJoin[Flatten[{root, assetDir0}]];

	If[!DirectoryQ[assetDir],
		DocsError["DocsAsset: no asset directory at path: ``", InputForm[assetDir]];
	];

	path = FileNameJoin[{assetDir, name}];

	If[!FileExistsQ[path],
		DocsError[
			"DocumentationBuild asset `` could not be found in asset dir ``.",
			InputForm[name],
			InputForm[assetDir]
		];
	];

	path
]

(************************  UseDocumentationBuildXMLTransforms[]  *************************)

UseDocumentationBuildXMLTransforms[] := (
  (* Add DocumentationBuild/XMLTransforms if it isn't already on path *)
  If[ !MemberQ[$DocumentationBuildTransformDirectory, Transmogrify`$XMLTransformPath],
    AppendTo[Transmogrify`$XMLTransformPath, $DocumentationBuildTransformDirectory];
  ];
)

(************************  FrontEnd Functions  *************************)


PubsStartFrontEnd[] := Module[{oldProt, x},
  If[
    Head[$PubsFELink] =!= LinkObject
    || (LinkReadyQ[$PubsFELink];First[LinkError[$PubsFELink]]) =!= 0
  ,
    (* Print["first run of PubsStartFrontEnd[]"]; *)
    $PubsFELink = Developer`InstallFrontEnd[];
    (* empty stylesheet cache to speed up build time *)
    EmptyStylesheetCache[];
    (* start a Kernel so that syntax coloring works *)
    (*Developer`UseFrontEnd @ FrontEndExecute[FrontEndToken["EvaluatorStart","Local"]];*)

    (* Global`AntLog["start a Kernel so that syntax coloring works"]; *)
    (* Global`AntLog["ExportString[Manipulate[x,{x,1,2}], \"GIF\"]..."]; *)
    (* ExportString[Manipulate[x,{x,1,2}], "GIF"]; *)
    (* Global`AntLog["...done."]; *)
    (* Global`AntLog["leaving PubsStartFrontEnd[]"]; *)

(*    oldProt = Unprotect[MathLink`CreateFrontEndLinks];
    Block[{MathLink`CreateFrontEndLinks, $ParentLink = Developer`InstallFrontEnd[]},
      $PubsFELink = $ParentLink;
      Get[ToFileName[{$InstallationDirectory,
        "SystemFiles", "FrontEnd", "TextResources"}, "GetFEKernelInit.tr"]]
    ];
    Protect[oldProt];
    Developer`UseFrontEnd[SetOptions[$FrontEndSession, Evaluator->"$ParentLink"]];*)
  ]
];

(* FE version of TimeConstrained *)
FETimeConstrained[func_, time_]:=
MathLink`CallFrontEnd @ FrontEnd`TimeConstrained[func, time];

PubsEvaluateWithFE[expr___] := (PubsStartFrontEnd[]; Developer`UseFrontEnd[expr]);
SetAttributes[PubsEvaluateWithFE, HoldAll];


(* Here we shrink notebooks by converting graphics to bitmaps (but only if the resulting
   bitmap is smaller than the original Graphics expression). Also, as an optimization
   for the time it takes for this conversion process, only operate on notebooks that
   are at least a certain size. This process modifies the notebooks in place.
*)

TagBitmapConditionalizedGraphics[nb_, ids_] := 
 Module[{newids},
	  (*NotebookFind[nb, "Output", All, CellStyle, AutoScroll -> False]; *)
	  newids = "CellID" /. Developer`CellInformation[nb];
	  SelectionMove[nb, Before, Notebook];
	  MapThread[(NotebookFind[nb, "Output", Next, CellStyle, AutoScroll -> False]; If[#1=!=#2, FrontEndExecute[FrontEnd`SelectionAddCellTags[nb, "RasterizedOutput"]]])&, {ids, newids}]
 ]

Options[ConvertGraphicsToBitmaps] = {};

ConvertGraphicsToBitmaps[nb_NotebookObject, entityType_, uri_] :=
	Which[
		MatchQ[entityType, "Workflow"],
			Module[{noinputinputs},
				(* "AnnotatedInput" cells that have no actual copyable input *)
				noinputinputs = Select[
					Cells[nb, CellStyle -> "AnnotatedInput"],
					FreeQ[NotebookRead[#], (Cell|StyleBox)[__, "Input", ___]]&
				];

				PubsEvaluateWithFE[
					NotebookFind[nb, "Output", All, CellStyle, AutoScroll -> False];
					oldids = "CellID" /. Developer`CellInformation[nb];
					FrontEndTokenExecute[nb, "SelectionConvert", "BitmapConditional"];
					TagBitmapConditionalizedGraphics[nb, oldids];
				];

				PubsEvaluateWithFE[
					NotebookFind[nb, #, All, CellStyle, AutoScroll -> False];
					FrontEndTokenExecute[nb, "SelectionConvert", "Bitmap"]
				] & /@ {"AnnotatedOutput", "AnnotatedExpression"};

				PubsEvaluateWithFE[
					(
						SelectionMove[#, All, Cell, AutoScroll -> False];
						FrontEndTokenExecute[nb, "SelectionConvert", "Bitmap"]
					) & /@ noinputinputs
				]
			],

		!MatchQ[entityType, "Guide" | "Workflow Guide"],
			If[
				MemberQ[uri][{
					(*"ref/StationaryWaveletPacketTransform", "ref/PairedHistogram",
					"ref/StationaryWaveletTransform", "ref/ContourPlot","ref/BarChart"*)
				}],
				PubsEvaluateWithFE[
					NotebookFind[nb, "Output", All, CellStyle, AutoScroll -> False];
					FrontEndTokenExecute[nb, "SelectionConvert", "Bitmap"]
				]
				,
				(* First, bitmap all ForceRaster cells *)
				(* Note:
					Check first if there are any "ForceRaster" cells to avoid a FE beep
					about "You used a cell conversion command when there were no suitable
					cells selected".
				*)
				If[Cells[nb, CellStyle -> "ForceRaster"] =!= {},
					PubsEvaluateWithFE[
						NotebookFind[nb, "ForceRaster", All, CellStyle, AutoScroll -> False];
						FrontEndTokenExecute[nb, "SelectionConvert", "Bitmap"]
					];
				];

				(* Now BitmapConditional the rest *)
				PubsEvaluateWithFE[
					NotebookFind[nb, "Output", All, CellStyle, AutoScroll -> False];
					oldids = "CellID" /. Developer`CellInformation[nb];
					FrontEndTokenExecute[nb, "SelectionConvert", "BitmapConditional"];
					TagBitmapConditionalizedGraphics[nb, oldids];
				]
			],

		True,
			Null
	];

(*========================================*)
(* Documentation stylesheet optimizations *)
(*========================================*)

(* These functions prevent the FE from constantly reloading stylesheets by creating a
   persistent notebook to keep the stylesheet in memory.  Also, creates a temporary
   inherited stylesheet which kills the docked cell.  Speeds things up. *)
CachedStylesheetPath[file_] := ToFileName[
  {$UserBaseDirectory, "SystemFiles", "FrontEnd", "StyleSheets"},
  PubsEvaluateWithFE[ToFileName[file]]];

CacheStylesheet[nbexpr_] := Module[{
	stylesheet = Lookup[Options[nbexpr], StyleDefinitions],
	stylesheetNbObj
},
	If[MissingQ[stylesheet],
		Return[Null, Module];
	];

	If[!AssociationQ[$CachedStylesheets],
		$CachedStylesheets = <||>;
	];

	If[Head[stylesheet] === Notebook || Head[stylesheet] === Symbol,
		(* This isn't a named stylesheet that we can cache, so skip it. *)
		Return[Null, Module];
	];

	If[KeyMemberQ[$CachedStylesheets, stylesheet],
		(* We've already cached this stylesheet, so skip caching it again. *)
		Return[Null, Module];
	];

	(*----------------------------------------*)
	(* Cache the stylesheet used by `nbexpr`. *)
	(*----------------------------------------*)

	(* Print["About to cache style sheet: ", stylesheet]; *)

	Quiet[CreateDirectory[DirectoryName[CachedStylesheetPath[stylesheet]]]];

	(* Temporarily override `stylesheet` (expecting it to be Reference.nb) to not
	   display a DockedCells toolbar. This slightly speeds up rendering of an opened
	   documentation notebook. *)
	Put[
		Notebook[{
			Cell[StyleData[StyleDefinitions -> stylesheet]],
			Cell[StyleData["Notebook"], DockedCells -> {}]
		}],
		CachedStylesheetPath[stylesheet]
	];

	(* Force a notebook with this stylesheet to remain open in the FrontEnd for the
	   remaining duration of the current documentation build. If we don't do this, the
	   FrontEnd will open and close the Reference.nb stylesheet each time we open and
	   close a documentation notebook. *)
	stylesheetNbObj = PubsEvaluateWithFE @ NotebookPut[
		Notebook[
			{},
			(* Note:
				If for some reason EmptyStylesheetCache[] is not called automatically at
				the end of the documentation build, it will be visible in Notebooks[].
				We use a descriptive name here so that it's easier for a user to determine
				where this notebook came from. *)
			WindowTitle -> "DocumentationBuild stylesheet cache (" <> ToString[stylesheet] <> ")",
			StyleDefinitions -> stylesheet,
			Visible -> False
		]
	];

	AssociateTo[$CachedStylesheets, stylesheet -> stylesheetNbObj];
];

EmptyStylesheetCache[] := (
	If[AssociationQ[$CachedStylesheets],
		Scan[
			stylesheet |-> Module[{path},
				path = CachedStylesheetPath[stylesheet];
				If[FileExistsQ[path],
					DeleteFile[path];
				];
			],
			Keys[$CachedStylesheets]
		];

		Scan[
			NotebookClose[#] &,
			Values[$CachedStylesheets]
		];
	];

	$CachedStylesheets = <||>;
);


(************************  Miscellaneous Utilities  *************************)

exportToXML[con_]:=
Switch[Head@con,
	XMLElement, ExportString[ con, "XML"],
	String, con,
	_, Message[exportToXML::val, Head@con]; con
];
exportToXML::val = "Expected String or XMLElement, found `1`.";



cleanFileNameString[str_String] := 
Module[{res=str},
  res = StringReplace[trimSpaces@res, RegularExpression["[^a-zA-Z0-9$_\\-\\./\\\\]"] -> ""];
  If[Head@res =!= String, Message[cleanFileNameString::reg, str]; str, res]
];
cleanFileNameString[str___] := (Message[cleanFileNameString::arg, str]; $Failed)
cleanFileNameString::reg = "StringReplace error in `1`";
cleanFileNameString::arg = "String expected, found `1`";


trimSpaces[s_String] :=
  StringReplace[s, { StartOfString ~~ Whitespace :> "", Whitespace ~~ EndOfString :> "" } ];


(* in need a stringifyer... *)
convertToString[c_String] := c;
convertToString[c_Cell] := convertToString@c[[1]];
convertToString[c_List] := StringJoin[convertToString /@ c];
convertToString[c_ButtonBox] := convertToString@c[[1]];
convertToString[c_BoxData] := convertToString@c[[1]];
convertToString[c_StyleBox] := convertToString@c[[1]];
convertToString[c_RowBox] := StringJoin[convertToString /@ c[[1]]];
convertToString[c_TextData] := convertToString@c[[1]];
convertToString[c_FormBox] := convertToString@c[[1]];
convertToString[c_AdjustmentBox] := convertToString@c[[1]];
convertToString[c_SuperscriptBox] := StringJoin[convertToString@c[[1]] <> "^"<>convertToString@c[[2]]];
convertToString[c_SubscriptBox] := StringJoin[convertToString@c[[1]] <> "_"<>convertToString@c[[2]]];
convertToString[c_] := ToString@c;
(* warn if a sequence *)
convertToString[c__] := (Message[convertToString::sequence]; StringJoin[convertToString /@ {c}]);
convertToString::sequence = "Warning: Sequence used as an argument, should be List";



(*
	ExcludePathsFromList[list, excludedPaths] 
	removes paths within list which match patterns defined in excludedPaths
*)
ExcludePathsFromList[paths_List, dirs_List] :=
 (DeleteCases[paths, 
     f_ /; ! StringFreeQ[f, Alternatives @@ #, 
        IgnoreCase -> False]] &[
   If[StringQ[#] && 
       StringFreeQ[#, $PathnameSeparator ~~ 
         EndOfString], # <> $PathnameSeparator, #] & /@ dirs])



(*
	ExcludeFilesFromList[list, excludedPaths] 
	removes files within list which match patterns defined in excludedPaths
*)
ExcludeFilesFromList[paths_List, files_List] := 
 DeleteCases[paths, f_ /;
      MemberQ[Alternatives@@files, StringReplace[f, {DirectoryName[f] -> ""}]]]



notebookFileNames[in_String, opts___?OptionQ]:=
Module[{inputDir=in, files={}, logfile, excluded, excludef, filelist, hoursAgo, total},

  {excluded, excludef, filelist, hoursAgo} = 
    {ExcludeDirectories, ExcludeFiles, FileList, HoursAgo} /. {opts} /. Options[notebookFileNames];

  (* quick ExcludeDirectories/ExcludeFiles error checking *)
  If[ !MatchQ[excluded,{(_String|_RegularExpression|_StringExpression)...}],
    Message[ExportOnlineNotebooks::exclude,excluded];
    excluded={}
  ];
  If[ !MatchQ[excludef,{(_String|_RegularExpression|_StringExpression)...}],
    Message[ExportOnlineNotebooks::exclude,excludef];
    excludef={}
  ];

  (* get rid of potentially dangerous pattern *)
  {excluded, excludef} = DeleteCases[#,"",1]&/@{excluded,excludef};
  
  (* add a trailing slash for consistency *)
  If[StringFreeQ[inputDir,$PathnameSeparator~~EndOfString],
    inputDir=inputDir<>$PathnameSeparator
  ];

  (* get /all/ the filenames in inputDir, including unwanted stuff
     this used to only get *.nb files, but it didn't work for directories
     without nbs in them.  this is much more liberal, and as such, is probably
     more prone to mistakes.  use ExcludeFiles (rather than ExcludeDirectories)
     to not create empty directories *)
  If[filelist===All,
    files = FileNames["*.*", inputDir, Infinity]; ,
    files = Union@Flatten@ModifiedFiles[inputDir, "*.*", Infinity, HoursAgo->hoursAgo]  ];
  
  (* now filter out unwanted files based on directories *)
  If[ Length@excluded > 0,
    files = DocumentationBuild`Common`ExcludePathsFromList[files, excluded];
    WriteLog[logfile,
      "Filtered "<>ToString[total-(total=Length[files])]<>" files with Directory filters",
      LineSeparator->None]
  ];

  (* remove non-notebooks *)
  files = Cases[files, f_ /; !StringFreeQ[f,".nb"~~EndOfString, IgnoreCase->True], Infinity];

  (* and any generic files we don't want to process *)
  If[ Length@excludef > 0,
    files = DocumentationBuild`Common`ExcludeFilesFromList[files, excludef];
    WriteLog[logfile,
      "Filtered "<>ToString[total-(total=Length[files])]<>" files with file filters",
      LineSeparator->None]
  ];

 files
];




pathToForwardSlash[s_String]:= StringReplace[s, "\\" -> "/"];
pathToForwardSlash[{s_String}]:= pathToForwardSlash[s];
pathToForwardSlash[s_]:= s;



FilepathToURI[filepath_String]:= FilepathToURI[filepath, "Mathematica"]
FilepathToURI[filepath_String, pacletname_String]:=
Module[{u, paclet=pacletname, lst, path=filepath},
  (* make sure that paclet is named *)
  paclet =
    If[StringLength[paclet] < 1,
      "Mathematica",
      StringReplace[pacletname, RegularExpression["(.+)\\s+Package"] -> "$1"]
    ];
  (* convert all slashes to foreward *)
  path = DocumentationBuild`Common`Private`pathToForwardSlash @ path;
  (* resolve any relative paths *)
  path = StringReplace[path, RegularExpression["/[^/]+/\\.\\./"] -> "/"];
  (* split path on 'paclet' and return second half *)
  (* Hack for webMathematica, which doesn't follow the standard layout *)
  If[StringCount[path, "/"<>paclet<>"/"] > 0,
    lst = StringSplit[path, "/"<>paclet<>"/"],
    lst = StringSplit[path, "/"<>paclet<>"Documentation/"]
  ];
  (* handle files which are named "<paclet>.nb" *)
  u = If[MemberQ[lst, ".nb"], StringJoin[lst[[-2]], paclet, lst[[-1]]], Last@lst];
  u = StringReplace[u, RegularExpression[".*Documentation/(English|Japanese|ChineseSimplified|Spanish)(.+)"] -> "$2"]; 
  (* make sure that incoming path starts with a separator *)
  u = If[StringMatchQ[u, "/*"], u, "/"<>u];
  u = DocumentationBuild`Common`FilePathToDocURI[ u, "/"];
  u = StringReplace[u, RegularExpression["/*(.+)"] -> "$1"];

  (* Add paclet name if appropriate *)
  u = If[paclet =!= "Mathematica", StringJoin[paclet, "/", u], u]
  ];

(*  *)
(* TODO: insanely stupid *)
FilePathToDocURI[p_String, l_String:"\\"]:=
  Module[{res, path = p, ll = If[l === "\\", "\\\\", l]},

dprint["Common.m: entering FilePathToDocURI, initial 'path' = "<>ToString[path, InputForm]];

  (* lots of replacement rules... *)
  res = FixedPoint[StringReplace[#, {

  (* FIXME: ugly, ugly mess *)
  l<>"ReferencePages"<>l<>"AppleScript"<>l               -> l<>"ref"<>l<>"applescript"<>l,
  l<>"ReferencePages"<>l<>"Blockchains"<>l               -> l<>"ref"<>l<>"blockchain"<>l,
  l<>"ReferencePages"<>l<>"C"<>l                         -> l<>"ref"<>l<>"c"<>l,
  l<>"ReferencePages"<>l<>"Callbacks"<>l                 -> l<>"ref"<>l<>"callback"<>l,
  l<>"ReferencePages"<>l<>"Characters"<>l                -> l<>"ref"<>l<>"character"<>l,
  l<>"ReferencePages"<>l<>"Classifiers"<>l               -> l<>"ref"<>l<>"classifier"<>l,
  l<>"ReferencePages"<>l<>"CompiledTypes"<>l             -> l<>"ref"<>l<>"compiledtype"<>l,
  l<>"ReferencePages"<>l<>"DatabaseConnections"<>l       -> l<>"ref"<>l<>"databaseconnection"<>l,
  l<>"ReferencePages"<>l<>"DataStructures"<>l            -> l<>"ref"<>l<>"datastructure"<>l,
  l<>"ReferencePages"<>l<>"Devices"<>l                   -> l<>"ref"<>l<>"device"<>l,
  l<>"ReferencePages"<>l<>"EmbeddingFormats"<>l          -> l<>"ref"<>l<>"embeddingformat"<>l,
  l<>"ReferencePages"<>l<>"Entities"<>l                  -> l<>"ref"<>l<>"entity"<>l,
  l<>"ReferencePages"<>l<>"ExcelFunctions"<>l            -> l<>"ref"<>l<>"excelfunction"<>l,
  l<>"ReferencePages"<>l<>"ExternalEvaluationSystems"<>l -> l<>"ref"<>l<>"externalevaluationsystem"<>l,
  l<>"ReferencePages"<>l<>"Files"<>l                     -> l<>"ref"<>l<>"file"<>l,
  l<>"ReferencePages"<>l<>"Formats"<>l                   -> l<>"ref"<>l<>"format"<>l,
  l<>"ReferencePages"<>l<>"FrontEndObjects"<>l           -> l<>"ref"<>l<>"frontendobject"<>l,
  l<>"ReferencePages"<>l<>"Indicators"<>l                -> l<>"ref"<>l<>"indicator"<>l,
  l<>"ReferencePages"<>l<>"Interpreters"<>l              -> l<>"ref"<>l<>"interpreter"<>l,
  l<>"ReferencePages"<>l<>"MenuItems"<>l                 -> l<>"ref"<>l<>"menuitem"<>l,
  l<>"ReferencePages"<>l<>"Messages"<>l                  -> l<>"ref"<>l<>"message"<>l,
  l<>"ReferencePages"<>l<>"Methods"<>l                   -> l<>"ref"<>l<>"method"<>l,
  l<>"ReferencePages"<>l<>"NetDecoders"<>l               -> l<>"ref"<>l<>"netdecoder"<>l,
  l<>"ReferencePages"<>l<>"NetEncoders"<>l               -> l<>"ref"<>l<>"netencoder"<>l,
  l<>"ReferencePages"<>l<>"Predictors"<>l                -> l<>"ref"<>l<>"predictor"<>l,
  l<>"ReferencePages"<>l<>"Programs"<>l                  -> l<>"ref"<>l<>"program"<>l,
  l<>"ReferencePages"<>l<>"ResourceObjectTypes"<>l       -> l<>"ref"<>l<>"resourceobject"<>l,
  l<>"ReferencePages"<>l<>"Services"<>l                  -> l<>"ref"<>l<>"service"<>l,
  l<>"ReferencePages"<>l<>"Symbols"<>l                   -> l<>"ref"<>l,
  l<>"ReferencePages"<>l<>"TextContents"<>l              -> l<>"ref"<>l<>"textcontent"<>l,
  l<>"ReferencePages"<>l<>"Widgets"<>l                   -> l<>"ref"<>l<>"widget"<>l,

  (* lowercase dirs after References (TODO: sigular v. plural?) *)
 	RegularExpression["[\\\\|/]ReferencePages[\\\\|/](\\w+)[\\\\|/]"] :> ToLowerCase[l<>"ref"<>l<>"$1"<>l],

	l<>"Notes"<>l -> l<>"note"<>l,
 	l<>"Guides"<>l -> l<>"guide"<>l,
 	l<>"ExamplePages"<>l -> l<>"example"<>l,
 	l<>"Tutorials"<>l -> l<>"tutorial"<>l,
 	l<>"HowTo"<>l -> l<>"howto"<>l,
 	l<>"HowTos"<>l -> l<>"howto"<>l,
 	l<>"Screencasts"<>l -> l<>"screencast"<>l,
 	l<>"Workflows"<>l -> l<>"workflow"<>l,
 	l<>"WorkflowGuides"<>l -> l<>"workflowguide"<>l,
 	
  l<>"System"<>l -> l,

  l<>"Packages"<>l -> l,
  RegularExpression[ll<>"*Documentation"<>ll<>"English"<>ll ] -> l,
  RegularExpression[ll<>"*Documentation"<>ll<>"Japanese"<>ll ] -> l,
  RegularExpression[ll<>"*Documentation"<>ll<>"ChineseSimplified"<>ll ] -> l,
  RegularExpression[ll<>"*Documentation"<>ll<>"Spanish"<>ll ] -> l,

	(* Remove nb extension *)
  RegularExpression["(.+)\\.nb$"] -> "$1"

}]&, path];

dprint["Common.m:FilePathToDocURI: final 'res' = "<>ToString[res, InputForm]];

res
];


(**********************************************************)

ToPacletURI[uri_?StringQ] :=
	If[StringStartsQ[uri, "paclet:"], uri, "paclet:" <> uri]

ToPacletURI[other_] :=
	DocsError["ToPacletURI: expected a String, got ``", InputForm[other]]

(**********************************************************)

If[$OperatingSystem === "MacOSX" || $OperatingSystem === "Linux",
	DocumentationBuild`$ffmpeg = Module[{ffmpeg},
		ffmpeg = StringReplace[
			RunProcess[
				$SystemShell,
				"StandardOutput",
				"/usr/bin/which ffmpeg",
				ProcessEnvironment -> Association[
				(* if ffmpeg is installed but this isn't finding it, append its location to this $PATH setting *)
				"PATH" -> "$PATH:/usr/local/bin:/opt/local/bin"
				]
			],
			"\n" -> ""
		];

		If[Not[StringQ[ffmpeg] && Positive[StringLength[ffmpeg]] && FileExistsQ[ffmpeg]],
			Global`AntLog["WARNING: ffmpeg is not installed but is required for conversion of Manipulate output to mp4 files in MarketingPages."];
			""
		,
			Global`AntLog["ffmpeg is required for MarketingPage builds and is installed at "<>ffmpeg];
			Global`AntLog["M- version: "<>ToString[SystemInformation["Small"]]];
			ffmpeg
		]
	];
	,
	DocumentationBuild`$ffmpeg = None;
]


End[] (* End Private Context *)

EndPackage[]
