(* ::Package:: *)

(* :Title: Info.m *)

(* :Authors:
	Andrew Hunt, Shaun McCance, Todd Gayley
	andy@wolfram.com, shaunm@wolfram.com, tgayley@wolfram.com
*)

(* :Package Version: 0.50 *)

(* :Mathematica Version: 6.0 *)

BeginPackage["DocumentationBuild`Info`", {
    "DocumentationBuild`Common`",
    "DocumentationBuild`Utils`",
    "Transmogrify`"
}];


(******************************************************************************)
(** Usage Statements **)

HasSourceInformation::usage = "";

GetSourceInformation::usage = "GetSourceInformation[uri, lang, keys___] looks up uri\
 in the source information for language lang.  If keys are given, it looks up information\
 for those keys.";

LookupSourceInformation::usage = "LookupSourceInformation[uri, lang, key, default] looks\
 up uri in the source information for language lang.  If a key is given, it looks up\
 information for that key.  If no value is found, Missing[..] or default (if provided) \
 will be returned."

GetNotebookInformation::usage = "";
GetNotebookTitle::usage = "";
GetNotebookURI::usage = "";
GetNotebookHistoryData::usage = "";
GetNotebookCategorization::usage = "";
GetNotebookDescription::usage = "";
GetNotebookMetaDescription::usage = "";
GetNotebookKeywords::usage = "";
GetNotebookSynonyms::usage = "";
GetNotebookFlag::usage = "";
GetNotebookSummary::usage = "";
GetNotebookKeywords::usage = "";
GetNotebookSynonyms::usage = "";
GetFunctionDevelopers::usage = "";
GetUnderdevelopment::usage = "";
GetButtonTextAndURL::usage = "";
GetTutorialCollectionLinks::usage = "";
GetWebResourcesLinks::usage = "";

ExportSourceInformation::usage = "ExportSourceInformation[dir,langs] exports a list of information about source notebooks to a standard location.";
ExportOnlineInformation::usage = "ExportOnlineInformation[dir,langs] exports a list of information about online notebooks to a standard location.";
ExportAllOnlineInformation::usage = "";
CreateSourceInformation::usage = "CreateSourceInformation[dir,lang] creates a list of information about source notebooks to a standard location.";
CreateOnlineInformation::usage = "CreateOnlineInformation[dir,lang] creates a list of information about online notebooks to a standard location.";
CreateCommonInformation::usage = "";
MakePrimaryGuideList::usage = "";

(* Load foreign function list *)
MakeForeignFunctionKeyword::usage = "ForeignFunctionKeywords[] creates a map from the functions in darta/keywords/ForeignFunctions.";
GetForeignFunctionKeywords::usage = "";
ForeignFunctionKeywords::usage = "";

CreateGuidesLists::usage = "CreateGuidesList[dir,langs] creates a map from the functions in dir/lang/System/ReferencePages to the guides in dir/lang/System/Guides they appear in, where lang is each element in langs";
GetDemonstrationSymbolList::usage = "GetDemonstrationSymbolList[] returns a list of current symbols categories used in demonstrations.w.c.";
GetDemonstrationKeywordList::usage = "";


$CoreInformation::usage = "";
SetCoreInformation::usage = "";
LoadCoreInformation::usage = "";
GetCoreInformation::usage = "";

SetSystemEnglishSourceInformationIfNeeded::usage = "";

GetSourceInformationLinkBases

Begin["`Private`"];

(******************************************************************************)

$CoreInformation = <||>

SetCoreInformation[
	language_?StringQ,
	info0 : _?ListQ | _?AssociationQ | _?StringQ | Automatic,
	(* Which link bases have information provided by `info0`. *)
	linkBases : {___?StringQ}
] := Module[{info},
	info = Replace[info0, {
		_?AssociationQ :> info0,
		_?ListQ :> (
			DocsAssert[MatchQ[info0, {(_?StringQ -> {(_?StringQ -> _) ...}) ...}]];

			If[info0 === {},
				DocsError["Core information provided as argument was empty: ``", info];
			];

			Association[info0]
		),
		file_?StringQ :> Module[{data},
			data = Get[file];

			If[FailureQ[data] || !ListQ[data],
				DocsError[
					"File '``' does not exist or contains invalid source information data: ``",
					InputForm[file], InputForm[data]
				];
			];

			DocsAssert[MatchQ[data, {(_?StringQ -> {(_?StringQ -> _) ...}) ...}]];

			If[data === {},
				DocsError["Core information data loaded from file was empty: ``", file];
			];

			Association[data]
		],
		Automatic :> (
			DocsAssert[linkBases === {"System"}];
			LoadCoreInformation[Automatic, language]
		),
		other_ :> DocsError["unexpected core information argument: ``", other]
	}];

	DocsAssert[AssociationQ[$CoreInformation]];

	DocsAssert[MatchQ[info, <| (_?StringQ -> { (_?StringQ -> _)... })... |>]];

	AssociateTo[
		$CoreInformation,
		language -> <|
			(* Which link bases (paclets, essentially) we have docs core information about. *)
			"LinkBases" -> linkBases,
			"Information" -> info
		|>
	];
]

SetCoreInformation[args___] :=
	DocsError["Bad arguments to SetCoreInformation: ``", InputForm[{args}]]

(******************************************************************************)

LoadCoreInformation[Automatic, language_String] := Module[{
	sourceInfoFile,
	data
},
	sourceInfoFile = StringJoin[
		"SourceInformation.",
		Replace[language, {
			"Japanese" -> "ja",
			"ChineseSimplified" -> "zh",
			"Spanish" -> "es",
			_ -> "en"
		}],
		".m"
	];

	data = Get[DocsAsset["Data:SystemSourceInfo", sourceInfoFile]];

	(* The source information files are stored as a Compress[..]'d string, because they
	   are quite large. Validate them and uncompress. *)
	DocsAssert[
		StringQ[data] && StringStartsQ[data, "1:"],
		"expected source info asset `` to contain a compressed string, got: ``",
		sourceInfoFile,
		Short[data]
	];

	data = Uncompress[data];

	If[!ListQ[data] || data === {},
		DocsError[
			"System source information data from '``' is malformed: ``",
			InputForm[assetName], InputForm[data]
		];
	];

	DocsAssert[MatchQ[data, {(_?StringQ -> {(_?StringQ -> _) ...}) ...}]];

	Association[data]
];

LoadCoreInformation[args___] :=
	DocsError["LoadCoreInformation: unexpected arguments: ``", Short[{args}]]

(******************************************************************************)

(*
	If this is a non-English build of a built-in system documentation link base,
	then also load the English System core information data.
	
	This is needed because DocumentationBuild has special behavior for System
	documentation, which is to check the document flag on the English version of
	a piece of documentation even when building a non-English translation of that
	page.
	
	(Which is a convenience, because it means only the English version of the
	page needs to be flagged to e.g. futurize or excise that page in every
	language. See ShippingFunctionQ.)
*)
SetSystemEnglishSourceInformationIfNeeded[
	linkBase : (_?StringQ | Automatic),
	language_?StringQ
] := Module[{},
	If[language =!= "English" (* && DocumentationBuild`Internal`SystemPacletQ[linkBase] *),
		SetCoreInformation["English", Automatic, {"System"}];
	];
]


SetSystemEnglishSourceInformationIfNeeded[args___] :=
	DocsError[
		"SetSystemEnglishSourceInformationIfNeeded: unexpected arguments: ``",
		Short[{args}]
	]

(******************************************************************************)
(** HasSourceInformation **)
HasSourceInformation[u_, lang_] :=
  Module[{uri = ToPacletURI[u]},
    Lookup[GetCoreInformation[lang], uri, None] =!= None
  ];

(******************************************************************************)

(* Returns the list of link bases that we have complete source information for. *)
GetSourceInformationLinkBases[lang_?StringQ] := Module[{linkBases},
	linkBases = NestedLookup[$CoreInformation, {lang, "LinkBases"}];

	If[MissingQ[info],
		DocsError["Source information link bases for language `` is not available.", lang];
	];

	DocsAssert[
		MatchQ[linkBases, {___?StringQ}],
		"unexpected source information link bases structure: ``",
		linkBases
	];

	linkBases
]

(******************************************************************************)

Attributes[GetCoreInformation] = {HoldAllComplete}

GetCoreInformation[
	lang0 : _?StringQ | Automatic
] := Module[{lang, info},
	DocsAssert[
		AssociationQ[$CoreInformation],
		"expected $CoreInformation to be AssociationQ, got: ``",
		Short[$CoreInformation]
	];
	
	(*
		If no language was explicitly specified, determine the language
		automatically by checking what language is set. This is an alternative
		to painstakingly passing the "Language" option value down as an explicit
		argument to every function in DocumentationBuild that needs it.
	*)
	lang = Replace[lang0, Automatic :> Module[{setting},
		setting = DocumentationBuild`Settings`$DocumentationBuildLanguage;

		If[!StringQ[setting],
			DocsError[
				"Unable to determine documentation build language. Invalid $DocumentationBuildLanguage setting. Expected string, got: ``",
				setting
			];
		];

		setting
	]];

	info = NestedLookup[$CoreInformation, {lang, "Information"}];

	If[MissingQ[info],
		DocsError["Core information for language `` is not available.", lang];
	];

	DocsAssert[
		AssociationQ[info],
		"Expected core information Association value, got: ``",
		Short[info]
	];

	info
]

GetCoreInformation[args___] :=
	DocsError["Bad arguments to GetCoreInformation: ``", InputForm[{args}]]

(******************************************************************************)

LookupSourceInformation[
	uri_?StringQ,
	lang : _?StringQ | Automatic,
	field : _?StringQ : All,
	default_ : Automatic
] := Module[{
	info,
	keys
},
	DocsAssert[
		StringStartsQ[uri, "paclet:"],
		"LookupSourceInformation: expected uri to start with \"paclet:\". Got: ``",
		InputForm[uri]
	];

	info = GetCoreInformation[lang];

	DocsAssert[AssociationQ[info]];

	keys = Replace[field, {
		All -> {uri},
		_ -> {uri, field}
	}];

	(* Note: Use With to work around the fact that NestedLookup is HoldAllComplete. *)
	With[{
		info0 = info,
		keys0 = keys
	},
		NestedLookup[info0, keys0, default]
	]
]

LookupSourceInformation[args___] :=
	DocsError["Bad arguments to LookupSourceInformation: ``", InputForm[{args}]]

(******************************************************************************)
(** GetSourceInformation **)
GetSourceInformation::nouri = "GetSourceInformation could not find `1`.";
GetSourceInformation::nokey = "GetSourceInformation could not find the key `1` for `2`.";
GetSourceInformation[u_, lang_, keys___] := Module[{
	uri = u,
	info
},
	If[!StringQ[uri],
		DocsError[
			"Unable to look up source information for non-string button data: ``",
			InputForm[uri]
		];
	];

	uri = ToPacletURI[uri];

    info = Lookup[GetCoreInformation[lang], uri, None];

    If[info === None,
(*
      Message[GetSourceInformation::nouri, uri];
*)
      Return[None]];
    If[Length[{keys}] === 0,
      info,
      Fold[
        Function[{l, k},
          If[MatchQ[l, {___, _[k, _], ___}],
            k /. l,
            (* Message[GetSourceInformation::nokey, {keys}, uri]; *)
            Return[None]
        ]],
        info, {keys}]]
  ];


(******************************************************************************)
(** GetNotebookInformation **)
GetNotebookInformation[file_String /; FileType[file] === File, opts___?OptionQ] := 
  GetNotebookInformation[ Get@file, opts]
GetNotebookInformation[origexpr_Notebook, opts___?OptionQ] := Module[{uri, type, title, contextval, contexttemp, pacletnameval, expr},
  
  (* sanitize nb expr from all tentative/future/excluded material to prevent leakage *)
  expr = DeleteCases[origexpr, Cell[__, "ExcludedMaterial"|"FutureExample"|"TentativeExample", ___], Infinity];
  
  (* get EntityType *)
  type = GetNotebookCategorization[expr, "Entity Type"];
  (* was EntityType found? If not, return $Failed *)
  If[type === $Failed, Return[$Failed] ];
  (* continue processing nb *)
  uri = GetNotebookURI[expr, opts];
  title = GetNotebookTitle[expr, type];

  If[type === "Guide" && MatchQ[uri, (* XX Root guide rename change XX *)"guide/WolframRoot"|"guide/Mathematica"], 
    type = "Root Guide";
    title = "Wolfram Mathematica";
  ];
  
    pacletnameval=GetNotebookCategorization[expr, "Paclet Name"];
  (*This code handles where the context is blank but is part of Mathematica*)
  contexttemp=GetNotebookCategorization[expr, "Context"];
  contextval=If[FreeQ[pacletnameval, "Mathematica" | ""], contexttemp, If[FreeQ[contexttemp, ""], contexttemp, "System`"]];
  
  {
    "Title" -> title,
    "TitleModifier" -> GetNotebookCategorization[expr, "Title Modifier"],
    "HistoryData" -> GetNotebookHistoryData[expr],
    "EntityType" -> type,
    "PacletName" -> pacletnameval,
    "Context" -> contextval,
    "URI" -> uri,
    "Keywords" -> GetNotebookKeywords[expr],
    "Underdevelopment" -> GetUnderdevelopment[expr],
    "TutorialCollectionLinks" -> GetTutorialCollectionLinks[expr],
    "ForeignFunctionKeywords" -> {},
    "Synonyms" -> GetNotebookSynonyms[expr],
    "Summary" -> GetNotebookSummary[expr, type, title],
    "MetaDescription" -> GetNotebookMetaDescription[expr],
    "Flag" -> GetNotebookFlag[expr],
    "IndexQ" -> GetNotebookIndexQ[type]
  }];



GetNotebookTitle::notitle= "The title of the notebook could not be found. It should be in a cell of style, `1`."

(******************************************************************************)
(** GetNotebookTitle **)
GetNotebookTitle[expr_] :=
  GetNotebookTitle[expr, GetNotebookCategorization[expr, "Entity Type"]];
GetNotebookTitle[expr_, type_] :=
  Module[{cellStyle, title},
    cellStyle = Switch[type,
      ("Featureset Page"),
        ("FeaturesetTitle"),
      ("Character" | "Character Name"),
        ("CharacterName"),
      ("Solution"),
        ("SolutionsTitle"),
      ("HowTo"),
        ("Title"),
      ("Guide"),
        ("GuideTitle" | "GuideTOCTitle"),
      ("Guide Overview"),
        ("TOCDocumentTitle"),
      ("Root Guide"),
        ("GuideTitle"),
      ("Note"),
        ("DataSourceTitle"),
      ("MarketingPage" | "ComparisonTitle"),
        ("MarketingTitle"),
      ("Tutorial" | "Upgrade Information" | "Overview" | "Monograph" | "Monograph Overview" | "Tech Note"),
        ("Title" | "Section" | "TOCDocumentTitle" | "DataSourceTitle" | "UpgradeObjectName"),
      ("Featured Example" | "FeaturedExample"),
        ("FeaturedExampleTitle"),
      ("Example Page" | "Example"),
        ("FeaturedExampleTitle"),
      ("Workflow"),
        ("WorkflowTitle" | "Title" | "GuideTitle"),
      ("Workflow Guide"),
        ("GuideTitle" | "WorkflowGuideTitle"),
      _,
        ("ObjectName" | "ObjectNameSmall" | "ObjectNameAlt" | "UpgradeObjectName" | "DataSourceTitle")
    ];
    title = Cases[expr, Cell[c_, cellStyle, ___] :> c, Infinity];
    title = DeleteCases[title, StyleBox[_, "FullListing"], Infinity];
	If[ Length[ title] === 0,
		Message[GetNotebookTitle::notitle, cellStyle];
		title = {"<missing title>"};
	];
    title = Transmogrify`ConvertToString[First[title]];
    title = If[StringFreeQ[title, "\n"], title, First@StringSplit[title, "\n"]];
    title = StringReplace[title, {
        "\[OpenCurlyDoubleQuote]" | "\[CloseCurlyDoubleQuote]" | "\[OpenCurlyQuote]" | "\[CloseCurlyQuote]" -> "'",
        "\[FilledRightTriangle]" -> ">",
        "\[Hyphen]" -> "-",
        "\"" -> "",
        " \[MediumSpace]" -> "",
        "\[MediumSpace]" -> "",
        " - Wolfram Mathematica" -> ""
    }];
    title
  ];


(******************************************************************************)
(** GetNotebookURI **)
GetNotebookURI[nb_Notebook, opts___?OptionQ] := Module[{
	uri,
	pname,
	fname,
	type
},
	uri = GetNotebookCategorization[nb, "URI"];

	(* check to see if uri is fake paclet:dir/XXX *)
	(* Formatting note: 7e6f5496: Fix Info.m:GetNotebookURI [explanation in commit message] *)
	uri = If[StringQ[uri] && StringContainsQ[uri, "XXX"],
		$Failed,
		uri
	];

	DocsAssert[
		StringQ[uri] || FailureQ[uri],
		"Unexpected notebook \"Categorization\" \"URI\" value: ``",
		uri
	];

	If[FailureQ[uri],
		pname = GetNotebookCategorization[nb, "Paclet Name"];
		type = GetNotebookCategorization[nb, "Entity Type"];

		If[FailureQ[pname] || FailureQ[type],
			(* FIXME:
				Fix the source notebooks that trigger this warning, and turn this back
				into a DocsError. Also, adjust CreateSourceInformation so that an error
				occurring while processing one notebook doesn't cause the whole source
				information creation function to fail.
			*)
			DocsWarning["Unable to determine notebook URI."];

			(* Note/TODO:
				This is what GetNotebookURI used to return in the event of an error,
				before the refactoring to use DocsError[..]. This value is simply a place
				holder. *)
			Return["$Failed//FIXME", Module];
		];

		uri = If[FreeQ[{"", "Mathematica"}, pname],
			pname <> "/",
			""
		];

		uri = uri <> Lookup[
			DocumentationBuild`Settings`$DocumentationPageTypeURIBases,
			type,
			(
				(* TODO: Make this an error? *)
				DocsWarning["Unknown documentation page type URI base: ``", type];
				""
			)
		];

		fname = Lookup[
			{opts},
			"FilePath",
			(
				(* TODO: Make this an error? *)
				DocsWarning["automatic URI construction failed because \"FilePath\" was not provided"];
				"FIXME"
			)
		];
		fname = FileBaseName[fname];

		uri = uri <> "/" <> fname;
	];

	uri
];


(******************************************************************************)
(** GetNotebookHistoryData **)
GetNotebookHistoryData::notfound = "Could not find the history data for `1`";
GetNotebookHistoryData[nb_Notebook] := Map[
  Function[{type},
    type -> If[# === {},
      Message[GetNotebookHistoryData::notfound, type]; "",
      First[#]
    ]&[
      Cases[nb,
        Cell[c_String, "HistoryData", ___, CellTags -> type, ___]
          :> StringReplace[c, " " -> ""],
        Infinity
    ]]
  ],
  {"New", "Modified", "Obsolete", "Excised"}];

GetNotebookHistoryData[___] := $Failed


(******************************************************************************)
(** GetNotebookCategorization **)
GetNotebookCategorization::val = "`1` value not found in notebook.";

GetNotebookCategorization[nb_String, cat_String] :=
	GetNotebookCategorization[GetQuietly[nb], cat];

GetNotebookCategorization[expr_Notebook, category_String] := Module[{
	values,
	taggingRules
},
	values = Cases[
		expr,
		Cell[
			t_,
			"Categorization",
			___,
			CellLabel -> category,
			___
		] :> t,
		Infinity
	];

	If[!MatchQ[values, {___?StringQ}],
		(* FIXME:
			Fix this in the source notebooks (CoreInformation build) and replace this
			warning with a DocsError. *)
		DocsWarning[
			"Notebook \"Categorization\" cell values are expected to be strings. Got: ``",
			InputForm[values]
		];
		values = {}
	];

	(* Quiet the message you get if `expr` doesn't have a TaggingRules option. *)
	taggingRules = Quiet[Check[
		Options[expr, TaggingRules],
		{}
	]];

	(* Formatting note: fd9f0542: minor CGP fix for TaggingRule detection *)
	If[TrueQ[NestedLookup[taggingRules, {TaggingRules, "CustomGuidePage", "CustomGuideQ"}]],
		values = Switch[category,
			"Entity Type",
				{"Guide"},
			"Paclet Name",
				{"Custom"},
			"URI",
				{"CustomGuidePages/guide/NEEDURI"},
			_,
				{"Guide"}
		]
	];

	If[category === "Title Modifier",
		(* Remove placeholder values. *)
		values = Map[Replace["XXXX" -> Nothing], values];

		Return @ DocumentationBuild`Utils`makeTitleModiferString[values];
	];

	Replace[values, {
		{} :> (
			Message[GetNotebookCategorization::val, category];
			$Failed
		),
		{only_} :> only,
		{first_, __} :> (
			(* TODO: Make this into a DocsError? *)
			DocsWarning[
				"Notebook contains multiple values for `` Categorization: ``",
				InputForm[category],
				InputForm[values]
			];
			first
		)
	}]
];


(******************************************************************************)
(** GetNotebookKeywords **)
GetNotebookKeywords[expr_Notebook] :=
  Cases[expr, cell:Cell[c_String /; !StringMatchQ[c, RegularExpression["X*"]], ___, "Keywords", ___] /; FreeQ[cell, "TentativeExample" | "TentativeWrapper"] :> 
    StringReplace[c, "," -> ""], Infinity];


(******************************************************************************)
(** GetNotebookSynonyms **)
GetNotebookSynonyms[expr_Notebook] :=
  Cases[expr, cell:Cell[c_String /; !StringMatchQ[c, RegularExpression["X*"]], ___, "Synonyms", ___] /; FreeQ[cell, "TentativeExample" | "TentativeWrapper"]:> 
    StringReplace[c, "," -> ""], Infinity];


(******************************************************************************)
(** GetFunctionDevelopers **)
GetFunctionDevelopers[expr_Notebook] :=
  Cases[expr, Cell[d_String, "Details", CellLabel->"Developers", ___]:> d, Infinity];

(******************************************************************************)
(** GetUnderdevelopment**)
GetUnderdevelopment[expr_Notebook] :=
  If[(Length[Cases[expr, "U N D E R   D E V E L O P M E N T", Infinity]]>0) ||
  	  (ScreenStyleEnvironment /. Options[expr]) === "ExperimentalObject", True, False];
  
(******************************************************************************)
(** GetButtonTextAndURL **)

GetButtonTextAndURL[cell_Cell] := Module[{buttontext, buttonurl}, 
       buttontext = First[Cases[cell, ButtonBox[aa_, ___] :> aa, Infinity]]; 
       buttonurl = First[Cases[cell, URL[bb_, ___] :> bb, Infinity]]; 
                 Flatten[{buttontext, buttonurl}]];

(******************************************************************************)
(** GetTutorialCollectionLinks **)

GetTutorialCollectionLinks[expr_Notebook] :=
Module[{guidetutorialcollectioncells, tutorialcollectiontitleandlinks},
	    guidetutorialcollectioncells = 
	       Cases[expr, Cell[_, "GuideTutorialCollection"|"TutorialCollection", ___], Infinity];
        tutorialcollectiontitleandlinks=
           If[Length[guidetutorialcollectioncells]>0, 
    	      Map[GetButtonTextAndURL, guidetutorialcollectioncells], 
    	    	{}]
    ];
    
(******************************************************************************)
(** GetWebResourcesLinks **)

GetWebResourcesLinks[expr_Notebook] := 
 Module[{links}, 
  links = Flatten[(Cases[expr, 
        Cell[l_, ___, "WebResources", ___, 
          CellLabel -> #, ___] :> {l, #}, 
        Infinity] & /@ {"Training Courses", "Videos", 
       "Demonstrations", "Community"}), 1];
  DeleteCases[links, {"XXXX", ___} | {}]
  ];


(******************************************************************************)
(** GetNotebookFlag **)

(* This is identical as DocumentationBuild`Utils`DocumentFlag, and used far less... *)
GetNotebookFlag = DocumentationBuild`Utils`DocumentFlag;
(*GetNotebookFlag[expr_Notebook]:=
Module[{f},
  f = Cases[expr, Cell[_, "ObsoleteFlag"|"InternalFlag"|"TemporaryFlag"|"PreviewFlag"|"FutureFlag"|"ExcisedFlag"|"AwaitingFutureDesignReviewFlag", ___], Infinity];
  If[Length@f > 0,
  	If[Length[Cases[expr, "U N D E R   D E V E L O P M E N T", Infinity]]>0, 
  		Switch[Part[List@@First@f,2], 
  			"ObsoleteFlag", None,
  			"InternalFlag", Part[List@@First@f,2],
  			"TemporaryFlag", None,
  			"PreviewFlag", None,
  			"FutureFlag", None,
  			"ExcisedFlag", Part[List@@First@f,2],
  			"AwaitingFutureDesignReviewFlag", None,  
  			_ , Part[List@@First@f,2]], 
  			Part[List@@First@f,2]
  			],  
  	None]
];*)


(******************************************************************************)
(** GetNotebookSummary **)
GetNotebookSummary[expr_Notebook, type_String, name___String] :=
Module[{text, summary},
  (* test to see if nb already has summery in TaggingRules *)
  summary = Cases[expr, Rule[TaggingRules, {___, Rule["Metadata", {___, Rule["summary", s_], ___}], ___}] :> s, Infinity];
  If[summary =!= {}, Return[summary] ];
  (* compute summary *)
  text = ExtractSummaryText[DeleteCases[expr, Cell[__, "ExcludedMaterial"|"FutureExample"|"TentativeExample", ___], Infinity], type, name];
  (* Fix weird characters bug (bug 246625 and others) -- edit: turns out this was fixed in Make.m;
     apparently the ButtonBox->TemplateBox conversion was messing this up *)
  (*text = StringReplace[text, "\"\!\(" ~~ Shortest[___] ~~ "StyleBox[\"" ~~ Shortest[content__] ~~ 
  "\", \"TI\"]" ~~ Shortest[___] ~~ "\"" :> content];*)
  (* Check for content. *)
  If[text === "", Return[" "] ];
  text = FixedPoint[StringReplace[#, {
             RegularExpression["^\"(.*)\"$"] -> "$1", 
             "\[OpenCurlyDoubleQuote]" | "\[CloseCurlyDoubleQuote]" | "\[OpenCurlyQuote]" | "\[CloseCurlyQuote]"->"'",
             "\[AliasIndicator]" -> "ESC", 
             "\[Hyphen]" -> "-",
             "\[FilledRightTriangle]" -> ">",
             "\[RightGuillemet]" -> ">",
             "\n" | "\\[FilledSmallSquare]" | "\\[FilledVerySmallSquare]" | "\r" | "\[LineSeparator]" | "\[ParagraphSeparator]" -> " ",
             (* Linear Syntax strings *)
             RegularExpression["StyleBox\\[\\s*\\\\\"(\\w+)\\\\\",\\s*\\\\\"\\w+\\\\\"\\s*]"] -> "$1",
             (*RegularExpression["StyleBox\\[\\s*\\\\\"(.+)\\\\\",\\s*\\\\\"\\w+\\\\\"\\s*]"] -> "$1", *)
             "\\!\\(\\*" -> "", "\\(" -> "", "\\)" -> "", "\\!" -> "", "\\*" -> "",
             RegularExpression["^\\s+"] -> "", 
             RegularExpression["\\s+"] -> " ",
             RegularExpression["\\[(\")*\\s(\\w)"] -> "[$1$2",
             RegularExpression[",\\s{\"\\s(\\w)"] -> ", {\"$1"
             (*  Remove SubscriptBox  *)
             (* Shortest["\"SubscriptBox[" ~~ a___ ~~ ", " ~~ b___ ~~ "]" ~~ c___ ~~ "\""] :> "\"" <> a <> "_" <> b <> c <> "\"" *)
           }]&, text];
   text = FEConvertString[text];
  (* safety fallthrough *)
    If[Head@text === String, StringReplace[StringReplace[text, {Shortest[
   "\"SubscriptBox["|"\"Subscript[" ~~ aa___ ~~ ", " ~~ bb___ ~~ "]" ~~ cc___ ~~ 
    "\""] :> "\"" <> aa <> "_" <> bb <> cc <> "\"", 
 Shortest["SubsuperscriptBox[" ~~ aa___ ~~ ", " ~~ bb___ ~~ ", " ~~ 
    cc___ ~~ "]" ~~ dd___ ~~ ""] :> 
  "" <> aa <> "_" <> bb <> "^" <> cc <> "", 
 Shortest["DynamicBox[ToBoxes[If[$OperatingSystem === MacOSX, " ~~ \
___ ~~ ", " ~~ a___ ~~ "], StandardForm], ImageSizeCache -> " ~~ ___ ~~
     "}]" ] :> a, 
 Shortest["TagBox[" ~~ a___ ~~ 
    ", EventHandlerTag[{ \"MouseDown\" -> Null, PassEventsDown -> \
Automatic, PassEventsUp -> True}]]"] :> a, 
 "GraphicsBox[TagBox[RasterBox[CompressedData[" ~~ ___ :> "", 
 "StyleBox[context`]" :> "context`", 
 "[\"StyleBox[\\\"context`\\\", \\\"TI\\\"]\"]" :> "[\"context\"]", 
 "StyleBox[x, TI]" :> "x", "StyleBox[package`]" :> "package`", 
 "StyleBox[\[Theta], TR]" :> "\[Theta]", 
 "\" StyleBox[\\\"context`\\\", \\\"TI\\\"]\"" :> "\"context`\"", 
 "StyleBox[\"E\"_FontWeight->\"Bold\"^FontSlant->\"Plain\", \
\"\\[Nu]\", \"\\[Mu]\"](z)" :> "E_\[Nu]^\[Mu](z)", 
 "GridBox[{ { RowBox[{ RowBox[{a_, 1}], ,, ..., ,, a_p}]}, { RowBox[{ \
RowBox[{b_, 1}], ,, ..., ,, b_q}]} }]). " :> 
  "{{a_1, ..., a_p}, {b_1, ... , b_p}}", 
 "UnderscriptBox[lim, m \\[Rule] 0, LimitsPositioning->True]" :> 
  "lim_(m->0)", 
 "UnderoverscriptBox[\"\\[Product]\", RowBox[{\"i\", \"=\", \
\"i_min\"}], \"i_max\", LimitsPositioning->True]\\ f" :> 
  "\[Product]_(i=i_min)^(i_max) f", 
 "UnderoverscriptBox[\"\\[Sum]\", RowBox[{\"i\", \"=\", \"i_min\"}], \
\"i_max\", LimitsPositioning->True]\\ f" :> 
  "\[Sum]_(i=i_min)^(i_max) f", 
 "UnderoverscriptBox[\"\\[Product]\", RowBox[{\"i\", \"=\", \"1\"}], \
\"i_max\", LimitsPositioning->True]expr" :> 
  "\[Product]_(i=1)^(i_max) expr", 
 "UnderoverscriptBox[\"\\[Product]\", RowBox[{\"i\", \"=\", \
\"i_min\"}], \"i_max\", LimitsPositioning->True] \
UnderoverscriptBox[\"\\[Product]\", RowBox[{\"j\", \"=\", \
\"j_min\"}], \"j_max\", LimitsPositioning->True]...expr" :> 
  "\[Product]_(i=i_min)^(i_max) \[Product]_(j=j_min)^(j_max) \
...expr", 
 "UnderoverscriptBox[\"\\[Sum]\", RowBox[{\"i\", \"=\", \"1\"}], \
\"i_max\", LimitsPositioning->True]expr" :> 
  "\[Sum]_(i=1)^(i_max) expr", 
 "UnderoverscriptBox[\"\\[Sum]\", RowBox[{\"i\", \"=\", \"i_min\"}], \
\"i_max\", LimitsPositioning->True] UnderoverscriptBox[\"\\[Sum]\", \
RowBox[{\"j\", \"=\", \"j_min\"}], \"j_max\", \
LimitsPositioning->True]...expr" :> 
  "\[Sum]_(i=i_min)^(i_max) \[Sum]_(j=j_min)^(j_max) ...expr", 
 "UnderoverscriptBox[\"\\[Product]\", RowBox[{\"i\", \"=\", \"1\"}], \
\"i_max\", LimitsPositioning->True]\\ f" :> 
  "\[Product]_(i=1)^(i_max) f", 
 "UnderoverscriptBox[\"\\[Product]\", RowBox[{\"i\", \"=\", \
\"i_min\"}], \"i_max\", LimitsPositioning->True] \
UnderoverscriptBox[\"\\[Product]\", RowBox[{\"j\", \"=\", \
\"j_min\"}], \"j_max\", LimitsPositioning->True]...\\ f" :> 
  "\[Product]_(i=i_min)^(i_max) \[Product]_(j=j_min)^(j_max) f", 
 "UnderscriptBox[\\[Product], i]f" :> "\[Product]_i f", 
  "\\[LeftBracketingBar]" :> "|", 
  "\\[RightBracketingBar]" :> "|",
 "TagBox[RowBox[{\"(\", GridBox[{ { TagBox[\"n\", Identity, \
Editable->True, Selectable->True]}, { TagBox[\"m\", Identity, \
Editable->True, Selectable->True]} }], \")\"}], \
InterpretTemplate[Cell$$986`QBinomial[#, #2, #3]& ], Editable->False, \
Selectable->False] q" :> "(n; m)_q", 
 "( GraphicsBox[RasterBox[CompressedData[ ], {Offset[{0.5, -21.5}, \
{0.5, 0.5}], Offset[{21.5, -0.5}, {0.5, 0.5}]}], ImageSize -> {19., \
Automatic}] \
1:eJyt0tENgCAMBNA7ncEBXMkRWID9/4wJjRTa0kSP8Pe40MBZ6lV2AEfbG/4L \
mXbM2MchYVsfVa9xRvqge+nZuZOuHe5JBlbNQ4rtl1jM8o3Vb0nXDTJwSoau \
kwvXz75K1rlvZMmkM//St9y9xwKJ , GraphicsBox[RasterBox[CompressedData[ \
], {Offset[{0.5, -21.5}, {0.5, 0.5}], Offset[{21.5, -0.5}, {0.5, \
0.5}]}], ImageSize -> {20., Automatic}] \
1:eJxTTMoPSmJmYGAQhWImBvoDRjAgRh0DGBJSy8gINRJZLVY9OFRiUYtdJSNO \
tWjuhOrEphbN73BbUCB2x6MBHOowVOJUh6YSjzoUlXjVIakkoA7Z74QAsepw \
xhE2lUSqw5GWKAEAwE0Cig== , and GraphicsBox[RasterBox[CompressedData[ \
], {Offset[{0.5, -21.5}, {0.5, 0.5}], Offset[{21.5, -0.5}, {0.5, \
0.5}]}], ImageSize -> {20., Automatic}] \
1:eJxTTMoPSmJmYGAQhWImBnoCRiAgViUDZWpx6MWilpFotSAecWoZGfGqRVcJ \
VosMcWlDAbjMR1eJUx2aSjzqUFTiVYekkoA6ZL8TAsSqwxtH6CqJVIczLZEP AHfnAm0= \
). Double-click the background and press the r key to make the \
Reshape tool ( GraphicsBox[RasterBox[CompressedData[ ], {Offset[{0.5, \
-21.5}, {0.5, 0.5}], Offset[{21.5, -0.5}, {0.5, 0.5}]}], ImageSize -> \
{15., Automatic}] \
1:eJxTTMoPSmJmYGAQhWImhmEKGElQSaxaRkZi1TIyEquWkRGmFhniUYkAOM0n \
Vh2aSrzuJVYdkkqC/idWHYnhSWwcER/vRKojGgAAJPYCTA== )" :> "()"}], 
{"StyleBox[\"E\"_FontWeight->\"Bold\"^FontSlant->\"Plain\", \"\\[Nu]\", \
\"\\[Mu]\"](z).":>"E_\\[Nu]^\\[Mu](z).", "Unicode: 2211. Alias: Esc\[ThinSpace]sum\[ThinSpace]Esc. Compound \
operator with built-in evaluation rules. UnderoverscriptBox[\[Sum], \
i, SubscriptBox[i, max], LimitsPositioning -> False]f is by default \
interpreted as Sum[f,{i,i_max}]. UnderoverscriptBox[\[Sum], \
SubscriptBox[StyleBox[RowBox[{i, =, i}], TI], min], SubscriptBox[i, \
max], LimitsPositioning -> False]f is by default interpreted as \
Sum[f,{i,i_min,i_max}]. Not the same as the Greek letter \
\\[CapitalSigma]. " :> "Unicode: 2211. Alias: Esc\[ThinSpace]sum\
\[ThinSpace]Esc. Compound operator with built-in evaluation rules. \
\[Sum]_(i)^i_max(f) is by default interpreted as Sum[f,{i,i_max}]. \
\[Sum]_(i=i_min)^i_max(f)f is by default interpreted as \
Sum[f,{i,i_min,i_max}]. Not the same as the Greek letter \
\\[CapitalSigma]. ", "Unicode: 220F. Alias: Esc\[ThinSpace]prod\[ThinSpace]Esc. Compound \
operator with built-in evaluation rules. \
UnderoverscriptBox[\[Product], i, SubscriptBox[i, max], \
LimitsPositioning -> False]f is by default interpreted as \
Product[f,{i,i_max}]. UnderoverscriptBox[\[Product], \
SubscriptBox[StyleBox[RowBox[{i, =, i}], TI], min], SubscriptBox[i, \
max], LimitsPositioning -> False]f is by default interpreted as \
Product[f,{i,i_min,i_max}]. Not the same as the Greek letter \
\\[CapitalPi]. " :> "Unicode: 220F. Alias: Esc\[ThinSpace]prod\
\[ThinSpace]Esc. Compound operator with built-in evaluation rules. \
\[Product]_(i)^i_max(f) is by default interpreted as \
Product[f,{i,i_max}]. \[Product]_(i=i_min)^i_max(f) is by default \
interpreted as Product[f,{i,i_min,i_max}]. Not the same as the Greek \
letter \\[CapitalPi]. ", "Unicode: F361. Alias: Esc\[ThinSpace]pw\[ThinSpace]Esc. Prefix \
operator with built-in evaluation rules. \
\[Piecewise]GridBox[{{SubscriptBox[StyleBox[e, TI], StyleBox[1, TR]], \
SubscriptBox[StyleBox[c, TI], StyleBox[1, TR]]}, \
{SubscriptBox[StyleBox[e, TI], StyleBox[2, TR]], \
SubscriptBox[StyleBox[c, TI], StyleBox[2, TR]]}}] is by default \
interpreted as Piecewise[{{e_1,c_1},{e_2,c_2}}]. Extensible \
character. Not the same as keyboard {. " :> "Unicode: F361. Alias: \
Esc\[ThinSpace]pw\[ThinSpace]Esc. Prefix operator with built-in \
evaluation rules. \[Piecewise]Grid[{{e_1, c_1}, {e_2, c_2}}] is by \
default interpreted as Piecewise[{{e_1,c_1},{e_2,c_2}}]. Extensible \
character. Not the same as keyboard {. ", "This constructs the sum UnderoverscriptBox[\[Sum], RowBox[{i, =, \
1}], 7, LimitsPositioning -> True] ( x^i ) / ( i ) . You can leave \
out the lower limit if it is equal to 1. This makes i increase in \
steps of 2, so that only odd-numbered values are included. " :> "This \
constructs the sum \[Sum]_(i=1)^7 (x^i)/(i) . You can leave out the \
lower limit if it is equal to 1. This makes i increase in steps of 2, \
so that only odd-numbered values are included. ", "Basic descriptive statistics operations. Given a list with n \
elements x_i, the mean Mean[list] is defined to be \
\[Mu](x)=OverscriptBox[x, _]=\[Sum]x_i/n. The variance Variance[list] \
is defined to be var(x)=\[Sigma]^2(x)=\[Sum](x_i-\[Mu](x))^2/(n-1), \
for real data. (For complex data \
var(x)=\[Sigma]^2(x)=\[Sum](x_i-\[Mu](x))(OverscriptBox[RowBox[{\
SubscriptBox[x, i], -, RowBox[{\[Mu], (, x, )}]}], _])/(n-1).) " :> \
"Basic descriptive statistics operations. Given a list with n \
elements x_i, the mean Mean[list] is defined to be \
\[Mu](x)=x^=\[Sum]x_i/n. The variance Variance[list] is defined to be \
var(x)=\[Sigma]^2(x)=\[Sum](x_i-\[Mu](x))^2/(n-1), for real data. \
(For complex data \
var(x)=\[Sigma]^2(x)=\[Sum](x_i-\[Mu](x))(x_i-\[Mu](x))^/(n-1).) ", "SqrtBox[RowBox[{-, 1}]]" :> "Sqrt[-1]", "Mathematical constants. Euler's constant EulerGamma is given by the \
limit \[Gamma]=UnderscriptBox[lim, RowBox[{m, \[Rule], \[Infinity]}], \
LimitsPositioning -> True](UnderoverscriptBox[\[Sum], RowBox[{k, =, \
1}], m, LimitsPositioning -> True] ( 1 ) / ( k ) - logm). It appears \
in many integrals, and asymptotic formulas. It is sometimes known as \
the Euler-Mascheroni constant, and denoted C. Catalan's constant \
Catalan is given by the sum UnderoverscriptBox[\[Sum], RowBox[{k, =, \
0}], \[Infinity], LimitsPositioning -> True](-1)^k(2k+1)^-2. It often \
appears in asymptotic estimates of combinatorial functions. It is \
variously denoted by C, K, or G. " :> "Mathematical constants. \
Euler's constant EulerGamma is given by the limit \[Gamma]=lim_(m\
->\[Infinity])(\[Sum]_(k=1)^\[Infinity]((1)/(k)-log(m))). It appears \
in many integrals, and asymptotic formulas. It is sometimes known as \
the Euler-Mascheroni constant, and denoted C. Catalan's constant \
Catalan is given by the sum \
\[Sum]_(k=0)^\[Infinity]((-1)^k(2k+1)^-2). It often appears in \
asymptotic estimates of combinatorial functions. It is variously \
denoted by C, K, or G. ", "Numerical integration functions. This finds a numerical \
approximation to the integral \[Integral]_0^\[Infinity]e^-x^3\
\[DifferentialD]x. Here is the numerical value of the double integral \
\[Integral]_RowBox[{-^1}, 1]dx \[Integral]_RowBox[{-^1}, 1]dy \
(x^2+y^2). " :> "Numerical integration functions. This finds a \
numerical approximation to the integral \[Integral]_(0)^\[Infinity] \
e^-x^3 \[DifferentialD]x. Here is the numerical value of the double \
integral \[Integral]_(-1)^1 dx \[Integral]_(-1)^1 dy (x^2+y^2). ", "Numerical sums and products. This gives a numerical approximation to \
UnderoverscriptBox[\[Sum], RowBox[{i, =, 1}], \[Infinity], \
LimitsPositioning -> True] ( 1 ) / ( i^3+i! ) . There is no exact \
result for this sum, so the Wolfram Language leaves it in a symbolic form. " \
:> "Numerical sums and products. This gives a numerical approximation \
to \[Sum]_(i=1)^\[Infinity]((1)/(i^3+i!)). There is no exact result \
for this sum, so the Wolfram Language leaves it in a symbolic form. ", "Simplifying with assumptions. The Wolfram Language does not automatically \
simplify this, since it is only true for some values of x. \
SqrtBox[SuperscriptBox[x, 2]] is equal to x for x\[GreaterEqual]0, \
but not otherwise. " :> "Simplifying with assumptions. The Wolfram Language \
does not automatically simplify this, since it is only true for some \
values of x. Sqrt[x^2] is equal to x for x\[GreaterEqual]0, but not \
otherwise. ", "Numerical sums, products and integrals. Here is a numerical \
approximation to UnderoverscriptBox[\[Sum], RowBox[{i, =, 1}], \
\[Infinity], LimitsPositioning -> True] ( 1 ) / ( i^3 ) . NIntegrate \
can handle singularities in the integration region. " :> "Numerical \
sums, products and integrals. Here is a numerical approximation to \
\[Sum]_(i=1)^\[Infinity]((1)/(i^3)). NIntegrate can handle \
singularities in the integration region. ", "file StyleBox[\\\".\\\", \\\"TI\\\"] ext\", expr]" :> "file.ext\", expr]"}], ""]
];


(******************************************************************************)
(** FEConvertString - Internal **)
(* This code is from jfultz, specifically used to remove box strings from summaries *)
FEConvertString[str_String]:=
  PubsEvaluateWithFE[FrontEndExecute[FrontEnd`ExportPacket[BoxData@str,"PlainText"]][[1]]];


(******************************************************************************)
(** ExtractSummaryText - Internal **)

ExtractSummaryText[nb_Notebook, 
	"Symbol" | "File" | "Menu Item" | "System Program" | "LibraryLink C Function" | "MathLink C Function" | "UnityLink C# Function" | "Format" | "Front End Object" | 
	"Widget" | "Net Encoder" | "Net Decoder" | "Compiled Type" | "Database Connection" | "Data Structure", ___]:=
Module[{n, args, forms, maxArgs=2},
  n = Cases[nb, Cell[c_, "Usage"|"FormatUsage"|"CFunctionUsage"|"ObsolescenceNote", ___], Infinity];
  If[n =!= {}, 
    (
    n = First@Flatten@n;
    n = n /. TextData[{a___, b_String, c___}] :> TextData[{a, StringReplace[b, "\[LineSeparator]" -> " "], c}];
    n = n /. ButtonBox[c_, ___] :> c;
    n = n /. StyleBox[a_, b_] :> a;
    n = n /. TemplateBox[{a_, b_}, ___] :> a;
    If[!FreeQ[n, "ObsolescenceNote"], 
      n = DeleteCases[n, Cell[BoxData[GraphicsBox[___]],___], Infinity]];
    n = DeleteCases[n, 
      Cell[c___] /; (MemberQ[{c}, TooltipBox, Infinity, Heads->True] && MemberQ[{c}, "ModInfo", Infinity]), Infinity];
    n = DeleteCases[n, Cell[_String, "ModInfo"], Infinity];
    n = n /. SuperscriptBox[a_, b_] :> 
      If[StringQ[a] && StringQ[b], 
        StringJoin[ StringReplace[a, RegularExpression["\\s*+"]->""], "^", StringReplace[b, RegularExpression["^\\s+"]->""] ], 
        RowBox@{a,b}
      ];
    n = n /. SubscriptBox[a_, b_] :> 
      If[StringQ[a] && StringQ[b], 
        StringJoin[ StringReplace[a, RegularExpression["\\s+$"]->""], (*"_"*)"", StringReplace[b, RegularExpression["^\\s+"]->""] ], 
        RowBox@{a,b}
      ];
          n = n /. UnderoverscriptBox[a_, b_, c_, ___] :> 
      If[StringQ[a] && StringQ[b] && StringQ[c], 
        StringJoin[ StringReplace[a, RegularExpression["\\s+$"]->""], (*"_"*)"", StringReplace[b, RegularExpression["^\\s+"]->""],"^", StringReplace[c, RegularExpression["^\\s+"]->""]], 
        RowBox@{a,b,c}
      ];
    (* take the first element *)
   If[ Cases[n, _GridBox, 3] =!= {},
       (* extact and flatten GridBox data *)
      n = Cases[n, GridBox[d_, ___] :> d, 2];
      n = Cell[TextData[Flatten@{n}]];
      (* attempt to fix usages running together without a space between them *)
      n = ReplaceAll[n, ""->" "];
    ];
   forms = CreateInputForm @ n;
   If[forms == $Failed, Message[CreateInputForm::err, n]; Return[""]];
   
   forms = StringReplace[forms, {
     RegularExpression["\\^\\s+"]->"^", RegularExpression["\\_\\s+"]->"_",
     "\\[Ellipsis]" -> "...", "\\[Cross]" -> "*"
     }]
    ),
    forms = "" ];
  forms
];

ExtractSummaryText[nb_Notebook, "Solutions Pages", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "Categorization",CellLabel->"Description", ___]:> c, Infinity];
  str = Transmogrify`ConvertToString[abstract];
  str
];

ExtractSummaryText[nb_Notebook, "Tutorial" | "Overview" | "Upgrade Information" | "Notes" | "Note" | "Monograph" | "Monograph Overview" | "Tech Note", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "UpgradeSummary" | "DataSourceSummary" | "Text" | "MonographText" | "TechNoteText" | "Caption"|"MathCaption"|"TOCChapter"|"MonographTOCChapter"|"Reference", ___]:> c, Infinity];
  abstract = If[Length@abstract > 3, Take[abstract,3], abstract];
  abstract = Riffle[abstract, " "];
  str = Transmogrify`ConvertToString[abstract];
  str
];

ExtractSummaryText[nb_Notebook, "Guide" | "Alphabetical Listing", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "GuideAbstract"|"GuideTOCAbstract", ___]:> c, Infinity];
  If[(Length@abstract > 0) && FreeQ[abstract, "XXXX", Infinity], 
    abstract = First@abstract, 
    (
      abstract = Cases[nb, Cell[c_, "GuideText" | "GuideTOCLink" | "GuideFunctionsSubsection" | "InlineGuideFunctionListing", ___]:> c, Infinity];
      abstract = abstract /. StyleBox[c_,"InlineSeparator",___] :> "";
      abstract = abstract /." \[EmptyVerySmallSquare] " :> ", ";
      abstract = Riffle[abstract, ", "];
    )
  ];
  str = Transmogrify`ConvertToString[abstract];
  str
];

ExtractSummaryText[nb_Notebook, "Character Name", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "Notes", ___]:> c, Infinity];
  abstract = Riffle[abstract, " "];
  str = StringJoin @ Transmogrify`ConvertToString @ abstract;
  str
];

ExtractSummaryText[nb_Notebook, "HowTo", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "HowToAbstract", ___]:> c, Infinity];
  If[Quiet[!StringQ[abstract[[1]]]],
    abstract = Cases[abstract, TextData[{c___}, ___] :> c, Infinity];
    abstract = Riffle[abstract, " "];
  ];
  str = StringJoin @ Transmogrify`ConvertToString @ abstract;
  str
];

ExtractSummaryText[nb_Notebook, "Service", ___]:=
Module[{abstract},
  abstract = Cases[nb, Cell[c_, "ServiceAbstract", ___]:> c, Infinity];
  If[(Length@abstract > 0) && FreeQ[abstract, "XXXX", Infinity], 
    abstract = First@abstract, 
    (
      abstract = Cases[nb, Cell[c_, "GuideText" | "GuideTOCLink" | "GuideFunctionsSubsection" | "InlineGuideFunctionListing", ___]:> c, Infinity];
      abstract = abstract /. StyleBox[c_,"InlineSeparator",___] :> "";
      abstract = abstract /." \[EmptyVerySmallSquare] " :> ", ";
      abstract = Riffle[abstract, ", "];
    )
  ];
  str = Transmogrify`ConvertToString[abstract];
  str
];

ExtractSummaryText[nb_Notebook, "Workflow", ___]:=
Module[{abstract, return},
	abstract = Cases[nb, Cell[c_, ___, "WorkflowDescription", ___]:>c, Infinity];
	Which[!MatchQ[abstract, {}],
		abstract = ReplaceRepeated[abstract, TemplateBox[s:{_String..}, _String, ___]:>StringJoin[Riffle[s, "+"]]];
		return = Transmogrify`ConvertToString[abstract],
		abstract = Cases[nb, Cell[TextData[{arrow_, spacer_, text_String},___], ___, "WorkflowHeader", ___]:>text, Infinity];!MatchQ[abstract, {}],
		abstract = ReplaceRepeated[abstract, TemplateBox[s:{_String..}, _String, ___]:>StringJoin[Riffle[s, "+"]]];
		return = Transmogrify`ConvertToString[StringJoin[Transmogrify`ConvertToString/@Riffle[abstract, ", "]]],
		abstract = Cases[nb, Cell[c_, ___, "WorkflowText", ___]:>c, Infinity, 5];!MatchQ[abstract, {}],
		abstract = ReplaceRepeated[abstract, TemplateBox[s:{_String..}, _String, ___]:>StringJoin[Riffle[s, "+"]]];
		return = Transmogrify`ConvertToString[StringJoin[Transmogrify`ConvertToString/@abstract]],
		True,
		return = " "
	];
	return
];

ExtractSummaryText[nb_Notebook, "Workflow Guide", ___]:=
Module[{abstract, return},
	abstract = Cases[nb, Cell[c_, ___, "WorkflowGuideSection", ___]:>c, Infinity];
	If[!MatchQ[abstract, {}],
		abstract = ReplaceRepeated[abstract, {TemplateBox[s:{_String..}, _String, ___]:>StringJoin[Riffle[s, "+"]], TemplateBox[{Cell[TextData[s_String]], uri_String},_String,___]:>s}];
		return = StringReplace[Transmogrify`ConvertToString[StringJoin[Transmogrify`ConvertToString/@Riffle[abstract, " . "]]], ">"|"\[RightGuillemet]" -> ""],
		return = " "];
	return
];

ExtractSummaryText[nb_Notebook, "Interpreter" | "Entity" | "Text Content", ___]:=
Module[{abstract, return},
	abstract = Cases[nb, (Cell[c_String, ___, "InterpreterUsage", ___] | Cell[c_String, ___, "EntityAbstract", ___]):>c, Infinity];
	If[!MatchQ[abstract, {}],
		abstract = ReplaceRepeated[abstract, {TemplateBox[s:{_String..}, _String, ___]:>StringJoin[Riffle[s, "+"]], TemplateBox[{Cell[TextData[s_String]], uri_String},_String,___]:>s}];
		return = Transmogrify`ConvertToString[StringJoin[Transmogrify`ConvertToString/@Riffle[abstract, " "]]],
		return = " "];
	return
];

ExtractSummaryText[nb_Notebook, "Method", ___]:= 
Module[{abstract, str}, 
  abstract = Riffle[Cases[nb, Cell[c_, "ObjectName" | "ObjectNameSmall" | "ObjectNameAlt" | "Notes", ___]:> c, Infinity], " "];
  str = StringJoin @ Transmogrify`ConvertToString @ abstract;
  str
];

ExtractSummaryText[nb_Notebook, ___]:= 
Module[{abstract}, 
  abstract = Cases[nb, Cell[c_, "ObjectName" | "ObjectNameSmall" | "ObjectNameAlt" | "Usage" | "GuideAbstract" | "Title" | "Section" | "Text", ___]:> c, Infinity];
  str = StringJoin @ Transmogrify`ConvertToString @ abstract;
  str
];

(******************************************************************************)
(** GetNotebookDescription **)
GetNotebookDescription[nb_Notebook]:=Module[{dscrptn}, 
  dscrptn=Cases[nb,
    Cell[c_String /; !StringMatchQ[c, RegularExpression["X*"]], ___, "Description", ___]
      :> c,
    Infinity]; If[Length[dscrptn]===0, "", dscrptn[[1]]]];
GetNotebookDescription[s___] := (Message[GetNotebookDescription::args, s]; $Failed);
GetNotebookDescription::args = "Expected Notebook but found: `1`";

(******************************************************************************)
(** GetNotebookMetaDescription **)
GetNotebookMetaDescription[nb_Notebook]:=Module[{dscrptn}, 
  dscrptn=Cases[nb,
    Cell[c_, ___, "MetaDescription", ___]
      :> If[FreeQ[c, "XXXX"], Transmogrify`ConvertToString[c], {}],
    Infinity]; If[Length[dscrptn]===0 || !MatchQ[dscrptn, {_String, ___}], "", dscrptn[[1]]]];
GetNotebookMetaDescription[s___] := (Message[GetNotebookMetaDescription::args, s]; $Failed);
GetNotebookMetaDescription::args = "Expected Notebook but found: `1`";

(******************************************************************************)
(** GetNotebookKeywords **)
GetNotebookKeywords[nb_Notebook]:=
  Cases[nb,
    cell:Cell[c_String /; !StringMatchQ[c, RegularExpression["X*"]], ___, "Keywords", ___] /; FreeQ[cell, "TentativeExample" | "TentativeWrapper"]
      :> StringReplace[c, "," -> ""],
    Infinity];
GetNotebookKeywords[s___] := (Message[GetNotebookKeywords::args, s]; $Failed);
GetNotebookKeywords::args = "Expected Notebook but found: `1`";


(******************************************************************************)
(** GetNotebookSynonyms **)
GetNotebookSynonyms[nb_Notebook]:=
  Cases[nb,
    cell:Cell[c_String /; !StringMatchQ[c, RegularExpression["X*"]], ___, "Synonyms", ___] /; FreeQ[cell, "TentativeExample" | "TentativeWrapper"]
      :> StringReplace[c, "," -> ""],
    Infinity];
GetNotebookSynonyms[s___]:= (Message[GetNotebookSynonyms::args, s]; $Failed);
GetNotebookSynonyms::args = "Expected Notebook but found: `1`";


(******************************************************************************)
(** GetDemonstrationKeywordList **)
GetDemonstrationKeywordList::failed = "Failed to import demonstration keyword list";
GetDemonstrationKeywordList[___] := (Message[GetDemonstrationKeywordList::failed]; $Failed);
GetDemonstrationKeywordList[] :=
Module[{lst, file},
  file = $TemporaryPrefix <> "demo-keyword-data.m";
  Export[file, Import["http://demonstrations.internal.wolfram.com/reports/keywords.jsp", "Byte"], "Byte"];
  lst = Import[file, "Package"];
  DeleteFile[file];
  If[lst =!= $Failed && Head[lst] =!= Import,
  	lst = lst /. Rule[a_, b_] :> {a, b};
  	DocumentationBuild`Info`$DemonstrationKeywordTitles = 
  		Union @ Flatten[{
  			{"http://demonstrations.wolfram.com/" -> {"Title" -> "Wolfram Demonstrations Project"}},
  			{Rule[#[[3]],{ Rule["Title", #[[2]] ]} ]} &/@ lst}, 1];
  	(* switch 2nd and 3rd columns for uniformity *)
  	lst = Union[{#[[1]], #[[3]]} & /@ lst];
    Return@lst,
    (Message[GetDemonstrationKeywordList::failed]; $Failed)
  ]
];

(******************************************************************************)
(** GetDemonstrationSymbolList **)
GetDemonstrationSymbolList::failed = "Failed to import demonstration list";
GetDemonstrationSymbolList[___] := (Message[GetDemonstrationSymbolList::failed]; $Failed);
GetDemonstrationSymbolList[] :=
Module[{lst},
  lst = Import["http://demonstrations.internal.wolfram.com/reports/symbols.jsp", "TEXT"];
  If[lst =!= $Failed && Head[lst] =!= Import,
    StringSplit[lst, "\n"],
    (Message[GetDemonstrationSymbolList::failed]; $Failed)
  ]
];

demonbs = {
  "CellularAutomaton",
  "ContourPlot3D",
  "Derivative",
  "DigitCount",
  "D",
  "Expand",
  "FactorInteger",
  "Factor",
  "FindFit",
  "FindRoot",
  "Fit",
  "Fourier",
  "FromCharacterCode",
  "IntegerPartitions",
  "NDSolve",
  "PadeApproximant",
  "ParametricPlot3D",
  "ParametricPlot",
  "Plot",
  "Plot3D",
  "PolarPlot",
  "PolyhedronData",
  "RiemannSiegelZ",
  "Series",
  "Solve",
  "TreePlot"
};




(******************************************************************************)
(** GetForeignFunctionKeywords **)

ForeignFunctionKeywordsComputed = False;
GetForeignFunctionKeywords[] = {};
GetForeignFunctionKeywords[uri_] :=
  Module[{in, files, data},
    If[ForeignFunctionKeywordsComputed, Return[{}]];
    ForeignFunctionKeywordsComputed = True;
	in = DocsAsset["InternalData", FileNameJoin[{"keywords", "ForeignFunctions"}]];
    files = FileNames["*.m", in, 2];
    data = {};
    Scan[
      Function[{file},
        data = Join[data,
          ReadList[StringToStream[Import[file, "Text", CharacterEncoding -> "UTF-8"]]]]
      ],
      files];
    data = DeleteCases[data, {"", _}];
    data = data /. {a_, b_} :> {StringReplace[b, "paclet:" -> ""], a};

    Scan[
      Function[{datum}, Module[{keywords},
        keywords = GetForeignFunctionKeywords[First[datum]];
        If[!MatchQ[keywords, {___, Last[datum], ___}], 
          GetForeignFunctionKeywords[First[datum]] = Append[keywords, Last[datum]] ];
      ]],
      data];

    GetForeignFunctionKeywords[uri]
  ];
  
  
(******************************************************************************)
(** MakeForeignFunctionKeyword - Internal **)
MakeForeignFunctionKeyword[] :=
Module[{in, files, data},
	in = ToFileName[{$DocumentationToolsDirectory, "Internal", "data", 
    	"keywords", "ForeignFunctions"}];
	files = FileNames["*.m", in, 2];
	data = {};
	Scan[(data = Join[ data,
		ReadList[
			StringToStream[
				Import[#, "Text", CharacterEncoding -> "UTF-8"]]] ]) &, files];
	(*remove entries without keywords*)
	data = DeleteCases[data, {"", _}];
	data = data /. {a_, b_} :> Rule[StringReplace[b, "paclet:" -> ""], a];
 
	ForeignFunctionKeywords[] := {};
	ForeignFunctionKeywords[_] := {};

	Scan[If[! MatchQ[ForeignFunctionKeywords[First[#]], {___, Last[#], ___}], 
		ForeignFunctionKeywords[First[#]] = 
			Append[ForeignFunctionKeywords[First[#]], Last[#]]] &, data];
];


(******************************************************************************)
(** ExportSourceInformation / ExportOnlineInformation - Internal **)
ExportSourceInformation[dir_, langs_List, pkg_String:""] :=
  Map[ExportSourceInformation[dir, #, pkg]&, langs];

ExportSourceInformation[dir_, lang_String:"English", pkg_String:""] :=
  Module[{outfile},
		outfile = FileNameJoin[{
			$DocumentationToolsDirectory, "Internal", "data",
			StringJoin["SourceInformation.", LanguageAbbreviation[lang], ".m"]
		}];
    If[ListQ[dir],
      Export[outfile, dir],
      Export[outfile, CreateSourceInformation[dir, lang, pkg]] ];
  ];

ExportOnlineInformation[___] :=
	DocsRemoved["ExportOnlineInformation"]


(******************************************************************************)
(** ExportAllOnlineInformation - Internal **)

(****************** This code will be refined later. ********)

CreateOnlineInformationForPackages[___] :=
	DocsRemoved["CreateOnlineInformationForPackages"]

CreateOnlineInformationForMathematicaLinks[___] :=
	DocsRemoved["CreateOnlineInformationForMathematicaLinks"]

ExportAllOnlineInformation[___] :=
	DocsRemoved["ExportAllOnlineInformation"]


(******************************************************************************)
(** CreateSourceInformation / CreateOnlineInformation **)

CreateSourceInformation[dir_, lang_String:"English", pkg_String:""] :=
  CreateCommonInformation[dir, lang, pkg, "Source"];

CreateOnlineInformation[dir_, lang_String:"English", pkg_String:""] :=
  CreateCommonInformation[dir, lang, pkg, "Online"];

CreateCommonInformation[indir_, lang_String:"English", pkg_String:"", type_String] := Module[{
  dir, sourceQ, symbols, refs, guides, tutorials, howtos, workflows, workflowguides, examples,
  info, getCommon, putCommon,
  isSystemDocs,
  systemDir, demoList, demonbs, title, mkuri, primaryGuideList, uridata = {},
  GetGuideRelevance, GetGuideWeight, GetWeightedRelatedGuides, OneLineListingQ,
  GetSectionPos, webguidevisits, guidepageexpr, GetGuideRanking,
  examplepageURLroot = "http://reference.wolfram.com/language/example/"
},
	dir = StringReplace[
		indir,
		RegularExpression["^(.+)(/|\\\\)"<>lang<>"(/|\\\\)*$"] -> "$1"
	];
	sourceQ = (type === "Source");

	webguidevisits = Get[DocsAsset["Data:SystemGuidePageWebVisitsRanks"]];

	isSystemDocs = pkg === "";

	(*----------------------------------------------------*)
	(*                                                    *)
	(* Define various utility functions                   *)
	(*                                                    *)
	(*----------------------------------------------------*)

	OneLineListingQ[guideuri_, refuri_] := Module[{
		nbdata = guidepageexpr
	},
		TrueQ[MemberQ[
			Cases[
				nbdata,
				Cell[c_, ___, "GuideText" | "InlineGuideFunctionListing", ___]
					/; MemberQ[c, ButtonBox[Last[StringSplit[refuri, "/"]], ___], Infinity],
				Infinity
			],
			"GuideText",
			Infinity
		]]
	];

	GetSectionPos[guideuri_, refuri_] := Module[{
		nbdata = guidepageexpr, group, count, pos, refgroup
	},
		group = First @ Cases[
			nbdata,
			Cell[CellGroupData[c_, ___]]
				/; MemberQ[
					c,
					Cell[cont_, ___, "GuideText" | "InlineGuideFunctionListing", ___]
						/; MemberQ[cont, ButtonBox[Last[StringSplit[refuri, "/"]], ___], Infinity],
					Infinity
				],
			Infinity
		];
		refgroup = Cases[
			group,
			ButtonBox[_String, ___, BaseStyle -> bs_ /; MemberQ[{bs}, "Link", Infinity], ___],
			Infinity
		];
		count = Max[Length[refgroup], 1];
		pos = Position[refgroup, Last[StringSplit[refuri, "/"]]];
		If[pos =!= {},
			pos = pos[[1, 1]]
			,
			pos = 1
		];
		(* larger is stronger *)
		N[(count - pos + 1)/count]
	];

	GetGuideRelevance[uri_, guideuri_List] :=
		Module[{guideuris = guideuri},
			Map[
				(
					guidepageexpr = Get[FileNameJoin[{
						dir, lang, systemDir, "Guides", Last[StringSplit[#, "/"]] <> ".nb"
					}]];
					{
						#,
						{"OneLine" -> OneLineListingQ[#, uri], "PositionInSection" -> GetSectionPos[#, uri]}
					}
				) &,
				guideuris
			]
		];

	GetGuideRanking[guideuri_String] :=
		Module[{name = Last[StringSplit[guideuri, "/"]], pos},
			pos = Position[webguidevisits, name];
			If[pos =!= {},
				pos = pos[[1, 1]];
				N[(Length[webguidevisits] - pos)/Length[webguidevisits]]
				,
				0
			]
		];

	GetGuideWeight[{uri_String, lis_List}] :=
		{
			uri,
			.4 ("PositionInSection" /. lis)
				+ .25 GetGuideRanking[uri]
				+ If[TrueQ[("OneLine" /. lis)], .25, 0]
		};

	GetWeightedRelatedGuides[refuri_, guideuris_List] :=
		Module[{relev = GetGuideRelevance[refuri, Last/@guideuris], lis},
			(lis = GetGuideWeight /@ relev);
			First /@ Reverse[SortBy[lis, Last]]
		];

	(*----------------------------------------------------*)
	(*                                                    *)
	(* Begin processing                                   *)
	(*                                                    *)
	(*----------------------------------------------------*)

	If[sourceQ && isSystemDocs,
		demoList = GetDemonstrationSymbolList[];
		If[demoList === $Failed,
			demoList = demonbs
		]
		,
		demoList = demonbs
	];
	systemDir = If[isSystemDocs, "System", ""];

	(* Populate uridata for use in tutorial navigation only if in System directory for now *)
	(* If in System directory, get Example Page data for backlinking to guides *)
	If[StringMatchQ[systemDir, "System"],
		uridata = Select[
			Flatten[
				Quiet[
					DocumentationBuild`SystemResourcesFromDocs`CreateTutorialBrowserData[
						{ToFileName[{dir, lang, systemDir, "Tutorials"}]},
						lang
					]
				]
			],
			StringMatchQ[ToString[#], "paclet:*tutorial/*"]
			&& StringFreeQ[ToString[#], "#"~~DigitCharacter..] &
		]
	];

	mkuri[args__] := StringJoin[
		"paclet:",
		If[!isSystemDocs, pkg<>"/", ""],
		Riffle[{args}, "/"]
	];

	(* Find "PrimaryObject" links in guide pages. This determines the link's primary guide page *)
	primaryGuideList = MakePrimaryGuideList[ToFileName[{dir, lang, systemDir, "Guides"}], pkg];

	(* Common *)
	Clear[linktypes, links, info];
	linktypes[_] := {};
	links[_, _] := {};
	info[_, "Title"] := "";
	info[_, "Categorization"] := "";
	info[_, "HistoryData"] := {" ", " ", " ", " "};
	info[_, "Flag"] := "";
	info[_, "PrimaryGuide"] := "";
	info[_, "LinkTrails"] := {};
	info[_, "URI"] := "";
	info[_, "ForwardBackNavData"] := {};
	info[_, "MetaDescription"] := "";

	(*----------------------------------------------------*)
	(*                                                    *)
	(* Define putCommon[] and getCommon[] helpers          *)
	(*                                                    *)
	(*----------------------------------------------------*)

	putCommon[uri_] := Join[
		If[sourceQ,
			Map[
				(# -> info[uri, #]) &,
				{"Title", "Categorization", "HistoryData", "Flag", "PrimaryGuide", "LinkTrails", "URI", "ForwardBackNavData", "MetaDescription"}
			]
			,
			{}
		],
		{"Links" -> Map[
			(# ->
				If[MemberQ[{"Guides"}, #],
					(* These link types get sorted and uniqued *)
					Module[{
						originalguides = First /@ Split[Sort[
							links[uri, #],
							If[#1[[1]] === #2[[1]],
								OrderedQ[{#1[[2]], #2[[2]]}]
								,
								OrderedQ[{#1[[1]], #2[[1]]}]
							] &
						]],
						newuris
					},
						If[isSystemDocs,
							(* If we're building the System documentation, sort the
							   related guide links so that guide pages that have a
							   higher number of web page visits get ranked higher. *)
							newuris = GetWeightedRelatedGuides[uri, originalguides];
							SortBy[originalguides, Position[newuris, #[[2]]] &]
							,
							originalguides
						]
					]
					, (* else *)
					(* But the rest don't *)
					links[uri, #]
				]
			)&,
			Union[linktypes[uri]]
		]}
	];

	getCommon[uri_, nb_] := Module[{cat,pglist},
		If[sourceQ,
			(* Source notebook *)
			cat = Cases[nb, Cell[s_, "Categorization", ___] :> s, Infinity, 1];
			If[Length[cat] =!= 0,
				info[uri, "Categorization"] = First[cat]
			];

			info[uri, "Title"] = title = GetNotebookTitle[nb];
			info[uri, "HistoryData"] = {"New", "Modified", "Obsolete", "Excised"} /. GetNotebookHistoryData[nb] /. {"New"->"", "Modified"->"", "Obsolete"->"", "Excised"->""};
			info[uri, "PrimaryGuide"] = DocumentationBuild`Utils`GetPrimaryGuide[uri, lang, primaryGuideList];
			pglist = Join[primaryGuideList, {uri -> {"PrimaryGuide" -> info[uri, "PrimaryGuide"]}}];
			info[uri, "LinkTrails"] = DocumentationBuild`Info`Private`GetLinkTrails[uri, pglist ];
			info[uri, "URI"] = GetNotebookURI[nb];

			(*If[MatchQ[nb[[1, 1]], Cell[_, s_/;StringMatchQ[s, "*Flag"], ___]],
				info[uri, "Flag"] = nb[[1, 1, 2]] ];*)
			info[uri, "Flag"] = GetNotebookFlag[nb];
			info[uri, "MetaDescription"] = GetNotebookMetaDescription[nb];

		,
			(* Online notebooks *)
			Module[{bboxFilter, sections},
				bboxFilter[str_String | {URL[str_String], None}] := str;
				sections = Union@Cases[nb,
					Cell[__, sty_String /; StringMatchQ[sty, (___~~"Section") | "Subsection" | "Title"], ___] :> sty,
					Infinity];
				Scan[
					Function[{linktype},
						linktypes[uri] = Append[linktypes[uri], linktype];
						links[uri, linktype] = (Union[Cases[
							Cases[
								nb,
								Cell[CellGroupData[{Cell[__, linktype, __], ___}, ___], ___],
								Infinity
							],
							ButtonBox[bt_, ___, ButtonData->bd_, ___] :>
								If[StringQ[bd] || MatchQ[bd, {URL[_String], None}],
									{bt, bboxFilter@bd}
									,
									Unevaluated[Sequence[]]
								],
							Infinity
						]]
							//. {a___, {b_String, u_}, {c_String, u_}, d___}
								:> {a, {b<>c, u}, d}
						)
					],
					sections
				];
			];
		]; (* end If[sourceQ, ...] *)
	];

	(*---------------------------------------------*)
	(*                                             *)
	(* Use Reap[Scan[...]] to collect information. *)
	(*                                             *)
	(*---------------------------------------------*)

	(* Symbols *)
	info[_, "Demonstrations"] := False;
	info[_, "HasOptions"] := False;

	symbols = Reap[Scan[
		Function[{file}, CheckAbort[
			Module[{base, uri, nb, hasopts},
				nb = DocumentationBuild`Utils`GetQuietly[file];
				base = FileBaseName[file];
				(* Print["BaseName: "<>ToString[base]]; *)
				uri = Sow[mkuri["ref", base]];
				(* Print["uri: "<>ToString[uri]]; *)

				(* TODO:
					If the implementation of HasOptionsQ instead determined this property
					by checking for options table documentation in the source `nb`
					expression, then the system and non-system cases below could be
					combined. HasOptionsQ used to do this, but it's not clear that its
					implementation could be changed back to doing that without a
					regression. *)
				hasopts = Which[
					MemberQ[{"GeodesicArrow", "GeodesicLine", "GeoDisk"}, ToString[base]],
						True,

					(* If we're building the system documentation, evaluate
					   Options[<symbol>] to determine if this symbol has any options. *)
					isSystemDocs,
						DocumentationBuild`Utils`HasOptionsQ[nb, base],

					(* If we're building non-System (presumably paclet) documentation,
					   assume that every symbol has options. This currently only controls
					   whether the Notes section title is 'Details' or
					   'Details and Options'. *)
					True,
						True
				];

				(* Print["Does this have Options: "<>ToString[hasopts]]; *)
				getCommon[uri, nb];

				If[sourceQ,
					If[MemberQ[demoList, base],
						info[uri, "Demonstrations"] = True
					];

					If[hasopts,
						info[uri, "HasOptions"] = True
					];
				];
			],
			Global`AntLog["skipping: "<>file];{}
		]],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Symbols"}]]
	]];

	refs = Reap[Scan[
		Function[{pair}, Module[{file, nb, base, uri},
			file = Last[pair];
			nb = DocumentationBuild`Utils`GetQuietly[file];
			base = FileBaseName[file];
			uri = Sow[mkuri["ref", First[pair], base]];
			getCommon[uri, nb];
		]],
		Join[
			{"applescript", #}              & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "AppleScript"}]],
			{"blockchain", #}               & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Blockchains"}]],
			{"c", #}                        & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "C"}]],
			{"character", #}                & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Characters"}]],
			{"classifier", #}               & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Classifiers"}]],
			{"compiledtype", #}             & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "CompiledTypes"}]],
			{"databaseconnection", #}       & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "DatabaseConnections"}]],
			{"datastructure", #}            & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "DataStructures"}]],
			{"device", #}                   & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Devices"}]],
			{"embeddingformat", #}          & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "EmbeddingFormats"}]],
			{"entity", #}                   & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Entities"}]],
			{"externalevaluationsystem", #} & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "ExternalEvaluationSystems"}]],
			{"file", #}                     & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Files"}]],
			{"format", #}                   & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Formats"}]],
			{"frontendobject", #}           & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "FrontEndObjects"}]],
			{"indicator", #}                & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Indicators"}]],
			{"interpreter", #}              & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Interpreters"}]],
			{"menuitem", #}                 & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "MenuItems"}]],
			{"message", #}                  & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Messages"}]],
			{"method", #}                   & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Methods"}]],
			{"netdecoder", #}               & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "NetDecoders"}]],
			{"netencoder", #}               & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "NetEncoders"}]],
			{"predictor", #}                & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Predictors"}]],
			{"program", #}                  & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Programs"}]],
			{"resourceobject", #}           & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "ResourceObjectTypes"}]],
			{"service", #}                  & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "Services"}]],
			{"textcontent", #}              & /@ FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ReferencePages", "TextContents"}]]
		]
	]];

	(* guides *)

	info[_, "WebResourcesLinks"] := {};

	guides = Reap[Scan[
		Function[{file}, Module[{nb, base, uri, exclude, sowButtonBoxURI},
			nb = DocumentationBuild`Utils`GetQuietly[file];
			base = FileBaseName[file];
			uri = Sow[mkuri["guide", base]];
			getCommon[uri, nb];

			sowButtonBoxURI = Function[{bbox},
				Module[{buri},
					buri = Which[
						MatchQ[bbox, ButtonBox[___, ButtonData -> _, ___]],
							ButtonData /. (List@@Cases[bbox, _~(Rule|RuleDelayed)~_]),
						StringQ[First[bbox]],
							mkuri["ref", First[bbox]],
						True, None
					];
					If[buri =!= None,
						Sow[buri]
					];
				]
			];

			If[sourceQ && FreeQ[{"paclet:guide/AlphabeticalListing","paclet:guide/NewIn60AlphabeticalListing"}, uri],
				exclude = Reap[
					Scan[
						Function[{cell}, Scan[
							sowButtonBoxURI,
							Cases[
								cell,
								ButtonBox[___, _->("Link"|{___,"Link",___}), ___],
								Infinity
							]
						]],
						Cases[
							nb,
							Cell[___, "ExcludedObject", ___],
							Infinity
						]
					];
					Scan[
						Function[{cell}, Scan[
							sowButtonBoxURI,
							Cases[
								cell,
								ButtonBox[___, _->{___,"Link",___,"ExcludedObject",___}, ___],
								Infinity
							]
						]],
						Cases[nb,
							Cell[
								_,
								"InlineFormula"|"InlineGuideFunction"|"InlineGuideFunctionListing"|"AlphabeticalListing"|"InlineCode",
								___
							],
							Infinity
						]
					];
				];
				If[Length[exclude[[2]]] > 0,
					exclude = exclude[[2, 1]],
					exclude = {}
				];

				Scan[
					Function[{cell},
						Scan[
							Function[{bbox}, Module[{buri},
								buri = Which[
									MatchQ[bbox, ButtonBox[___, ButtonData -> _, ___]],
										ButtonData /. (List@@Cases[bbox, _~(Rule|RuleDelayed)~_]),
									StringQ[First[bbox]],
										mkuri["ref", First[bbox]],
									True, None
								];
								If[buri =!= None && FreeQ[exclude, buri],
									linktypes[buri] = Append[linktypes[buri], "Guides"];
									links[buri, "Guides"] = Append[
										links[buri, "Guides"],
										{info[uri, "Title"], uri}
									]
								];
							]],
							Cases[
								cell,
								ButtonBox[___, _->("Link"|{___,"Link",___}), ___],
								Infinity
							]
						]
					],
					Cases[
						nb,
						Cell[
							_,
							"InlineFormula"|"InlineGuideFunction"|"InlineGuideFunctionListing"|"AlphabeticalListing"|"InlineCode",
							___
						],
						Infinity
					]
				];

				info[uri, "WebResourcesLinks"] = GetWebResourcesLinks[nb];

			]; (* end If[sourceQ, ...] *)

		]],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "Guides"}]]
	]];

	(* tutorials *)

	info[_, "TutorialSections"] := {};

	tutorials = Reap[Scan[
		Function[{file}, Module[{base, uri, nb, cat},
			nb = DocumentationBuild`Utils`GetQuietly[file];
			base = FileBaseName[file];
			uri = Sow[mkuri["tutorial", base]];
			getCommon[uri, nb];

			If[sourceQ,
				If[MatchQ[info[uri, "Categorization"], "Overview" | "Monograph Overview"], Module[{getSects},
					getSects[expr_, types_] := Module[{poss},
						poss = Position[
							expr,
							Cell[_, First@types, ___]
								| Cell[CellGroupData[{Cell[_, First@types, ___], ___}, ___], ___],
							Infinity
						];
						poss = Last /@ Split[
							poss,
							(Length[#1] > Length[#2] && Take[#1, Length[#2]] === #2) &
						];
						Map[
							Function[{pos}, Module[{cell, bbox, buri},
								cell = First@Extract[expr, {pos}];
								If[MatchQ[cell, Cell[CellGroupData[___], ___]],
									{
										cell[[1, 1, 1, 1]],
										getSects[Rest[cell[[1, 1]]], Rest[types]]
									}
									,
									(
										bbox = First @ Cases[cell, _ButtonBox, Infinity, 1];
										buri = ButtonData /. (List@@Cases[bbox, _~(Rule|RuleDelayed)~_]);
										linktypes[buri] = Append[linktypes[buri], "Overview"];
										links[buri, "Overview"] = {info[uri, "Title"], uri};
										{First[bbox], buri}
									)
								]
							]],
							poss
						]
					];
					linktypes[uri] = Append[linktypes[uri], "Chapters"];
					links[uri, "Chapters"] = getSects[
						nb,
						{"TOCChapter", "TOCSection", "TOCSubsection", "TOCSubsubsection"}
					];
				]];

				info[uri, "ForwardBackNavData"] = DocumentationBuild`Navigation`GetNavDestinations[uri, uridata, lang];
				info[uri, "TutorialSections"] = DocumentationBuild`Navigation`GetTutorialAnchors[uri, nb, lang];
			]; (* end If[sourceQ, ...] *)
		]],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "Tutorials"}]]
	]];

	howtos = Reap[Scan[
		Function[{file}, Module[{base, uri, nb, cat},
			nb = DocumentationBuild`Utils`GetQuietly[file];
			base = FileBaseName[file];
			uri = Sow[mkuri["howto", base]];
			getCommon[uri, nb];

			If[sourceQ,
				If[info[uri, "Categorization"] === "HowTo", Module[{getSects},
					getSects[expr_, types_] := Module[{poss},
						poss = Position[
							expr,
							Cell[_, First@types, ___]
								| Cell[CellGroupData[{Cell[_, First@types, ___], ___}, ___], ___],
							Infinity
						];
						poss = Last /@ Split[
							poss,
							(Length[#1] > Length[#2] && Take[#1, Length[#2]] === #2) &
						];
						Map[
							Function[{pos}, Module[{cell, bbox, buri},
								cell = First@Extract[expr, {pos}];
								If[MatchQ[cell, Cell[CellGroupData[___], ___]],
									{
										cell[[1, 1, 1, 1]],
										getSects[Rest[cell[[1, 1]]], Rest[types]]
									}
									,
									(
										bbox = First @ Cases[cell, _ButtonBox, Infinity, 1];
										buri = ButtonData /. (List@@Cases[bbox, _~(Rule|RuleDelayed)~_]);
										linktypes[buri] = Append[linktypes[buri], "HowTo"];
										links[buri, "HowTo"] = {info[uri, "Title"], uri};
										{First[bbox], buri}
									)
								]
							]],
							poss
						]
					];
					linktypes[uri] = Append[linktypes[uri], "Chapter"];
					links[uri, "Chapters"] = getSects[
						nb,
						{"TOCChapter", "TOCSection", "TOCSubsection", "TOCSubsubsection"}
					];
				]];
			]; (* end If[sourceQ, ...] *)
		]],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "HowTos"}]]
	]];

	workflows = Reap[Scan[
		Function[{file},
			Module[{base, uri, nb, cat},
				nb = DocumentationBuild`Utils`GetQuietly[file];
				base = FileBaseName[file];
				uri = Sow[mkuri["workflow", base]];
				getCommon[uri, nb];
			]
		],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "Workflows"}]]
	]];

	workflowguides = Reap[Scan[
		Function[{file},
			Module[{base, uri, nb, cat},
				nb = DocumentationBuild`Utils`GetQuietly[file];
				base = FileBaseName[file];
				uri = Sow[mkuri["workflowguide", base]];
				getCommon[uri, nb];
			]
		],
		FileNames["*.nb", ToFileName[{dir, lang, systemDir, "WorkflowGuides"}]]
	]];

	(* featured examples --- only for System directory *)
	If[StringMatchQ[systemDir, "System"] && sourceQ,
		examples = Reap[Scan[
			Function[{file},
				Module[{base, uri, nb},
					nb = DocumentationBuild`Utils`GetQuietly[file];
					base = FileBaseName[file];
					uri = Sow[mkuri["example", base]];

					info[uri, "Title"] = GetNotebookTitle[nb, "Featured Example"];
					info[uri, "WebURL"] = StringJoin[examplepageURLroot, base, ".html"];
					info[uri, "URI"] = GetNotebookURI[nb];
					info[uri, "Flag"] = GetNotebookFlag[nb];
				]
			],
			FileNames["*.nb", ToFileName[{dir, lang, systemDir, "ExamplePages"}]]
		]],
		examples = {Null, {}}
	];

	(*---------------------------------------------*)
	(*                                             *)
	(* Join collected information into the source/ *)
	(* online information structure.               *)
	(*                                             *)
	(*---------------------------------------------*)

	Assert[AllTrue[
		{symbols, refs, guides, tutorials, howtos, workflows, workflowguides, examples},
		Function[var, MatchQ[var, {Null, {{___?StringQ}} | {} }]]
	]];

	symbols        = Flatten @ symbols[[2]];
	refs           = Flatten @ refs[[2]];
	guides         = Flatten @ guides[[2]];
	tutorials      = Flatten @ tutorials[[2]];
	howtos         = Flatten @ howtos[[2]];
	workflows      = Flatten @ workflows[[2]];
	workflowguides = Flatten @ workflowguides[[2]];
	examples       = Flatten @ examples[[2]];

	Join[
		Map[
			Function[{uri}, uri -> Join[
				putCommon[uri],
				If[sourceQ,
					{
						"Demonstrations" -> (info[uri, "Demonstrations"] === True),
						"HasOptions" -> (info[uri, "HasOptions"] === True)
					},
					{}
				]
			]],
			symbols
		],

		Map[
			Function[{uri}, uri -> putCommon[uri]],
			refs
		],

		Map[
			Function[{uri}, uri -> Join[
				putCommon[uri],
				If[sourceQ,
					{
						"WebResourcesLinks" -> info[uri, "WebResourcesLinks"]
					},
					{}
				]
			]],
			guides
		],

		Map[
			Function[{uri}, uri -> Join[
				putCommon[uri],
				If[sourceQ,
					{"TutorialSections" -> (info[uri, "TutorialSections"])},
					{}
				]
			]],
			tutorials
		],

		Map[
			Function[{uri}, uri -> putCommon[uri]],
			howtos
		],

		Map[
			Function[{uri}, uri -> putCommon[uri]],
			workflows
		],

		Map[
			Function[{uri}, uri -> putCommon[uri]],
			workflowguides
		],

		Map[
			Function[{uri},
				uri -> Map[
					(# -> info[uri, #] &),
					{"Title", "WebURL", "URI", "Flag"}
				]
			],
			examples
		]
	]
];


(******************************************************************************)
(** MakePrimaryGuideList **)

(*
	MakePrimaryGuideList creates rule list of function->guide pages
	based on "PrimaryObject" links in guide pages 

	in: guide page dir
	out: {.., paclet:ref/Sin->{PrimaryGuide->paclet:guide/TrigonometricFunctions},...}
*)
MakePrimaryGuideList[guideDir_String, linkBase_String:""] := 
Module[{files, expr, pkg=linkBase,files2,pos,pos2,excisedFiles,newfiles,test},
  pkg = If[ MemberQ[{"Mathematica", ""}, pkg], "", pkg];
  files = FileNames["*.nb", guideDir];
  files2 = Map[
   Cases[Import[#, "NB"], Cell[_String, "ExcisedFlag", _], Infinity] &, 
   files];
   pos = Position[files2, Cell[_, "ExcisedFlag", _] ];
   pos2 = Map[{First[#]} &, pos];
   excisedFiles = Extract[files, pos2];
   newfiles = Complement[files, excisedFiles];
   
Flatten@{{expr = DocumentationBuild`Utils`GetQuietly[#]; 
       Function[{guide, functions}, 
         Flatten@{Rule[#, {Rule["PrimaryGuide",guide]}] & /@ functions}][
        "paclet:"<>DocumentationBuild`Common`FilepathToURI[#, pkg], 
        Union@Flatten@
          Cases[expr, 
           bb : ButtonBox[func_String, ___, BaseStyle -> {___, "PrimaryObject", ___}, ___] :> pacletFromButtonBox@bb, 
           Infinity]]} & /@ newfiles}];



(******************************************************************************)
(** pacletFromButtonBox - Internal **)
pacletFromButtonBox[ButtonBox[func_String, opts___?OptionQ]] :=
 Module[{bd},
  bd = ButtonData /. {opts};
  If[Head@bd =!= String, "paclet:ref/" <> func, bd]
  ];
pacletFromButtonBox[___] := {}


(******************************************************************************)
(** GetLinkTrails - Internal **)

GetLinkTrails[u_String /; !StringMatchQ[u, "paclet:*"], primaryGuides_]:= 
	GetLinkTrails["paclet:"<>u, primaryGuides];
GetLinkTrails[u_String /; StringMatchQ[u, "paclet:*"], primaryGuides_]:= 
Module[{trail, uri=u, info, primaryGuide, pg, lst, res},
	LogIt["-- GetLinkTrails: "<>u<>", "<>ToString[Length@primaryGuides]];
	(* get the primary guide *)
	pg = DocumentationBuild`Utils`GetPrimaryGuide[uri, primaryGuides];
	(* If no primary guide, then return root guide *)
	If[uri==="paclet:tutorial/VirtualBookOverview", Return[{{"paclet:tutorial/VirtualBookOverview"}}]];
	If[pg === uri, (* XX Root guide rename change XX *)Return[{{"paclet:guide/WolframRoot"}}] ];
	LogIt["-- "];
	trail =
	If[DocumentationBuild`Utils`CoreMathematicaPacletQ[uri], 
		(* wri guide/ref/tutorial... *)
		If[StringMatchQ[uri, "paclet:*"], 
			(* compute link trail for wri guides *)
			If[StringMatchQ[uri, "paclet:tutorial/*"],
				res = DocumentationBuild`Info`GetLinkTrailFromFilename[uri, DocumentationBuild`Navigation`$TutPages, primaryGuides];
			If[res === {}, {{"paclet:tutorial/VirtualBookOverview"}}, res],
				res = DocumentationBuild`Info`GetLinkTrailFromFilename[uri, DocumentationBuild`Navigation`$Pages, primaryGuides];
			If[res === {}, (* XX Root guide rename change XX *){{"paclet:guide/WolframRoot"}}, res]]
		,
			(* uri is not a wri guide... *)
			(* check to see if paclet uri has a primary guide *)
			If[(pg === (* XX Root guide rename change XX *)"paclet:guide/WolframRoot") || (pg === ""),
				(* return link to root guide *)
				(* XX Root guide rename change XX *){ {"paclet:guide/WolframRoot"} }
			,
				(* return root + primary guide *)
				DocumentationBuild`Utils`UnsortedUnion @ { Flatten @ DocumentationBuild`Utils`UnsortedUnion @
					(* XX Root guide rename change XX *)DeleteCases[{"paclet:guide/WolframRoot", DocumentationBuild`Utils`GetPrimaryGuide[pg, primaryGuides], pg}, ""|Null|$Failed]  }
			]
		]
	,
		(* return pakage's primary guide *)
		{{pg }}

	]

];

(* error check *)
GetLinkTrails[u___]:= (Message[GetLinkTrails::arg, u]; $Failed)
GetLinkTrails::arg = "Paclet uri expected, but found `1`";

GetLinkTrailFromFilename[filename_String /; StringMatchQ[filename, "paclet:ref/" ~~ __], data_List, primaryGuides_List] :=
Module[{list, primaryguide, fn = filename},
    Which[

      (* normal Symbol ref page *)
      StringFreeQ[fn, "ref/message/"],
        primaryguide = Quiet["PrimaryGuide" /. (filename /. primaryGuides)];
        (* next line does a check to see if ref page has an entry in the $CoreInformation data set *)
        If[primaryguide === Quiet[("PrimaryGuide" /. filename)], Message[GetLinkTrailFromFilename::nodat,filename];Return[{}]];
        (* continue as normal *)
        If[primaryguide === (* XX Root guide rename change XX *)"paclet:guide/WolframRoot", {primaryguide}, 
        list = Cases[data,
            {StringReplace[primaryguide, {"paclet:guide/"->"", ".nb"->""}] /. DocumentationBuild`Navigation`$NameToNumberRules, trail_} :> trail, Infinity] /. DocumentationBuild`Navigation`$NumberToNameRules;
        Join[#, {primaryguide, filename}] & /@ Flatten[list //. {a___, b_String /; ! StringMatchQ[b, "paclet:*"], c___} :> {a, "paclet:guide/" <> b, c}, 1]],

      (* message page code goes here eventually... *)
      !StringFreeQ[fn, "ref/message/"],
        (*code here*)
        {},

      (* catch all default value to prevent docs breaks for now *)
      True, {}]
];    

GetLinkTrailFromFilename[filename_String /; StringMatchQ[filename, "paclet:guide/" ~~ __], data_List, primaryGuides_List] := 
Module[{list},
	list = Cases[data, 
			{StringReplace[filename, {"paclet:guide/"->"", ".nb"->""}] /. DocumentationBuild`Navigation`$NameToNumberRules, trail_} :> trail, Infinity] /. DocumentationBuild`Navigation`$NumberToNameRules;
(*Print["List: "<>ToString@list];*)
	Append[#, filename]& /@ Flatten[list //. {a___, b_String /; ! StringMatchQ[b, "paclet:*"], c___} :> {a, "paclet:guide/" <> b, c}, 1]
];

GetLinkTrailFromFilename[filename_String /; StringMatchQ[filename, "paclet:tutorial/" ~~ __], data_List, primaryGuides_List] :=
Module[{list},
	list = Cases[data, 
			{StringReplace[filename, {"paclet:tutorial/"->"", ".nb"->""}] 
				/. DocumentationBuild`Navigation`$TutNameToNumberRules, trail_} :> trail, Infinity] 
				/. DocumentationBuild`Navigation`$TutNumberToNameRules;
(*Print["List: "<>ToString@list];*)
	Append[#, filename]& /@ Flatten[list //. {a___, b_String /; ! StringMatchQ[b, "paclet:*"], c___} :> {a, "paclet:tutorial/" <> b, c}, 1]
];

GetLinkTrailFromFilename[filename_String /; StringMatchQ[filename, "paclet:howto/" ~~ __], data_List, primaryGuides_List] :=
{};

GetLinkTrailFromFilename::list = "Navigation'$Pages not a list: `1`";
GetLinkTrailFromFilename::nodat = "The specified URI, `1`, is missing a Primary Guide.";


End[]
EndPackage[]
