BeginPackage["DocumentationBuild`Internal`"]

(**
 * This file contains code which is used only during internal documentation builds; those
 * run by WRI as part of building the Wolfram System documentation.
 **)

LoadInternalSettings::usage = "LoadInternalSettings[] loads settings suitable for \
performing internal builds of Wolfram System documentation. This function will fail if \
you are not a WRI developer connected to VPN."

CheckShouldUseInternalHTMLAssets::usage = "CheckShouldUseInternalHTMLAssets[] heuristically \
determines if the package being built is one of a set of specially handled packages."

InternalAssetsDirectory
SystemPacletQ::usage = "SystemPacletQ[name] returns true if name is a paclet built-in to the Wolfram System."

UpdateLinkDestination::usage = "UpdateLinkDestination updates links to internal pages which have been moved."

Begin["`Private`"]

Needs["DocumentationBuild`Common`"]
Needs["DocumentationBuild`Utils`"]
Needs["DocumentationBuild`Settings`"]

InternalAssetsDirectory[] := With[{
	path = FindFile["DocumentationAssets`"]
},
	If[FailureQ[path],
		None,
		FileNameDrop[path, -2]
	]
]

LoadInternalSettings[] := Module[{},
	(*------------------*)
	(* $FunctionPrereqs *)
	(*------------------*)

	DocumentationBuild`Settings`$FunctionPrereqs = <|
		"URIsWithoutDesktopSupport" -> {
			"ref/CloudBase", "ref/CloudConnect", "ref/CloudDeploy",
			"ref/CloudDirectory", "ref/CloudDisconnect", "ref/CloudEvaluate",
			"ref/CloudExport", "ref/CloudFunction", "ref/CloudGet", "ref/CloudImport",
			"ref/CloudObject", "ref/CloudObjectInformation",
			"ref/CloudObjectInformationData", "ref/CloudObjects", "ref/CloudPut",
			"ref/CloudSave", "ref/CloudSymbol", "ref/SetCloudDirectory",
			"ref/CloudPublish", "ref/$CloudBase", "ref/$CloudConnected",
			"ref/$CloudEvaluation", "ref/$CloudRootDirectory", "ref/$CloudSymbolBase",
			"ref/APIFunction", "ref/FormFunction", "ref/ExternalBundle",
			"ref/ExportForm", "ref/ResponseForm", "ref/Delayed", "ref/HTTPResponse",
			"ref/HTTPRedirect", "ref/Permissions", "ref/$Permissions",
			"ref/CloudAccountData", "ref/$EvaluationEnvironment", "ref/HTTPRequestData",
			"ref/$RequesterAddress", "ref/EmbeddedHTML", "ref/EmbeddedService",
			"ref/CloudExpression", "ref/CreateCloudExpression", "ref/CloudExpressions",
			"ref/CloudExpressionObject", "ref/DeleteCloudExpression"
		},
		"URIsWithoutCloudSupport" -> Map[
			StringJoin["ref/", #] &,
			Lookup[Get[DocsAsset["CloudPrerequisiteData.m"]], "Unsupported", {}]
		],
		"URIsWithoutPublicCloudSupport" -> Map[
			StringJoin["ref/", #] &,
			Lookup[Get[DocsAsset["CloudPrerequisiteData.m"]], "Restricted", {}]
		],
		"URIsWithServiceCredits" -> {"ref/WebSearch", "ref/WebImageSearch", "ref/TextTranslate"}
	|>;


	(*-------------------*)
	(* $CitationCallback *)
	(*-------------------*)

	$CitationCallback = Function[citationInfo, Module[{
		lang,
		name, url,
		new, mod,
		prodVersion = ToString[IntegerPart[$VersionNumber]] <> "." <> ToString[$ReleaseNumber],
		styles
	},
		lang = citationInfo["Language"];
		(* Remove parenthetical parts of the title from the citation. *)
		name = StringTrim @ StringReplace[citationInfo["Title"], "(" ~~ __ ~~ ")" -> ""];
		url = "https://reference.wolfram.com/language/" <> citationInfo["URI"] <> ".html";

		new = citationInfo["NewInYear"];
		mod = citationInfo["ModifiedInYear"];

		DocsAssert[StringQ[lang]];
		DocsAssert[MatchQ[new, _?StringQ]];
		DocsAssert[MatchQ[mod, _?StringQ | None]];

		styles = <|
			"Text" -> TemplateApply[
				If[mod =!= None,
					Localized["BuildCitationTemplate", lang],
					Localized["BuildCitationTemplateNoUpdate", lang]
				],
				{new, name, url, mod}
			]
		|>;

		AssociateTo[
			styles,
			Function[# -> TemplateApply[
				If[mod =!= None,
					StringTemplate[Localized["BuildCitation"<>#<>"Template", lang]],
					StringTemplate[Localized["BuildCitation"<>#<>"TemplateNoUpdate", lang]]
				],
				{new, name, Replace[mod, None -> new], url}
			]] /@ {"CMS", "APA"}
		];

		(* Formatting note: df6e6cb5: fix BibTeX citation per bug 400416 *)
		AssociateTo[
			styles,
			Function[# -> TemplateApply[
				StringTemplate[Localized["BuildCitation" <> # <> "Template", lang]],
				{
					name,
					Replace[mod, None -> new],
					DateString["Year"],
					url,
					ToLowerCase[name],
					prodVersion
				}
			]] /@ {"BibTeX", "BibLaTeX"}
		];

		styles
	]];

] ~Catch~ $DocsErrorTag

(******************************************************************************)
(* Helper functions                                                           *)
(******************************************************************************)

(**************************************)
(* importCloudPrereqData              *)
(**************************************)

$cloudPrereqData = None

importCloudPrereqData[] := Module[{
	urlbase = "https://stash.wolfram.com/projects/CLOUD/repos/techdocs/raw/PublicDocs/SymbolLists/",
	types = {"PartiallySupported", "Restricted", "Supported", "Unsupported"},
	filetype = ".txt",
	urlpost = "",
	data,
	CleanImport
},
	(* Check if we've already cached this value. *)
	If[AssociationQ[$cloudPrereqData],
		Return[$cloudPrereqData];
	];

	CleanImport[url_String] := (
		data = Import[url, "Lines"];

		If[FailureQ[data],
			(* Import will also have issued a message, providing more detail. *)
			DocsError["Unable to import SymbolLists from stash.wolfram.com: ``", data];
		];

		DeleteCases[
			Select[data, !StringMatchQ[#, "#"~~___]&],
			""
		]
	);

	$cloudPrereqData = AssociationMap[
		CleanImport[urlbase <> # <> filetype <> urlpost] &,
		types
	];

	Assert[AssociationQ[$cloudPrereqData]];

	$cloudPrereqData
]

(**************************************)
(* CheckShouldUseInternalHTMLAssets   *)
(**************************************)

CheckShouldUseInternalHTMLAssets[inputDir_?StringQ] := Module[{
	webmQ = StringContainsQ[inputDir, "webMathematica", IgnoreCase -> True],
	lwgQ  = StringContainsQ[inputDir, "LightweightGrid", IgnoreCase -> True],
	wsmQ  = StringContainsQ[inputDir, "MathModelica", IgnoreCase -> True],
	customBundle
},
	If[webmQ || lwgQ || wsmQ, (
		(* The assets to use if we're building one of a handful of Wolfram products. *)
		customBundle = Which[
			webmQ,
				"webm-assets",
			lwgQ,
				"lwg-assets",
			(* Formatting note: a410487c:
				Updates for in-product WSM documentation
				- use wsmQ=True when completeHTMLQ=True to trigger use of the new
					wsm-assets for WSM in-product docs
			*)
			wsmQ,
				"wsm-assets",
			True,
				DocsError["Expected unreachable Which case"]
		];

		{DocsAsset["Internal:Web", customBundle], customBundle}
	),
		Automatic
	]
]

(******************************************************************************)

SystemPacletQ[name_] := MemberQ[
	{
		(* Main system documentation *)
		"", "Mathematica",
		(* Packages *)
		"ANOVA", "Audio", "BarCharts", "Benchmarking", "BlackBodyRadiation",
		"Calendar", "CCompilerDriver", "ClusterIntegration", "Combinatorica",
		"Compatibility", "ComputationalGeometry", "ComputerArithmetic",
		"DatabaseLink", "Developer", "DLLTools", "EquationTrekker", "ErrorBarPlots",
		"Experimental", "FiniteFields", "FourierSeries", "FunctionApproximations",
		"Geodesy", "GraphUtilities", "GUIKit", "HierarchicalClustering", "Histograms",
		"HypothesisTesting", "JLink", "LightweightGridClient", "LinearRegression",
		"Messages", "MultivariateStatistics", "Music", "NETLink",
		"NonlinearRegression", "Notation", "NumericalCalculus", "NumericalDifferentialEquationAnalysis",
		"Parallel", "ParallelTools", "PhysicalConstants", "PieCharts", "PlotLegends",
		"PolyhedronOperations", "Polytopes", "PrimalityProving", "Quaternions",
		"RawGuides", "RegressionCommon","RemoteSLightweightGridCliental",
		"ResonanceAbsorptionLines","Splines", "StandardAtmosphere", "StatisticalPlots",
		"Symbols", "Units", "VariationalMethods", "VectorAnalysis", "VectorFieldPlots",
		"WebServices", "WorldPlot", "XML"
	},
	name
]

(******************************************************************************)

UpdateLinkDestination[destination0_?StringQ] := Module[{
	destination = destination0
},
	(* If `destination` is already written with the updated link destination, return it. *)
	(* Note: This is necessary to avoid the string replacement logic below generating links
	         like "JLink/ref/java/com/wolfram/jlink/com/wolfram/jlink/". *)
	If[
		StringContainsQ[
			destination,
			"JLink/ref/java/com/wolfram/jlink/"
			| "NETLink/ref/net/Wolfram.NETLink."
		],
		Return[destination, Module]
	];

	(* Note:
		My assumption about the reason for the existance of this code is that, over the
		years, certain internal documentation pages have been moved from one location to
		another. Given that there were many existing documentation pages with links
		pointing to the previous location, it made sense to force all documentation
		pages to point their links at the new destination, and so this logic was added to
		DocumentationBuild's link handling code. *)
	StringReplace[
		StringReplace[
			destination,
			{
				"JLink/ref/java/" -> "JLink/ref/java/com/wolfram/jlink/",
				"NETLink/ref/net/" -> "NETLink/ref/net/Wolfram.NETLink.",
				"IKernelLink/" -> "IKernelLink."
			}
		],
		{
			(*---------------*)
			(* JLink updates *)
			(*---------------*)

			"JLink/ref/java/com/wolfram/jlink/BracketMatcher"  -> "JLink/ref/java/com/wolfram/jlink/ui/BracketMatcher",
			"JLink/ref/java/com/wolfram/jlink/ConsoleWindow"   -> "JLink/ref/java/com/wolfram/jlink/ui/ConsoleWindow",
			"JLink/ref/java/com/wolfram/jlink/InterruptDialog" -> "JLink/ref/java/com/wolfram/jlink/ui/InterruptDialog",
			"JLink/ref/java/com/wolfram/jlink/MathSessionPane" -> "JLink/ref/java/com/wolfram/jlink/ui/MathSessionPane",
			"JLink/ref/java/com/wolfram/jlink/SyntaxTokenizer" -> "JLink/ref/java/com/wolfram/jlink/ui/SyntaxTokenizer",
			"JLink/ref/java/com/wolfram/jlink/LinkSnooper"     -> "JLink/ref/java/com/wolfram/jlink/util/LinkSnooper",
			"JLink/ref/java/com/wolfram/jlink/MathematicaTask" -> "JLink/ref/java/com/wolfram/jlink/util/MathematicaTask",

			(*-----------------*)
			(* NETLink updates *)
			(*-----------------*)

			"NETLink/ref/net/Wolfram.NETLink.NETLink" -> "NETLink/ref/net/Wolfram.NETLink",
			"Wolfram.NETLink.IKernelLink.EvaluateToOutputForm" -> "Wolfram.NETLink.IKernelLink.EvaluateToOutputForm_overloads",
			"Wolfram.NETLink.IKernelLink.EvaluateToImage" -> "Wolfram.NETLink.IKernelLink.EvaluateToImage_overloads",
			"Wolfram.NETLink.IKernelLink.Evaluate" -> "Wolfram.NETLink.IKernelLink.Evaluate_overloads",
			"Wolfram.NETLink.IKernelLink.GetInteger" -> "Wolfram.NETLink.IMathLink.GetInteger",
			"Wolfram.NETLink.ConsoleWindow" -> "Wolfram.NETLink.UI.ConsoleWindow",
			"Wolfram.NETLink.DoubleBufferedPanel" -> "Wolfram.NETLink.UI.DoubleBufferedPanel",
			"Wolfram.NETLink.MathPictureBox" -> "Wolfram.NETLink.UI.MathPictureBox",
			"Wolfram.NETLink.MathLinkFactory/CreateKernelLink" -> "Wolfram.NETLink.MathLinkFactory.CreateKernelLink_overloads",
			(*"Wolfram.NETLink.NETLink"<>""->"index",*)
			"NETLink/ref/net/DotNETInterface" -> "NETLink/guide/DotNETInterface"
		}
	]
]


End[]

EndPackage[]