BeginPackage["DocumentationBuild`Settings`"]

(**
 * This file contains user-configurable settings, to alter how DocumentationBuild
 * processes documentation notebooks. The default values for these settings should be
 * suitable for building external user-authored documentation.
 *
 * Some of these settings are likely only useful for internal users.
**)

$FunctionPrereqs = <||>
$StandardBundleHTMLTemplateIncludes
$InternalBundleHTMLTemplateIncludes
$CitationCallback = None
$DocumentationBuildLanguage = None;
$InvisibleNotebooks = False
$DocumentationPageTypeURIBases

FunctionPrereqs
ResolveHTMLTemplateIncludes

Begin["`Private`"]

Needs["DocumentationBuild`Common`"] (* DocsError *)

(* Return a list of prerequisite conditions which need to be satisfied in order for a WL
   function to be supported in a particular computing environment. E.g. only works in
   cloud, only works on desktop, requires service credits, etc. See WLDOCS-1001 and
   WLDOCS-1210. *)
FunctionPrereqs[uri_?StringQ] := (
	If[!AssociationQ[$FunctionPrereqs],
		DocsError["Value of $FunctionPrereqs should be an association."];
	];

	If[
		!AllTrue[
			Keys[$FunctionPrereqs],
			MemberQ[{"URIsWithoutDesktopSupport", "URIsWithoutCloudSupport",
			         "URIsWithoutPublicCloudSupport", "URIsWithServiceCredits"}, #] &
		],
		DocsError["Unrecognized Keys in value of $FunctionPrereqs"]
	];

	Keys @ Select[
		$FunctionPrereqs,
		Function[uris,
			Assert[MatchQ[uris, {___?StringQ}]];
			StringMatchQ[uri, uris]
		]
	]
)

(**************************************)
(* Documentation Page Types           *)
(**************************************)

(* Mapping of documentation page type names (also called the 'Entity Type'; no connection
   to Entity[..]) to the URI base used to refer to pages of this type. *)
$DocumentationPageTypeURIBases = <|
	"AppleScript"                   -> "ref/applescript",
	"Character Name"                -> "ref/character",
	"Character"                     -> "ref/character",
	"Classifier"                    -> "ref/classifier",
	"Compiled Type"                 -> "ref/compiledtype",
	"Data Structure"                -> "ref/datastructure",
	"Database Connection"           -> "ref/databaseconnection",
	"Device Connection"             -> "ref/device",
	"Embedding Format"              -> "ref/embeddingformat",
	"Entity"                        -> "ref/entity",
	"Example"                       -> "example",
	"External Evaluation System"    -> "ref/externalevaluationsystem",
	"File"                          -> "ref/file",
	"Format"                        -> "ref/format",
	"Front End Object"              -> "ref/frontendobject",
	"Guide Overview"                -> "guide",
	"Guide"                         -> "guide",
	"HowTo"                         -> "howto",
	"Indicator"                     -> "ref/indicator",
	"Interpreter"                   -> "ref/interpreter",
	"LibraryLink C Function"        -> "ref/callback",
	"MathLink C Function"           -> "ref/c",
	"Menu Item"                     -> "ref/menuitem",
	"Message"                       -> "ref/message",
	"Method"                        -> "ref/method",
	"Monograph Overview"            -> "tutorial",
	"Monograph"                     -> "tutorial",
	"Net Decoder"                   -> "ref/netdecoder",
	"Net Encoder"                   -> "ref/netencoder",
	"Note"                          -> "note",
	"Overview"                      -> "tutorial",
	"Predictor"                     -> "ref/predictor",
	"Program"                       -> "ref/program",
	"Remote Batch Computation Provider" -> "ref/batchcomputationprovider",
	"Resource Object"               -> "ref/resourceobject",
	"Root Guide"                    -> "guide",
	"Screencast"                    -> "screencast",
	"Service Connection"            -> "ref/service",
	"Symbol"                        -> "ref",
	"System Program"                -> "ref/program",
	"Tech Note"                     -> "tutorial",
	"Text Content"                  -> "ref/textcontent",
	"Tutorial"                      -> "tutorial",
	"Upgrade Information"           -> "tutorial",
	"Workflow Guide"                -> "workflowguide",
	"Workflow"                      -> "workflow"
|>

Protect[$DocumentationPageTypeURIBases];

(**************************************)
(* HTML Templates                     *)
(**************************************)

ResolveHTMLTemplateIncludes[
	templateIncludes_?AssociationQ,
	params_?AssociationQ
] := Module[{value},
	Map[
		Function[{value0},
			value = value0;

			If[MatchQ[value, Function[___]],
				value = Check[
					value[params]
					,
					(* Expecting this will occur rarely when Function::slota is issued
					when the hard-coded `params` passed from DocumentationBuildHTML get
					out of sync with the usage of named `Slot[".."]` values. *)
					DocsError["Unexpected messages while resolving HTML template include value"]
				];
			];

			If[!StringQ[value],
				DocsError["Malformed HTML include value: ``", InputForm[value]];
			];

			value
		],
		templateIncludes
	]
]

$StandardBundleHTMLTemplateIncludes = <|
	"doc-start" -> Function["
		<!DOCTYPE html>
		<html lang=\""<> #la <>"\">

		<head>
		<meta http-equiv=\"x-ua-compatible\" content=\"ie=edge\">
		<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">
		<meta charset=\"utf-8\">
	"],
	"head-tag-end" -> Function[
		StringJoin @ Map[
			Function[file,
				"<link rel=\"stylesheet\" href=\"" <> #webAssetsPath <> file <> "\">"
			],
			{
				"common/framework/css/framework.en.css",
				"common/framework/css/gui.en.css",
				"common/js/clipboard/2.0/clipboard.css",
				"fonts/source-code-pro/1.0/all.css",
				"fonts/source-sans-pro/1.0/global.css",
				"language-assets/css/entities.css",
				"language-assets/css/footer.css",
				"language-assets/css/global.css",
				"language-assets/css/guide.css",
				"language-assets/css/header-language.css",
				"language-assets/css/iconography.css",
				"language-assets/css/icons.css",
				"language-assets/css/in-out.css",
				"language-assets/css/language-specific.css",
				"language-assets/css/language.css",
				"language-assets/css/man-page.css",
				"language-assets/css/related-links.css",
				"language-assets/css/root-guide-language.css",
				"language-assets/css/special-styles.css",
				"language-assets/css/tables.css",
				"language-assets/css/technotes.css",
				"language-assets/css/tooltips.css",
				"language-assets/css/workflows.css"
			}
		]
	],
	"body-tag-start" -> "<main class=\"main\">"
	,
	"body-tag-end" -> Function[
		"
		</main>
		<div id=\"_footer-offset\"></div>
		<div id=\"_footer\"></div>
		" <> StringJoin @ Map[
			Function[file,
				"<script src=\"" <> #webAssetsPath <> file <> "\"></script>"
			],
			{
				"common/js/jquery/2.2.4/jquery.min.js",
				"common/js/jquery/plugins/hoverintent/1.8.0/hoverintent.min.j2",
				"common/js/clipboard/2.0/clipboard.js",
				"common/js/clipboard/2.0/code-clipboard.js",
				"common/framework/js/body.en.js",
				"common/framework/js/head.en.js",
				"language-assets/js/dropdowns.js",
				"language-assets/js/global.js",
				"language-assets/js/language-specific.js.en",
				"language-assets/js/language.js",
				"language-assets/js/temporary.js",
				"language-assets/js/toggles.js",
				"language-assets/js/tooltips.js",
				"language-assets/js/workflows.js"
			}
		]
	],
	"root-guide-head-tag-start" -> Function["
		<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">
		<meta charset=\"utf-8\">
		<link rel=\"stylesheet\" href=\"//www.wolframcdn.com/fonts/source-sans-pro/1.0/global.css\">
		<link rel=\"stylesheet\" href=\"" <> #webAssetsPath <> "common/framework/css/framework.en.css\">
		<link rel=\"stylesheet\" href=\"" <> #webAssetsPath <> "common/framework/css/gui.en.css\">
		<script src=\"" <> #webAssetsPath <> "common/framework/js/head.en.js\"></script>
		<script src=\"" <> #webAssetsPath <> "common/js/jquery/2.2.4/jquery.min.js\"></script>
		<script src=\"" <> #webAssetsPath <> "common/js/jquery/plugins/hoverintent/1.8.0/hoverintent.min.js\"></script>
	"],
	"root-guide-body-tag-end" -> Function["
		<script src=\"" <> #webAssetsPath <> "language-assets/js/root-guide.js\"></script>
	"]
|>;

(* Formatting note: bbc7de70:
	Update webM docbuild to v12 styles

	https://bugs.wolfram.com/show?number=372505

	New files/directories:

		Internal/web/webm-assets.sh
			- shell script to copy the latest live cloud-assets and lightly process
			them to handle webm docs.

		Internal/web/webm-assets/
			- the result of running webm-assets.sh on the live cloud-assets from
			May 22, 2019 (post-v12 release)

	XMLTransforms/HTML/Components/Defaults.m
		- when setting relativeBase, make special arrangements for
		webMathematica to use the new webm-assets dir at its local address in
		a tomcat/apache server.
		- update filledinlayoutexpr with all-new resolved includes for webm
		pages.  The basic idea is to load a page that uses cloud-assets, then
		View Source to see how the server resolves all the includes.  Then
		replace the page's virtual includes with resolved paths.
*)
(* Formatting note: 736980733d5a5f17794b1809bc485a65c8c6770a:
	Fix web wsm docs

	https://jira.wolfram.com/jira/browse/SYSTEMMODELER-2295

	We currently generate stand-alone in-product documentation for three
	products: Wolfram SystemModeler, webMathematica, and Lightweight Grid.
	Wolfram SystemModeler also has a web version at reference.wolfram.com.

	In Defaults.m, this modification of filledinlayoutexpr should happen
	only for in-product docs: webm, lwg, and in-product wsm. Web wsm docs
	should not be modified in this way.
*)
(* These includes are used when bundling the wsm-assets/lwg-assets/webm-assets files. *)
$InternalBundleHTMLTemplateIncludes = <|
	"doc-start" -> Function["
		<!DOCTYPE html>
		<html lang=\"" <> #la <> "\">

		<head>
		<meta http-equiv=\"x-ua-compatible\" content=\"ie=edge\">
		<meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">
		<meta charset=\"utf-8\">
	"],
	"head-tag-end" -> Function["
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/framework.en.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/fonts.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/global.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/iconography.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/header.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/icons.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/root-guide.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/special-styles.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/language-specific.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/man-page.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/clipboard.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/tables.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/related-links.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/footer.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/guide.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/in-out.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/technotes.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/workflows.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/cloud.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/workflows-cloud-only.css\">
		<link rel=\"stylesheet\" href=\""<> #webAssetsPath <>"css/sticky-top-header.css\">
		<script src=\""<> #webAssetsPath <>"js/head.js.en\"></script>
		<script src=\""<> #webAssetsPath <>"js/jquery.min.js?_v=2.2.4\"></script>
	"],
	"body-tag-start" -> "<main class=\"main\">"
	,
	"body-tag-end" -> Function["
		<script async=\"true\" src=\""<> #webAssetsPath <>"js/dropdowns.js\"></script>
		<script async=\"true\" src=\""<> #webAssetsPath <>"js/cloud.js\"></script>
		<script src=\""<> #webAssetsPath <>"js/clipboard.js.en\"></script>
		<script src=\""<> #webAssetsPath <>"js/language-specific.js.en\"></script>
		<script src=\""<> #webAssetsPath <>"js/global.js\"></script>
		<script src=\""<> #webAssetsPath <>"js/body.en.js\"></script>
		<script src=\""<> #webAssetsPath <>"js/toggles.js\"></script>
		<script src=\""<> #webAssetsPath <>"js/workflows.js\"></script>
		<script src=\""<> #webAssetsPath <>"js/temporary.js\"></script>
		"
		(* Close the <div class="inner main-content"> *)
		<>If[TrueQ[#wsmQ] && TrueQ[#completeHTMLQ], "</div>", ""]<>"
		</main>
	"]
|>

End[]

EndPackage[]