(* ::Package:: *)

(*
  TemplateTools (Experimental)
  Routines for generating function templates

  Stand-alone version of code extracted from FEKernelInit.tr

  T. Sherlock
*)

BeginPackage["DocumentationBuild`TemplateTools`"]

(*
*)

GenerateFunctionTemplate::usage="GenerateFunctionTemplate[functionName] returns a cell expression for the function template";

Begin["`Private`"]

(* start template extraction routines -- set up transformation rules
   handling special case templates *)

alphabetrules = Map[FromCharacterCode[#] ->0 &,Range
        [65,65+25]~Join~Range[97,97+25]];

allrules = Dispatch[Join[alphabetrules,
			{" "-> Null, "\t" -> Null, "\n" -> Null, "\r" -> Null,
                       _String?(35839 >= First[ToCharacterCode[#]] >= 12228 &) -> 1}]];

(* use of utilfun is a kludge around a bug in previous versions of
   FEKernelInit.tr. When incompatable FEKernelInit packages are loaded,
   the FE removes all symbols in FE`; old versions do a blind remove
   which can cause an error message if temporary symbols are involved.
   Use of utilfun is to prevent anomalous creation of temporary pattern
   symbol. *)
utilfun[{c___,"="|".",".",(Null | 1),___}] := {c,".","."}
utilfun[any_] := any

tr = {{a:("_"|"#")..,(Null | 1 | "(" | "\:ff08"),___}:>{a},
      {a___, Null..., 1, ___} :> {a},
      {a___,"-","-",(Null | 1),___} :> {a,"-","-"},
      {a___, "+","+",(Null | 1),___} :> {a,"+","+"},
      {a__, b:("!"..), (Null | 1), ___} :> {a,b},
      {_,"'",___} :> {" "," "},
      {a___,"<",">",Null,".",".",".",___} :>{a," "," "," "," "," "," "},
      {a___, b:(0)..,(Null | 1)..,(0)..,___} :> utilfun[{a,b}]};

(* extract a special case template (one in operator form, like the one
   for Plus) *)
getSpecialTemplate[usage_String] :=
	StringTake[usage,Length[Replace[(Characters[usage]/.allrules),tr]]]

(* pre-processing -- some usage messages may contain information which 
	gets stripped when converting to plain text so preserves original form. *)
getTemplateFromUsage[msg_String /; $eTemplate =!= True, name_String] := 
 Block[{$eTemplate = True, template}, 
  template = makePlainText[ToBoxes[msg]];
  template = getTemplateFromUsage[template, name];
  If[ StringQ[template] && StringMatchQ[template, StringJoin[name, "*"]], 
	ToBoxes[template],
	getTemplateFromUsage[msg, name] ]
 ]

(* extract a regular template; calls the special case if none of the
   other possibilities can be handled. *)
getTemplateFromUsage[msg_String, name_String] :=
  If[MatchQ[#, {(0).., Null, ___}],
        If[StringMatchQ[StringTake[msg,
                First[First[Position[#, Null]]]-1
               ], name | StringExpression[WordCharacter..]],
           name,
           getSpecialTemplate[msg]
        ],
        If[# === {},
            $Failed,
            StringTake[msg, #[[1,1]]]
        ]&[Position[Partition[
                FoldList[Plus, 0, #/.Null -> 0],
             2, 1], {1,0}], #]
    ]&[Characters[msg <> " "]/.
         {"[" -> 1, "{" -> 1, "(" -> 1, "\(" -> 1, "\)" -> -1, ")" -> -1,
          "}" -> -1, "]" -> -1, " " -> Null, 
          "\t" -> Null, "\n" -> Null, "\r" -> Null,
           _String?(35839 >= First[ToCharacterCode[#]] >= 12228 &) -> Null,
           _String -> 0}
  ]

(* fallback -- just return the name of the function *)
getTemplateFromUsage[_, name_String] := name

splitUsage[usage_String] := 
With[{marker = "<USAGESPLITMARKER>"},
	Select[StringSplit[StringReplace[
			(* This would be simpler if Except[{"\*\n", ",\n"}, "\n"] worked in StringSplit *)
			StringReplace[usage, {"\n" -> marker}],
			{("/*"<>marker) -> "\*\n", (","<>marker) -> ",\n"}], marker],
		!StringMatchQ[#, WhitespaceCharacter...]&] ]

splitUsage[usage_List, template_List, name_] :=
	splitUsage[{First[usage]}, Rest[usage], Rest[template], name]

splitUsage[{new___, a_}, {usage_, b___}, {template_, c___}, name_] :=
	splitUsage[
		If[template =!= name, {new, a, usage}, {new, a <> "\n" <> usage}], {b}, {c}, name]
		
splitUsage[new_, {}, {}, name_] := new;

$nonArg = ":>" | "->" | "=" | "==" | "::" | "\[Rule]" | "\[RuleDelayed]" | "\[LeftAssociation]" | "\[RightAssociation]";
nonargQ[s_] := StringMatchQ[s, RegularExpression["[[:punct:]]|\\s"] | $nonArg]
systemQ[s_] := StringMatchQ[Quiet[Check[Context[s], ""]], "System`"]
lowcaseQ[s_] := StringMatchQ[s, RegularExpression["[\[Alpha]\[Beta]\[Gamma]\[Delta]\[Epsilon]\
\[CurlyEpsilon]\[Zeta]\[Eta]\[Theta]\[CurlyTheta]\[Iota]\[Kappa]\
\[CurlyKappa]\[Lambda]\[Mu]\[Nu]\[Xi]\[Omicron]\[Pi]\[CurlyPi]\[Rho]\
\[CurlyRho]\[Sigma]\[FinalSigma]\[Tau]\[Upsilon]\[Phi]\[CurlyPhi]\
\[Chi]\[Psi]\[Omega]]+"] | RegularExpression["[[:lower:]]([[:alnum:]]*)"] | "\[Ellipsis]" | "\"*\""]
argumentQ[s_] := lowcaseQ[s] || Not[nonargQ[s] || systemQ[s]]

(* special case needed for CodeAssist FunctionTemplates. CodeAssist wraps the box structure in a BoxData Cell.
   The boxes are extracted from BoxData before the TagBoxes are applied.
*)
insertTagBox[clXpr:Cell[BoxData[boxes_],___], name_String] := Block[{boxestagged=insertTagBox[boxes, name]},
  {clXpr /. boxes -> First@boxestagged, Last@boxestagged}
]

(* if template does not begin with function name, assume that it is in operator form *)
insertTagBox[boxes_, name_String] := 
 insertTagBox[boxes, name, 
  If[MatchQ[boxes, _[{name, __}]], argumentQ, lowcaseQ]] /.
	{ ":>" /; !StringMatchQ[name, "RuleDelayed"] -> "\[RuleDelayed]", 
	  "->" /; !StringMatchQ[name, "Rule"] -> "\[Rule]" }

(* tag all strings which match the pattern regex *)
insertTagBox[boxes_, name_String, regex_] := 
 Block[{pos = After},
     {boxes //. {Except[TagBox, h_][x___, 
        Except[name, s_String], y___] :> 
       h[x, TagBox@s, y] /; regex[s]} //. {h_[s : {TagBox[_] ..}] :> 
      h /@ Thread[s, TagBox], 
     s : (SubscriptBox | SuperscriptBox)[TagBox[_], _] :> 
      TagBox[s /. TagBox[h_] :> h] /; 
        !StringMatchQ[name, "Subscript"], 
     TagBox[s_] :> (pos = Placeholder;
       TagBox[FrameBox@s, "Placeholder"])}, pos}]

reparseBoxStructure[boxes_, name_] := 
 MathLink`CallFrontEnd[FrontEnd`ReparseBoxStructurePacket[boxes]] /. 
  {RowBox[{x_, "^", y_}] /; !StringMatchQ[name, "Power"] :> 
	SuperscriptBox[x, y],
   s : _[{"Subscript", __}] /; !StringMatchQ[name, "Subscript"] :> 
    MakeBoxes @@ MakeExpression@s, 
   s_String /; StringMatchQ[s, "\"Subscript[*]\""] :> 
    MakeBoxes @@ MakeExpression@Evaluate[StringTake[s, {2, -2}]]}

makePlainText[usage_] :=
 Block[{template = usage},
  template = Cell[BoxData[template], ShowStringCharacters -> False];
  template = MathLink`CallFrontEnd[FrontEnd`ExportPacket[template, "PlainText"]];
  If[template === $Failed, usage, First@template]]

fixUp[template_, name_String]:=
	insertTagBox[template //. {StyleBox[x_,y_]:>x, GridBox[{{x_}}]:>x}, name];

GenerateFunctionTemplate[name_String]:=
Block[{usage, usages, templates, strippedName}, 
  strippedName = Last[StringSplit[name, "`"]];
  If[!SymbolQ[name], Return[]];
  usage = ToExpression[name <> "::usage"];
  If[! StringQ[usage], Return[]];
  If[usage == "", usage = "No usage message present when building template."];
  usages = splitUsage[usage];
  templates = getTemplateFromUsage[#, name] & /@ usages;
  
  (* combine clauses of the usage message that have no custom template *)
  usages = splitUsage[usages, templates, getTemplateFromUsage[name, name]];
  templates = getTemplateFromUsage[#, strippedName] & /@ usages;
  templates = If[!FreeQ[templates , $Failed], Return[], makePlainText[#]& /@ templates];
  templates = If[!FreeQ[templates , $Failed], Return[], reparseBoxStructure[#, name]& /@ templates];
  templates = Map[Cell[#, "InlineFormula", Rule[CellTags, name <>"_templates"]]&, BoxData /@ GridBox /@ List /@ List /@ templates];
  templates = fixUp[#, name]&/@templates;
  templates[[All,1]]
]


End[]
EndPackage[]
