#!/Applications/Mathematica.app/Contents/MacOS/WolframScript -script

(* teamcity logs for web builds *)
dir = "/Users/billw/wri2/logs/";
manifest = Sort @ FileNames["*.log", dir, 1];

(* relative standard deviation: how large is SD in relation to the data? the SD
is rSD% of the Mean of the data.
https://www.statisticshowto.com/relative-standard-deviation/ *)

rSD[list_List] := 100*StandardDeviation[list]/Mean[list];

(* * step 1: harvest data from logs *)

res = Reap[Module[{log, data, machine, items},
  log = #;
  data = Import[log];
  machine = Quiet @ Check[First[StringCases[data, "Started " ~~ Shortest[__] ~~ " on agent '" ~~ Shortest[m:__] ~~ "'" -> m]], "undetermined"];
  (* split each log into one record for each file and harvest its cvs filename
  and build time.  we need the cvs filename because when the build manifest is
  partitioned in MakeMathematicaHTML1.m, the manifest consists of cvs filenames.
  The sorted list to be partitioned here needs to match the sorted list there so
  this partitioning will be valid for that list of files. *)

  items = StringSplit[data, "[for] sequential"];

  Module[{item, file, time}, Catch[
    item = #;

    file = Quiet @ Check[First[
      StringReplace[
        StringCases[item, "webinFile: " ~~ f:Except[WhitespaceCharacter].. -> f]
      , ___ ~~ "/Mathematica/Documentation/" -> "/"
      ]
    ], Throw[$Failed]];

    file = StringReplace[
      file
    , $PathnameSeparator ~~ Alternatives["English", "Japanese", "ChineseSimplified"] ~~ $PathnameSeparator -> $PathnameSeparator<>"Language"<>$PathnameSeparator
    ];

    time = Quiet @ Check[First[StringCases[item, "ExportWebPageFromResultNB timing: " ~~ t:NumberString.. -> t]], $Failed];

    If[StringQ[file] && StringQ[time],
      Sow[Rule[ToExpression[time], {machine, log}], StringReplace[file, "\"" -> ""]]
    ];

 ]]& /@ items;
]& /@ manifest, _, Rule][[2]];

res[[All, 2]] = Flatten /@ res[[All, 2]];

Print["Exporting res.m and res-lines.m"];
Export["res.m", res];
Export["res-lines.m", res, "Lines"];

(* res entry format:

"/Language/System/ReferencePages/TextContents/Tunnel.nb" -> {
  29.152 -> {"mac-doc-9", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_099.log"},
  17.6616 -> {"mac-doc-14", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_101.log"},
  16.4996 -> {"mac-doc-3", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_102.log"},
  16.0251 -> {"mac-doc-2", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_103.log"},
  15.977 -> {"mac-doc-8", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_104.log"},
  16.6387 -> {"mac-doc-3", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_105.log"},
  14.7455 -> {"mac-doc-3", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_107.log"},
  12.4467 -> {"mac-doc-9", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/English_Reference_-_Part_08_109_Web_Build.Reference.Part-08.English_13.0.0_7498471_.log"},
  15.3436 -> {"mac-doc-2", "/wri/logs/WolframLanguage_Web_English_Reference/logs-mean-partitioning/Japanese_Reference_-_Part_07_20_Web_Build.Reference.Part-07.Japanese_13.0.0_7523980_.log"}
}
*)

(* res = Import["res.m"]; *)

(* * step 2: compute data for analysis of each file's list of timings *)

(* for each file's list of build times, compute the mean, various clipped means
to reduce the effect of outliers, and the relative standard deviation.  My
current choice for the mean is WinsorizedMean[list, 0.49]. *)

res2 = Reap[Module[{pagedata, page, timings, relativeSD},
  pagedata = #;
  page = pagedata[[1]];
  timings = pagedata[[2, All, 1]];
  If[
    Length[timings] == 1
  ,
    Sow["Timings" -> timings, page]
  ,
    Sow[{
      "Timings" -> timings
    , "Mean" -> Mean[timings]
    , "TrimmedMean (0.1)" -> TrimmedMean[timings, 0.1]
    , "TrimmedMean (0.2)" -> TrimmedMean[timings, 0.2]
    , "TrimmedMean (0.3)" -> TrimmedMean[timings, 0.3]
    , "TrimmedMean (0.4)" -> TrimmedMean[timings, 0.4]
(* the second argument for the clipped Means must be < 0.5 *)
    , "TrimmedMean (0.49)" -> TrimmedMean[timings, 0.49]
    , "WinsorizedMean (0.1)" -> WinsorizedMean[timings, 0.1]
    , "WinsorizedMean (0.2)" -> WinsorizedMean[timings, 0.2]
    , "WinsorizedMean (0.3)" -> WinsorizedMean[timings, 0.3]
    , "WinsorizedMean (0.4)" -> WinsorizedMean[timings, 0.4]
    , "WinsorizedMean (0.49)" -> WinsorizedMean[timings, 0.49]
    , "RelativeStandardDeviation" -> rSD[timings]
    }, page]
  ];
]& /@ res, _, Rule][[2]];

res2[[All, 2]] = (Flatten /@ res2[[All, 2]]);

Print["Exporting res2.m and res2-lines.m"];
Export["res2.m", res2];
Export["res2-lines.m", res2, "Lines"];

(* res2 entry format:

"/Language/System/ReferencePages/Symbols/Abs.nb" -> {
  "Timings" -> {140.513, 126.538, 127.014, 147.657, 97.9644, 150.682, 90.5107, 162.012, 161.508, 113.262, 112.336},
  "Mean" -> 129.99973636363637,
  "TrimmedMean (0.1)" -> 130.8304888888889,
  "TrimmedMean (0.2)" -> 131.14314285714286,
  "TrimmedMean (0.3)" -> 130.9968,
  "TrimmedMean (0.4)" -> 131.355,
  "TrimmedMean (0.49)" -> 127.014,
  "WinsorizedMean (0.1)" -> 130.63152727272728,
  "WinsorizedMean (0.2)" -> 131.2761818181818,
  "WinsorizedMean (0.3)" -> 130.70372727272726,
  "WinsorizedMean (0.4)" -> 132.93354545454545,
  "WinsorizedMean (0.49)" -> 127.01400000000001,
  "RelativeStandardDeviation" -> 18.946121702446742
}
*)

(* res2 = Import["res2.m"]; *)

(* * step 3: associate each filename with the mean of its build times *)

res3 = Reap[Module[
  {file,data,wmean}
,
  file = First[#];
  data = Last[#];   
  wmean = "WinsorizedMean (0.49)" /. data;

  If[wmean == "WinsorizedMean (0.49)", wmean = Ceiling[data[[1, 2, 1]]]];
  Sow[Ceiling[wmean], file]
]& /@ res2, _, Rule][[2]]

res3[[All, 2]] = (First /@ res3[[All, 2]]);
res3 = Sort@res3;

Print["Exporting build-time-db.m and build-time-db-lines.m"];
Export["build-time-db.m", res3];
Export["build-time-db-lines.m", res3, "Lines"];
  
(* res3 entry format:
"/Language/System/ReferencePages/Symbols/WebColumn.nb" -> 83 *)

Quit[]
======================================================================

(* * fragments of commentary *)

the statistics for files with small means are irrelevant - the mean will suffice
perfectly well as the partitioning runtime for the file.

files with a small relative standard deviation don't differ much in their
runtimes, so the mean will work for those, too.

So files with higher rSDs are a problem when partitioning builds - what should
be used as their mean build time for partitioning?

For sets with large outliers, Mean provides a number that is less
"representative" of the main body of the set.  For timing sets with large
relative standard deviations, a trimmed or winsorized mean provides a more
representative number.  See https://en.wikipedia.org/wiki/Winsorizing

so... how many files have a Mean > 20 seconds and a rSD of over 50%?

112:

In[73]:= y = Select[res, Module[{data, mean, rsd}, Quiet[data = Last[#]; mean = "Mean" /. data; rsd = "RelativeStandardDeviation" /. data; mean > 20 && rsd > 50]]&]; y//Length

Out[73]= 112

sample those:

In[74]:= RandomChoice[y, 3]//format

Out[74]= {
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/workflow/EvaluateAnExpressionInPlace.html.en" ->
             {
               "Timings" -> {11.2679, 82.9954, 11.7607, 14.1241, 7.94424, 5.46686, 10.9672},
               "Mean" -> 20.64662857142857,
               "TrimmedMean (0.1)" -> 20.64662857142857,
               "TrimmedMean (0.2)" -> 11.212827999999998,
               "TrimmedMean (0.3)" -> 11.331933333333334,
               "WinsorizedMean (0.1)" -> 20.646628571428568,
               "WinsorizedMean (0.2)" -> 11.161782857142855,
               "WinsorizedMean (0.3)" -> 11.350228571428572,
               "RelativeStandardDeviation" -> 133.84957770792673
             },

it makes sense to allocate 11 seconds [WinsorizedMean (0.3)] instead of 20 [Mean] for workflow/EvaluateAnExpressionInPlace

           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/ref/entity/WeightTrainingExercise.html.en" ->
             {
               "Timings" -> {59.4148, 60.0905, 64.3677, 53.5166, 52.8102, 44.0263, 157.675, 142.946},
               "Mean" -> 79.3558875,
               "TrimmedMean (0.1)" -> 79.3558875,
               "TrimmedMean (0.2)" -> 72.19096666666667,
               "TrimmedMean (0.3)" -> 59.34740000000001,
               "WinsorizedMean (0.1)" -> 79.3558875,
               "WinsorizedMean (0.2)" -> 78.61275,
               "WinsorizedMean (0.3)" -> 59.144774999999996,
               "RelativeStandardDeviation" -> 55.93433206932678
             },
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/tutorial/DSolveClassificationOfDifferentialEquations.html.en" ->
             {
               "Timings" -> {74.9577, 29.9377, 26.0983, 42.5137, 31.9604, 27.8192, 17.0053, 17.3985},
               "Mean" -> 33.461349999999996,
               "TrimmedMean (0.1)" -> 33.461349999999996,
               "TrimmedMean (0.2)" -> 29.287966666666662,
               "TrimmedMean (0.3)" -> 28.953899999999997,
               "WinsorizedMean (0.1)" -> 33.461349999999996,
               "WinsorizedMean (0.2)" -> 29.455000000000002,
               "WinsorizedMean (0.3)" -> 28.991625,
               "RelativeStandardDeviation" -> 55.72070561446255
             }
         }

the scheme doesn't make much difference here - 29sec vs 33sec

what about a mean > 40 seconds with the same rSD > 50%?

In[75]:= y = Select[res, Module[{data, mean, rsd}, Quiet[data = Last[#]; mean = "Mean" /. data; rsd = "RelativeStandardDeviation" /. data; mean > 40 && rsd > 50]]&]; y//Length

Out[75]= 78

In[76]:= RandomChoice[y, 3]//format

Out[76]= {
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/tutorial/NDSolveVectorNorm.html.en" ->
             {
               "Timings" -> {94.6492, 44.1488, 27.974, 50.3879, 34.3843, 35.2318, 22.6583, 22.5418},
               "Mean" -> 41.497012500000004,
               "TrimmedMean (0.1)" -> 41.497012500000004,
               "TrimmedMean (0.2)" -> 35.79751666666667,
               "TrimmedMean (0.3)" -> 35.434725,
               "WinsorizedMean (0.1)" -> 41.4970125,
               "WinsorizedMean (0.2)" -> 35.9789125,
               "WinsorizedMean (0.3)" -> 35.7480625,
               "RelativeStandardDeviation" -> 56.867388173944086
             },
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/ref/entity/FiniteGroup.html.en" ->
             {
               "Timings" -> {46.3679, 13.1572, 13.0737, 35.4381, 43.0751, 37.0768, 119.511, 124.352},
               "Mean" -> 54.006474999999995,
               "TrimmedMean (0.1)" -> 54.006474999999995,
               "TrimmedMean (0.2)" -> 49.10434999999999,
               "TrimmedMean (0.3)" -> 40.489475,
               "WinsorizedMean (0.1)" -> 54.006475,
               "WinsorizedMean (0.2)" -> 53.41178749999999,
               "WinsorizedMean (0.3)" -> 40.696237499999995,
               "RelativeStandardDeviation" -> 80.99155577154815
             },
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/tutorial/Calculus.html.en" ->
             {
               "Timings" -> {953.853, 451.783, 414.63, 684.348, 486.41, 404.223, 222.391, 226.829},
               "Mean" -> 480.558375,
               "TrimmedMean (0.1)" -> 480.558375,
               "TrimmedMean (0.2)" -> 444.7038333333333,
               "TrimmedMean (0.3)" -> 439.2615,
               "WinsorizedMean (0.1)" -> 480.558375,
               "WinsorizedMean (0.2)" -> 447.425,
               "WinsorizedMean (0.3)" -> 442.28900000000004,
               "RelativeStandardDeviation" -> 50.21384360973501
             }
         }

only 78 files have a Mean > 40 seconds with rSD > 50%, and might need special
handling to reduce the effect of outliers on the Mean.

at any rate, this shows me how to refine build partitioning.

In the past, I've used the timings in the most recently completed build logs as
allocations for a new round of partitioning.  These timings can vary wildly: the
timings of 3000 files (20% of the files) have a relative standard deviation >
30%, 800 have an rSD > 40%.  If some of those happen to have large outliers in
the most recent run, partitioning based on those numbers will be worthless.  As
we've seen - a fresh partitioning just results in different builds timing out.

Will the Mean of a set of timings suffice, or should a clipped mean be brought
into action when it might make a difference?  In other words, how much should
large outliers be allowed to affect partitioning?

I don't have a complete set of numbers to support this, but I'm leaning toward a
clipped mean when it will contribute a significant reduction of outlier
influence.  What is "significant"?  Perhaps when the Mean and [TrimmedMean (0.3)
or WinsorizedMean (0.3)] differ by a meaningful amount.

Also look into even stronger clipping of means.

Mean is less useful (or more misleading) as rSD increases and as the Mean itself
increases.  This doesn't matter at all for files with mean runtimes < 10
seconds.  Maybe consider it for files with

mean > 30 seconds: 4000

mean > 30 seconds and rsd > 25%: 1424 files

mean > 60 seconds: 2200 files

mean > 60 seconds and rsd > 25%: 747 files

let's go with mean > 30 and rsd > 25%.  look at those files, see what effect
clipping has on the various kinds of means.

what about files for which Mean minus (a clipped mean) is greater than, say, 30
seconds?

In[14]:= y = Select[res, Module[{d,m,r,t,w}, Quiet[d=Last[#];
{m,r,t,w} = {"Mean","RelativeStandardDeviation","TrimmedMean (0.49)","WinsorizedMean (0.49)"}/.d;
m-t > 30]]&]; y//Length
                  
Out[14]= 75

In[15]:= RandomChoice[y, 3]//format

Out[15]= {
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/tutorial/RealPolynomialSystems.html.en" ->
             {
               "Timings" -> {897.738, 375.544, 365.919, 610.715, 455.063, 424.598, 281.467, 281.782},
               "Mean" -> 461.60325000000006,
               "TrimmedMean (0.1)" -> 461.60325000000006,
               "TrimmedMean (0.2)" -> 418.93683333333337,
               "TrimmedMean (0.3)" -> 405.28100000000006,
               "TrimmedMean (0.4)" -> 400.071,
               "TrimmedMean (0.49)" -> 400.071,
               "WinsorizedMean (0.1)" -> 461.60325,
               "WinsorizedMean (0.2)" -> 425.76475000000005,
               "WinsorizedMean (0.3)" -> 407.88599999999997,
               "WinsorizedMean (0.4)" -> 400.071,
               "WinsorizedMean (0.49)" -> 400.071,
               "RelativeStandardDeviation" -> 44.488218991095295
             },
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/ref/Plot3D.html.en" ->
             {
               "Timings" -> {654.057, 425.94, 390.396, 495.82, 524.597, 389.055, 410.756, 659.907},
               "Mean" -> 493.816,
               "TrimmedMean (0.1)" -> 493.816,
               "TrimmedMean (0.2)" -> 483.5943333333333,
               "TrimmedMean (0.3)" -> 464.27825,
               "TrimmedMean (0.4)" -> 460.88,
               "TrimmedMean (0.49)" -> 460.88,
               "WinsorizedMean (0.1)" -> 493.81600000000003,
               "WinsorizedMean (0.2)" -> 493.25237500000003,
               "WinsorizedMean (0.3)" -> 465.977375,
               "WinsorizedMean (0.4)" -> 460.88,
               "WinsorizedMean (0.49)" -> 460.88,
               "RelativeStandardDeviation" -> 22.62666649826936
             },
           "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/tutorial/ComplexPolynomialSystems.html.en" ->
             {
               "Timings" -> {1342.61, 625.869, 302.548, 524.466, 401.566, 472.127, 284.008, 285.77},
               "Mean" -> 529.8705,
               "TrimmedMean (0.1)" -> 529.8705,
               "TrimmedMean (0.2)" -> 435.391,
               "TrimmedMean (0.3)" -> 425.17674999999997,
               "TrimmedMean (0.4)" -> 436.8465,
               "TrimmedMean (0.49)" -> 436.8465,
               "WinsorizedMean (0.1)" -> 529.8705,
               "WinsorizedMean (0.2)" -> 440.498125,
               "WinsorizedMean (0.3)" -> 419.34187499999996,
               "WinsorizedMean (0.4)" -> 436.8465,
               "WinsorizedMean (0.49)" -> 436.8465,
               "RelativeStandardDeviation" -> 66.19990736888309
             }
         }


only 75 files.

Use the mean of many past build times rather than the times from the most recent
build.  but to reduce the effect of outliers (often due to a transient problem
with a build machine or the m- installed on it), use a clipped mean.
WinsorizedMean[timings, 0.49] looks good in the data I've collected.  It's the
farthest away from the Mean when large outliers are present, yet it's still a
meaningful sort of Mean.

So here are two refinements to partitioning:

 - rather than use runtimes for each file in the last build, use the mean
   runtime for each file from the last 10 or so builds.

 - but to filter out the effect of outliers due to transient problems with build
   machines, use a clipped mean.  TrimmedMean[timings, 0.49] provides reasonable
   results.

----------------------------------------------------------------------

y = Select[
  res
, Module[
    {element, data, mean, rsd}
  , Quiet[
      element = #;
      data = Last[element];
      mean = "Mean" /. data;
      rsd = "RelativeStandardDeviation" /. data;
      rsd > 40 && mean > 50
    ]
  ]&
];

In[67]:= y = Select[res, Module[{element, data, mean, rsd}, Quiet[element = #; data = Last[element]; mean = "Mean" /. data; rsd = "RelativeStandardDeviation" /. data; rsd > 40 && mean > 50]]&]; y//Length
Out[67]= 144

In[68]:= RandomChoice[y]//format
Out[68]= "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/ref/WaveletFilterCoefficients.html.en" ->
           {
             "Timings" -> {59.6355, 96.7749, 38.2094, 36.0953, 75.7058, 55.8334, 45.1357, 30.1206},
             "Mean" -> 54.688825,
             "TrimmedMean (0.1)" -> 54.688825,
             "TrimmedMean (0.2)" -> 51.76918333333333,
             "TrimmedMean (0.3)" -> 49.7035,
             "WinsorizedMean (0.1)" -> 54.688825,
             "WinsorizedMean (0.2)" -> 52.802025,
             "WinsorizedMean (0.3)" -> 49.312975,
             "RelativeStandardDeviation" -> 41.18323442254701
           }

In[69]:= RandomChoice[y]//format
Out[69]= "/Volumes/TeamCity/work/1e95b97cf0aa68fc/output/Files/ref/FunctionAnalytic.html.en" ->
           {
             "Timings" -> {187.644, 296.291, 177.979, 387.302, 208.7, 184.373, 500.603, 305.226},
             "Mean" -> 281.01475,
             "TrimmedMean (0.1)" -> 281.01475,
             "TrimmedMean (0.2)" -> 261.58933333333334,
             "TrimmedMean (0.3)" -> 249.46525,
             "WinsorizedMean (0.1)" -> 281.01475000000005,
             "WinsorizedMean (0.2)" -> 267.65137500000003,
             "WinsorizedMean (0.3)" -> 247.950125,
             "RelativeStandardDeviation" -> 41.28927844562774
           }
