(* 
    Title: Layout.m
    $Id: Layout.m,v 1.9 2007/07/23 20:19:34 shaunm Exp $

    Context: Transmogrify`Layout`

    Authors: stephen layland 

    Copyright: © February 2005 by stephen layland

    Requirements: 
    Mathematica 5.1 or later for RegularExpression capabilities
    
    Resources: 
    Friedl, Jeffrey E. F. "Mastering Regular Expressions"    

    TODO:

    o  Fix block scoping implementation.  there's probably a better way.

    o  clean up implementation, perhaps break up code to two separate private
       functions, one for Strings the other for XML.
*)

(* -------------------------------------------------------------------------- *)
(*  this was originally written for the Transmogrify nb->XML converter, but   *)
(*  it can be used for any number of generic text based layouts. See        *)
(*  documentation in help browser                                             *)
(* -------------------------------------------------------------------------- *)

(* changed from BeginPackage to create a subcontext of Transmogrify` *)
Begin["`Layout`"];


(* ----------------------------- *)
(*   public methods              *)
(* ----------------------------- *)

(* export usage statements to Transmogrify` context *)
(* <!-- Public 
FillInLayout::usage = "FillInLayout is a utility which allows you to create a\
 structure based file of your choosing, and interpolate specially formatted text\
 as arbitary Mathematica code.  The following different methods exist for interpolating\
 variables from the default scope, from a list, or from a context, respectively:\n
 FillInLayout[\"layoutfile\",\"format\",opts]\n\
 FillInLayout[\"layoutfile\",\"format\",{varlist},opts]\n\
 FillInLayout[\"layoutfile\",\"format\",\"context`\",opts],\n\
 where {varlist} is of the form {\"var1\"->var1,...}, or {\"var1\", \"var2\"->val2,...}.";
 
EmbeddedCodeRule::usage = "This options is a rule that will convert either a\
 string, or a Mathematica expression, of your choice into a _String_ that denotes\
 arbitrary Mathematica code that will be passed off to ToExpression.  Setting this to\
 Automatic causes FillInLayout to recognize <!-- {mcode} --> as Mathematica code for all\
 files imported as SymbolicXML (default), and uses ${mcode} for regular text files. See\
 FillInLayout.nb for examples.";
-->*)

(* ----------------------------- *)
(*   BEGIN IMPLEMENTATION        *)
(* ----------------------------- *)

Begin["`Private`"]

(* default options *)

Options[FillInLayout] = {
    ConversionOptions->Automatic,
    EmbeddedCodeRule->Automatic
}

(* error messages *)
FillInLayout::error = "``";


(* FillInLayout["filename", "format", opts] *)
(* this is a Block[] because it can get variables defined in the scope
   that FillInLayout[] was called in *)

FillInLayout[
	fn_String,
	formatin_String : "",
	varlist_List : {},
	opts___?OptionQ
] := Module[{
	format = formatin,
	convopts,
	structure,
	templateParameters,
	vars,
	rule
},
  Catch[
    (* snag known options *)
    {convopts, rule} = {ConversionOptions, EmbeddedCodeRule} /. {opts} /.
        Options[FillInLayout];

	If[rule =!= Automatic,
		Message[FillInLayout::error, "FillInLayout: use of obsolete EmbeddedCodeRule option."];
		Throw[$Failed, FillInLayout];
	];

    (*--- get "varN" \[Rule] valN options for replacement ---*)
    (*
       basically, any Rule that is of the form "string" goes to something
       these will take precedence over any of the same variables in varlist
    *)

    vars = Cases[
		Flatten @ {opts},
		HoldPattern[(Rule | RuleDelayed)[_String, _]]
	];

	vars = Join[vars, varlist];

    (* If we didn't get any vars passed in, [obsolete:] try to get stuff from
	   FillInLayout's scope *)
	If[Length[vars] === 0,
		Message[FillInLayout::error, "FillInLayout: obsolete empty variable list behavior."];
		Throw[$Failed, FillInLayout];
	];

	(* otherwise use only the variables we are asking for that are passed
	   in as "var" -> val options or as a varlist argument *)

	(* first format all the var types we have *)
	Module[{badvars},
		badvars = Cases[vars, Except[_?StringQ ~(Rule|RuleDelayed)~ _?StringQ]];

		If[Length[badvars] > 0,
			Message[FillInLayout::error, ToString[StringForm @@ {
				"FillInLayout: variable -> value must both be strings: ``",
				InputForm[badvars]
			}]];
			Throw[$Failed, FillInLayout];
		];
	];

	Assert[MatchQ[vars, {RepeatedNull[_?StringQ ~(Rule|RuleDelayed)~ _?StringQ]}]];

	(* "varname" -> "val" *)
	vars = Association[vars];

	Assert[AssociationQ[vars]];

    (* use file extension if format wasn't given *)
	If[format === "",
		format = StringReplace[fn, RegularExpression@".*\\.(.+)$" -> "$1"]
	];

    (* set up automatic conversion options depending on format *)

    If[convopts === Automatic,
      convopts =
        If[
          !StringFreeQ[format, RegularExpression@"ml$", IgnoreCase->True],
          {"IncludeEmbeddedObjects" -> "Comments"}
          ,
          {}
        ]
    ];

    structure = Import[fn, format, Sequence @@ convopts];

    If[structure === $Failed,
		Throw[$Failed]
	];

    (* handle string replacements *)

	(* rule to turn ${mcode} into mcode.  i love regexes! *)
	rule = RegularExpression[
  "(?x)
     (?<!\\\\)\\${                   # match an opening ${ not preceded by a slash   \n
     (                               # and start capturing the stuff that follows    \n
         (?:[^\\\\\\n](?!\\${))*     # then any non \ or newline character that is not\n
         (?:                         #    followed by an escaped ${                  \n
             \\\\                    # now match any \ or newline that we got to earlier \n
             (?:[^\\\\\\n](?!\\${))* # and any amount of valid embedded code that follows\n
         )*                          # as many times as you can                      \n
     )
     }                               # lastly, make sure it ends with a closing } \n
  "
	] :> "$1";

	(*
		Create a mapping of positions of template parameters appearing in `structure` to
		the name of template parameter. E.g.:

			<|
				{start_, end_} -> "${paramFoo}",
				{_, _} -> "${paramBar}",
				...
			|>
	*)
	templateParameters = AssociationMap[
		StringTake[structure, #] &,
		StringPosition[structure, First @ rule]
	];

	(* Remove the ${..} wrapper. E.g.:
		<| {_, _} -> "paramFoo", {_, _} -> "paramBar", ...|> *)
	templateParameters = Map[StringReplace[rule], templateParameters];

	(* Resolve the named template parameters to the user provided values. *)
	templateParameters = Map[
		Function[param,
			Lookup[vars, param, (
				(* Throw a hard error if the caller has not provided a value for a
				   template parameter. *)
				(* This makes it more difficult to accidentally delete a template parameter
				   without noticing content was missing from the filled in template. *)
				Message[
					FillInLayout::error,
					ToString[StringForm @@ {
						"FillInLayout: Error processing layout '``': "
							<> "no value for template parameter '``'",
						fn,
						param
					}]
				];
				Throw[$Failed, FillInLayout]
			)]
		],
		templateParameters
	];

    (* leave now if we didn't find anything to replace *)
	If[templateParameters === {},
		Throw[structure]
	];

	(* Substitute template parameters for template values in the original string. *)
	structure = StringReplacePart[
		structure,
		Values @ templateParameters,
		Keys @ templateParameters
	];

    structure
  ]
];

FillInLayout[
	fn_String,
	format_String,
	c_String /; !StringFreeQ[c, RegularExpression["`\\*?$"]],
	opts___?OptionQ
] := (
	Message[FillInLayout::error, "Unsupported call to obsolete variant of FillInlayout[..]"];
	Throw[$Failed, FillInLayout]
);

(* FillIn and export to file *)
FillInLayout[outfile_String, layoutfile_String, format_String, opts___?OptionQ]:=
  Export[
    outfile,
    FillInLayout[layoutfile,format,opts],
    format, "AttributeQuoting"->"\""
  ];

End[] (*`Private`*)

End[] (*`Layout`*)

