(* Wolfram Language package *)

CoveringArray::usage = "CoveringArray[{{val11,val12,...},{val21,val22,...},...}] constructs a parameter test set where all pairwise parameter interactions occur.
CoveringArray[{{val11,...},...}, n] constructs a parameter test set where all n-wise parameter interactions occur.";


Begin["`CoveringArrays`"]


(* Check if two parameter interactions are comptible, e.g. {1,_,2} and {_,3,2}*)
ClearAll[compatibleQ];
compatibleQ[q_List, r_List] := Module[{len = Length[q]},
  And @@ (q[[#]] === r[[#]] || q[[#]] === _ || r[[#]] === _ & /@ Range[len])
  ]

(* Combine two compatible parameter interactions *)
ClearAll[combine];
combine[q_List, r_List] := Module[{res = q},
   If[((res[[#]] === _) && IntegerQ[r[[#]]]), res[[#]] = r[[#]], 
      res[[#]] = q[[#]]] & /@ Range[Length[res]]
   ] /; compatibleQ[q, r]

(* Check if r would cover q *)
ClearAll[subsetQ];
subsetQ[q_List, r_List] := 
  compatibleQ[q, r] && (SubsetQ[Position[q, Verbatim[_]], Position[r, Verbatim[_]]]);

(* check if q is uncovered in P *)
ClearAll[uncoveredQ];
uncoveredQ[q_, P_, Pcoverage_] := 
	MemberQ[Flatten[Pick[P, Pcoverage, 1], 1], Verbatim@q]

(* Update coverage matrix for new test case r *)
ClearAll[updateCoverageList];
updateCoverageList[r_, P_, Pcoverage_] :=
    Module[ {indices, newPcoverage},
        newPcoverage = Pcoverage;
        indices = Position[Map[subsetQ[#, r] &, #] & /@ P, True];
        If[ Length[indices] > 0,
            (Part[newPcoverage, Sequence @@ #] = 0) & /@ indices
        ];
        newPcoverage
    ]

(* Verify the correctness of a covering array *)
ClearAll[verifyCoverage];
verifyCoverage[groupLengths_, R_, n_:2] :=
    Module[ {groupCount, groupIndices, tuples, P},
        groupCount = Length[groupLengths];
        groupIndices = Range /@ groupLengths;
        tuples = Subsets[Range[groupCount], {n}];
        P = Tuples[Fold[ReplacePart, Table[{_}, groupCount], Map[# -> groupIndices[[#]] &, #]]] & /@ tuples;
        P = Flatten[P, 1];
        And @@ (MemberQ[R, #] & /@ P)
    ]


IndexArray[groupLengths_List, n_Integer:2] :=
    Module[ {groupCount, groupIndices, tuples, P, Pcoverage, R, r, p, Q, Qcoverage, existsUncovererd, u, adds, choice, uncoveredLengths, potential},
    	groupCount = Length[groupLengths];
        groupIndices = Range /@ groupLengths;
        If[n===groupCount,
        	(* Corner case: return all possible interactions *)
        	Return[Flatten[Outer[List, Sequence @@ groupIndices], groupCount-1]]
        ];
        tuples = Subsets[Range[groupCount], {n}];
        (* Construct matrix P of all parameter interactions to cover *)
        P = Tuples[Fold[ReplacePart, Table[{_}, groupCount], Map[# -> groupIndices[[#]] &, #]]] & /@ tuples;
        Pcoverage = ConstantArray[1, #] & /@ Length /@ P; (* 1=uncovered, 0=covered *)
        R = {}; (* set of test cases *)
        While[! FreeQ[Pcoverage, 1],(*  while there are uncovered interactions *)
         
         (* Choose a parameter interaction p from P with most uncovered slots
            Pick the first uncovered combination from p *)
       	 uncoveredLengths = Length /@ (Select[# === 1 &] /@ Pcoverage);
         p = First@FirstPosition[uncoveredLengths, Max[uncoveredLengths]];
         r = P[[p, First@FirstPosition[Pcoverage[[p]], 1]]];
         While[! FreeQ[r, Verbatim[_]], (* While there are parameters with no values in r *)
          
          (* Look at subset Q of P that covers at least one parameter with no representation in r
             Look at slots in Q that have values consistent with already chosen values in r *)
          Q = Select[Transpose@{Flatten[P, 1], Flatten[Pcoverage]}, 
          				compatibleQ[r, First[#]] && (! FreeQ[First[#][[Flatten@Position[r, Verbatim[_]]]], _Integer]) &];
          Qcoverage = Q[[All,2]];
          Q = Q[[All,1]];
          
          (* If there exist uncovered combinations:
             Pick a slot with values which when added to r would cover the most uncovered combinations *)
		  existsUncovererd = Total[Qcoverage]>0;
          r = If[ existsUncovererd,
              u = Pick[Q, Qcoverage, 1];
              adds = (potential = combine[r, #];
                      Length@Cases[subsetQ[#, potential] & /@ Flatten[Pick[P, Pcoverage, 1], 1], True]
                      ) & /@ u;
              choice = First@u[[First@Position[adds, Max[adds]]]];
              combine[r, choice],
              (* Else: Pick randomly a covered combination to add to r *)
              SeedRandom[1]; (* TODO : add a RandomSeeding option to control this *)
              choice = RandomChoice[Q];
              combine[r, choice]
          ];
           
          (* Mark the chosen combination in P as covered *)
          Pcoverage = updateCoverageList[r, P, Pcoverage];
          ];
          AppendTo[R, r];
        ];
    	R
    ] /; (n>=2 && n<=Length@groupLengths)

IndexArray[___] := $Failed


CoveringArray[parValues : {__List}, n_Integer:2] :=
    Block[ {groupLengths, selections},
    	If[n<2 || n>Length@parValues || Min[Length/@parValues]===0,
    		Return[$Failed]
    	];
        groupLengths = Length /@ parValues;
        selections = IndexArray[groupLengths, n];
        MapThread[#1[[#2]] &, {parValues, #}] & /@ selections
    ];


End[(* `CoveringArrays` *)]
