
(* 
	
	TestEvaluate is a pure function that takes an association representing user input and returns an association rappresenting a TestResultObject. 
	You decide how to evaluate Input/Output by simply using RuleDelayed, this is much better than using HoldForm because rule is SequenceHold and using Rule here eliminates an entire class of bugs.

	Input will evaluate exactly once, everything that happens during the evaluation (abort, throw catch, will be wrapped).
	ExpectedOutput will also evaluate at most once, however it can be aborted.
	ExpectedOutput might never evaluate if input was aborted.

	TestEvaluate @ <|
		"Input" :> 2+2,
		"ExpectedOutput" :> 4,
		"ExpectedMessages" :> {},
	|>

	This function will ALWAYS return a new association with more computed keys, it can't fail or be aborted, it will return Output and Input WITHOUT HoldForm, VerificationTest will need to add extra layers.

	## Coding style

	The style of programming here is not done by chance. We have 3 requirements here:

	1. Evaluate expressions at most once.
	2. We don't want to create any side effects during the test execution, we should not impact over memory usage.
	3. Construct small user faced immutable expressions that are easy to read / debug. 
	4. The implementation needs to do as little as possibile.
	
	Applying functions in a pipeline is one effective way to archive it.

	## How this function should be used:

	Use this function the moment you want to actually run a test.

	Always feed this function with immutable expressions, because the input of this function should be parallelizable, internally ALWAYS pass to this function immutable expressions and not unevaluated references to symbols.

	This function should be used as the central point of a middleware.
	Call this function with a preprocessor in order to wrap the input in other functions, or with postprocessor to mutate the output.
	
	MUnit`Test in the future can simply use this implementation, the fact that this implementation is PURE means that it can be successfully re-used by MUnit because the implementation is not going to change.

	## What this function should NOT become:

	This function should not be used to hardcode WRI test suite specifics, there are a lot of MUnit` shortcuts that have design for TestOutputWrapper, and other undocumented options. All those preprocessing can be done by simply adding preprocessors that are wrapping input / output.

	In short the implementation of this function should never change, the job of this function is to simply run the test, prevent the test from escaping and collect all side effect caused in the meantime. Anything else must be done using post/pre processors.

*)


(*  short utility function that is computing the approximated execution time using SessionTime. 
	We are not using AbsoluteTime directly at every test run because it's terribly slow:
	1 call of AbsoluteTime is currently accounting for 1/3 of all timing spent in running a VerificationTest.
*)

utcTime := utcTime = Block[
	{$TimeZone = "UTC"},
	Function @@ {#1 + AbsoluteTime[] - SessionTime[]}
]


TestEvaluate[association_?AssociationQ] := 
	$TestEvaluatePostProcessor @ processData[

		(* first we set all default keys, that includes True as input / output of the test and everything we are going to use in the future. *)
		<|
			"TestKey"          -> Inherited,
			"TestID"           -> Inherited,
			"TestFileName"     -> Missing["NotComputed"],
			"SameTest"         -> Inherited,
			"SameMessages"     -> Inherited,
			"MemoryConstraint" -> Inherited,
			"TimeConstraint"   -> Inherited,
			"Outcome"          -> Missing["NotComputed"],
			"Input"            -> HoldForm[True],
			"ExpectedOutput"   -> HoldForm[True],
			"ActualOutput"     -> HoldForm[Missing["NotComputed"]],
			"ExpectedMessages" -> HoldForm[{}],
			"ActualMessages"   -> {},

			(* now we are going to merge the input association with all default values *)
			association
		|>,
		(*  
			At this point we process all user input by replacing Inherited with the right default
			The default is hardcoded in order to prevent other developers (and myself) from being lazy and just block the internals.

			provideDefault will take care of taking the user input, evaluate the result exactly once and prevent Sequence from causing troubles.

			We also overwrite Outcome in case that was already there in the original association
		 *)
		Function @ {
			"SameTest"         -> provideDefault[#SameTest, SameQ],
			"SameMessages"     -> provideDefault[#SameMessages, MessageMatchQ],
			"MemoryConstraint" -> provideDefault[#MemoryConstraint, Infinity],
			"TimeConstraint"   -> provideDefault[#TimeConstraint, Infinity],
			"Outcome"          :> Missing["NotComputed"]
		},
		$TestEvaluatePreProcessor,
		
		Function @ {
			"AbsoluteTime"	   -> 0,	
			"AbsoluteTimeUsed" -> SessionTime[],
			"CPUTimeUsed"      -> TimeUsed[],
			"MemoryUsed"       -> MemoryInUse[]
		},

		(* this is the main logic it will take care of making sure #Input can't escape the test by using CheckAll. 
			rule is used all over the place to prevent Sequence from escaping
		*)
		Function @ {

			CheckAll[
				quietMessages @ extractUnevaluated @ MemoryConstrained[
					extractUnevaluated @ TimeConstrained[
						collectMessages["ActualOutput" -> Map[Evaluate, #Input]],
						#TimeConstraint,
						"ActualOutput" -> constraintFailure[TimeConstrained, HoldForm[#TimeConstraint]]
					], 
					#MemoryConstraint,
					"ActualOutput" -> constraintFailure[MemoryConstrained, HoldForm[#MemoryConstraint]]
				],
				Function @ Replace[
					{##}, {
						{a_, Hold[]} :> a,
						{_, h_Hold} :> {"ActualOutput" :> h, "Outcome" -> "Error"}
					}
				]
			],

			"AbsoluteTimeUsed" -> SessionTime[] - #AbsoluteTimeUsed,
			"CPUTimeUsed"      -> TimeUsed[]    - #CPUTimeUsed,
			"MemoryUsed"       -> MemoryInUse[] - #MemoryUsed,
			"AbsoluteTime"     -> utcTime[#AbsoluteTimeUsed],

			"ExpectedOutput"   -> Map[Evaluate, #ExpectedOutput],

			"TestKey"          -> provideDefault[#TestKey, Hash @ CreateUUID[]],
			"TestID"           -> provideDefault[#TestID, CreateUUID[], Automatic],
			"TestFileName"     -> provideDefault[#TestFileName, $TestFileName]

		},
		Function @ {

			(* now we can calculate the outcome *)
			"Outcome" -> If[
				(* if this was already set, we need to confirm it *)
				StringQ[#Outcome], 
				#Outcome,
				If[
					trueQ @ Apply[#SameTest, Join[#ActualOutput, #ExpectedOutput]],
					If[
						trueQ @ Apply[#SameMessages, Join[HoldForm[#ActualMessages], #ExpectedMessages]],
						"Success",
						"MessagesFailure",
						Message[VerificationTest::uneval, "SameMessages"]; "Error"
					],
					"Failure",
					Message[VerificationTest::uneval, "SameTest"]; "Error"
				]
			]
		}
	]



SetAttributes[constraintFailure, HoldRest]

constraintFailure[h_, constraint_, msg_: VerificationTest::cnsexd] :=
	HoldForm @ Evaluate @ Failure[
		SymbolName[h], <|
			"MessageTemplate" :> msg,
			"MessageParameters"-> <|"Function" :> h, "Constraint" :> constraint|>
		|>
	]

extractUnevaluated[(h:MemoryConstrained|TimeConstrained)[_, c_, ___]] := {
	"ActualOutput" -> constraintFailure[h, HoldForm[c], VerificationTest::cnsnvl],
	"Outcome" -> "Error"
}
extractUnevaluated[h_] := h

(* 
	message collecting logic is dead simple:
	
	1. we collect all messages using HandlerBlock and Internal`Bag
	2. we evaluate the expression by removing OutputStream from $Messages, that should allow other the user to test a custom backend for messages
	3. we need to restore all global variables and prevent TestReport and friends from collecting Tests happening in the first argument of VerificationTest. by doing that this implementation is able to test itself.
*)

trueQ[bool:True|False] := bool

SetAttributes[{collectMessages, quietMessages}, {HoldAllComplete, SequenceHold}]

quietMessages[expr_] :=
	Block[
		{
			$collected = Null,
			$Messages = {},
			$MessageList = {},
			$DefaultOptions = {},
			$TestEvaluatePostProcessor := System`TestResultObject,
			$TestEvaluatePreProcessor := Identity,
			$TestReportPostProcessor := System`TestReportObject,
			$TestReportPreProcessor := Identity
		},
		expr
	]
collectMessages[expr_] := 
	Block[
		{
			$collected = Internal`Bag[]
		}, 
		{
			Internal`HandlerBlock[
				{
					"MessageTextFilter", 
					Function @ If[
						ListQ[$Messages],
						Internal`StuffBag[$collected, HoldForm @@ #3]
					]
				}, 
				StackBegin @ expr
			],
			"ActualMessages" -> Internal`BagPart[$collected, All]
		}
	]


SetAttributes[provideDefault, {HoldRest, SequenceHold}]

provideDefault[s_Sequence, ___] := s
provideDefault[expr_, default_, patt_:Automatic|None|Null|_Missing|Inherited] := If[MatchQ[expr, patt], default, expr]
