
(* 
	TestReport needs to do two thing before evaluating his first argument:
	
	1. Evaluate all options and store them in a global variable so that every time VerificationTest evaluates they are available.
	2. Create a global mode that is collecting all VerificationTest happening during the evaluation of the first argument, in order to do that we are blocking a global processor that is going to evaluate every time a VerificationTest is produced.

	MUnit during the execution of a test suite will do exactly the same, it will block $TestEvaluatePostProcessor in the same way, but with another function.

	The fact that we are rely on a Composition means that this method should NOT be used outside MUnit / Testing because we will find a better way to do this in the future.
 *)

SetAttributes[TestReport, HoldFirst]

Options[TestReport] = {
	MemoryConstraint -> Inherited, 
	SameTest         -> Inherited, 
 	TimeConstraint   -> Inherited
}
 

HoldPattern[e_TestReport]:=
	With[
		{parse = System`Private`ArgumentsWithRules[e, 1, Hold]},
		Replace[
			parse,
			{Hold[a_], Hold[o___]} :> iTestReport[a, o]
		] /; Length[parse] == 2
	]


(* collect and merge logic *)
SetAttributes[iTestReport, HoldFirst]
iTestReport[expr_, opts___] := 
	$TestReportPostProcessor @ Block[
		{
			$DefaultOptions = Flatten @ {opts, $DefaultOptions},
			$TestEvaluatePostProcessor := Composition[parseTestReport, TestResultObject],
			$TestEvaluatePreProcessor := Identity,
			$TestReportPostProcessor := Composition[parseTestReport, TestReportObject],
			$TestReportPreProcessor := Identity
		},
		processData[
			$TestReportPreProcessor @ <|
				"Title"                       -> Automatic, 
				"Aborted"                     -> False, 
				"TestResults"                 -> <||>, 
				"TestsSucceededKeys"          -> {}, 
				"TestsFailedWrongResultsKeys" -> {}, 
				"TestsFailedWithMessagesKeys" -> {}, 
				"TestsFailedWithErrorsKeys"   -> {},
				Reap[
					parseTestReport @ expr; {},
					_tag,
					mergeTestReport
				]
			|>,
			Function @ GroupBy[
				#[["TestResults", All, 1, "Outcome"]],
				<|
					"Success"         -> "TestsSucceededKeys", 
					"Failure"         -> "TestsFailedWrongResultsKeys", 
					"MessagesFailure" -> "TestsFailedWithMessagesKeys", 
					"Error"           -> "TestsFailedWithErrorsKeys"
				|>,
				Keys /* Developer`ToPackedArray
			]
		]
	]

mergeTestReport[tag[s_], rest___] := mergeTestReport[s, rest]
mergeTestReport["Title", {Automatic|None..., t_, ___}] := "Title" -> t
mergeTestReport["Title", _] := {}
mergeTestReport["TestResults", results_] := "TestResults" -> Association[results]
mergeTestReport["Aborted", aborted_] := "Aborted" -> Or @@ aborted
mergeTestReport[title_, any_] := title -> any


(* Case of TestReport working on a list of VerificationTest/MUnit`Test - see bug 297800 *)
parseTestReport[tr_TestResultObject] := 
	Last @ Sow[tr["TestKey"] -> tr, tag["TestResults"]]


parseTestReport[tr:_List|_Association?AssociationQ] :=
	Map[parseTestReport, tr]

parseTestReport[tr:TestReportObject[KeyValuePattern[{"TestResults" -> tests_?AssociationQ, "Aborted" -> aborted_, "Title" -> title_}]]] := (
	Sow[tests,   tag["TestResults"]];
	Sow[aborted, tag["Aborted"]];
	Sow[title,   tag["Title"]];
	tr
)

(* Main TestReport case : Given a test file or test notebook this is only running Get *)
parseTestReport[testSource:_String?StringQ|File[_String?StringQ]] := 
	Block[{$ContextPath = Prepend[$ContextPath, "MUnit`"]},
		Replace[
			FindFile[testSource], {
			full_String :> MUnit`TestRun[
				full, 
				"Loggers" -> {<|
					"LogSuccess"         -> parseTestReport, 
					"LogFailure"         -> parseTestReport, 
					"LogMessagesFailure" -> parseTestReport, 
					"LogError"           -> parseTestReport,
					"LogBegin"           -> Function @ Sow[#, tag["Title"]]
				|>}
			],
			any_:>	Message[TestReport::fnfnd, testSource]
		}]
	]

parseTestReport[nb:Notebook[__]] := safeNotebookToTest[nb]
parseTestReport[any_] := Message[TestReport::invalidexpr, HoldForm[any]]


(* 
	this is a function that is using MUnit notebook parsing functionality to extract a list of verification test from a notebook 
	NotebookToTests might be throwing errors, the function is adding a generic Catch and returning all errors to the user as messages
	instead of silently failing.

	Error messages and handling might be improved in the future.
*)
safeNotebookToTest[nb_Notebook, filename_:"Notebook.nb"] := 
	Replace[
		Catch[MUnit`NotebookToTests[nb, filename], _], {
			(* the function was able to parse correctly the notebook *)
			s_String :> ToExpression[s],
			(* MUnit errors have a Message list we can display *)
			{___, "Messages" -> {title_, msg_}, ___} :> 
				Message[TestReport::nberr, title, msg],
			(* this is a catchall safety valve in theory we should never display this *)
			error_ :> 
				Message[TestReport::nbunxerr, error]
		}
	]
