(* ::Package:: *)

(* ::Title:: *)
(*TestRun*)

(* ::Section:: *)
(*Information*)

`Information`$TestRunId = "$Id: TestRun.m,v 1.21 2009/12/08 17:20:19 brenton Exp $"

(* ::Section:: *)
(*Public*)

TestRun::usage =
"TestRun[tests] runs tests and returns True if all tests pass, and False otherwise."

TestRunTitle::usage =
"TestRunTitle is an option for TestRun."

BeginTestSection::usage =
"BeginTestSection[section] indicates the beginning of a related group of tests that can be skipped by giving TestFailureAction -> \"SkipSection\" in a Test.\n
BeginTestSection[section, require] gives a requirement for the section to run. If require is False, the section is skipped."

EndTestSection::usage =
"EndTestSection[] marks the end of a test section."

TestLog::usage =
"TestLog[msg] writes msg to the test loggers."

TestTerminate::usage =
"TestTerminate[msg] logs the given message and terminates the current test run."


TestFileSet::usage =
"TestFileSet[ file] sets a file to be the current test file."

Loggers::usage =
"Loggers is an option to TestRun that specifies the loggers to use."

CanaryCondition::usage =
"CanaryCondition is an option to TestRun that specfies if the first test of a section should skip the rest of the section on failure."

$CurrentFile::usage =
"$CurrentFile is similar to $Input, in that it is the name of the current test file."

$CurrentFileChild::usage =
"$CurrentFileChild is the child of $CurrentFile if one exists and Null otherwise."

MUnitErrorTag::usage =
"Throw tag used internally by MUnit"

(* ::Section:: *)
(*Package*)

Begin["`Package`"]

End[]

(* ::Section:: *)
(*Private*)

Begin["`TestRun`Private`"]

TestRun::nostack =
"EndTestSection[] was called without a matching call to BeginTestSection."

TestRun::strnbobj = 
"String or NotebookObject expected in position `1` in `2`."

Options[TestRun] =
	{
		Loggers -> {PrintLogger[]},
		CanaryCondition -> False,
		TestRunTitle -> None
	}

(* given a string, TestRun interprets it as a test file name *)
TestRun[testFileName_String, opts:OptionsPattern[]] :=
	Module[{testStream, $loggers = OptionValue[Loggers], title = OptionValue[TestRunTitle]},
		Block[{$CurrentFile = testFileName, $testRunByteCount, $CurrentFileChild = Null},
			Catch[
				(* the DOSTextFormat -> False option is to treat \r\n as 2 bytes wrt StreamPosition.
				Normally, DOSTextFormat is True and \r\n counts as 1 byte, which is pretty useless *)
				testStream = Quiet[OpenRead[testFileName, DOSTextFormat -> False], {OpenRead::noopen, OpenRead::fstr}];
				If[testStream === $Failed,
					Throw[testFileName, MUnitErrorTag[{"TestRun", "TestRun could not open the test file"}]]
				];
				$testRunByteCount = FileByteCount[testFileName];
				TestRun[testStream, opts]
				,
				_MUnitErrorTag
				,
				Function[{value, tag},
					Scan[LogStart[#, title]&, $loggers];
					(*
					TODO: fix the crazy 9's problem
					*)
					Scan[LogFatal[#, ToString[Column[{tag[[1,1]], Style[tag[[1,2]], Bold], value}]]]&, $loggers];
					Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
					False
				]
			]
		]
	]

(* given a NotebookObject, TestRun interprets as a test notebook *)
TestRun[testNotebookObject_NotebookObject, opts:OptionsPattern[]] :=
	Module[{notebookTests, testStream, $loggers = OptionValue[Loggers], title = OptionValue[TestRunTitle]},
		Block[{$CurrentFile = None, $testRunByteCount, $CurrentFileChild = Null},
			Catch[
				notebookTests = NotebookToTests[testNotebookObject];
				$testRunByteCount = StringLength[notebookTests];
				testStream = StringToStream[notebookTests];
				TestRun[testStream, opts]
				,
				_MUnitErrorTag
				,
				Function[{value, tag},
					Scan[LogStart[#, title]&, $loggers];
					(*
					TODO: fix the crazy 9's problem
					*)
					Scan[LogFatal[#, ToString[Column[{tag[[1,1]], Style[tag[[1,2]], Bold], value}]]]&, $loggers];
					Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
					False
				]
			]
		]
	]

(* TestRun ultimately uses an InputStream *)
TestRun[testStream_InputStream, opts:OptionsPattern[]] :=
	Module[
		{
			$loggers = OptionValue[Loggers],
			title = OptionValue[TestRunTitle],
			$abort = False, $fatal = False,
			ccondition = OptionValue[CanaryCondition],
			$canary = 1,
			$sectionStack = {},
			$testRunExpressionCount = 0,
			$testRunByteCounter = 0,
			expr,
			testids,
			results,
			testResults,
			cpuTimeUsed,
			absTimeUsed,
			memoryUsed,
			successCnt = 0,
			failureCnt = 0,
			msgFailureCnt = 0,
			secSkippedCnt = 0,
			errorCnt = 0,
			wasSuccessful
		}
		,
		Block[{
			$TestIndex = 0,
			TestLog, TestTerminate,
			TestFileSet
			},
			(* TestLog should accept anything *)
			TestLog[e_] :=
				TestLog[ToString[e, OutputForm]];			
			TestLog[msg_String] :=
				Scan[LogMessage[#, msg]&, $loggers];			
			(* TestTerminate should accept anything *)
			TestTerminate[e_] :=
				TestTerminate[ToString[e, OutputForm]];			
			TestTerminate[msg_String] :=
				(
					Scan[LogMessage[#, "Stopping test run on TestTerminate: " <> msg]&, $loggers];
					$abort = True
				);
			TestFileSet[ file_] := ($CurrentFileChild = file);
			cpuTimeUsed = TimeUsed[];
			absTimeUsed = SessionTime[];
			memoryUsed = MemoryInUse[];
			
			Scan[LogStart[#, title]&, $loggers];
			
			While[True,
				Scan[LogNewTestRunIteration, $loggers];
				
				Quiet[
					Check[
						expr = Read[testStream, HoldComplete[Expression]];
						,
						Scan[LogFatal[#, "Syntax error in test file"]&, $loggers];
						$fatal = True;
						Break[]
						,
						{Syntax::sntunc, Syntax::sntufn, Syntax::com, Read::readt}
					]
					,
					{Syntax::sntunc, Syntax::sntufn, Syntax::com, Read::readt}
				];
				
				$testRunByteCounter = StreamPosition[testStream];
				
				If[IntegerQ[$testRunByteCount],
					(* only log progress if $testRunByteCount has been set, $testRunByteCount may not be set if TestRun was called
					directly with a stream *)
					Scan[LogTestRunProgress[#, $testRunByteCounter, $testRunByteCount]&, $loggers];
				];
				
				If[expr == EndOfFile, Break[]];
				
				$testRunExpressionCount++;
				
				(*preprocess*)
				expr = Replace[expr, {
							HoldComplete[CompoundExpression[b:(_BeginTestSection | _TestRequirement), Null]] :> HoldComplete[b],
							HoldComplete[CompoundExpression[e:(_EndTestSection | _EndRequirement), Null]] :> HoldComplete[e]}];
				
				Which[
					MatchQ[expr, HoldComplete[_BeginTestSection | _TestRequirement]],
					Module[{res, section, require},
						res = ReleaseHold[expr];
						section = res[[1]];
						require = If[Length[res] == 2, res[[2]], True];
						If[And @@ $sectionStack && require,
							$canary = $TestIndex + 1
							,
							secSkippedCnt++
						];
						(* unlike the LogMessages, which are not touched in a completely skipped section, we want
						to log EVERY section with LogBeginTestSection and LogEndTestSection *)
						$sectionStack = Append[$sectionStack, require];
						Scan[LogBeginTestSection[#, section, require]&, $loggers]
					]
					,
					
					MatchQ[expr, HoldComplete[_EndTestSection | _EndRequirement]],
					Scan[LogEndTestSection, $loggers];
					If[Length[$sectionStack] <= 0,
						Scan[LogFatal[#, ToString[StringForm[TestRun::nostack]]]&, $loggers];
						$fatal = True;
						,
						$sectionStack = Most[$sectionStack]
					]
					,
					
					And @@ $sectionStack,
					
					(* log testids of tests before they are run, to help with crashes *)
					testids = Cases[expr, (TestID -> id_) :> id, Infinity];
					Outer[LogTestID[#1, #2]&, $loggers, testids];
					
					(* The actual evaluation of the input and comparing of result to expected is done here. *)
					results = Reap[ReleaseHold[expr], "MUnitTest"][[2]];
					
					If[results == {},
						(* expr was a non-Test expression, possibly setup or teardown code *)
						Continue[]
					];
					
					(* we are only interested in the "MUnitTest"-tagged results, so grab first list *)
					testResults = results[[1]];
					
					Function[{preRes},
						Module[{res = preRes, mode},
							If[!TestResultQ[res],
								(* test index count will possibly be off by one, but this is a very strange case *)
								res = testError["Expected a TestResultObject, but received " <> ToString[preRes]]
							];
							mode = FailureMode[res];
							If[ StringQ[TestFileChild[ res]], Scan[ LogTestFile[#, TestFileChild[ res]]&, $loggers]];
							Switch[mode,
								"Success",
								successCnt++;
								Scan[LogSuccess[#, res]&, $loggers]
								,
								"Failure",
								failureCnt++;
								Scan[LogFailure[#, res]&, $loggers];
								Switch[TestFailureAction[res],
									"Abort",
									$abort = True;
									Scan[LogMessage[#, "Stopping test run on TestFailureAction"]&, $loggers];
									Continue[]
									,
									"SkipSection",
									$sectionStack[[-1]] = False;
									secSkippedCnt++;
									Scan[LogMessage[#, "Skipping test section on TestFailureAction"]&, $loggers];
									Continue[]
								]
								,
								"MessagesFailure",
								msgFailureCnt++;
								Scan[LogMessagesFailure[#, res]&, $loggers]
								,
								"Error",
								errorCnt++;
								Scan[LogError[#, res]&, $loggers];
								Switch[TestErrorAction[res],
									"Abort",
									$abort = True;
									Scan[LogMessage[#, "Stopping test run on TestErrorAction"]&, $loggers];
									Continue[]
									,
									"SkipSection",
									$sectionStack[[-1]] = False;
									secSkippedCnt++;
									Scan[LogMessage[#, "Skipping test section on TestErrorAction"]&, $loggers];
									Continue[]
								]
							];
							(*
							the canary condition: if the first test of a section fails,
							skip the rest of the section (this includes the 'default' first section)
							*)
							If[ccondition && $TestIndex == $canary && mode != "Success",
								$sectionStack[[-1]] = False;
								secSkippedCnt++;
								Scan[LogMessage[#, "Skipping Test Section on failure of canary test"]&, $loggers];
								Continue[]
							]
						]
					] ~Scan~ testResults
					,
					True,
					(* ignore everything else, since it is not allowed to run (e.g., it's in a test section that is being skipped) *)
					Null
				];
				
				If[$abort || $fatal, Break[]]
				
			]; (* end While *)
			
			cpuTimeUsed = TimeUsed[] - cpuTimeUsed;
			absTimeUsed = SessionTime[] - absTimeUsed;
			memoryUsed = MemoryInUse[] - memoryUsed;
			
			Close[testStream];
			
			Scan[LogExpressionCount[#, $testRunExpressionCount]&, $loggers];
			Scan[LogCPUTimeUsed[#, cpuTimeUsed]&, $loggers];
			Scan[LogAbsoluteTimeUsed[#, absTimeUsed]&, $loggers];
			Scan[LogMemoryUsed[#, memoryUsed]&, $loggers];
			Scan[LogTestRunProgress[#, $testRunByteCounter, $testRunByteCount]&, $loggers];
			Scan[LogEnd[#, $TestIndex, successCnt, failureCnt, msgFailureCnt, secSkippedCnt, errorCnt, $abort || $fatal]&, $loggers];
			
			wasSuccessful = successCnt == $TestIndex && errorCnt == 0 && !$abort && !$fatal;
			
			wasSuccessful
		]
	]
	
TestRun[args___, opts:OptionsPattern[]] :=
	Module[{
		$loggers = OptionValue[Loggers],
		title = OptionValue[TestRunTitle]},
		Scan[LogStart[#, title]&, $loggers];
		Scan[LogFatal[#, ToString[StringForm[TestRun::strnbobj, 1, HoldForm[TestRun[args, opts]]]]]&, $loggers];
		Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
		False
	]


End[]
