(* ::Package:: *)

(* ::Title:: *)
(*TestRun*)

(* ::Section:: *)
(*Information*)

`Information`$TestRunId = "$Id: TestRun.m,v 1.53.4.3 2010/12/21 17:53:09 brenton Exp $"

(* ::Section:: *)
(*Public*)

TestRun::usage =
"TestRun[tests] runs tests and returns True if all tests pass, and False otherwise."

TestRunTitle::usage =
"TestRunTitle is an option for TestRun."

BeginTestSection::usage =
"BeginTestSection[section] indicates the beginning of a related group of tests that can be skipped by giving TestFailureAction -> \"SkipSection\" in a Test.\n
BeginTestSection[section, require] gives a requirement for the section to run. If require is False, the section is skipped."

EndTestSection::usage =
"EndTestSection[] marks the end of a test section."

TestLog::usage =
"TestLog[msg] writes msg to the test loggers."

TestTerminate::usage =
"TestTerminate[msg] logs the given message and terminates the current test run."

Loggers::usage =
"Loggers is an option to TestRun that specifies the loggers to use."

CanaryCondition::usage =
"CanaryCondition is an option to TestRun that specfies if the first test of a section should skip the rest of the section on failure."

$CurrentFile::usage =
"$CurrentFile is similar to $Input, in that it is the name of the current test file.
$CurrentFile is deprecated, please use $CurrentTestSource."

$CurrentTestSource::usage =
"$CurrentTestSource is the current source of running tests, either a file or a notebook."

MUnitErrorTag::usage =
"Throw tag used internally by MUnit"

(* ::Section:: *)
(*Package*)

Begin["`Package`"]

(* API for internals of Test *)
$actualTestIndex
(* $allTestIndex counts all tests, including skipped ones *)
$allTestIndex
testSourceStack
logTestResult

(*
ReadableForm is like InputForm, but maintains \n and things as single bytes
*)
codePointToReadableForm

End[]

(* ::Section:: *)
(*Private*)

Begin["`TestRun`Private`"]

(*
set this here so that Tests can be run outside of TestRun
*)
$actualTestIndex = 0
$allTestIndex = 0
testSourceStack = {None}

(*
$CurrentFile is deprecated, but must keep around for backward-compatibility
*)
$CurrentFile := Last[testSourceStack] /. _NotebookObject -> None
$CurrentTestSource := Last[testSourceStack]


If[ListQ["GetFileEvent" /. Internal`Handlers[]],
	Internal`AddHandler["GetFileEvent", MUnitGetFileHandler]
]

TestRun::badsec =
"Invalid section primitive: `1`"

TestRun::nostack =
"EndTestSection[] was called without a matching call to BeginTestSection."

TestRun::strnbobj = 
"String or NotebookObject expected in position `1` in `2`."

BeginTestSection::require =
"Section requirement did not evaluate to True or False: `1`"

Options[TestRun] =
	{
		Loggers :> {PrintLogger[]},
		CanaryCondition -> False,
		TestRunTitle -> None
	}

(* given a string, TestRun interprets it as a test file name *)
TestRun[testFileName_String, opts:OptionsPattern[]] :=
	Module[{testStream, $loggers, title, res},
		{$loggers, title} = OptionValue[{Loggers, TestRunTitle}];
		Block[{testSourceStack = {testFileName}},
			Scan[LogStart[#, title]&, $loggers];
			Scan[LogBegin[#, title]&, $loggers];
			Catch[
				testStream = Quiet[OpenRead[testFileName], {OpenRead::noopen, OpenRead::fstr}];
				If[testStream === $Failed,
					Throw[{"Value" -> testFileName, "Messages" -> {"TestRun", "TestRun could not open the test file"}}, MUnitErrorTag]
				];
				Scan[LogBeginTestSource[#, testFileName]&, $loggers];
				res = testRun[testStream, opts];
				Close[testStream];
				Scan[LogEndTestSource[#]&, $loggers];
				Scan[LogEnd[#]&, $loggers];
				res
				,
				MUnitErrorTag
				,
				Function[{rules, tag},
					Module[{msgs = "Messages" /. rules, val = "Value" /. rules},
						(*
						TODO: fix the crazy 9's problem
						*)
						Scan[LogFatal[#, ToString[Column[{msgs[[1]], Style[msgs[[2]], Bold], val}]]]&, $loggers];
						Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
						Scan[LogEnd[#]&, $loggers];
						False
					]
				]
			]
		]
	]

(* given a NotebookObject, TestRun interprets as a test notebook *)
TestRun[testNotebookObject_NotebookObject, opts:OptionsPattern[]] :=
	Module[{notebookTests, testStream, $loggers, title, res},
		{$loggers, title} = OptionValue[{Loggers, TestRunTitle}];
		Block[{testSourceStack = {testNotebookObject}},
			Scan[LogStart[#, title]&, $loggers];
			Scan[LogBegin[#, title]&, $loggers];
			Catch[
				notebookTests = NotebookToTests[testNotebookObject];
				testStream = StringToStream[notebookTests];
				Scan[LogBeginTestSource[#, testNotebookObject]&, $loggers];
				res = testRun[testStream, opts];
				Close[testStream];
				Scan[LogEndTestSource[#]&, $loggers];
				Scan[LogEnd[#]&, $loggers];
				res
				,
				MUnitErrorTag
				,
				Function[{rules, tag},
					Module[{msgs = "Messages" /. rules, val = "Value" /. rules},
						Scan[LogFatal[#, rules]&, $loggers];
						(*
						TODO: fix the crazy 9's problem
						*)
						Scan[LogFatal[#, ToString[Column[{msgs[[1]], Style[msgs[[2]], Bold], val}]]]&, $loggers];
						Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
						Scan[LogEnd[#]&, $loggers];
						False
					]
				]
			]
		]
	]

TestRun[testNotebookFileName_String /; StringMatchQ[testNotebookFileName, "*.nb"], opts:OptionsPattern[]] :=
	Module[{testNotebook, notebookTests, testStream, $loggers, title, res},
		{$loggers, title} = OptionValue[{Loggers, TestRunTitle}];
		Block[{testSourceStack = {testNotebookFileName}},
			Scan[LogStart[#, title]&, $loggers];
			Scan[LogBegin[#, title]&, $loggers];
			Catch[
				testNotebook = Quiet[Get[testNotebookFileName], {Get::noopen}];
				If[testNotebook === $Failed,
					Throw[{"Value" -> testNotebookFileName, "Messages" -> {"TestRun", "TestRun could not open the test notebook file"}}, MUnitErrorTag]
				];
				notebookTests = NotebookToTests[testNotebook];
				testStream = StringToStream[notebookTests];
				Scan[LogBeginTestSource[#, testNotebookFileName]&, $loggers];
				res = testRun[testStream, opts];
				Close[testStream];
				Scan[LogEndTestSource[#]&, $loggers];
				Scan[LogEnd[#]&, $loggers];
				res
				,
				MUnitErrorTag
				,
				Function[{rules, tag},
					Module[{msgs = "Messages" /. rules, val = "Value" /. rules},
						Scan[LogFatal[#, ToString[Column[{msgs[[1]], Style[msgs[[2]], Bold], val}]]]&, $loggers];
						Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
						Scan[LogEnd[#]&, $loggers];
						False
					]
				]
			]
		]
	]

Options[testRun] = Options[TestRun]

(* TestRun ultimately uses an InputStream *)
testRun[testStream_InputStream, opts:OptionsPattern[]] :=
	Module[
		{
			$loggers,
			$abort = False, $fatal = False,
			ccondition,
			$canary = 1,
			$sectionStack = {},
			$testRunExpressionCount = 0,
			pos,
			expr,
			testids,
			cpuTimeUsed,
			absTimeUsed,
			memoryUsed,
			successCnt = 0,
			failureCnt = 0,
			msgFailureCnt = 0,
			skippedTestCnt = 0,
			errorCnt = 0,
			wasSuccessful
		}
		,
		{$loggers, ccondition} = OptionValue[{Loggers, CanaryCondition}];
		Block[{
			(* API for test files *)
			TestLog, TestTerminate, MUnitGetFileHandler,
			(* API for internals of Test *)
			$actualTestIndex = 0, $allTestIndex = 0, logTestResult}
			,
			(* TestLog should accept anything *)
			TestLog[e_] :=
				Scan[LogMessage[#, ToString[e, OutputForm]]&, $loggers];			
			(* TestTerminate should accept anything *)
			TestTerminate[e_] :=
				(
					Scan[LogMessage[#, "Stopping test run on TestTerminate: " <> ToString[e, OutputForm]]&, $loggers];
					$abort = True
				);
			MUnitGetFileHandler[HoldComplete[file_String, wrapper_, mode:First|Last]] :=
				Switch[mode,
					First,
					testSourceStack = Append[testSourceStack, file];
					Scan[LogBeginTestSource[#, file]&, $loggers];
					,
					Last,
					testSourceStack = Most[testSourceStack];
					Scan[LogEndTestSource[#]&, $loggers];
				];
			logTestResult[tr_] :=
				Module[{fm = FailureMode[tr], action = ""},
					Switch[fm,
						"Success",
							successCnt++;
							Scan[LogSuccess[#, tr]&, $loggers]
						,
						"Failure",
							failureCnt++;
							Scan[LogFailure[#, tr]&, $loggers];
							action = TestFailureAction[tr]
						,
						"MessagesFailure",
							msgFailureCnt++;
							Scan[LogMessagesFailure[#, tr]&, $loggers];
							action = TestFailureAction[tr]
						,
						"Error",
							errorCnt++;
							Scan[LogError[#, tr]&, $loggers];
							action = TestErrorAction[tr]
					];
					If[action != "",
						Which[
							action == "Abort",
							$abort = True;
							Scan[LogMessage[#, "Aborting test run on TestFailureAction or TestErrorAction"]&, $loggers];
							Throw["Abort", "internalMUnitTestRunTag"]
							,
							action == "SkipSection" || ccondition && $actualTestIndex == $canary,
							$sectionStack[[-1]] = False;
							Scan[LogMessage[#, "Skipping test section on TestFailureAction, TestErrorAction, or canary condition"]&, $loggers];
							Throw["SkipSection", "internalMUnitTestRunTag"]
						]
					]
				];
			
			cpuTimeUsed = TimeUsed[];
			absTimeUsed = SessionTime[];
			memoryUsed = MemoryInUse[];
			
			While[True,
				
				pos = StreamPosition[testStream];
				Check[
					Quiet[
						expr = Read[testStream, HoldComplete[Expression]]
						,
						{General::spell1, General::spell}
					]
					,
					Module[{len, chars},
						len = StreamPosition[testStream] - pos;
						SetStreamPosition[testStream, pos];
						chars = ReadList[testStream, Character, len];
						Scan[LogFatal[#, "Syntax error in test file: " <> chars]&, $loggers];
						$fatal = True;
						Break[]
					]
				];
				
				If[expr == EndOfFile, Break[]];
				
				$testRunExpressionCount++;
				
				(*
				preprocess
				1. replace foo; with just foo
				2. add 2nd arg to BeginTestSection, normalize to BeginTestSection and EndTestSection
				3. handle errors
				*)
				expr = Replace[expr, HoldComplete[CompoundExpression[e:(_BeginTestSection|_TestRequirement|_EndTestSection|_EndRequirement|_TestIgnore|_EndIgnore), Null]] :> HoldComplete[e]];
				expr = Replace[expr, {
					(* valids *)
					HoldComplete[BeginTestSection[sec_, require_]] :>
						HoldComplete[BeginTestSection[sec, require]],
					HoldComplete[EndTestSection[]] :>
						HoldComplete[EndTestSection[]],
					(* processes *)
					HoldComplete[BeginTestSection[sec_]] :>
						HoldComplete[BeginTestSection[sec, True]],
					HoldComplete[TestRequirement[require_, ___]] :>
						HoldComplete[BeginTestSection["", require]],
					HoldComplete[TestIgnore[ignore_, ___]] :>
						HoldComplete[BeginTestSection["", !ignore]],
					HoldComplete[(EndRequirement|EndIgnore)[]] :>
						HoldComplete[EndTestSection[]],
					(* errors *)
					HoldComplete[BeginTestSection[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;),
					HoldComplete[TestRequirement[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;),
					HoldComplete[TestIgnore[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;),
					HoldComplete[EndTestSection[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;),
					HoldComplete[EndRequirement[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;),
					HoldComplete[EndIgnore[___]] :>
						(Scan[LogFatal[#, ToString[StringForm[TestRun::badsec, expr]]]&, $loggers];$fatal = True;)}];
				
				If[$fatal, Break[]];
				
				Switch[expr,
					HoldComplete[_BeginTestSection],
					Module[{section, require},
						{section, require} = Extract[expr[[1]], {{1}, {2}}];
						If[And @@ $sectionStack && require,
							$canary = $actualTestIndex + 1
							,
							Null
							,
							Scan[LogFatal[#, ToString[StringForm[BeginTestSection::require, require]]]&, $loggers];
							$fatal = True;
						];
						(* unlike the LogMessages, which are not touched in a completely skipped section, we want
						to log EVERY section with LogBeginTestSection and LogEndTestSection *)
						$sectionStack = Append[$sectionStack, require];
						Scan[LogBeginTestSection[#, section, require]&, $loggers]
					]
					,
					
					HoldComplete[_EndTestSection],
					Scan[LogEndTestSection, $loggers];
					If[Length[$sectionStack] <= 0,
						Scan[LogFatal[#, ToString[StringForm[TestRun::nostack]]]&, $loggers];
						$fatal = True;
						,
						$sectionStack = Most[$sectionStack]
					]
					,
					
					_,
					testids = Cases[expr, (TestID -> id_) :> id, Infinity];
					Block[{
						(* special handling for DefectiveTest. Treat it like a miniature test section *)
						$sectionStack = If[MatchQ[expr, HoldComplete[_DefectiveTest]], Append[$sectionStack, False], $sectionStack]},
						Outer[LogTestInfo[#1, #2, ++$allTestIndex, And @@ $sectionStack]&, $loggers, testids];
						If[And @@ $sectionStack,
							(* The actual evaluation of the input and comparing of result to expected is done here. *)
							Catch[
								ReleaseHold[expr]
								,
								"internalMUnitTestRunTag"
							]
							,
							skippedTestCnt+=Length[testids]
						]
					]					
				];
				
				If[$abort || $fatal, Break[]]
				
			]; (* end While *)
			
			cpuTimeUsed = TimeUsed[] - cpuTimeUsed;
			absTimeUsed = SessionTime[] - absTimeUsed;
			memoryUsed = MemoryInUse[] - memoryUsed;
			
			Scan[LogExpressionCount[#, $testRunExpressionCount]&, $loggers];
			Scan[LogCPUTimeUsed[#, cpuTimeUsed]&, $loggers];
			Scan[LogAbsoluteTimeUsed[#, absTimeUsed]&, $loggers];
			Scan[LogMemoryUsed[#, memoryUsed]&, $loggers];
			Scan[LogTestCount[#, $actualTestIndex]&, $loggers];
			Scan[LogSuccessCount[#, successCnt]&, $loggers];
			Scan[LogFailureCount[#, failureCnt]&, $loggers];
			Scan[LogMessagesFailureCount[#, msgFailureCnt]&, $loggers];
			Scan[LogSkippedTestCount[#, skippedTestCnt]&, $loggers];
			Scan[LogErrorCount[#, errorCnt]&, $loggers];
			Scan[LogWasAborted[#, $abort]&, $loggers];
			Scan[LogWasFatal[#, $fatal]&, $loggers];
			
			wasSuccessful = successCnt == $actualTestIndex && errorCnt == 0 && !($abort || $fatal);
			Scan[LogWasSuccessful[#, wasSuccessful]&, $loggers];
			
			Scan[LogEnd[#, $actualTestIndex, successCnt, failureCnt, msgFailureCnt, skippedTestCnt, errorCnt, $abort||$fatal]&, $loggers];
			
			wasSuccessful
		]
	]
	
TestRun[args___, opts:OptionsPattern[]] :=
	Module[{$loggers, title},
		{$loggers, title} = OptionValue[{Loggers, TestRunTitle}];
		Scan[LogBegin[#, title]&, $loggers];
		Scan[LogFatal[#, ToString[StringForm[TestRun::strnbobj, 1, HoldForm[TestRun[args, opts]]]]]&, $loggers];
		Scan[LogEnd[#, 0, 0, 0, 0, 0, 0, True]&, $loggers];
		Scan[LogEnd[#]&, $loggers];
		False
	]

End[]
