(* * Export.m *)

(* :Title: Make.m *)

(* :Authors:
	Andrew Hunt, Shaun McCance, Todd Gayley
	andy@wolfram.com, shaunm@wolfram.com, tgayley@wolfram.com
*)

(* :Package Version: 0.50 *)

(* :Mathematica Version: 6.0 *)
                     
BeginPackage["DocumentationBuild`Export`", {
  "DocumentationBuild`Common`",
  "DocumentationBuild`Info`",
  "DocumentationBuild`Utils`",
  "DocumentationBuild`Make`",
  "Transmogrify`"
}];


(* * Usage Statements *)

(******************************************************************************)
(** Usage Statements **)

ExportDocumentation::usage = "ExportDocumentation[file, file.nb | Notebook, ExportFormats -> {...}] creates\
 documentation files from source notebook file.nb.";

(* Web *)
ExportWebPage::usage = "ExportWebPage[file.html, file.nb | Notebook] takes a production notebook and exports to xml.";
BuildLog;
ExportWebPageFromResultNB::usage="";
CreateFeaturedPage::usage="";
ExportNavigatorWebPage::usage = "ExportNavigatorWebPage[file.html, file.nb] converts a function navigator notebook into a web page.";
ExportLandingPage::usage = "ExportLandingPage[path] creates landing page for future site from Bill's reports.";
GetFutureLandingContent::usage = "Generate data for future landing page in http://reference.future.wolfram.com/mathematica.html.";
MakeLandingContent::usage="";
(* XMLTransform functions *)
htmlBoxesQ;
inputQueue;
ReplaceSpecialCharacters;
WrapInOpenCloseButton;
AddFirstAndLastToexpandWraps::usage="";
CreateLearningResources::usage="";
deleteNonASCII::usage="";

BuildDocPacletIndex::usage = "BuildDocPacletIndex builds the PcletDB.m file \
 files from a documentation directory. It is typically run after ExportOnlineNotebooks\
 and operates on the output directory of converted notebook files produced by that function.";
  
MakeMarketingThumbnailList::usage = "";
debugOn::usage = "Manually activate debugging messages";
debugOff::usage = "Manually deactivate debugging messages";
dprint::usage = "dprint[string] prints \"[debug] \"<>string via Print[] and Global`AntLog[].";
dfile::usage = "dfile[expr] exports expr to a uuid-named file and prints the filename via dprint. \
Options: \
 - \"Format\": string or list of strings: export to file(s) in the given format(s).  Default value is \".nb\". \
 - \"Directory\": directory in which to write the file(s).  Default value is $TemporaryDirectory. \
 - \"DeleteLargeItems\" -> True
";

DocLinkURL

(* load external data *)

(* Get walsh's url encode/decoder *)
Get[ToFileName[{DocumentationBuild`Common`$DocumentationBuildDirectory, "Kernel"}, "ShortenIdentifier.m"]];

Assert[MemberQ[$ContextPath, "DocumentationBuild`Common`"]];


Begin["`Private`"];

debugOn[] := (
  Print["dprint/dfile are now on."];
  Global`AntLog["dprint/dfile are now on."];
  optDebug = True
);
debugOff[] := (
  Print["dprint/dfile are now off."];
  Global`AntLog["dprint/dfile are now off."];
  optDebug = False
);

dprint[s_String] := Module[{timestamp},
  If[optDebug
  , timestamp = DateString[];
    Print["debug["<>DateString["ISODateTime"]<>"]: "<>ToString[s]];
    Global`AntLog["debug["<>DateString["ISODateTime"]<>"]: "<>ToString[s]];
  , Null
  ]];

(* when writing a .nb file, the fe "helpfully" intervenes and rewrites
expressions to keep everything nice and neat.  I need to see the actual
unchanged expressions, which are obtainable when writing .m files. *)
Options[dfile] = {"Format" -> ".m", "Directory" -> $TemporaryDirectory, "DeleteLargeItems" -> True};
dfile[expr_, OptionsPattern[]] := Module[
  {
    expr1 = expr, optFormat, optDirectory, optDeleteLargeItems
  }
,
  nForm[num_?NumberQ] := NumberForm[num, {Infinity, 10}, NumberPoint -> ".", DigitBlock -> 3, NumberSeparator -> ",", ExponentFunction -> (Null &)];

  If[optDebug
  , {optFormat, optDirectory, optDeleteLargeItems} = OptionValue[{"Format", "Directory", "DeleteLargeItems"}];
    (* Global`AntLog["dfile: initial ByteCount[expr] = "<>ToString[ByteCount[expr1]//nForm]]; *)
    If[
      optDeleteLargeItems
    ,
      (* Global`AntLog["dfile: \"DeleteLargeItems\" -> True"]; *)
      expr1 = expr1 //. {
        _GraphicsComplex3DBox -> graphicscomplex3dbox
      , _RasterBox -> rasterBox
      , _GeometricTransformationBox -> geometricTransformationBox
      , _PolygonBox -> polygonBox
      , _NumericArray -> numericArray
      }
    ];
    (* Global`AntLog["dfile: final ByteCount[expr] = "<>ToString[ByteCount[expr1]//nForm]]; *)
    dfile[expr1, optFormat, optDirectory];
  , Null
  ]];

dfile[expr_, formats_List, dir_String] :=
  If[optDebug, dfile[expr, #, dir]& /@ formats, Null];

dfile[expr_, format_String, dir_String] := Module[{doutfile},
  If[optDebug
  , doutfile = FileNameJoin[{dir, CreateUUID[]<>format}];
    Export[doutfile, expr];
    dprint["dfile: expr exported to "<>doutfile];
  , Null
  ]];

(* * ExportDocumentation: run MakeNotebook and ExportWebPage *)

(******************************************************************************)
(** ExportDocumentation **)

Options[ExportDocumentation] = {
    "ExportFormats" -> {"Notebook" -> "OutputDirectory", "HTML"},
    "Cache" -> False,
    "CacheLocation"->"",
    "Language" -> "English",
    "BuildNumber" -> "0000000",
    "Debug" -> False
};

ExportDocumentation::support = "Format not supported: `1`";

ExportDocumentation[outFile_String, file_String, opts___?OptionQ]:=
  ExportDocumentation[outFile, GetQuietly[file], "FilePath" -> file, opts];

ExportDocumentation[file_String, nb_Notebook, opts___?OptionQ]:=
  Module[{dir, outdir, outfile, expr, info, formats, language, res},

    info = GetNotebookInformation[nb, opts];

    ServerLog[ separator <> "\nExportDocumentation:\n"];
    ServerLog[ nb ];
    ServerLog[ info ];
    (* debugOn[]; *)
    dprint["dprint Export.m: " <> separator <> "\nExportDocumentation:\n"];
    dprint["dprint Export.m: " <> ToString[info, InputForm]];
    Global`AntLog["AntLog Export.m: " <> separator <> "\nExportDocumentation:\n"];
    Global`AntLog["AntLog Export.m: " <> ToString[info, InputForm]];

    {formats, language} = {"ExportFormats", "Language"} /. {opts} /. Options[ExportDocumentation];

    formats = Flatten@{formats};
    formats = Map[If[StringQ[#], # -> {}, #]&, formats];

    BuildLog["** Entering ExportDocumentation **"];
    BuildLog["   Language: "<>language ];
    BuildLog["   'formats' = "<>ToString[formats] ];
      dprint["   Language: "<>language ];
      dprint["   'formats' = "<>ToString[formats, InputForm]];

    BuildLog[separator ];

    res = {};

    (* create in-product notebook *)
    dprint["ExportDocumentation: calling MakeNotebook"];
    expr = MakeNotebook[
		nb,
		info,
		Sequence @@ FilterRules[{opts}, Options[MakeNotebook]]
	];

    dprint["Export.m: after MakeNotebook, Head[expr] = "<>ToString[Head[expr], InputForm]];

    dprint["ExportDocumentation: doing the Notebook Put/Get dance to normalize cellgroups in the generated notebook expr"];
    expr = Module[{nbobj, nbexpr},
      nbobj = NotebookPut[expr, Visible -> False];
      nbexpr = NotebookGet[nbobj];
      NotebookClose[nbobj];
      nbexpr
    ];
    dprint["Export.m: after NotebookPut/Get dance, Head[expr] = "<>ToString[Head[expr], InputForm]];

    (* set default export directory *)
    dir = DirectoryName[file];
    BuildLog["   Default export directory: "<>dir ];


    (* write in-product notebook to a file *)    
    If[
      MemberQ[formats, ("Notebook" -> _)]
    ,
      dprint["Export.m:ExportDocumentation: writing intermediate nb to a file because MemberQ[formats, (\"Notebook\" -> _)]"];
      dprint["ExportDocumentation: checking formats: \"Notebook\" /. formats = "<>ToString["Notebook" /. formats, InputForm]];

      outfile = FileBaseName[file];
      dprint["  'outfile' .1 = "<>ToString[outfile, InputForm]];

      If[Transmogrify`GetFileExtension[outfile] != "nb", outfile = FileBaseName[outfile] <> ".nb"];
      dprint["  'outfile' .2 = "<>ToString[outfile, InputForm]];

      If[MemberQ["Notebook" /. formats, ("OutputDirectory" -> _)],
        outdir = "OutputDirectory" /. ("Notebook" /. formats),
        outdir = DirectoryName[file]
      ];
      dprint["  'outdir' = "<>ToString[outdir, InputForm]];

      (* make sure directory exists *)
      CreateDirectory[ outdir ];

      (* export notebook and add notebook's path to result list *)
      BuildLog["  Notebook path: "<>ToFileName[{outdir}, outfile] ];
        dprint["  Notebook path: "<>ToFileName[{outdir}, outfile] ];
        dprint["  running Export["<>ToFileName[{outdir}, outfile]<>", expr]"];
      res = Append[res, "Notebook" -> Export[ToFileName[{outdir}, outfile], expr]];
        dprint["  'res' = "<>ToString[res, InputForm]];

    ];

    BuildLog[separator ];

    (* create on-web page *)
    If[
      MemberQ[formats, ("HTML" -> _)]
    ,
      outfile = FileNameTake[file];

      If[MemberQ["HTML" /. formats, ("OutputDirectory" -> _)]
      , outdir = "OutputDirectory" /. ("HTML" /. formats)
      , outdir = DirectoryName[file]
      ];

      (* make sure directory exists *)
		Quiet[
			CreateDirectory[outdir],
			(* Note: `filex` is emitted in WL 12.2-, eexist in 12.3+. *)
			{CreateDirectory::filex, CreateDirectory::eexist}
		];

      (* export html and add mlht's path to result list *)
      BuildLog["   HTML path: "<>FileNameJoin[{outdir, outfile}] ];
      
      res = Append[res,
        "HTML" -> ExportWebPage[FileNameJoin[{outdir, outfile}], expr, info, opts]];
    ];

    (* Print["ExportDocumentation output 'res' is being exported to /tmp/res.m"]; *)
    (* Export["/tmp/res.m", res]; *)
  res
];


(* * ExportWebPage: munge the ip nb expr and call Transmogrify *)

(******************************************************************************)
(** ExportWebPage **)
(* 
    ExportWebPage[outfile, expr, opts] 
    this is the main Transmogrify[] wrapper.
    NOTE - outfile = None just returns the raw Transmogrify output
           ExportTest->True causes ExportWebPage to use TransmogrifyString[]
           the shorthand functions rawmogtest[] and mogtest[], respectively,
           are wrappers for these two functionalities.
*)

(* ** Options[ExportWebPage] *)

Options[ExportWebPage] = {
    "DefaultParameters" -> {},
    "BuildMode" -> Normal,
    "ExportTest" -> False,
    "Version" -> "6.0",
    "CompleteHTMLQ" -> False,
	"HTMLTemplateIncludes" -> <||>,
    "BuildNumber" -> "0000000",
    "HistoryData"->{"New"->"", "Modified"->"", "Obsolete"->"", "Excised"->""},
    "IsManipulate" -> True,
    "Cache" -> False,
    "CacheLocation"->"",
    "Language" -> "English",
    "FunctionPaclet" -> False,
    "URLBaseRemote" -> Automatic,
	"WebAssetsPath" -> Automatic,
    "Debug" -> False
};

(* ** ExportWebPageFromResultNB[outfile_String, nbfile_String, opts___?OptionQ] *)

(*
Notes on the NotebookOpen/Get/Put dance found in this function:

In recent years there have been two ways to reach ExportWebPage's munging and
export: 
 - via ExportWebPageFromResultNB, which does the "Open/Get/Put dance", and
 - via ExportDocumentation, which *does not* do the dance.  Why?  The answer is
   lost in the mists of time. [update Dec 2019: I've implemented the dance in
   ExportDocumentation immediately after MakeNotebook.]

The Open/Get/Put dance allows the frontend to repair and normalize cellgroups
in a notebook expression:

  nbobj = NotebookOpen[nbfile, Visible -> False];
  nbget = NotebookGet[nbobj];
  NotebookClose[nbobj];
  nbobj2 = NotebookPut[nbget, Visible -> False];
  nb=NotebookGet[nbobj2];
  NotebookClose[nbobj2];

Possible modifications to these paths in the future:

1. ExportDocumentation needs to do the Open/Get/Put dance after running
   MakeNotebook [update Dec 2019 - I have implemented that], or...
2. Extract the dance into a separate step called either before ExportWebPage or
   first thing inside it.
3. TODO: add a dance interlude at the end of ExportWebPage to clean up any
   cellgroups broken by its munging.

Webdoc builds that use ExportDocumentation skip the really quite important
NotebookOpen/Get/Put dance:

                       -------------------------------------------------------
                       |                    cvs notebooks                    |
                       -------------------------------------------------------
                                   ||                     ||
             path for package docs ||                     || path for main docs
                                   ||                     \/
                                   ||            --------------------------------
                                   ||            | a round of MakeNotebook in a |
                                   ||            | separate build that produces |
                                   ||            | in-product notebooks         |
                                   ||            --------------------------------
                                   ||                     ||
                                   \/                     \/
                 html.xml/DocumentationBuildHTML    MakeMathematicaHTML.xml/HTMLConversion.m
                ----    -----------------------  -----------------------------    ----
call MakeNotebook  | => | ExportDocumentation |  | ExportWebPageFromResultNB | <= | NotebookOpen/Get/Put dance
Open/Get/Put dance |    -----------------------  -----------------------------    | call ExportWebPage
call ExportWebPage |                      ||        ||                            ----
                ----                      ||        ||
                        ....................................................
                    future: move the post-MakeNotebook Open/Get/Put dance here 
                        ....................................................
                                          ||        ||             
                                          \/        \/                                                                 
                                       -----------------    ----
                                       | ExportWebPage | <= | do all the munging              
                                       -----------------    | call Transmogrify               
                                              ||            ----
                                              ||
                                              \/
                           ............................................
                            add a post-munging Open/Get/Put dance here 
                           ............................................
*)

ExportWebPageFromResultNB[outfile_String, nbfile_String, opts___?OptionQ] := Module[
  {lang, metainfo, failure, nbInfo, nbobj, nb, nbget, nbobj2}
, 
  Print["Entering ExportWebPageFromResultNB"];

  metainfo = Quiet @ DocumentationBuild`Utils`GetSearchMetaDataList[nbfile];
  lang = Switch[("language" /. metainfo),
    "en", "English",
    "ja", "Japanese",
    "zh", "Chinese",
    "es", "Spanish",
    _, "English"
  ];

  nbInfo = {
    "Title" -> ("windowtitle" /. metainfo),
    "TitleModifier" -> ("titlemodifier" /. metainfo),
    "HistoryData" -> Apply[{"New" -> #1, "Modified" -> #2, "Obsolete" -> #3, "Excised" -> #4} &, ("history" /. metainfo)],
    "EntityType" -> ("type" /. metainfo),
    "PacletName" -> ("paclet" /. metainfo),
    "Context" -> ("context" /. metainfo),
    "URI" -> ("uri" /. metainfo),
    "Keywords" -> ("keywords" /. metainfo),
    "TutorialCollectionLinks" ->  ("tutorialcollectionlinks" /. metainfo),
    "ForeignFunctionKeywords" -> {},
    "Synonyms" -> ("synonyms" /. metainfo),
    "Summary" -> ("summary" /. metainfo),
    "TMJTest" -> False,
    (* work around a failure with some 11.1 message pages *)
	"Flag" -> LookupSourceInformation[
		"paclet:" <> ("uri" /. metainfo),
		lang,
		"Flag",
		""
	],
    "IndexQ" -> ("index" /. metainfo)
  };

  (* Print["Export.m: at beginning of ExportWebPageFromResultNB, 'nbInfo' = "<>ToString[nbInfo]]; *)
  (* Print["and outfile = "<>ToString[InputForm[outfile]]]; *)

  System`StrokeForm;

  (* This is because of a problem seen with the version 10 build not making CellGroups right *)

  (* nbobj = NotebookOpen[nbfile, Visible -> False]; *)
  (* nbget = NotebookGet[nbobj]; *)
  (* NotebookClose[nbobj]; *)
  (* nbobj2 = NotebookPut[nbget, Visible -> False]; *)
  (* nb=NotebookGet[nbobj2]; *)
  (* NotebookClose[nbobj2]; *)

	failure = Catch[
		nbobj = NotebookOpen[nbfile, Visible -> False];
		If[Head[nbobj] =!= NotebookObject,
			DocsError["ExportWebPageFromResultNB: NotebookOpen 1: failure"];
		];

		nbget = NotebookGet[nbobj];
		If[Head[nbget] =!= Notebook,
			DocsError["ExportWebPageFromResultNB: NotebookGet 1: failure"];
		];

		NotebookClose[nbobj]; Clear[nbobj];

		nbobj = NotebookPut[nbget, Visible -> False];
		If[Head[nbobj] =!= NotebookObject,
			DocsError["ExportWebPageFromResultNB: NotebookOpen 2: failure"];
		];

		Clear[nbget];

		nb = NotebookGet[nbobj];
		If[Head[nb] =!= Notebook,
			DocsError["ExportWebPageFromResultNB: NotebookGet 2: failure"];
		];

		NotebookClose[nbobj];
		,
		$DocsErrorTag
	];

	If[FailureQ[failure],
		(* Refactor note: 1f38fba9:
			Handle NotebookOpen/Get/Put/Get failures

			ExportWebPageFromResultNB: when doing the NotebookOpen/Get/Put/Get
			expression cleaner, check each step for failure.  One cause of failures is
			erroneous expressions in the in-product notebooks (pinking).

			When failure is detected <...> reset the fe and inform the build system of
			page failure.  The fe is reset because the errors I'm seeing leave the build
			system in a state in which every subsequent page fails to build.  A reset
			cleans up whatever has gone wrong.  TODO: this may be a bug.
		*)
		Developer`UninstallFrontEnd[];
		(* Re-throw the failure. *)
		Throw[failure, $DocsErrorTag];
	];

	(*--------------------*)
	(* Call ExportWebPage *)
	(*--------------------*)

	ExportWebPage[outfile, nb, nbInfo, opts]
];

(* ** ExportWebPage[_, $Failed, ___] *)

ExportWebPage[_, $Failed, ___] := (
  Developer`UninstallFrontEnd[];
  Return[$Failed];
)

(* ** ExportWebPage[outfile_String, nbfile_String, nbInfo_List, opts___?OptionQ] *)

ExportWebPage[outfile_String, nbfile_String, nbInfo_List, opts___?OptionQ] :=
  ExportWebPage[outfile, GetQuietly[nbfile], nbInfo, opts];

(* ** final chance for munging: ExportWebPage[file_String, nb_?(!AtomQ[#]&), info_List, opts___?OptionQ] *)
(* *** initial diagram, declaration of variables *)

(*********************************************************************
billw 20151112: this definition of ExportWebPage is our final chance
to munge a notebook expression before sending it off to Transmogrify.

  |------------------|
  | source notebooks |
  |------------------|
          ||
          || <= make.m
          ||
          \/
|----------------------|
| in-product notebooks |
|----------------------|
          ||
          || <= export.m: do anything needed to 
          ||    make html transformation easier
          ||    [you are here]
          \/
|- - - - - - - - - - - - -+
| <<intermediate massaged | <-- this is ephemeral; 
|    expression>>         | it's also what gets transmogrified
|- - - - - - - - - - - - -+
          ||
          || <= XMLTransforms
          ||
          \/
       |------|
       | html |
       |------|

*********************************************************************)

ExportWebPage[
	outfile0_String,
	nb_?(!AtomQ[#]&),
	info_List,
	opts___?OptionQ
] := Block[{
	expr, outfile, language, langext, la, params, test, isManipulate,
	cache, cachelocation,
	completeHTMLQ,
	buildNumber,
	history,
	contextName,
	keywordList = {},
	transform, layoutfile,
	pageTitle = "None",
	pageTitleModifier = "",
	specialKeywords,
	tutorialFunctionList = {},
	tutorialMoreAboutList = {},
	relatedTutorialsList = {},
	guideFunctionList = {},
	guideMoreAboutList = {},
	guideTutorialsList = {},
	htmlBase = "",
	flag = None,
	urlBaseRemote, wsmlinkQ, wsmlibQ, wsmQ, webmQ, lwgQ,
	uri, summaryText,
	pacletQ,
	isTMJQ, isSolutionsQ,
	rootCaptions,
	nbobj, nbget, nbobj2,
	pacletName, entityType, synonymList, nbObj, featuredexampledata,
	seeAlsoFunctionList, moreAboutGuideList, exampleSectionList,  tutorialList,
	relativeBase,
	htmlTemplateIncludes,
	properties = {},
	r,
	errMessageList = {},
	buildmode, additionalstyles, defaultstylesheet,
	taggingRules,
	workflowQ = False,
	workflowOverviewQ = False,
	newin, updatedin = "", experimentalQ,
	exprBefore, exprAfter, packageQ,
  prerequisite
},

	dprint["Export.m: entering ExportWebPage: opts = "];dfile[ToString[{opts}, InputForm]];

    (* Check to see if a WSMLink build *)

	dprint["ExportWebPage, start of munging: info = "<>ToString[info, InputForm]]; dfile[info];
	dprint["ExportWebPage, start of munging: expr = "]; dfile[expr];

    uri = "URI" /. info;
    pacletName = "PacletName"/.info;

    wsmlinkQ = StringMatchQ[uri, "*WSMLink/*"]||StringMatchQ[pacletName, "*SystemModeler*"|"*System Modeler*"];
    wsmlibQ = StringMatchQ[pacletName, "Wolfram SystemModeler Library"];
    wsmQ = Or[wsmlinkQ, wsmlibQ];
    webmQ = StringMatchQ[pacletName, "webMathematica"];
    lwgQ = StringMatchQ[pacletName, "LightweightGridManager"];

	urlBaseRemote = "URLBaseRemote" /. {opts} /. Options[ExportWebPage] /. {Automatic -> "/language/"};

	If[!StringQ[urlBaseRemote],
		DocsError["Invalid value for urlBaseRemote: ``", InputForm[urlBaseRemote]];
	];

    Check[

    BuildLog["** Entering ExportWebPage **"];

    Off[CreateDirectory::filex];
    (* turn off error message limit *)
    Off[General::stop];

    dprint["Export.m: head[nb] = "<>ToString[Head[nb], InputForm]];
    nbobj2 = NotebookPut[nb, Visible -> False];
    dprint["Export.m: Head[nbobj2] = "<>ToString[Head[nbobj2], InputForm]];
    nbget=NotebookGet[nbobj2];
    dprint["Export.m: Head[nbget] = "<>ToString[Head[nbget], InputForm]];
    NotebookClose[nbobj2];

    (* Which language? *)

    language = "Language" /. {opts} /. Options[ExportWebPage];
    (* la will always be a language's 2-letter ISO 639-1 code as used in our documentation *)
    la = DocumentationBuild`Utils`LanguageAbbreviation[language];

    expr = nbget;
    dprint["Export.m: Head[expr] = "<>ToString[Head[expr], InputForm]];

	dprint["earliest possible expr immediately after getting it"];
	dfile[expr];

    SetOptions[$FrontEndSession, Rule[AutoStyleOptions, {Rule["HighlightUndefinedSymbols", False]}]];

	additionalstyles = {
		Cell[StyleData["InlineFormula"], FontFamily->"Source Sans Pro", FontSize->Inherited 1.08],
		Cell[StyleData["MR"], FontFamily->"Source Sans Pro"],
		Cell[StyleData["StandardForm"],FontFamily->"Source Sans Pro"],
		Cell[StyleData["InputForm"],FontFamily->"Source Sans Pro"],
		Cell[StyleData["Input"],  FontFamily->"Source Sans Pro", FontSize->13, LineSpacing->{1.2, 0}],
		Cell[StyleData["Output"], FontSize->13],
		Cell[
			StyleData["ServiceSubsection", StyleDefinitions -> StyleData["Section"]],
			Editable -> False,
			CellFrame -> False,
			CellGroupingRules -> {"SectionGrouping", 40},
			FontSize -> 16,
			FontWeight -> "Bold",
			FontColor -> GrayLevel[0.305882]
		],
		(*********************************************************************
		https://jira.wolfram.com/jira/browse/WLDOCS-870?focusedCommentId=2358054&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-2358054
		*********************************************************************)
		Cell[StyleData["TextFrame"], Thickness -> .25],
		Cell[StyleData["WorkflowInput"],
			ShowCellBracket -> True,
			CellMargins -> {{120, 10}, {5, 11}},
			ShowAutoSpellCheck -> False,
			SpanMaxSize -> 1,
			MenuSortingValue -> 1900,
			MenuCommandKey -> "9",
			FontFamily -> "Source Sans Pro",
			FontSize -> 13.85,
			FontWeight -> "DemiBold",
			FontSlant -> "Plain",
			FontTracking -> "Plain"
		]
	};

    defaultstylesheet= Cell[StyleData[StyleDefinitions -> FrontEnd`FileName[{"Wolfram"}, "Reference.nb"]]];

    (*
    grab the expr's tagging rules and convert the lists to associations for easy retrieval and updating:

    In[4]:= taggingRules["Metadata"]["context"]
    Out[4]= System`

    In[5]:= taggingRules["Metadata"]["context"] = "Gronk`"
    Out[5]= Gronk`

    In[6]:= taggingRules["Metadata"]["context"]
    Out[6]= Gronk`
    *)

	taggingRules = (Flatten @ Cases[expr, TaggingRules ~ (Rule|RuleDelayed) ~ a_ :> a, All])
		//. List[x___, r:_Rule, y___] -> Association[x, r, y];

    (* temporary kludge to make CloudConnectForExcel do the right thing since it has to happen right now *)
    (* If[StringContainsQ[taggingRules["Metadata"]["context"], "CloudConnectorForExcel"],
      taggingRules["Metadata"]["context"] = "CloudConnectorForExcel`"]; *)

	expr = expr /. {
		Rule[StyleDefinitions, FrontEnd`FileName[___]] :> Rule[
			StyleDefinitions,
			Notebook[
				Flatten[{defaultstylesheet, additionalstyles}],
				AutoStyleOptions -> {"HighlightUndefinedSymbols" -> False}
			]
		],
		Rule[StyleDefinitions, Notebook[{contents___}, options___]] :> Rule[
			StyleDefinitions,
			Notebook[
				Flatten[{contents, additionalstyles}],
				options,
				AutoStyleOptions -> {"HighlightUndefinedSymbols" -> False}
			]
		]
	};

	dprint["xxx1"];
	dfile[expr];

	rootCaptions = Cases[
		Options[expr],
		Rule["RootCaptions", a__] :> a,
		Infinity
	] /. {} -> {""};
	rootCaptions = If[Length @ Map[Cell[#, "GuideText"] &, rootCaptions[[1]]] >= 8,
		Take[
			Map[Cell[#, "GuideText"] &, rootCaptions[[1]]],
			8
		]
	];

    UseDocumentationBuildXMLTransforms[];

	dprint["xxx2"];
	dfile[expr];

    outfile = outfile0;

    dprint["ExportWebPage: earliest possible value of 'outfile' is "<>ToString[outfile, InputForm]];

    (* parse user opts *)
    params = "DefaultParameters" /. {opts} /. Options[ExportWebPage];
    test = "ExportTest" /. {opts} /. Options[ExportWebPage];
    isManipulate = "IsManipulate" /. {opts} /. Options[ExportWebPage];
    cache = "Cache" /. {opts} /. Options[ExportWebPage];
    cachelocation = "CacheLocation" /. {opts} /. Options[ExportWebPage];
    completeHTMLQ = "CompleteHTMLQ" /. {opts} /. Options[ExportWebPage];
    buildNumber = "BuildNumber" /. {opts} /. Options[ExportWebPage];
    pacletQ = "FunctionPaclet" /. {opts} /. Options[ExportWebPage];
    buildmode = "BuildMode" /. {opts} /. Options[ExportWebPage];
	specialKeywords = If[Length @ Cases[expr, Rule["specialkeywords", a_] :> a, Infinity] > 0,
		Cases[expr, Rule["specialkeywords", a_] :> a, Infinity][[1]]
		,
		{}
	];
    isTMJQ = "TMJTest" /. {opts} /. Options[ExportTMJWebPage];
    isSolutionsQ = "SolutionsTest" /. {opts} /. Options[ExportSolutionsWebPage];


	(* get paclet name *)
    pacletName = "PacletName" /. info;
	(* get entity type *)
	entityType = "EntityType" /. info;
	(* Formatting note: 21cac84cbed14bb8c988f5ed966124da7cac456e
		- a kludge for the (temporary, let it be hoped) problem described in
		  the first comment of WLDOCS-2305. *)
	If[uri == "UnityLink/guide/GettingStarted",
		entityType = "Workflow Guide"
	];
	(* get notebook flag *)
	flag = "Flag" /. info;
	(* get notebook history *)
	history = "HistoryData" /. info;
	(* get notebook summary *)
    summaryText = "Summary" /. info;

    pageTitle = If[entityType =!= "RootGuidePage",
        "Title" /. info
		,
		{"Wolfram Mathematica: Documentation Center"}
	];

    (* https://jira.wolfram.com/jira/browse/WLDOCS-1537: don't use the Title
    Modifier categorization information until it's revamped & approved

    pageTitleModifier ="TitleModifier" /. info; *)

    pageTitleModifier = "";

	(* capture keywords *)
	keywordList = "Keywords" /. info;
	synonymList = "Synonyms" /. info;

	(* page uri *)
	uri = "URI" /. info;

	(* get classification *)
    contextName = GetClassificationBoxes[info, language];

	(*contextName = If[(entityType === "Guide") && (pacletName === "Mathematica"), "",
      AddSpaceToCamelCase[contextName]];*)
    contextName = AddSpaceToCamelCase[contextName];

	(* Add the base name of the file as a keyword. *)
	(* Repeatedly apply FileBaseName to remove any '.html.la' suffixes. *)
	keywordList = Flatten[{
		Append[keywordList, FixedPoint[FileBaseName, outfile]]
	}];
	keywordList = If[Length[keywordList] > 0,
		StringDrop[StringDrop[ToString[keywordList], 1], -1],
		keywordList
	];

	(* Print["exporting unchanged expr (before munging) to /tmp/expr-01-before.m"]; *)
	(* Export["/tmp/expr-01-before.m", expr]; *)

	(* *** here beginneth the munging *)

	(*********************************************************************
	here beginneth our last-chance expression munging before running
	Transmogrify
	*********************************************************************)

	(* Print["Preparing notebook expression for Transmogrify..."]; *)

	(* **** misc munging *)

	(* ***** harvest and remove PrerequisiteTag *)
  (* will be passed to transmog as a parameter/property *)

  prerequisite = Quiet @ Check[First @ Cases[expr, StyleBox[a:_, "PrerequisiteTag", ___] -> a, Infinity], ""];
  dprint["'prerequisite' = "<>ToString[prerequisite, InputForm]];
	expr = expr /. StyleBox[_, "PrerequisiteTag", ___] -> Nothing;

(* ***** find the in-product nb's "New in" & "Updated in" version numbers and experimental status, if available *)

	(* these will be passed to Transmogrify as parameters/properties *)

	(* "11   \:306e\:65b0\:6a5f\:80fd" *)
	(* "11.1 \:306e\:65b0\:6a5f\:80fd" *)
	(* "11.2 \:306e\:65b0\:6a5f\:80fd" *)
	(* "11.3 \:306e\:65b0\:6a5f\:80fd" *)

	(* Formatting note: 983868c2:
		https://jira.wolfram.com/jira/browse/WLDOCS-2416
		https://jira.wolfram.com/jira/browse/WLDOCS-2426

		- Rework the newin/updatedin banners to simply follow what's in the
		in-product notebook.  This code uses the new taggingRules setup.
	*)
	(* This must be one of "", "t" or x_NumberString and is used in filling out the
	html file's iconography information.  "" produces no banner, "t" produces the
	"NEW" banner, x produces the "NEW IN X" banner. *)

  dprint["Export.wl: expr before messing with NewIn stuff: "]; dfile[expr];

(* ****** new in *)

	newin = Module[{newinString, res,
    newinPattern = StyleBox[GraphicsBox[InsetBox[StyleBox[new:_String, ___], ___], ___], "NewInGraphic", ___]
  },
		newinString = Quiet @ Check[
			First @ Flatten @ Cases[
				expr,
				newinPattern -> new,
				Infinity
			],
			""
		];

		dprint["Export.wl newin processing: newinString = "<>ToString[newinString, InputForm]];

(* all values of newinString in v13 in-product notebooks:
List["\"NEW\"", "\"NEW IN ??\"", "\"NEW IN 13\"", "\"NEW IN 13.0\"", "\"UPDATED IN 13\""] *)

		res = Which[
      StringContainsQ[newinString, {
        "UPDATED",
        (*zh:*) "的更新", (*which is:*) "\:7684\:66f4\:65b0",
        (*ja:*) "で更新", (*which is:*) "\:3067\:66f4\:65b0"
      }],
      updatedin = newinString; "" (* UPDATED does not indicate NEWness; carry this over to 'updatedin' processing *)
    , StringContainsQ[newinString, NumberString],
      (* canonicalize 13.0 to 13 *)
      StringReplace[First @ StringCases[newinString, NumberString], ".0" -> ""]
    , StringMatchQ[newinString, ""],
      ""
    , True,
      "t"
		];

    expr = expr /. Cell[BoxData[TooltipBox[StyleBox[_TagBox, "NewInGraphic", ___], ___], ___], ___] -> Sequence[];

		dprint["Export.wl newin processing: newin res = "<>ToString[res, InputForm]];

		res
	];

(* ****** updated, updated in *)

(* note that 'updatedin' may have been given a value above when looking in the
"NewInGraphic" structure, which as of v13 can contain "UPDATED IN NN" *)

	(* if the page has mod highlights, arrange to show an "UPDATED/show changes" banner that will turn them yellow when clicked *)
	updatedin = Module[{updatedVersionString, res},
		updatedVersionString = Quiet @ Check[
			First @ Flatten @ Cases[
				expr,
				StyleBox[
					s_String /; StringContainsQ[s, {"UPDATED IN", "\:3067\:66f4\:65b0", "\:7684\:66f4\:65b0"}],
					___
				] :> StringReplace[StringCases[s, NumberString], ".0" -> ""],
				Infinity
			],
			""
		];

		dprint["Export.wl update/version processing: updatedVersionString = "<>ToString[updatedVersionString, InputForm]];

    dprint["before finding UPDATED info, expr is "]; dfile[expr];

		res = Which[
      StringContainsQ[updatedin, {"UPDATED IN", "\:3067\:66f4\:65b0", "\:7684\:66f4\:65b0"}],
        StringReplace[First @ StringCases[updatedin, NumberString], ".0" -> ""],
			Positive @ Count[expr, TooltipBox[_TogglerBox, "Toggle changes", ___], Infinity],
				"t",
			StringContainsQ[updatedVersionString, NumberString],
				updatedVersionString,
			True,
				""
		];

		dprint["Export.wl update/version processing: res = "<>ToString[res, InputForm]];

		res
	];

(* ****** experimental *)

	experimentalQ = Positive @ Count[expr, Cell[_, "ExperimentalBanner", ___], All];
  expr = expr /. Cell[_, "ExperimentalBanner", ___] -> Sequence[];

(* ****** final results of new/updated/experimental *)

	dprint["Export.m, top of munging: {newin, updatedin, experimentalQ} = "<>ToString[{newin, updatedin, experimentalQ}, InputForm]];

(* ***** continuing *)

	(* Formatting note: a37e7104:
		http://bugs.wolfram.com/show?number=348362

		- in Method pages, wrap bare Notes cells in a FormatUsage cell (may have been
		  overlooked in Make.m)
	*)
	(* overlooked in Make.m? *)
	expr = If[entityType == "Method",
		expr /. Cell[a__, "Usage", b___] :> Cell[a, "FormatUsage", b]
		,
		expr
	];
	expr = If[entityType == "Method",
		expr /. Cell[a__, "Notes", b___] :> Cell[a, "FormatNotes", b]
		,
		expr
	];

	(* Filter out the bad cells in function page examples causing double traingles. *)
	expr = expr /. Cell[
		BoxData[InterpretationBox[
			Cell[
				TextData[Cell[BoxData[InterpretationBox[
						Cell[TextData[{a1__, Cell[a2_,"ExampleCount"]}], "ExampleSection"],
						b__
				]]]],
				"ExampleSection"
			],
			c__
		]],
		"ExampleSection",
		d__
	] :> Cell[
		BoxData[InterpretationBox[
			Cell[TextData[{a1, Cell[a2, "ExampleCount"]}], "ExampleSection"],
			b
		]],
		"ExampleSection"
	];

	(* **** root guide page,  *)

	(* There used to be one kind of tile on the root guide page: click and get a dropdown menu.
		As of v12 there are four kinds of items scattered here and there around the page:
			- click to get a menu
			- click to go to a target
			- an inert section header providing grouping
			- ordinary text hyperlinks

		The goals of the following code:
			- easily add even more kinds of structures
			- handle structures in any order and location without worries
			- put all the pieces back into notebook order after working on them
			- don't try to write and maintain One Pattern to match them all and in the docbuild bind them. *)

	If[entityType === "Root Guide", (
		(* remove 99.5% of the expression's ByteCount by removing images - they're handled via css: *)
		expr = expr //. {
			_GraphicsBox -> graphicsBox[],
			_RasterBox -> rasterBox[]
		};

		dprint["Export.m: initial root guide expr is"]; dfile[expr];

		expr = Module[{res = {}, positions},
			(* ***** DONE main tiles, each with a dropdown list of paclet links *)
			(*       CLOSED: [2019-05-10 Fri 00:36] *)

			dprint["step 1: main tiles, each with a dropdown list of paclet links"];

			positions = Position[
				expr,
				TagBox[
					ActionMenuBox[
						PaneSelectorBox[titleIsInside:_List, ___],
						listOfButtons:_List,
						___
					],
					___
				],
				Infinity
			];
			AppendTo[
				res,
				(* TODO: Replace this Reap/Sow with a Map -- the Sow operation is always
				         the very last expression. *)
				Reap[
					Module[{pos, x, linkList},

						pos = #;
						x = Extract[expr, pos];

						(* TODO: Use FirstCase here? Alternatively, validate that there is exactly one
								"RootSectionTitle", for future-proofing. *)
						title = Quiet @ Check[
							First @ DeleteDuplicates @ Cases[
								x,
								StyleBox[s:_String, "RootSectionTitle", ___]
									:> StringReplace[s, {"\\n" -> " ", "\"" -> "", Whitespace.. -> " "}],
								Infinity
							],
							""
						];
						dprint["title = "<>ToString[title, InputForm]];

						linkList = Cases[
							x,
							Delimiter | RuleDelayed[
								_String,
								FEPrivate`FrontEndExecute[FrontEndToken["OpenHelpLink", {Part[{_String, _String}, _], ___}]]
							],
							Infinity
						];
						dprint["linkList 1 = "]; dfile[linkList];

						linkList = linkList /. RuleDelayed[
							_String,
							FEPrivate`FrontEndExecute[FrontEndToken[
								"OpenHelpLink",
								{Part[{s1:_String, s2:_String}, _], ___}
							]]
						] :> Cell[
							BoxData[ButtonBox[s1, BaseStyle -> "Link", ButtonData -> s2]],
							"RootItem"
						];

						(* split the link list into a list of cellgroups *)
						linkList = SplitBy[linkList, # =!= Delimiter&] /. {Delimiter} -> Sequence[];
						linkList = linkList /. {c:__Cell} :> Cell[CellGroupData[{c}]];

						If[linkList == {},
							(* Formatting note: cc3aad19:
								- accommodate special styling for Fast Introduction links and link lists *)
							linkList = Cases[
								x,
								{s1:_String, {URL[s2:_String], None}} :>
									If[StringContainsQ[s2, "/fast-introduction"],
										Cell[BoxData[ButtonBox[s1, BaseStyle -> "Link", ButtonData -> s2]], "RootItemFastIntroductions"],
										Cell[BoxData[ButtonBox[s1, BaseStyle -> "Link", ButtonData -> s2]], "RootItem"]
									],
								Infinity
							]
						];
						dprint["linkList 2 = "]; dfile[linkList];

						(* on Sowing with a list as the second argument: https://mathematica.stackexchange.com/a/44121/76 *)
						Sow[
							Cell[CellGroupData[Flatten[{
								Cell[title, "RootSectionTitle"],
								Cell[CellGroupData[linkList]]
							}]]],
							{pos}
						];

					]& /@ positions,
					_,
					Rule
				][[2]]
			];

			(* ***** DONE a tile that is itself a link with no dropdowns. *)
			(*       CLOSED: [2019-05-10 Fri 00:36] *)

			(* TagBox[PaneSelectorBox[titleAndLinkAreInside:_List, ___], ___] *)

			positions = Position[
				expr,
				TagBox[PaneSelectorBox[titleAndLinkAreInside:_List, ___], ___],
				Infinity
			];
			AppendTo[
				res,
				(* TODO: Replace this Reap/Sow with a Map -- the Sow operation is always
				         the very last expression. *)
				Reap[
					Module[{pos, x, title, link},
						pos = #;
						x = Extract[expr, pos];

						(* TODO: Validate that there is always exactly one "RootSectionTitle" *)
						title = Quiet @ Check[
							First @ DeleteDuplicates @ Cases[
								x,
								StyleBox[
									s:_String /; StringFreeQ[s, StartOfString~~"\"»\""~~EndOfString],
									"RootSectionTitle",
									___
								] :> StringReplace[s, {
									"\\n" -> " ",
									"\"" -> "",
									Whitespace.. -> " "
								}],
								Infinity
							]
							,
							""
						];

						link     = Cases[x, Rule[ButtonNote, s:_String       ] -> s, Infinity];
						If[link === {},
							link = Cases[x, Rule[ButtonData, {URL[s:_String]}] -> s, Infinity]
						];
						If[link === {},
							link = Cases[x, Rule[ButtonData, s:_String       ] -> s, Infinity]
						];

						link = Quiet @ Check[First[link], {}];

						(* this is a single tile so it's in its own cellgroup *)
						cell = Cell[CellGroupData[{Cell[title, "RootSectionTitle", "SingleLink" -> link]}]];

						Sow[cell, {pos}];
					]& /@ positions,
					_,
					Rule
				][[2]]
			];

			(* ***** DONE headings: "WORKFLOWS", "RESOURCES", empty junk headings to be removed: *)
			(*       CLOSED: [2019-05-10 Fri 00:36] *)

			(* keep: ItemBox[StyleBox["\"WORKFLOWS\"", ___], ___]
			toss: ItemBox[StyleBox["\"\"", ___], ___] *)

			(* TODO: Does this work? Should it have Infinity as the third argument,
			         like the similar calls to Position[..] in this section of the code? *)
			positions = Position[expr, ItemBox[StyleBox[s:_String, ___], ___]];
			AppendTo[
				res,
				(* TODO: Replace this Reap/Sow with a Map -- the Sow operation is always
				         the very last expression. *)
				Reap[
					Module[{pos, x, title},

						pos = #;

						x = Extract[expr, pos];
						x = x /. ItemBox[StyleBox["\"\"", ___], ___] -> Sequence[];

						title = Quiet @ Check[
							(* TODO: Validate that this is the ItemBox[StyleBox[..]] value
							         we expect, somehow? Is there a specific cell style
									 we're expecting, e.g. "WORKFLOWS" as mentioned in the
									 prior comment? *)
							First @ Cases[
								x,
								StyleBox[s:_String, ___]
									:> StringReplace[s, {
										"\\n" -> " ",
										"\"" -> "",
										Whitespace.. -> " "
									}],
								Infinity
							],
							""
						];

						If[title =!= "",
							Sow[Rule[Cell[title, "RootSectionHeading"], "heading"], {pos}]
						];
					]& /@ positions,
					_,
					Rule
				][[2]]
			];

			(* ***** TODO row of links at bottom of page: *)

			(* TemplateBox[{TemplateBox[linksAreInside:_List, ___]..}, "RowWithSeparator"] *)

			(* Formatting note: 3a29e04d:
				language updates for root guide page

				WBI work: https://jira.wolfram.com/jira/browse/WLDOCS-2480
				DQA work: https://jira.wolfram.com/jira/browse/WLDOCS-2481

				- handle tile-less links at the bottom of the root guide page, dubbed RootFooterLinks
				- add ja and zh versions of headings for nb manipulation
			*)
			dprint["Export.m: expr before row of links at bottom of page:"]; dfile[expr];

			positions = Position[
				expr,
				TemplateBox[
					{PaneSelectorBox[{Rule[False, _], True -> _}, ___], _},
					"RefLink",
					BaseStyle -> "RootFooterLink"
				],
				Infinity
			];
			AppendTo[
				res,
				Reap[
					Module[{pos, x},
						pos = #;
						x = Extract[expr, pos];
						{text, link} = Flatten @ Cases[
							x,
							TemplateBox[
								{PaneSelectorBox[{_, Rule[True, StyleBox[s1:_String, ___]]}, ___], s2:_String},
								"RefLink",
								BaseStyle -> "RootFooterLink"
							] :> {
								StringReplace[s1, {
									"\\n" -> " ",
									"\"" -> "",
									Whitespace.. -> " "
								}],
								s2
							},
							All
						];
						Sow[
							Cell[
								TextData[{ButtonBox[
									text,
									BaseStyle -> "Link",
									ButtonData -> link
								]}],
								"RootFooterLink"
							],
							{pos}
						];
					]& /@ positions,
					_,
					Rule
				][[2]]
			];

			(* Also, the last link is different in-product and on the web.  In-product it
			points to a dynamically-generated page listing the user's installed packages.
			Web should point to https://www.wolfram.com/products/applications/ *)

			res = res /. "paclet:guide/InstalledAddOns" -> "https://www.wolfram.com/products/applications/";

			(* ***** DONE sort page elements into their original notebook order *)
			(*       CLOSED: [2019-05-10 Fri 00:38] *)

			res = Flatten[res];
			(* turn Positions into sorting keys: *)
			res[[All, 1]] = Map[
				ArrayPad[#, {0, Max[Length /@ res[[All, 1]]] - Length[#]}]&,
				res[[All, 1]]
			];
			res = Sort[res]; (* put elements back into their original notebook order sorting by Position *)
			res = Flatten[res[[All, 2]]]; (* remove sorting keys and flatten *)
			res = Notebook[res];

			(* ***** http:// -> https:// throughout *)

			res = res /. s:_String :> StringReplace[s, "http://" -> "https://"];

			(* ***** DONE wrap sections as needed *)
			(*       CLOSED: [2019-05-10 Fri 00:38] *)

			res = res /.
				Notebook[{first:___, cells:Longest[Cell[_, "RootFooterLink"]..]}] ->
				Notebook[{first, RootFooterCell[CellGroupData[{cells}]]}];

			res = res /.
				Notebook[{cells:Longest[_Cell..], rest:___}] ->
				Notebook[{Cell[CellGroupData[{Cell["DEFAULT", "RootSectionHeading"], cells}]], rest}];

			res = res /.
				Notebook[{first:___, Rule[head:Cell["WORKFLOWS"|"\:30ef\:30fc\:30af\:30d5\:30ed\:30fc"|"\:5de5\:4f5c\:6d41\:7a0b", "RootSectionHeading"], _], cells:Longest[_Cell..], rest:___}] ->
				Notebook[{first, Cell[CellGroupData[{head, cells}]], rest}];

			res = res /.
				Notebook[{first:___, Rule[head:Cell["RESOURCES"|"\:30ea\:30bd\:30fc\:30b9"|"\:8d44\:6e90", "RootSectionHeading"], _], cells:Longest[_Cell..], rest:___}] ->
				Notebook[{first, Cell[CellGroupData[{head, cells}]], rest}];

			res = res /. RootFooterCell -> Cell;

			dprint["Export.m: final value of root guide page res = "]; dfile[res];
			res
		]
	)];


	(* Formatting note: c4c90ed0:
		Fix Usage structure in GUIKit/ref/widget
		https://jira.wolfram.com/jira/browse/WLDOCS-3385#comment-3330832
	*)
	(* **** supply an ObjectNameGrid to those pages missing it *)

	If[
		StringContainsQ[uri, "GUIKit/ref/widget/"]
		&& Count[expr, Cell[_, "ObjectNameGrid", ___], Infinity] == 0
		,
		expr = expr /. c:Cell[_, "ObjectName", ___]
			:> Cell[
				BoxData[GridBox[{{
					Cell[TextData[{
						c,
						Cell[BoxData[TemplateBox[{8}, "Spacer1"]]],
						Cell[BoxData[""], "ObjectNameTranslation"]
					}]],
					"\[SpanFromLeft]"
				}}]],
				"ObjectNameGrid"
			]
	];

	(* **** continuing *)

	(* Formatting note: c7b84a91: character image issue fixed *)
	expr = expr /. {
		Cell[
			BoxData[GridBox[{{
				a___,
				GridBox[
					b___,
					GridBoxItemSize -> {
						"Columns" -> {{Scaled[0.75]}},
						"Rows" -> {{0.4}}
					},
					c___
				]
			}}]],
			"CharacterImage"
		] :> Cell[
			BoxData[GridBox[{{
				a,
				GridBox[
					b,
					GridBoxItemSize -> {
						"Columns" -> {{}},
						"Rows" -> {{0.4}}
					},
					c
				]
			}}]],
			"CharacterImage"
		]
	};

	(* Formatting note: 5edf07c6:
		adding ExampleSection and ExampleSubsection twice per jfultz request *)
    expr = expr /.
      Cell[a_,"ExampleSection", "ExampleSection", b___]
        :> Cell[a, "ExampleSection", b];
     expr = expr /.
      Cell[a_,"ExampleSubsection", "ExampleSubsection", b___]
        :> Cell[a, "ExampleSubsection", b];

	(* Formatting note: 8edf6da7:
		Fix for bug #76118 to fix cell tags which were getting stripped from
		Example Section heads. *)
	expr = expr /.
		Cell[
			BoxData[InterpretationBox[stuff:Cell[a_,"ExampleSection",b___], $Line = 0;]],
			"ExampleSection",
			c___
		] :> Cell[a, "ExampleSection", b, c];

	expr = expr /.
		Cell[
			BoxData[InterpretationBox[stuff:Cell[a_,"ExampleSubsection",b___], $Line = 0;]],
			"ExampleSubsection",
			c___
		] :> Cell[a, "ExampleSubsection", b, c];

	expr = expr /.
		Cell[
			BoxData[InterpretationBox[stuff:Cell[a_,"ExampleSubsubsection",b___], $Line = 0;]],
			"ExampleSubsubsection",
			c___
		] :> Cell[a, "ExampleSubsubsection", b, c];

	expr = expr /.
		Cell[
			BoxData[InterpretationBox[stuff_, $Line = 0;]],
			"PrimaryExamplesSection",
			a___
		] :> Cell[
			Localized["PrimaryExamplesSectionText", language],
			"PrimaryExamplesSection",
			a
		];

	expr = expr /.
		Cell[
			BoxData[InterpretationBox[stuff_, $Line = 0;]],
			"ExampleDelimiter",
			s___
		] :> Cell["\t", "ExampleDelimiter", s];

	expr = expr /.
		Cell[
			cellcontent_,
			"InlineFormula",
			"TemplateInclusion",
			cellOptions___
		] :> Cell[cellcontent, "InlineFormula", cellOptions];

	(* TODO: Are the following two rules identical to the previous rule? Why are these
	         repeated? Is this just a copy-paste mistake, or is there a reason? Or are
			 they the same and I'm blind? *)
	expr = expr /.
		Cell[
			cellcontent_,
			"InlineFormula",
			"TemplateInclusion",
			cellOptions___
		] :> Cell[cellcontent, "InlineFormula", cellOptions];

	expr = expr /.
		Cell[
			cellcontent_,
			"InlineFormula",
			"TemplateInclusion",
			cellOptions___
		] :> Cell[cellcontent, "InlineFormula", cellOptions];

	(* Formatting note: 020757c5:
		https://jira.wolfram.com/jira/browse/RELEASE-3575
		Add patterns to catch various arcane and malformed Return/Enter buttons. *)
    (* Symbols/AutoIndent.nb revision 1.70, Symbols/Piecewise.nb revision 1.91 *)
	expr = expr /.
		Cell[
			BoxData[FormBox[
				DynamicBox[
					ToBoxes[
						If[$OperatingSystem === "MacOSX", a_, b_],
						c__
					],
					d___
				],
				TraditionalForm
			]],
			"KeyEvent",
			e___
		] :> StyleBox[b, "KeyEvent"];

	expr = expr /.
		Cell[
			BoxData[FormBox[
				StyleBox[
					DynamicBox[
						ToBoxes[
							If[$OperatingSystem === "MacOSX", a_, b_],
							c__
						],
						d___
					],
					"KeyEvent"
				],
				TraditionalForm
			]],
			"KeyEvent",
			e___
		] :> StyleBox[b, "KeyEvent"];

	expr = expr /.
		StyleBox[
			DynamicBox[
				ToBoxes[
					If[$OperatingSystem === "MacOSX", a_, "\[NegativeMediumSpace]"],
					b___
				],
				c___
			],
			"KeyEvent"
		] :> " ";

	expr = expr /.
		Cell[
			BoxData[FormBox[
				RowBox[{
					a___,
					StyleBox[
						DynamicBox[
							ToBoxes[
								If[$OperatingSystem === "MacOSX", aa_, bb_],
								c__
							],
							___
						],
						"KeyEvent"
					]
				}],
				Style_
			]],
			NotebookDefault,
			cellOptions___
		]:> Sequence[a, StyleBox[bb, "KeyEvent"]];

	expr = expr /.
		Cell[
			BoxData[FormBox[
				RowBox[{
					a___,
					StyleBox[
						DynamicBox[ToBoxes[b_, _], ___],
						"KeyEvent"
					]
				}],
				Style_
			]],
			NotebookDefault,
			cellOptions___
		] :> Sequence[a, StyleBox[b, "KeyEvent"]];

	expr = expr /.
		Cell[
			BoxData[FormBox[
				DynamicBox[
					If[$OperatingSystem === "MacOSX", a_String, b_String],
					__
				],
				TextForm
			]],
			"KeyEvent"
		] :> Sequence[
			StyleBox[b, "KeyEvent"],
			" (",
			StyleBox[a, "KeyEvent"],
			" on MacOSX)"
		];

	expr = expr /.
		StyleBox[
			DynamicBox[
				ToBoxes[If[$OperatingSystem === "MacOSX", a_, b_], StandardForm],
				c___
			],
			"KeyEvent"
		] :> StyleBox[ToBoxes[b, StandardForm], "KeyEvent"];

	expr = expr /. DynamicBox[If[$OperatingSystem === "MacOSX", a_, b_], c___] :> b;

	expr = expr /. DynamicBox[ToBoxes[If[$OperatingSystem === "Windows", a_, b_], c___], d___] :> a;

	(* System/Piecewise.nb revision 1.91 *)
	expr = expr /. DynamicBox[ToBoxes[If[$OperatingSystem === "MacOSX", a_, b_], c___], d___] :> a;

	expr = expr /.
		StyleBox[
			DynamicBox[
				ToBoxes[
					If[$OperatingSystem === "MacOSX", a__, "\[NegativeMediumSpace]"],
					StandardForm
				],
				c___
			],
			"TableText" | "MonographTableText"
		] :> " ";

	expr = expr /. StyleBox[" ", "KeyEvent"] :> " ";

	dprint["xxx6"];
	dfile[expr];

	(* Formatting note: a725a8c1: begin tracking down InlineFormula errors *)
	expr = expr /. {
		Cell[
			BoxData[FormBox[a_, b_]],
			"InlineFormula",
			rest:___
		] :> Cell[
			BoxData[FormBox[a, TraditionalForm]],
			"InlineFormula",
			rest
		],
		Cell[
			BoxData[FormBox[a_, b_]],
			"InlineMath",
			rest:___
		] :> Cell[
			BoxData[FormBox[a, TraditionalForm]],
			"InlineMath",
			rest
		]
	};

	(* Formatting note: d4c9687a:
		ExportWebPage: add more styles to InlineFormula math spacing mix. *)
	expr = expr /. {
		Cell[
			BoxData[RowBox[{a__ButtonBox, b__}]],
			d_String /; MemberQ[{"InlineFormula","InlineCode","InlineOutput","InlineMath","FunctionTemplate"}, d]
		] :> Sequence[
			Cell[BoxData[a], d],
			Cell[BoxData[RowBox[{b}]], d]
		],
		Cell[
			BoxData[RowBox[{a__, b__ButtonBox, c__}]],
			d_String /; MemberQ[{"InlineFormula","InlineCode","InlineOutput","InlineMath","FunctionTemplate"}, d]
		] :> Sequence[
			Cell[BoxData[RowBox[{a}]], d],
			Cell[BoxData[b], d],
			Cell[BoxData[RowBox[{c}]], d]
		],
		Cell[
			BoxData[RowBox[{a__, b__ButtonBox}]],
			d_String /; MemberQ[{"InlineFormula","InlineCode","InlineOutput","InlineMath","FunctionTemplate"}, d]
		] :> Sequence[
			Cell[BoxData[RowBox[{a}]], d],
			Cell[BoxData[b], d]
		]
	};

	expr = expr /.
		Cell[
			TextData[{a___,Cell[aa__, "KeyEvent"],b___}],
			"Text",
			c___
		] :> Cell[
			TextData[{a,StyleBox[aa, "KeyEvent"],b}],
			"Text",
			c
		];

	expr = expr /.
		Cell[
			cellcontent__,
			"NumberedEquation",
			cellOpts__
		] :> Cell[
			cellcontent,
			"DisplayFormula",
			cellOpts
		];

	expr = expr /.
		Cell[
			BoxData[GridBox[{{
				a___,
				Cell[TextData[{
					b___,
					Cell[c__, "InlineMath", d___],
					e___
				}]],
				f___
			}}]],
			"Usage",
			g___
		] :> Cell[
			BoxData[GridBox[{{a,
				Cell[TextData[{
					b,
					Cell[
						c,
						"InlineMath",
						"GridBoxOptions" -> {
							BaseStyle -> {FontFamily -> "Times"},
							GridBoxSpacings -> {},
							GridBoxDividers -> {}
						},
						d
					],
					e
				}]],
				f
			}}]],
			"Usage",
			g
		];

	expr = expr /.
		Cell[a___, "DeviceUsage", b___] :> Cell[
			MapAll[If[StringQ[#], StringReplace[#, "\n" -> "\[LineSeparator]"], #] &, a],
			"DeviceUsage",
			b
		];

	dprint["xxx7"];
	dfile[expr];

	(* Formatting note: 7f64502a0620ce4c2a963cf67df6a5634c849fbf:
		Fix the margins in some ref/device/{GPIO,I2C}

		- select only the "DeviceNotes" and "1ColumnTableMod" cells under ObjectName for special margin
		fixes via DeviceListA, which has extra apparatus compared to the regular DeviceList cellgroup.
	*)
	(* extending per https://jira.wolfram.com/jira/browse/WLDOCS-1994 *)
	expr = FixedPoint[
		Replace[
			#
		,
			Notebook[
				{
					aaa___,
					Cell[
						CellGroupData[
							{
								a:Cell[___, "ObjectNameGrid" | "ObjectNameAlt", ___],
								b___,
								c:Longest@PatternSequence[Cell[___, "DeviceNotes" | "1ColumnTableMod", ___] ..],
								d___
							},
							aa___
						],
						bb___
					],
					bbb___
				},
				ccc___
			] :> Notebook[
				{
					aaa,
					(* DeviceListA: a special version of DeviceList for use in the ObjectName cellgroup of a device notebook *)
					Cell[
						CellGroupData[
							{a, b, Cell[CellGroupData[{Cell["", "DeviceListA"], c}]], d},
							aa
						],
						bb
					],
					bbb
				},
				ccc
			]
		]&
		,
		expr
	];

	dprint["Export.m: before FormatList cellgroup insertion, expr = "];
	dfile[expr];
	dfile[expr, "Format" -> ".nb"];

	(* Formatting note: c15152c7b1f6dc4a7d355ecfe3073b4b708056ea:
		Fix for longstanding ref/format bug
		https://bugs.wolfram.com/show?number=387876

		The initial list of items on the page need to be in a gray box with bulleted list
		items. *)
	expr = expr /. {
		first:___,
		a:Cell[_, "ObjectNameGrid" | "ObjectNameAlt", ___],
		Longest[b:Cell[_, "FormatNotes" | "FormatUsage", ___]..],
		rest:___
	} :> {
		first,
		a,
		Cell[CellGroupData[{
			Cell["", "FormatList"],
			b
		}]],
		rest
	};

	dprint["Export.m: after FormatList cellgroup insertion, expr = "];
	dfile[expr];
	dfile[expr, "Format" -> ".nb"];

	(* Formatting note: 8d3f77e9:
		Update formatting for the ref/format Background section

		https://jira.wolfram.com/jira/browse/WLDOCS-2076

		- rewrite & simplify the FormatUsage section for conversion to suef's improved
		web structures.
		- new styles FormatUsageWrapper, FormatUsageColumnOne, FormatUsageColumnTwo;
		repurposed FormatUsage
		- remove FormatUsageSeparator - these are inserted by css.
		- retain the "\n" -> LineSeparator converter

		XMLTransforms/HTML/Components/Defaults.m:
			- treat FormatUsage separately from InterpreterUsage and DeviceUsage, at least
			for now. I'm not sure yet whether any instances of those extend to two
			columns.
		*)
    expr = expr //.
      Cell[a_String, "DeviceDoubleUsage", d___]
        :> Cell[a, "DeviceUsage", d];

	(* **** FormatUsage munging *)

	dprint["entering FormatUsage munging, expr = "];
	dfile[expr];

	(* remove degenerate cells *)

	expr = expr /. {
		Cell["", "FormatUsage", ___] -> Sequence[]
	};

	(*
		FormatUsageWrapper
			- FormatUsageColumnOne
				- FormatUsage
				- FormatUsage
				- FormatUsage
				- FormatUsage
			- FormatUsageColumnTwo
				- FormatUsage
				- FormatUsage
				- FormatUsage
	*)

	expr = expr /.
		Cell[
			BoxData[
				GridBox[{{a_GridBox, b_GridBox}, ___}, ___],
				___
			],
			"FormatUsage",
			___
		] :> Cell[
			TextData[{
				Cell[
					TextData[
						Flatten @ {
							Cases[a, Cell[_, "FormatUsage", ___], Infinity]
						}
					],
					"FormatUsageColumnOne"
				],
				Cell[
					TextData[
						Flatten @ {
							Cases[b, Cell[_, "FormatUsage", ___], Infinity]
						}
					],
					"FormatUsageColumnTwo"
				]
				(* are there any FormatUsage enclosing cells =!= 2 columns? *)
			}],
			"FormatUsageWrapper"
		];

	expr = expr /.
		Cell[
			TextData[{StyleBox[_, "FormatUsageSeparator", ___], a__}, ___],
			b__
		] :> Cell[TextData[{a}], b];

	(* expr = expr /.
	Cell[
		BoxData[GridBox[{{a___, Cell[TextData[{b___}]], c___}}]],
		"FormatUsage",
		d___
	] :>
	Cell[TextData[{b}], "FormatUsage"]; *)

	expr = expr /.
		Cell[a_, "FormatUsage", b___]
		:> Cell[
			MapAll[
				If[StringQ[#], StringReplace[#, "\n" -> "\[LineSeparator]"], #]&,
				a
			],
			"FormatUsage",
			b
		];

	dprint["leaving FormatUsage munging, expr = "];
	dfile[expr];

	(* ***** IndicatorUsage *)
	(* Formatting note: 86759629:
		WLDOCS-1397: proper paragraphing for IndicatorUsage

		- reworked IndicatorUsage munging from scratch: harvested all from ip
		& developed patterns to match all cases.  Fortunately,
		IndicatorUsage cells occur only in Indicator ip nbs.
	*)

	(* 11.3 ref/indicator/AccumulationDistributionLine, et al. *)
	(* an IndicatorUsage cell with no child cells *)
	expr = expr /.
		Cell[a_String, "IndicatorUsage", ___]
		:> Cell[
			BoxData[Cell[#, "IndicatorUsageItem"]& /@ StringSplit[a, "\n"]],
			"IndicatorUsage"
		];

	(* 11.3 ref/indicator/CloseLocationValue, et al. *)
	(* an IndicatorUsage cell with one {"", Cell[a___]} in a GridBox *)
	expr = expr /.
		Cell[
			BoxData[GridBox[{{"", Cell[a___]}}]],
			"IndicatorUsage",
			d___
		] :> Cell[
			BoxData[Cell[#, "IndicatorUsageItem"]& /@ StringSplit[a, "\n"]],
			"IndicatorUsage"
		];

	(* 11.3 ref/indicator/AbsolutePriceOscillator, et al. *)
	(* an IndicatorUsage cell with two {"", Cell[a___]} in a GridBox *)
	expr = expr /.
		Cell[
			BoxData[GridBox[{{"", Cell[a___]}, {"", Cell[b___]}}]],
			"IndicatorUsage",
			d___
		] :> Cell[
			BoxData[Flatten@{
				(Cell[#, "IndicatorUsageItem"]& /@ StringSplit[a, "\n"]),
				Cell[b, "IndicatorUsageItem"]
			}],
			"IndicatorUsage"
		];

	(* 11.3 ref/indicator/DoubleExponentialMovingAverage only *)
	(* an IndicatorUsage cell with three {"", Cell[a___]} in a GridBox *)
	expr = expr /.
		Cell[
			BoxData[GridBox[{{"", Cell[a___]}, {"", Cell[b___]}, {"", Cell[c___]}}]],
			"IndicatorUsage",
			d___
		] :> Cell[
			BoxData[Flatten@{
				(Cell[#, "IndicatorUsageItem"]& /@ StringSplit[a, "\n"]),
				Cell[b, "IndicatorUsageItem"],
				Cell[c, "IndicatorUsageItem"]
			}],
			"IndicatorUsage"
		];

	(* ***** UpgradeSummary *)

	(* does UpgradeSummary need the same treatment as IndicatorUsage?  start with a census *)

	expr = expr /.
		Cell[a___, "UpgradeSummary", b___]
		:> Cell[
			MapAll[If[StringQ[#], StringReplace[#, "\n" -> "\[LineSeparator]"], #] &, a],
			"UpgradeSummary",
			b
		];

	(* ***** TutorialCollection *)

	(* Formatting note: b34c74b3:
		Fix tutorial collection head in ref html pages. *)
	expr = expr /.
		Cell[a_, "TutorialCollectionSection", aa___]
		:> Cell[
			Localized["TutorialCollectionSectionText", language],
			"TutorialCollectionSection",
			aa
		];

	(* Formatting note: 4eae81fc:
		Add filters for Tutorial Collection section capitalization and grouping. *)
	expr = expr /.
		Cell[CellGroupData[
			{
				Cell[a_,"TOCChapter",aa__],
				Cell[CellGroupData[
					{
						Cell[TextData[b_ButtonBox], "TOCSection", bb__],
						c__
					},
					Open
				]],
				d__,
				Cell[CellGroupData[{Cell[e_,"TutorialCollectionSection"],ee__},Open]]
			},
			Open
		]] :> {
			Cell[CellGroupData[
				{
					Cell[a,"TOCChapter", aa],
					Cell[CellGroupData[
						{
							Cell[TextData[b], "TOCSection", bb],
							c
						},
						Open
					]],
					d
				},
				Open
			]],
			Cell[CellGroupData[
				{Cell[e,"TutorialCollectionSection"],ee},
				Open
			]]
		};

	(* dprint["Export.m transform test 1: is expr changed by this?"]; *)
	(* exprBefore = expr; *)
	expr = expr /.
		Cell[CellGroupData[
			{
				Cell[a_String, "TOCDocumentTitle", b___],
				x___,
				Cell[CellGroupData[
					{
						Cell[aa__, "TOCChapter", bb___],
						Cell[CellGroupData[
							{
								Cell["TUTORIAL COLLECTION", "TutorialCollectionSection", c1___],
								c2___
							},
							c3___
						]]
					},
					c4___
				]]
			},
			c5___
		]] :> {
			Cell[
				CellGroupData[
					{
						Cell[a, "TOCDocumentTitle", b],
						x,
						Cell[CellGroupData[
							{Cell[aa, "TOCChapter", bb]},
							c4
						]]
					},
					c5
				]
			],
			Cell[CellGroupData[
				{
					Cell["TUTORIAL COLLECTION", "TutorialCollectionSection", c1],
					c2
				},
				c3
			]]
		};
	(* exprAfter = expr; *)
	(* dprint["Export.m transform test 1: "<>If[exprBefore === exprAfter, "expr is unchanged", "expr is changed"]]; *)
	(* Clear[exprBefore, exprAfter]; *)

	(* ***** TeX and LaTeX *)

	expr = expr /.
		Cell[BoxData[StyleBox[RowBox[{
			"L",
			StyleBox[
				AdjustmentBox["A", cellOptions___],
				moreOptions___
			],
			"T",
			AdjustmentBox["E", otherOpts___],
			"X"
		}]]]] :> Cell[
			BoxData[StyleBox[RowBox[{
				"L",
				StyleBox[
					AdjustmentBox["A", cellOptions],
					moreOptions
				],
				"T",
				AdjustmentBox["E", otherOpts],
				"X"
			}]]],
			"LaTex"
		];

	expr = expr /.
		Cell[BoxData[StyleBox[RowBox[{
			"T",
			AdjustmentBox["E", cellOptions___],
			"X"
		}]]]] :> Cell[
			BoxData[StyleBox[RowBox[{"T", AdjustmentBox["E", cellOptions], "X"}]]],
			"LaTex"
		];

(* ***** continuing *)

	(* Formatting note: f2195e9b:
		Rule to reverse PaneSelectorBox in NewInRootLink for proper new in 6/new in 7 on
		root guide. *)
	expr = expr /.
		PaneSelectorBox[
			{
				Rule[a_, Cell[TextData[StyleBox[aa_String, "RootNewInLink"]], "RootLinkText"]],
				Rule[b_, Cell[TextData[StyleBox[bb_String, "RootLink"]], "RootLinkText"]]
			},
			c_
		] :> PaneSelectorBox[
			{
				Rule[b, Cell[TextData[StyleBox[bb, "RootLink"]], "RootLinkText"]],
				Rule[a, Cell[TextData[StyleBox[aa, "RootNewInLink"]], "RootLinkText"]]
			},
			c
		];

	(*expr = expr /.  Cell[TextData[{x_String,
					ButtonBox[btext_, BaseStyle -> "RootNewInLink", opts1___],
					ButtonBox[" \[RightGuillemet]", BaseStyle -> "RootLink", ___] }], "RootLinkText"] :>
		Cell[TextData[{x,
					ButtonBox[{StyleBox[btext, "RootNewInLink"]," \[RightGuillemet]"}, BaseStyle -> "RootLink", opts1] }], "RootLinkText"] ;
	expr = expr /.  PaneSelectorBox[{rl1:Rule[False, c1:Cell[TextData[{x__}]]], rl2:Rule[True, c2:Cell[TextData[{x__}],"RootDarkLink"]]},_] :>c2;
	expr = expr /.  PaneSelectorBox[{rl1:Rule[False, c1:Cell[TextData[{StyleBox[x_,"RootNewInLink"]}],"RootLinkText"]], rl2__},_] :>c1;

	expr = expr /.  PaneSelectorBox[{rl1:Rule[False, c1:Cell[x__]],
		rl2:Rule[True,c2:Cell[TextData[{x__}],"RootDarkLink"]]},__]:>rl2;*)

	(* fix complex subscript boxes with italic numbers *)
	expr = expr /.
		Cell[
			BoxData[SubscriptBox[
				StyleBox[a_, "TI"],
				StyleBox[
					StyleBox[RowBox[{
						b_,
						StyleBox[c_, FontSlant -> "Plain"]
					}]],
					"TI"
				]
			]],
			"InlineFormula"
		] :> Cell[
			BoxData[SubscriptBox[
				StyleBox[a, "TI"],
				RowBox[{
					StyleBox[b, "TI"], StyleBox[c , FontSlant -> "Plain"]
				}]
			]],
			"InlineFormula"
		];

	expr = expr /.
		Cell[a_String, "TableRowIcon", "BrighterFlag", b___]
		:> Cell[a, "BrighterFlag", b];

	(* Formatting note: 3a8041ff:
		Filter for InlineMath in ref/LUDecomposition More Info---block level -> inline element. *)
	expr = expr /.
		Cell[
			BoxData[FormBox["m", Cell$$38228`TraditonalForm]],
			"InlineMath"
		] :> Cell[
			StyleBox["m", FontSlant -> "Italic"],
			"InlineMath"
		];

	(* Formatting note: 7530e71f:
		Pre-processing for ["Hyperlink", FontBeight->Bold] inside ButtonBox. *)
	expr = expr /.
		ButtonBox[
			a1_,
			BaseStyle -> {"Hyperlink", FontWeight -> a3_},
			b_Rule
		] :> StyleBox[
			ButtonBox[a1, BaseStyle -> "Hyperlink", b],
			FontWeight -> a3
		];

	expr = expr /.
		Cell[
			a__,
			"Usage",
			GridBoxOptions -> {
				GridBoxBackground -> Dynamic[{
					b__,
					"Rows" -> {
						c___,
						If[
							_
							,
							co_RGBColor
							,
							None
						],
						d___
					},
					e___
				}]
			},
			f___
		] :> Cell[
			a,
			"Usage",
			GridBoxOptions -> {GridBoxBackground -> {b, "Rows" -> {c, co, d}, e}},
			f
		];

	expr = expr /. {
		Cell[
			TextData[Cell[BoxData[GridBox[a___]]]],
			style: "TableText" | "MonographTableText"
		] :> Cell[
			TextData[Cell[BoxData[GridBox[a]], style]],
			style
		]
	};

	expr = expr /.
		Cell[
			BoxData[RowBox[{
				a___String,
				StyleBox[b_String, c_String, d___],
				e___String
			}]],
			"InlineFormula",
			f___
		] :> Cell[
			BoxData[RowBox[{a, StyleBox[b, c, d], e}]],
			"InlineFormulaTableText",
			f
		];

	expr = expr /.
		Cell[
			BoxData[StyleBox[a_String,b_String,c___]],
			"InlineFormula",
			d___
		] :> Cell[
			BoxData[StyleBox[a,b,c]],
			"InlineFormulaTableText",
			d
		];

	expr = expr /.
		Cell[
			BoxData[StyleBox[a_String, b_String]],
			"InlineFormula"
		] :> Cell[
			BoxData[StyleBox[a, b]],
			"InlineFormulaTableText"
		];

	expr = expr /.
		Cell[
			BoxData[a_GridBox],
			"DefinitionBox",
			"TableHeaderLine",
			b___
		] :> Cell[
			BoxData[a],
			"DefinitionBox",
			"TableHeaderLine",
			GridBoxOptions -> {
				GridBoxDividers -> {
					"Columns" -> {{False}},
					"ColumnsIndexed" -> {},
					"Rows" -> {False, True, {False}, False},
					"RowsIndexed" -> {}
				}
			},
			b
		];

	expr = expr /.
		Cell[
			TextData[Cell[BoxData[a_String]]],
			"TableText" | "MonographTableText"
		] :> Cell[a, "InlineFormulaTableText"];

	(* Formatting note: 5e87453a: fixing a bug with TextData merges *)
	expr = expr /. a_TextData :> ConvertTextData[a];
	expr = expr /. ConvertTextData[a_] :> a;

	(* ***** setting variables based on entityType: long list of possible types *)

	(* Formatting note:463e44291824ea89d9c5d6741f8e1e963bd1e9cb:
		WLDOCS-2221: fix monographs; debugging updates
		https://jira.wolfram.com/jira/browse/WLDOCS-2221

		- Check whether the current page being processed is a package and store
		  results in packageQ.
	*)
	packageQ = StringMatchQ[
		uri,
		StartOfString ~~ CharacterRange["A", "Z"] ~~ Except["/"].. ~~ "/" ~~ __
	];

	(* Formatting note: fe3ba46f *)
	If[
		entityType === "Guide"
		&& !StringFreeQ[
			"URI" /. info,
			{
				"ComputableDataOverview",
				"CoreLanguageOverview",
				"DataManipulationOverview",
				"DynamicInteractivityOverview",
				"MathematicsAndAlgorithmsOverview",
				"NotebooksAndDocumentsOverview",
				"SystemsInterfacesAndDeploymentOverview",
				"VisualizationAndGraphicsOverview"
			}
		]
		,
		entityType = "Guide Overview"
	];

	If[
		entityType === "Guide"
		&& !StringFreeQ["URI" /. info, {"StandardExtraPackages"}]
		,
		entityType = "Standard Extra Packages"
	];

	If[
		entityType === "Guide"
		&& !StringFreeQ[
			"PacletName" /.info,
			{"Compatibility Package"}
		]
		,
		entityType = "Guide Overview"
	];

    (* package refpages should use a special layoutfile, so make that distinction here temporarily *)
    If[StringMatchQ[uri, StartOfString ~~ CharacterRange["A", "Z"] ~~ Except["/"].. ~~ "/ref/" ~~ __],
		entityType = "Package Symbol"
	];
	(* Formatting note:ffb149d615aaa44152a00b406e56dc51c7eda4c0:
		Fix styles for package message pages
		https://jira.wolfram.com/jira/browse/WLDOCS-2313 *)
    If[entityType === "Package Symbol" && StringContainsQ[uri, "/message/"],
		entityType = "Package Message"
	];

	(* Formatting note: d62ccc4c:
		Add blockchain class to body tag
		https://jira.wolfram.com/jira/browse/WLDOCS-2273
		For the new ref/blockchain pages
	*)
    If[StringContainsQ[uri, {"ref/blockchain/", "ref/blockchains/"}],
		entityType = "Blockchain"
	];

	dprint["Export.m: entityType = "<>ToString[entityType, InputForm]];

    (* indentify parameters to send to Transmogrify *)
    Switch[entityType,
        (
          "AppleScript"
          | "File"
          | "LibraryLink C Function"
          | "Symbol"
          | "System Program"
          | "Widget"
          | "Screencast"
        ),
            (* Print["using transforms in Functions.m"]; *)
            transform = If[buildmode==="Future", {"HTML","FunctionsInternalBuild.m"}, {"HTML","Functions.m"}];

            layoutfile = DocsAsset["HTMLTransformLayout", "Functions.html"];

            layoutfile = DocsAsset["HTMLTransformLayout","Functions.html"];
            (* Temporary so that CreateInputForm can show in error log *)
            Global`PageTitleForCreateInputForm = pageTitle;
        ,
        ("MathLink C Function"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "MathLinkCFunctions.html"];
        ,
        ("Workflow"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Workflows.html"];
        ,
        ("Workflow Guide"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "WorkflowGuides.html"];
        ,
        ("Package Symbol"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "PackageFunctions.html"];
        ,
        ("Package Message"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "PackageMessages.html"];
        ,
        ("Message"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Messages.html"];
        ,
        ("Program"), (* ref/program/wolframscript *)
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML", "Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "ManPage.html"];
        ,
        ("Embedding Format"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "EmbeddingFormats.html"];
        ,
        ("Device Connection"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Devices.html"];
        ,
        ("Database Connection"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "DatabaseConnections.html"];
        ,
        ("Blockchain"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Blockchains.html"];
        ,
		(* Formatting note: d54d2b65: https://jira.wolfram.com/jira/browse/WLDOCS-2683 *)
        ("Data Structure"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Datastructures.html"];
        ,
		(* Formattign note: c3252a9d: https://jira.wolfram.com/jira/browse/WLDOCS-2129 *)
        ("Text Content"),
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Textcontents.html"];
        ,
        ("Entity"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Entities.html"];
        ,
        ("Classifier" | "Method" | "Interpreter" | "Predictor" | "Format" | "Compiled Type" | "Question Interface"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Formats.html"];
        ,
        ("Service Connection"),
            (* Print["using transforms in Services.m"]; *)
            transform = {"HTML","Services.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Services.html"];
        ,
        ("Example"),
            (* Print["using transforms in ExamplePages.m"]; *)
            transform = {"HTML","ExamplePages.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "ExamplePages.html"];
        ,
        ("HowTo"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "HowTos.html"];
        ,
         ("Front End Object"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "FrontEndObjects.html"];
        ,
        ("Indicator"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Services.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Indicators.html"];
        ,
        ("External Evaluation System"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "ExternalEvaluationSystem.html"];
        ,
		(* Formatting note: 0d33d4b4: https://jira.wolfram.com/jira/browse/WLDOCS-3194 *)
        ("Remote Batch Computation Provider"),
            (* Print["using transforms in Functions.m"]; *)
			(* Formatting note: ec5aa603
				https://jira.wolfram.com/jira/browse/WLDOCS-3403
				ensure that batchcomputation section heads work the same way as those in
				ref/service pages
			*)
            transform = {"HTML","Services.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "RemoteBatchComputationProvider.html"];
        ,
        ("MLModel"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Services.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "MLModel.html"];
        ,
        ("Net Decoder"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "NetDecoder.html"];
        ,
        ("Net Encoder"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "NetEncoder.html"];
        ,
        ("ResourceObject"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "ResourceObject.html"];
        ,
        ("Menu Item"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "MenuItems.html"];
        ,
        ("Character" | "Character Name"),
            (* Print["using transforms in Functions.m"]; *)
            transform = {"HTML","Functions.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Characters.html"];
        ,

        ("Application"),
			(* TODO:
				Verify that "Application" type pages are in fact dead and upgrade
				this to an error. *)
			DocsWarning["Documentation notebook has deprecated \"Application\" Entity Type value."];

            (* Print["using transforms in Applications.m"]; *)
            transform = {"HTML","Applications.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Applications.html"];
            pageTitle = Cases[nb, Cell[c_, "ObjectName"|"ObjectNameSmall", ___]:> c, Infinity];
            contextName = "Built-in <i>Mathematica</i> Object";
            (* add routine to literalize CounterBoxes before exporting to html*)
            nbObj = NotebookPut[nb];
            FrontEndExecute[{FrontEnd`NotebookDynamicToLiteral[nbObj]}];
            nb = NotebookGet@nbObj;
            NotebookClose[nbObj];
            (* end routine to literalize CounterBoxes *)
        ,
        ("Guide"),
          (* Print["using transforms in Guides.m"]; *)
          transform = {"HTML","Guides.m"};
			(* Formatting note: 4dc79dce:
				2017 redesign: special treatment for latest Summary guide *)
          If[StringContainsQ[FileNameTake[outfile], "SummaryOfNewFeaturesIn"<>StringReplace[ToString[$VersionNumber], "." -> ""]],
            layoutfile = DocsAsset["HTMLTransformLayout", "GuideSummaryOfNewFeatures.html"],
            layoutfile = DocsAsset["HTMLTransformLayout", "Guides.html"]
          ]
        ,
        ("Root Guide"),
            dprint["Layout selection: using RootGuide.html"];
            Print["using transforms in Guides.m"];
            transform = {"HTML","Guides.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "RootGuide.html"];
            flag = "RootGuidePage";
        ,
        ("Guide Overview"),
            (* Print["using transforms in Guides.m"]; *)
            transform = {"HTML","Guides.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "GuideOverview.html"];
            ,
        ("Standard Extra Packages"),
            (* Print["using transforms in SpecialGuides.m"]; *)
            transform = {"HTML","SpecialGuides.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Guides.html"];
        ,
        ("Tutorial" | "Upgrade Information"),
            (* Print["using transforms in Tutorials.m"]; *)
            transform = {"HTML","Tutorials.m"};
            layoutfile = If[TrueQ[wsmQ],
              If[TrueQ[completeHTMLQ],
                DocsAsset["HTMLTransformLayout", "WSMTutorialsStandalone.html"],
                DocsAsset["HTMLTransformLayout", "WSMTutorials.html"]
              ],
              DocsAsset["HTMLTransformLayout", "Tutorials.html"]
            ];
        ,
		(* Formatting note: c6f6c83e:
			Assign guide styles to notes pages
			https://jira.wolfram.com/jira/browse/WLDOCS-1941

			The "Notes on sources" pages (notes/* pages) should be styled like
			guides with <body id="guide">.  The cleanest way to do that is to assign
			those pages their own layout file, Notes.html.
		*)
        ("Note"),
            transform = {"HTML","Tutorials.m"};
            layoutfile = DocsAsset["HTMLTransformLayout","Notes.html"];
        ,
		(* Formatting note: 39197ccb:
			Regularize tutorial <body> tags

			https://jira.wolfram.com/jira/browse/WLDOCS-2198

			<body id="tutorial"> for tutorials
			<body id="monograph"> for monographs
			<body id="monographOverview"> for monograph overviews
			<body id="techNote"> for technotes
		*)
        ("Tech Note"),
            transform = {"HTML","Tutorials.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "TechNotes.html"];
        ,
		(* Formatting note: 614fae9d: https://jira.wolfram.com/jira/browse/WLDOCS-2142 *)
        ("Monograph"),
            (* Print["using transforms in Tutorials.m"]; *)
            transform = {"HTML","Tutorials.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Monographs.html"];
        ,
        ("Monograph Overview"),
            (* Print["using transforms in Tutorials.m"]; *)
            transform = {"HTML","TutorialOverviews.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "MonographOverviews.html"];
        ,
        ("Overview"),
            (* Print["using transforms in TutorialOverviews.m"]; *)
            transform = {"HTML","TutorialOverviews.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "TutorialOverviews.html"];
        ,
        ("Note" | "Tutorial"),
            (* Print["using transforms in Tutorials.m"]; *)
            transform = {"HTML","Tutorials.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Tutorials.html"];
        ,

        "generic",
			DocsError[
				"Use of deprecated \"EntityType\" value: ``. Use a more specific documentation notebook type.",
				InputForm[entityType]
			];
        ,
        "alphabeticalpages",
            (* Print["using transforms in Guides.m"]; *)
            transform = {"HTML","Guides.m"};
            layoutfile = DocsAsset["HTMLTransformLayout", "Guides.html"];
        ,

		_,
			(* TODO: Support custom, user-provided values for `transform` and `layoutfile`
			         in this case? *)
			DocsError[
				"Unrecognized \"EntityType\" value: ``. Unable to determine transform and HTML template.",
				InputForm[entityType]
			];
	];

	dprint["Export.m: transform = "<>ToString[transform, InputForm]];
	dprint["Export.m: layoutfile = "<>ToString[layoutfile, InputForm]];

    (* restore package refpages to their original entityType *)
    If[StringMatchQ[uri, StartOfString ~~ CharacterRange["A", "Z"] ~~ Except["/"].. ~~ "/ref/" ~~ __],
		entityType = "Symbol"
	];

(* ***** continuing *)

    BuildLog["   Replace various patterns"];

    (* Combine {Cell, Input, Deemphasis} into one style
       TODO: Allow Transmogrify templates to take multiple matcher arguments
    *)
    expr = expr /. Cell[c_, "Input", "Deemphasis", o___] :> Cell[ c, "InputDeemphasis", o];
    expr = expr /. Cell[c_, "Program", "InlineCode", o___] :> Cell[ c, "Program", o];

	expr = expr //.
		Cell[
			TextData[a_, b:(ButtonBox[__]|{__})],
			"RootMoreAbout",
			c___
		] :> Cell[b, "RootMoreAbout", c];

	expr = expr //.
		Cell[
			TextData[{a_, b:(ButtonBox[__]|{__})}],
			"RootMoreAbout",
			c___
		] :> Cell[b, "RootMoreAbout", c];

	(*   *)
	expr = expr /.
		Cell[c_, "NotesSection", o___]
		:> Cell[
			If[
				1 === Length[Cases[
					c,
					Cell[TextData[{a_, b_String}], "NotesFrameText"] :> b,
					Infinity
				]]
				,
				Cases[
					c,
					Cell[TextData[{a_, b_String}], "NotesFrameText"] :> b,
					Infinity
				][[1]]
				,
				(* TODO: Is it ever valid the the Length of the above condition to be
				         anything other than 1 (or just >1)? *)
				Localized["NotesSectionText", language]
			],
			"NotesSection",
			o
		];

    (* add output cell for Input followed by ExampleDelimiter --next 2 rules  *)
    expr = expr //.
		{
			pre___,
			c:Cell[_, "Input", ___],
			Cell[CellGroupData[{Cell[cc1_, "ExampleDelimiter", cc2___], cgd___}, oc_], cc___],
			post___
		} :> {
			pre,
			c,
			Cell[" ", "Spacer"],
			Cell[CellGroupData[{Cell[cc1, "ExampleDelimiter", cc2], cgd}, oc], cc],
			post
		};

    expr = expr //.
		{pre___,
			Cell[
				CellGroupData[
					{
						Cell[cc1_, "ExampleDelimiter", cc2___],
						c:Cell[_, "ExampleText", ___],
						Cell[a_, "Input", b___]
					},
					oc_
				],
				cc___
			],
			post___
		} :> {
			pre,
			Cell[
				CellGroupData[
					{
						Cell[cc1, "ExampleDelimiter", cc2],
						c,
						Cell[a, "Input", b],
						Cell[" ", "Spacer"]
					},
					oc
				],
				cc
			],
			post
		};

    expr = expr //.
		{
			pre___,
			Cell[
				CellGroupData[
					{
						Cell[cc1_, "ExampleDelimiter", cc2___],
						c:Cell[_, "ExampleText", ___]
					},
					oc_
				],
				cc___
			],
			post___
		} :> {
			pre,
			Cell[
				CellGroupData[
					{
						Cell[cc1, "ExampleDelimiter", cc2],
						c,
						Cell[" ", "Spacer"]
					},
					oc
				],
				cc
			],
			post
		};

    (* Fix local page linking *)
	expr = expr /.
		(*ButtonBox[con_, BaseStyle->"Hyperlink", Appearance ->__List, ButtonData->t_String/;StringTake[t, 5]=!="http:", o___] :> *)
		ButtonBox[
			con_,
			BaseStyle -> "Hyperlink",
			Appearance -> __List,
			ButtonData -> t_String /; StringFreeQ[t, "http://"],
			o___
		] :> ButtonBox[
			con,
			BaseStyle -> "Hyperlink",
			ButtonData -> {URL["#"<>t], None},
			o
		];

	expr = expr /.
		(*ButtonBox[con_, BaseStyle->"Hyperlink", ButtonData->t_String/;StringTake[t, 5]=!="http:", o___]*)
		ButtonBox[con_, BaseStyle->"Hyperlink", ButtonData->t_String/;StringFreeQ[t, "http://"], o___]:>
		ButtonBox[con, BaseStyle->"Hyperlink", ButtonData->{URL["#"<>t], None}, o];

	(* Formatting note: 3ab2cf33: Fix some issues in converting guide notebooks *)
	expr = expr /.
		StyleBox[
			ButtonBox[str_String, o___],
			"MenuName"
		] :> StyleBox[
			ButtonBox[
				StringSplit[
					str,
					RegularExpression["\\s*\[FilledRightTriangle]\\s*"]
						-> StyleBox[" \[FilledRightTriangle] ", "MenuNameDelimiter"]
				],
				o
			],
			"MenuName"
		];

    (* Wrap FilledRightTriangle in StyleBox in order to decrease size on web *)
	expr = expr //.
	Cell[TextData[{a___, " \[EmptyVerySmallSquare] ", b___}], "RelatedLinks", o___]
		:> Cell[TextData[{a, StyleBox["*", "SeeAlsoBullet"], b}], "RelatedLinks", o];
	expr = expr //.
	Cell[TextData[{a___, " \[EmptyVerySmallSquare] ", b___}], "SeeAlso", o___]
		:> Cell[TextData[{a, StyleBox["*", "SeeAlsoBullet"], b}], "SeeAlso", o];

	expr = expr //.
	Cell[TextData[{a___, " \[EmptyVerySmallSquare] ", b___}], o___]
		:> Cell[TextData[{a, "\[NonBreakingSpace]",
			StyleBox["\[FilledVerySmallSquare]", "InlineSeparator"], " ", b}], o];

    (* Remove MSG second style in Messages -- TODO: need to recognize multiple styles *)
	expr = expr //.
		Cell[a_, "Message", b___, "MSG", c___]
		:> Cell[
			a,
			"Message",
			b,
			FontFamily -> CurrentValue["PanelFontFamily"]
		];

	expr = expr //.
		Cell[
			CellGroupData[
				{
					Cell[a___, "NotesSection", b___],
					c___,
					Cell[d___, "Message", e___],
					f___
				},
				g___
			],
			h___
		] :> Cell[
			CellGroupData[
				{
					Cell[a, "NotesSection", b],
					c,
					Cell[{Cell[d, "Message", e]}, "MessageNote"],
					f
				},
				g
			],
			h
		];

	expr = expr //.
		Cell[
			CellGroupData[
				{
					Cell[a___, "ImportExportSection", b___],
					c___,
					Cell[d___, "WeakDivider", e___],
					f___
				},
				g___
			],
			h___
		] :> Cell[
			CellGroupData[
				{
					Cell[a, "ImportExportSection", b],
					c,
					Cell[{Cell[d, "WeakDivider", e]}, "FormatNotes"],
					f
				},
				g
			],
			h
		];

	expr = expr //.
		Cell[
			CellGroupData[
				{
					Cell[a___, "NotebookInterfaceSection", b___],
					c___,
					Cell[d___, "WeakDivider", e___],
					f___
				},
				g___
			],
			h___
		] :> Cell[
			CellGroupData[
				{
					Cell[a, "NotebookInterfaceSection", b],
					c,
					Cell[{Cell[d, "WeakDivider", e]}, "FormatNotes"],
					f
				},
				g
			],
			h
		];

	expr = expr //.
		Cell[
			CellGroupData[
				{
					Cell[a___, style:"ElementsSection"|"ExportElementsSection", b___],
					c___,
					Cell[d___, "WeakDivider", e___],
					f___
				},
				g___
			],
			h___
		] :> Cell[
			CellGroupData[
				{
					Cell[a, style, b],
					c,
					Cell[{Cell[d, "WeakDivider", e]}, "FormatNotes"],
					f
				},
				g
			],
			h
		];

	expr = expr //.
		Cell[
			CellGroupData[
				{
					Cell[a___, "OptionsSection", b___],
					c___,
					Cell[d___, "WeakDivider", e___],
					f___
				},
				g___
			],
			h___
		] :> Cell[
			CellGroupData[
				{
					Cell[a, "OptionsSection", b],
					c,
					Cell[{Cell[d, "WeakDivider", e]}, "FormatNotes"],
					f
				},
				g
			],
			h
		];



	expr = expr //.
		Cell[a_, "Print", b___, "PrintUsage", c___]
		:> Cell[
			a,
			"Print",
			c,
			PageWidth -> 550,
			CellFrame -> {{0, 0}, {0, 3}},
			CellFrameColor -> RGBColor[0.947448, 0.91873, 0.880354],
			Background -> RGBColor[0.972732, 0.966506, 0.947112]
		];

	expr = expr //.
		Cell[a_, "Input", b___, "Deemphasis", c___]
		:> Cell[a, "Deemphasis", c,  FontColor -> GrayLevel[0.75]];

	expr = expr //.
		Cell[a_, "DefinitionBox", b___, "TableHeaderLine", c___]
			:> Cell[a, "DefinitionBox", c];

	expr = expr //.
		Cell[TextData[a_, b:(ButtonBox[__]|{__})], "RootMoreAbout", c___]
			:> Cell[b, "RootMoreAbout", c];

	expr = expr //.
		Cell[a_, "Input", b___, "ConversionLockdown", c___]
		:> Cell[a, "Input", c];

	expr = expr //.
		Cell[a_, "Output", b___, "ConversionLockdown", c___]
		:> Cell[a, "Output", c];

	expr = expr //.
		Cell[a_, style: "MathCaption" | "MonographMathCaption", b___, "ConversionLockdown", c___]
		:> Cell[a, style, c];

	expr = expr //.
		Cell[a_, "Print", b___, "ConversionLockdown", c___]
		:> Cell[a, "Print", c];

	expr = expr //.
		Cell[__, CellTags->"UnderDevelopment", ___]
		:> Cell["U N D E R  D E V E L O P M E N T", "UnderDevelopment"];

	(* Formatting note: ce287a1e:
		A large transparent image has started appearing under most h1 heads because the
		relevant Export.m munging no longer works with current ip nbs.  I've updated the
		munging with new patterns and also deleted a separate attempt to deal with the
		problem. *)
	expr = expr /.
		Cell[
			TextData[{
				a:Cell[_, s_String /; StringContainsQ[s, "ObjectName"], ___],
				Cell[BoxData[InterpretationBox[__, Spacer[8], ___]], ___],
				b___
			}],
			c___
		] :> Cell[TextData[{a, b}], c];

	expr = expr /.
		Cell[
			TextData[{
				a:Cell[_, s_String /; StringContainsQ[s, "ObjectName" | "GuideTitle"], ___],
				Cell[BoxData[InterpretationBox[__, Spacer[10], ___]], ___],
				b___
			}],
			c___
		] :> Cell[
			TextData[{
				a,
				Cell["[[UNFINISHED DRAFT]]", "UnfinishedDraft"]
			}],
			c
		];

	expr = expr //.
		Cell[a__, CellTags->"TextAnnotation", b___]
		:> Cell[a, b];

	expr = expr //.
		StyleBox[
			FormBox[
				x_String /; StringMatchQ[x, RegularExpression["[0-9]+"]],
				TraditionalForm
			],
			"TR"
		] :> StyleBox["1", "TR"];

    (*Deal with Input/LegacyInput in Compatibility pages *)
    expr = expr //.
      Cell[a_, "Input", b___, "LegacyInput", c___]
        :> Cell[a, "LegacyInput", b,
          PageWidth->550,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          CellFrame->{{1, 1}, {1, 1}},
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
            Cell[TextData[{
              StyleBox["Version", FontSize -> 9], "\n  5.2"}], "LegacyInputLabel"], None}, {
            None, None}},
          FontColor->GrayLevel[0.5]];

     expr = expr //.
       Cell[a_, "Input", b___, "LegacyInputV7", c___]
         :> Cell[a, "LegacyInputV7", b,
           PageWidth->550,
           CellFrame->True,
           CellMargins->{{28, 4}, {3, 8}},
           CellFrame->{{1, 1}, {1, 1}},
           CellFrameMargins->{{12, 12}, {12, 12}},
           CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
           CellFrameLabels->{{
             Cell[
               TextData[{
                StyleBox["Version", FontSize -> 9], "\n  6.0"}], "LegacyInputLabel"], None}, {
             None, None}},
           FontColor->GrayLevel[0.5]];

     expr = expr //.
       Cell[a_, "Input", b___, "LegacyInputV8", c___]
         :> Cell[a, "LegacyInputV8", b,
           PageWidth->550,
           CellFrame->True,
           CellMargins->{{28, 4}, {3, 8}},
           CellFrame->{{1, 1}, {1, 1}},
           CellFrameMargins->{{12, 12}, {12, 12}},
           CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
           CellFrameLabels->{{
             Cell[
               TextData[{
                StyleBox["Version", FontSize -> 9], "\n  7.0"}], "LegacyInputLabel"], None}, {
             None, None}},
           FontColor->GrayLevel[0.5]];

       expr = expr //.
       Cell[a_, "Input", b___, "LegacyInputV9", c___]
         :> Cell[a, "LegacyInputV9", b,
           PageWidth->550,
           CellFrame->True,
           CellMargins->{{28, 4}, {3, 8}},
           CellFrame->{{1, 1}, {1, 1}},
           CellFrameMargins->{{12, 12}, {12, 12}},
           CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
           CellFrameLabels->{{
             Cell[
               TextData[{
                StyleBox["Version", FontSize -> 9], "\n  8.0"}], "LegacyInputLabel"], None}, {
             None, None}},
           FontColor->GrayLevel[0.5]];

    expr = expr //.
      Cell[a_, "UpgradeDetail" | "Input", b___, "LegacyInputError", c___]
        :> Cell[a, "LegacyInputError", b,
          PageWidth-> 500,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          Evaluatable->False,
          GeneratedCellStyles->{"Output"->"LegacyOutput"},
          ShowCellLabel->False,
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
            Cell[TextData[{
              StyleBox["Version", FontSize -> 9], "\n  5.2"}], "LegacyInputLabel"],
              Cell["Error", "LegacyInputLabel"]}, {None, None}},
          ShowAutoStyles->False,
          MenuPosition->800,
          MenuCommandKey->"8",
          FontColor->GrayLevel[0.5],
          Background->RGBColor[1, 0.9, 0.9]];

    expr = expr //.
      Cell[a_, "UpgradeDetailLabel", b___, "LegacyInputError", c___]
        :> Cell[a, "LegacyInputError", b,
          PageWidth-> 500,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          Evaluatable->False,
          GeneratedCellStyles->{"Output"->"LegacyOutput"},
          ShowCellLabel->False,
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
            Cell[TextData[{
              StyleBox["Version", FontSize -> 9], "\n  5.2"}], "LegacyInputLabel"],
              Cell["Error", "LegacyInputLabel"]}, {None, None}},
          ShowAutoStyles->False,
          MenuPosition->800,
          MenuCommandKey->"8",
          FontColor->GrayLevel[0.5],
          Background->RGBColor[1, 0.9, 0.9]];
          
     expr = expr//.
       Cell[a_, "UpgradeDetail" | "UpgradeDetailLabel" | "Input", b___, "LegacyInputErrorV7", c___]
         :> Cell[a, "LegacyInputErrorV7", b,
          PageWidth-> 500,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          Evaluatable->False,
          GeneratedCellStyles->{"Output"->"LegacyOutput"},
          ShowCellLabel->False,
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
          Cell[
          TextData[{
                 StyleBox["Version", FontSize -> 9], "\n  6.0"}], "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None], 
              Cell["returns\nunevaluated", "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None]}, {None, None}},
           ShowAutoStyles->False,
           MenuPosition->800,
           MenuCommandKey->"8",
           FontColor->GrayLevel[0.5],
           Background->RGBColor[1, 0.9, 0.91]]; 

     expr = expr//. 
       Cell[a_, "UpgradeDetail" | "UpgradeDetailLabel" | "Input", b___, "LegacyInputErrorV8", c___] 
         :> Cell[a, "LegacyInputErrorV8", b,
          PageWidth-> 500,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          Evaluatable->False,
          GeneratedCellStyles->{"Output"->"LegacyOutput"},
          ShowCellLabel->False,
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
          Cell[
          TextData[{
                 StyleBox["Version", FontSize -> 9], "\n  7.0"}], "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None], 
              Cell["returns\nunevaluated", "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None]}, {None, None}},
           ShowAutoStyles->False,
           MenuPosition->800,
           MenuCommandKey->"8",
           FontColor->GrayLevel[0.5],
           Background->RGBColor[1, 0.9, 0.91]]; 
           
      expr = expr//. 
       Cell[a_, "UpgradeDetail" | "UpgradeDetailLabel" | "Input", b___, "LegacyInputErrorV9", c___] 
         :> Cell[a, "LegacyInputErrorV9", b,
          PageWidth-> 500,
          CellFrame->True,
          CellMargins->{{28, 4}, {3, 8}},
          Evaluatable->False,
          GeneratedCellStyles->{"Output"->"LegacyOutput"},
          ShowCellLabel->False,
          CellFrameMargins->{{12, 12}, {12, 12}},
          CellFrameColor->RGBColor[0.55, 0.650000000000003, 0.55],
          CellFrameLabels->{{
          Cell[
          TextData[{
                 StyleBox["Version", FontSize -> 9], "\n  8.0"}], "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None], 
              Cell["returns\nunevaluated", "LegacyInputLabel", FontColor->RGBColor[0.3000000000000003, 0.6000000000000006, 0.5], Background->None]}, {None, None}},
           ShowAutoStyles->False,
           MenuPosition->800,
           MenuCommandKey->"8",
           FontColor->GrayLevel[0.5],
           Background->RGBColor[1, 0.9, 0.91]]; 

	(* **** remove buttons that open frontend elements in in-product nbs *)

	(* https://jira.wolfram.com/jira/browse/WLDOCS-2001 *)

	(* intercept this before changing all "active" ButtonBoxes *)
	expr = expr /.
	(
		StyleBox[
			b:ButtonBox[s:_String, ___],
			"DialogElementName",
			___
		] /; Positive @ Count[b, _FrontEnd`FrontEndToken, All]
	):> StyleBox[s, "DialogElementName"];

(* **** continue *)

    (*  *)
    (*expr = expr /.
      ButtonBox[con_, a___, ButtonFunction~(Rule|RuleDelayed)~f_, b___]
        :> ButtonBox[con, a, ButtonFunction->"DEADFUNCTION", b];*)

	expr = expr /.
		ButtonBox[
			con_,
			a___,
			ButtonFunction ~(Rule|RuleDelayed)~ f_,
			b___
		] :> ButtonBox[con, a, b];

	(* Formatting note: 2e3f3fdb:
		Fixed ButtonBox/FIXME links on headers for PES, SAS, etc in HTML *)
	expr = expr /.
		ButtonBox[
			con_,
			BaseStyle ~(Rule|RuleDelayed)~ None,
			ButtonFunction ~(Rule|RuleDelayed)~ "DEADFUNCTION",
			ButtonFrame ~(Rule|RuleDelayed)~ "None",
			Evaluator ~(Rule|RuleDelayed)~ None,
			Method ~(Rule|RuleDelayed)~ "Preemptive"
		] :> con;

	expr = expr /.
		ButtonBox[
			con_,
			BaseStyle ~(Rule|RuleDelayed)~ None,
			_Rule,
			Evaluator ~(Rule|RuleDelayed)~ None,
			Method ~(Rule|RuleDelayed)~ "Preemptive",
			ButtonFunction ~(Rule|RuleDelayed)~ "DEADFUNCTION"
		] :> con;

	expr = expr /.
		ButtonBox[
			aa___,
			BaseStyle -> {
				"Link",
				"GuideFunctionsSubsection" | "MenuName" | "FullListing"
			},
			bb___
		] :> ButtonBox[aa, BaseStyle -> "Link", bb];

	expr = expr /. {
		Cell[
			BoxData[GridBox[{
				{a___, Cell[b___]}
			}]],
			"ObsolescenceNote",
			c___
		] :> Cell[b, "ObsolescenceNote"]
	};

	expr = expr /. {
		Cell[
			BoxData[GridBox[{
				{a___, Cell[b___]}
			}]],
			"AwaitingReviewNote",
			c___
		] :> Cell[b, "AwaitingReviewNote"]
	};

	expr = expr /. {
		Cell[
			TextData[{
				Cell[BoxData[
					InterpretationBox[
						StyleBox[
							GraphicsBox[
								{},
								BaselinePosition -> Baseline,
								ImageSize -> {6, 0}
							],
							CacheGraphics -> False
						],
						Spacer[6]
					]
				]],
				w___
			}],
			"FunctionEssaySection",
			WholeCellGroupOpener -> True
		] :> Cell[TextData[{w}], "FunctionEssaySection", WholeCellGroupOpener -> True]
	};

    (* bug 309600 *)
    expr = expr /. Cell[BoxData[""], "Input"] -> Sequence[];

    (* https://jira.wolfram.com/jira/browse/WLDOCS-936 *)
	expr = Block[{pos, usage},
		pos = Position[expr, Cell[_, "Usage", ___]];
		usage = Extract[expr, pos];
		usage = usage /. Cell[BoxData[""], "Input", ___] -> Sequence[];
		ReplacePart[expr, pos -> usage]
	];

	(* Formatting note: 407d3c5b:
		note that EchoLabel's \[RightGuillemet] needs to be base-aligned with the text,
		if any, of the EchoLabel.  this rules out the standard two-<td> table approach
		since it can't ensure base-alignment.
	*)
	expr = expr /.
		Cell[
			BoxData[RowBox[list_List]],
			"Echo",
			rest___
		] :> Cell[
			BoxData[RowBox[Flatten @ {
				TagBox["\[RightGuillemet]", "EchoLabel", TagBoxOptions -> {BaseStyle -> Large}],
				" ",
				list
			}]],
			"Echo",
			rest
		];

	expr = expr /.
		Cell[BoxData[s_String], "Echo", rest___]
		:> Cell[
			BoxData[RowBox[{
				TagBox["\[RightGuillemet]", "EchoLabel", TagBoxOptions -> {BaseStyle -> Large}],
				" ",
				s
			}]],
			"Echo",
			rest
		];

    (* wolframscript, the new Program entity - v11, July 2016 *)

	expr = Block[{pos, synopsis},
		pos = Position[
			expr,
			Cell[CellGroupData[{
				Cell["SYNOPSIS", "ProgramSection", ___],
				Longest[Cell[_, "ProgramCodeText", ___]..],
				___
			}, ___], ___],
			Infinity
		];

		synopsis = Extract[expr, pos];
		synopsis = synopsis /.
			a:Cell[_, "ProgramCodeText", ___]
				:> Cell[TextData[{a}], "ProgramNotes"];

		If[Length[synopsis] > 0,
			ReplacePart[expr, pos -> First[synopsis]]
			,
			expr
		]
	];

	(* Formatting note: 433ec3ed4733d18d9455c28639cca15c98878578:
		WLDOCS-866: some ProgramNotes should not be wrapped in a code paragraph *)
	expr = Block[{pos, target},
		pos = Position[
			expr,
			Cell[CellGroupData[{Cell["DESCRIPTION", "ProgramSection", ___], ___}, ___], ___],
			Infinity
		];

		target = Extract[expr, pos];
		target = target /.
			Cell[first:_, "ProgramNotes", rest:___]
				:> Cell[first, "ProgramNotesDescriptionSection", rest];

		If[Length[target] > 0,
			ReplacePart[expr, pos -> First[target]]
			,
			expr
		]
	];

	expr = Block[{pos, target},
		pos = Position[
			expr,
			Cell[CellGroupData[{Cell["EXAMPLES", "ProgramSection", ___], ___}, ___], ___],
			Infinity
		];

		target = Extract[expr, pos];
		target = target /.
			Cell[first:_, "ProgramNotes", rest:___]
				:> Cell[first, "ProgramNotesExamplesSection", rest];

		If[Length[target] > 0,
			ReplacePart[expr, pos -> First[target]]
			,
			expr
		]
	];

	expr = Block[{pos, target},
		pos = Position[
			expr,
			Cell[CellGroupData[{Cell["DETAILS", "ProgramSection", ___], ___}, ___], ___],
			Infinity
		];

		target = Extract[expr, pos];
		target = target /.
			Cell[first:_, "ProgramNotes", rest:___]
				:> Cell[first, "ProgramNotesDetailsSection", rest];

		If[Length[target] > 0,
			ReplacePart[expr, pos -> First[target]]
			,
			expr
		]
	];

	expr = expr /. Cell[s_String, "ProgramCodeText", rest___] :> StyleBox[s, "ProgramCodeText", rest];

    (* the *Standalone transforms insert the right <ul> element around these
    tables because wolframscript's tables are not already contained in an
    environment that inserts the <ul>.  i.e., these are standalone tables,
    outside the usual table-containing structures. *)

	If[Positive @ Count[expr, "uri" -> "ref/program/wolframscript", Infinity],
		expr = expr /. {
			Cell[a_, "2ColumnTable", rest___] :> Cell[a, "2ColumnTableStandalone", rest],
			Cell[a_, "3ColumnTable", rest___] :> Cell[a, "3ColumnTableStandalone", rest],
			Cell[a_, "1ColumnTableMod", rest___] :> Cell[a, "1ColumnTableModStandalone", rest],
			Cell[a_, "2ColumnTableMod", rest___] :> Cell[a, "2ColumnTableModStandalone", rest],
			Cell[a_, "3ColumnTableMod", rest___] :> Cell[a, "3ColumnTableModStandalone", rest],
			Cell[a_, "3ColumnIllustratedTableMod", rest___] :> Cell[a, "3ColumnIllustratedTableModStandalone", rest]
		}
	];

    expr = expr /.
		{
			s1_String,
			"\[NonBreakingSpace]\[NonBreakingSpace]",
			c:Cell[_, "ExampleCount", ___]
		} :> {s1, " ", c};

	(* Formatting note: 29120bbc: 2017 redesign - remove ObjectNameTranslation cells for now *)
    expr = expr /. Cell[_, "ObjectNameTranslation", ___] -> Nothing;

	expr = expr /. {
		a:___,
		Cell[
			CellGroupData[
				{
					b1:Cell[_, "GuideTitle" | "GuideTitleGrid", ___],
					b2:Cell[_, "GuideAbstract", ___]..,
					b3:__
				},
				boc:___
			]
		],
		c:___
	} :> {
		a,
		Cell[CellGroupData[{b1, b2}, boc]],
		Cell[CellGroupData[Flatten[{Cell["", "GuideReferenceLinksSection"], b3}], Open]],
		c
	};

    (* remove the 11.1 redesign's literal section opener/closers - on the web those are done via css *)
	patt = {
		first:___,
		Cell[BoxData[DynamicBox[ToBoxes[If[_ , _Style , _Rotate]], ___]], ___],
		Cell[BoxData[InterpretationBox[StyleBox[__], _Spacer, ___]], ___],
		s_String,
		rest:___
	};
	expr = expr /. patt -> {first, s, rest};

    (* a new thing not needed for the web build *)
    expr = expr /. Cell[_, "PageDelimiter", ___] :> Unevaluated[Sequence[]];

    (* removing this for clarity while examining NotesThumbnails to implement it *)
    (* expr = expr /. Cell[CellGroupData[{Cell[_, "PrimaryExamplesSection", ___], ___}, ___], ___] :> Unevaluated[Sequence[]]; *)
    (* Print["CAUTION: REMOVING ALL EXAMPLES FOR FASTER TESTING!  RESTORE WHEN TESTING IS COMPLETE!"]; *)

    (* Print["entering NotesSection handling"]; *)
    (* myres = DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"]; *)
    (* Print["uri = "<>ToString[uri, InputForm]]; *)
    (* Print["language = "<>ToString[language, InputForm]]; *)
    (* Print["DocumentationBuild`Info`GetSourceInformation[uri, language, \"HasOptions\"] = "<>ToString[DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"]]]; *)

	(*
	uri = "ref/AbsoluteTime"
	language = "English"
	DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"] = True
	NotesSection: has options; using Details and Options

	uri = "ref/Abs"
	language = "English"
	DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"] = False
	NotesSection: does not have options; using Details

	uri = "FEMDocumentation/ref/BoundaryConditionData"
	language = "English"
	DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"] = None
	NotesSection: does not have options; using Details
	*)

	expr = expr /.
		Cell[a:_, "NotesSection", b:___]
		:> Module[
			{hasoptions}
			,
			hasoptions = DocumentationBuild`Info`GetSourceInformation[uri, language, "HasOptions"];
			hasoptions = If[MatchQ[hasoptions, {} | None | "None"],
				False,
				hasoptions
			];

			If[
				hasoptions
			,
				(* Print["NotesSection: has options; using "<>ToString[Localized["NotesSectionText", language]]]; *)
				Cell[Localized["NotesSectionText", language], "NotesSection", b]
			,
				(* Print["NotesSection: does not have options; using "<>ToString[Localized["NotesSectionNoOptsText", language]]]; *)
				Cell[Localized["NotesSectionNoOptsText", language], "NotesSection", b]
			]
		];

	(* **** sections with thumbnail previews *)
	(* Formatting note: 61efb33d:
		Part of the implementation of ref/entity pages in WLDOCS-2023

		generalize thumbnail sections to handle different kinds and multiple
		instances in one page, as found in entity pages.  Note that this requires no
		changes to the related transforms in Functions.m.
	*)
	(*
	here we're looking for refpage sections that are represented by thumbnail
	previews when closed.  the structure is:
		{
			stuff
			╭────[ referred to as the "boxed pair" in code below ]
			│ NotesSection|EntitySection
			│ NotesThumbnails
			╰────
			more stuff
		}

	as of Feb 2019, pre-v12.0, the boxed pair appears once in ref/ pages and
	twice consecutively in ref/entity pages.
	*)

	dprint["Export.m thumbnail processing: incoming expr is:"];
	dfile[expr, "Format" -> ".nb"];


	(* Formatting note: d755c966:
		Fix thumbnail previews: https://jira.wolfram.com/jira/browse/WLDOCS-2798 *)
	expr[[1]] = SequenceReplace[
		(* SequenceReplace expects a _List here, so use the Notebook[] expr's initial list of content *)
		expr[[1]]
		,
		(* each of the <section, thumbs> boxed pairs *)
		{section:_, thumbs:Cell[_, "NotesThumbnails", ___]} :> Module[{
			title, content, images, originalCellStyle
		},
			(* Formatting note: 77a2d205: Fix "Updated" highlighting: https://bugs.wolfram.com/show?number=402532 *)
			originalCellStyle = First @ Cases[
				{section},
				Cell[CellGroupData[List[Cell[_, s:_String, ___], ___], ___], ___] -> s,
				Infinity
			];

			dprint["Export.m thumbnail processing: originalCellStyle = "<>ToString[originalCellStyle, InputForm]];

			(* take the title from the section cell *)
			title = Quiet @ Check[
				First[Flatten[Cases[
					section,
					Cell[
						TextData[{__, title:_}],
						_String,
						___,
						WholeCellGroupOpener -> True,
						___
					] -> title,
					Infinity
				]]],
				"FIXME"
			];

			dprint["Export.m thumbnail processing: title step 1 = "<>ToString[title, InputForm]];

			(* TODO: Issue a hard error if this happens instead of grabbing something
			         arbitrary. *)
			If[title === "FIXME",
				(* then grab whatever is in the first cell in that cellgroup *)
				title = Quiet @ Check[
					First[Cases[
						section,
						Cell[title_, _String, ___] -> title,
						Infinity
					]],
					"FIXME2"
				]
			];

			dprint["Export.m thumbnail processing: final title = "<>ToString[title, InputForm]];

			(* take the content of the section cellgroup by deleting its heading *)
			content = DeleteCases[
				section,
				Cell[_, _String, ___, Rule[WholeCellGroupOpener, True], ___],
				Infinity
			];

			dprint["Export.m thumbnail processing: content = "];
			dfile[content];

			(* and grab the images from the NotesThumbnails cell *)
			images = Cases[
				thumbs,
				Grid[{{_Spacer, img:__, _Spacer}}] :> img,
				Infinity
			];

			dprint["Export.m thumbnail processing: images = "];
			dfile[images];

			(* then replace each <section, thumbs> boxed pair with this CellGroup expression: *)
			Cell[CellGroupData[
				Flatten @ {
					(* "OriginalCellStyle" is not currently used in the transform in Functions.m, but keep it for future work *)
					Cell[
						title,
						"NotesThumbnailsSection",
						"OriginalCellStyle" -> originalCellStyle
					],

					Cell[CellGroupData[{
						Cell["", "NotesThumbnails"],
						Sequence @@ Flatten[{
							Cell[BoxData[ToBoxes[#]], "NotesThumbnail"]& /@ images
						}]
					}]],

					content
				},
				Closed
			]]
		]
	];

(* **** continue *)

    (* the new PrimaryExamplesSection styles are handled in Examples.m *)
	expr = Module[{count},
		count = Count[expr, Cell[_, "ExampleSection", ___], Infinity];
		expr /.
			Cell[
				first:_,
				s_String /; StringMatchQ[s, "PrimaryExamplesSection"|"IndicatorExampleSection"],
				rest:___
			] :> Cell[
				first,
				If[count > 1,
					"PrimaryExamplesSectionWithOpener",
					"PrimaryExamplesSectionWithoutOpener"
				],
				rest
			]
	];

    (* expr = expr /. Cell[s_String /; StringMatchQ[s, Whitespace...], "GuideDelimiterSubsection"|"GuideReferenceLinksSection", ___] :> Unevaluated[Sequence[]]; *)
    expr = expr /. Cell[s_String /; StringMatchQ[s, Whitespace...], "GuideReferenceLinksSection", ___] :> Unevaluated[Sequence[]];

    (* ExperimentalBanner -> UnfinishedDraft *)
    expr = expr /. Cell[_, "ExperimentalBanner", ___] -> Cell["[[UNFINISHED DRAFT]]", "UnfinishedDraft"];

    (* https://jira.wolfram.com/jira/browse/WLDOCS-1260?focusedCommentId=2544496&page=com.atlassian.jira.plugin.system.issuetabpanels:comment-tabpanel#comment-2544496 *)
	expr = expr /.
		{
			first:___,
			a:Cell[_, "DeviceImage", ___],
			Cell[b1:_, "DeviceNotes", b2:___],
			rest:___
		} -> {
			first,
			a,
			Cell[b1, "DeviceImageCaption", b2],
			rest
		};

    (* remove empty guide abstracts per https://jira.wolfram.com/jira/browse/WLDOCS-1334 *)
    expr = expr /. Cell[s_String /; StringMatchQ[s, Whitespace..], "GuideAbstract", ___] :> Unevaluated[Sequence[]];

	(* Formatting note: 5b934d68:
		https://bugs.wolfram.com/show?number=348426
		https://bugs.wolfram.com/show?number=350792 *)
    expr = expr /. Cell[BoxData[DynamicBox[FrontEndResource["FEBitmaps", s_String], ___]], "InlineFormula", ___] :> Cell[s, "feBitmap"];

    (* https://jira.wolfram.com/jira/browse/WLDOCS-1418 *)
	(* https://jira.wolfram.com/jira/browse/WLDOCS-2200 *)
    (* treat a series of tutorial Notes cells as one entity to be turned into a ul list *)
	If[
		MemberQ[{"Tutorial", "Note", "Tech Note", "Monograph"}, entityType]
		&& Positive @ Count[expr, Cell[_, "Notes", ___], All]
	,
		expr = expr /.
			{
				first:___,
				c:Longest@PatternSequence[Cell[_, "Notes", ___]..],
				last:___
			} -> {
				first,
				Cell[CellGroupData[{Cell["", "TutorialNotesList"], c}, Open]],
				last
			}
	];

    (* clear out new ip section openers *)
	expr = expr /.
		{
			first:___,
			Cell[BoxData[DynamicBox[ToBoxes[__], ___]], ___],
			Cell[BoxData[TemplateBox[_, s_String /; StringContainsQ[s, "Spacer"]]], ___],
			rest:___
		} :> {first, rest};
    (* and automated open/close directives for cell groups *)
    expr = expr /. Dynamic[CurrentValue[EvaluationNotebook[], {TaggingRules, "Openers", __}, _]] :> Unevaluated[Sequence[]];

	(* **** begin Tutorial processing *)

	If[entityType == "Tutorial",
		expr = expr /. Cell[a:_, "Item", b:___] :> Cell[a, "BulletedText", b];
	];

	(* **** entityType = "Monograph" or "Monograph Overview" *)

	If[MemberQ[{"Monograph", "Monograph Overview"}, entityType],

		dprint["entering Monograph/Monograph Overview processing; expr is"];
		dfile[expr];
		(* /tmp/expr1.m.m *)

		(* ***** monograph body styles *)
		(* Formatting note: a71b8b1153ff20fe4f346dbf347bbef1f6c9443a:
			Update Monograph Overview builds
			https://jira.wolfram.com/jira/browse/WLDOCS-2142

			- moved the monograph body styles munging to the first step in
			monograph processing so all the names will be settled for following
			code.
		*)

		expr = expr //. {
			Cell[a:_, "MonographText", b:___] :> Cell[a, "Text", b]
			, Cell[a:_, "MonographMathCaption", b:___] :> Cell[a, "MCap", b]
			, Cell[a:_, "MonographSection", b:___] :> Cell[a, "Section", b]
			, Cell[a:_, "MonographSubsection", b:___] :> Cell[a, "Subsection", b]
			, Cell[a:_, "MonographTOCDocumentTitle", b:___] :> Cell[a, "TOCDocumentTitle", b]
		};

		(* "!@#$@!$" --Yosemite Sam *)

		(* Formatting note: c746380e: Handle MonographBulletedText *)
		(* the main problem with "MonographBulletedText" cells is that the last one in a
		set is segregated as the first cell in a sub-CellGroupData.  try moving it out
		of there back into the previous level of the expr. *)

		If[Positive @ Count[expr, Cell[_, "MonographBulletedText", ___], Infinity],

			dprint["MonographBulletedText manipulation, initial positions:"];
			dprint /@ (ToString /@ Position[expr, Cell[_, "MonographBulletedText", ___], Infinity]);

			expr = expr //.
				{
					a:___,
					Cell[CellGroupData[{b:Cell[_, "MonographBulletedText", ___], c:___}, d:___], e:___],
					f:___
				} :> {
					a,
					b,
					Cell[CellGroupData[{c}, d], e],
					f
				};

			expr = expr //.
				{
					first:___,
					cells:Longest[Cell[_, "MonographBulletedText", ___]..],
					rest:___
				} :> {
					first,
					Cell[CellGroupData[{
						Cell["", "MonographBulletedTextContainer"],
						cells /.
							(* TODO: This is just a === comparision between strings? *)
							s_String /; StringMatchQ[s, StartOfString~~"MonographBulletedText"~~EndOfString]
								-> "MonographBulletedTextProcessed"
					}]],
					rest
				};

			expr = expr /. "MonographBulletedTextProcessed" -> "MonographBulletedText";

			dprint["MonographBulletedText manipulation, final positions:"];
			dprint /@ (ToString /@ Position[expr, Cell[_, "MonographBulletedText", ___], Infinity]);
		];

		dprint["after Monograph body styles, expr is:"]; dfile[expr];
		(* /tmp/expr2.m.m *)

		(* (ediff-files "/tmp/expr1.m.m" "/tmp/expr2.m.m") *)

		(* ***** handle <nav> top-of-page items: Monograph only (not overviews) *)

		(* - put MonographHeader, MonographSubheader at the end of AnchorBarGrid *)
		(* - harvest the links from the hamburger menu and put them after the monograph headers *)

		dprint["Export.m: entering hamburger processing; incoming expr is: "];dfile[expr];

		If[MemberQ[{"Monograph"}, entityType],

			Module[{
				all, header, subhead,
				backLink, foreLink,
				hamburgerLinks, firstHamburgerLink, lastHamburgerLink,
				backLinkStyle, foreLinkStyle
			},
				all = Cases[expr, Cell[_, "MonographHeaderGrid", ___], Infinity];
				expr = DeleteCases[expr, Cell[_, "MonographHeaderGrid", ___], Infinity];

				header = Cases[all, Cell[_, "MonographHeader", ___], Infinity];
				subhead = Cases[all, Cell[_, "MonographSubheader", ___], Infinity];

				hamburgerLinks = Flatten @ Cases[
					all,
					TagBox[ActionMenuBox[_PaneSelectorBox, list:_List, ___], ___] -> list,
					Infinity
				];
				hamburgerLinks = hamburgerLinks /. Documentation`HelpLookup[s:_String] -> s;

				firstHamburgerLink = First[hamburgerLinks[[All, 2]]];
				lastHamburgerLink = Last[hamburgerLinks[[All, 2]]];

				dprint["Export.m:monograph - hamburgerLinks = "<>ToString[hamburgerLinks, InputForm]];
				dprint["Export.m:monograph - firstHamburgerLink = "<>ToString[firstHamburgerLink, InputForm]];
				dprint["Export.m:monograph - lastHamburgerLink = "<>ToString[lastHamburgerLink, InputForm]];

				(* TODO: Validate that one any only one of these links exists. *)
				backLink = Quiet @ Check[
					First @ Cases[
						all,
						ButtonBox[
							a:_PaneSelectorBox /; Positive@Count[a, s_String /; StringMatchQ[s, "Monograph-Back.png"], Infinity],
							Rule[BaseStyle, _],
							Rule[ButtonData, u:_String]
						] -> u,
						Infinity
					],
					""
				];
				dprint["Export.m:monograph - backLink = "<>ToString[backLink, InputForm]];
				dprint["Export.m:monograph: uri = "<>ToString[uri, InputForm]];
				(* Formatting note: 445f8312:
					Fix monograph hamburgers: https://bugs.wolfram.com/show?number=376450

					The case of a "<" or ">" link without a target hadn't been implemented.
					It's supposed to be gray and have no link. *)
				backLinkStyle = If[Or[uri == firstHamburgerLink, backLink == ""],
					"MonographInactiveBackLink",
					"MonographBackLink"
				];
				dprint["Export.m:monograph - backLinkStyle = "<>ToString[backLinkStyle, InputForm]];

				(* TODO: Validate that one any only one of these links exists. *)
				foreLink = Quiet @ Check[
					First @ Cases[
						all,
						ButtonBox[
							a:_PaneSelectorBox /; Positive@Count[a, s_String /; StringMatchQ[s, "Monograph-Forward.png"], Infinity],
							Rule[BaseStyle, _],
							Rule[ButtonData, u:_String]
						] -> u,
						Infinity
					]
					,
					""
				];

				dprint["Export.m:monograph - foreLink = "<>ToString[foreLink, InputForm]];
				dprint["Export.m:monograph: uri = "<>ToString[uri, InputForm]];

				foreLinkStyle = If[Or[uri == lastHamburgerLink, foreLink == ""],
					"MonographInactiveForeLink",
					"MonographForeLink"
				];

				dprint["Export.m:monograph - foreLinkStyle = "<>ToString[foreLinkStyle, InputForm]];
				dprint["Export.m:monograph - current uri = "<>ToString[uri, InputForm]];

				expr = expr /.
					Cell[a:_, style:"AnchorBarGrid", b:___]
					:> Cell[
						Flatten[
							{
								a,
								Cell[CellGroupData[Flatten@{header, subhead}]],
								Cell[CellGroupData[{
									Cell["", "MonographHamburgerGroup"],
									Cell[TextData[{ButtonBox["", BaseStyle->"Link", ButtonData->backLink]}], backLinkStyle],
									Cell[TextData[{ButtonBox["", BaseStyle->"Link", ButtonData->foreLink]}], foreLinkStyle],
									Cell[
										TextData[Flatten @ {
											Cell[
												TextData[{
													ButtonBox[
														#[[1]],
														BaseStyle -> "Link",
														ButtonData -> #[[2]]
													]
												}],
												"MonographHamburgerLink"
											]& /@ hamburgerLinks
										}],
										"MonographHamburgerLinks"
									]
								}]]
							},
							1
						],
						style,
						b
					]
			]
		];

		dprint["Export.m: leaving hamburger processing"];
		dprint["after Monograph <nav> items, expr is:"]; dfile[expr];

		(* ***** handle Monograph Overview structures *)

		(* In-product Monograph Overview nbs have their PacletNameCell outside the
		AnchorBarGrid, but the ABG is the only mechanism set up to handle the PNC in web
		conversion.  So we move PNC (-> badge -> <nav>) into AnchorBarGrid to use
		existing processing that puts the PNC into <nav>.  Note that the
		MonographHamburgerGroup processing above happens *only* for Monographs, not
		Monograph Overviews.  So here in a Monograph Overview we'll be starting with the
		ABG in the same state it was at the start of that processing - we don't need to
		take into account the structures introduced with the HamburgerGroup. *)

		(* dprint["Export.m transform test 3: is expr changed by this?"]; *)
		(* exprBefore = expr; *)
		If[MemberQ[{"Monograph Overview"}, entityType],
			Module[{pnc},
				pnc = Quiet @ Check[
					First @ Cases[expr, Cell[_, "PacletNameCell", ___], Infinity],
					{}
				];
				dprint["pnc = "<>ToString[pnc, InputForm]];
				expr = DeleteCases[expr, Cell[_, "PacletNameCell", ___], Infinity];

				expr = expr /.
					Cell[a:_, style:"AnchorBarGrid", b:___]
					:> Cell[
						Flatten[
							{a, pnc},
							1
						],
						style,
						b
					];

				(* and then clean up the remaining Monograph Overview TOCDocumentTitle structure *)

				expr = expr /.
					Cell[
						BoxData[
							GridBox[
							{
								{},
								{
									cell:Cell[_, "TOCDocumentTitle", ___],
									___
								},
								___
							},
							___
							],
							___
						],
						"TOCDocumentTitle",
						___
					] :> cell;

				dprint["before wrapping monograph overview links in a cellgroup, expr = "];
				dfile[expr];

				(* and then wrap the list of links in a cellgroup that can be transformed to an enclosing <section> *)

				expr = expr /.
					{a:___, x:Longest@PatternSequence[Cell[_, "MonographTOCChapter", ___]..], b:___}
					->
					{a, FlattenCellGroups[Cell[CellGroupData[{x}]]], b};
			]
		];
		(* exprAfter = expr; *)
		(* dprint["Export.m transform test 3: "<>If[exprBefore === exprAfter, "expr is unchanged", "expr is changed"]]; *)
		(* Clear[exprBefore, exprAfter]; *)

		dprint["after Monograph Overview <nav> items, expr is:"]; dfile[expr];

		(* ***** wrap body (but not footers) in special markup, part 1 *)
		(* Formatting note: d6eed3a6:
			Implement monograph body wrapper: https://jira.wolfram.com/jira/browse/WLDOCS-2142

			In a monograph the entire body after MonographTitle and before any bottom-of-page
			link lists needs to be wrapped in:

			<section>
			<div class="inner">
				...
			</div>
			</section>

			Export.m:
			- wrap all contents after title in MonographBodyWrapper
			- then remove the bottom-of-page link lists from the wrapper and insert
			them immediately after.
			- note that we limit the possible link lists to those associated with
			monographs in Make.m

			XMLTransforms/HTML/Tutorials.m
			- implement the new MonographBodyWrapper to emit the correct html
			markup requested in WLDOCS-2142.
		*)

		expr = expr /.
			Cell[
				CellGroupData[
					{
						title:Cell[_, "MonographTitle", ___],
						rest1:___
					},
					rest2:___
				],
				rest3:___
			]
			:>
			Cell[
				CellGroupData[
					{
						title,
						Cell[CellGroupData[{
							Cell["", "MonographBodyWrapper"],
							rest1
						}]]
					},
					rest2
				],
				rest3
			];

		dprint["after Monograph processing step 2, expr is:"]; dfile[expr];

		(* ***** temporarily remove hamburger menus from package monographs *)

		(* Formatting note: 463e4429: *)
		If[packageQ,
			expr = expr /.
				Cell[CellGroupData[{Cell[_, "MonographHamburgerGroup"], ___}]]
					-> Sequence[]
		];

		(* ***** end monograph-specific processing *)

		dprint["leaving Monograph processing; expr is"]; dfile[expr];

	]; (* closes If[MemberQ[{"Monograph", "Monograph Overview"}, entityType] *)

	(* **** begin HowTo processing *)

	(* Formatting note: 8221f49e and aee970e0:
		https://jira.wolfram.com/jira/browse/WLDOCS-1903

		- fix some HowTo margins by grouping the abstract & any following
		Subsubsections, all to be wrapped by a <section>
		- better section grouping in HowTos: HowTos have an odd thing going on where
		the one Subsubsection immediately before a HowToDelimiter opens a cell group
		for itself and the delimiter.  Fix that.
		- which is the final fix for https://jira.wolfram.com/jira/browse/WLDOCS-1902
	*)
	expr = expr /.
		Cell[
			CellGroupData[
				{
					a:Cell[_, "Subsubsection", ___],
					b:Cell[_, "HowToDelimiter", ___]
				},
				___
			],
			___
		] -> Unevaluated[Sequence[a, b]];

	If[
		Positive @ Count[expr, Cell[_, "HowToAbstract", ___], Infinity]
		,
		expr = expr /.
			{
				first:___,
				a:Cell[_, "HowToAbstract", ___],
				b:Longest[Cell[_, "Subsubsection"|"HowToDelimiter", ___]..],
				rest:___
			} :> {
				first,
				Cell[CellGroupData[{a, b}]],
				rest
			}
	];

	(* **** begin Workflow processing *)

	(* debugOff[]; *)

	dprint["just before workflow processing, 'uri' = "<>ToString[uri, InputForm]<>", entityType = "<>ToString[entityType, InputForm]];

	dprint["======================================================================"];
	dprint["Export.m: entering workflows & workflowguides..."];
	dfile[expr];

	(* ***** part 1: entityType = Workflow *)

    (* first handle the "main" workflows with entityType "Workflow" (as distinguished from "Workflow Guide" et al.) *)
    If[entityType == "Workflow",

		workflowQ = True;

		dprint["entering Workflow processing, expr ="];
		dfile[expr];

		(* ****** handle the top-of-page stuff from badge to platform tabs list *)

		Module[{
			anchorBarGridCell, workflowTitleCell, platformList, workflowTabsList = {}
		},

			anchorBarGridCell = Cases[expr, Cell[_, "AnchorBarGrid", ___], All];
			expr = DeleteCases[expr, Cell[_, "AnchorBarGrid", ___], All];

			workflowTitleCell = Cases[expr, Cell[_, "WorkflowTitle", ___], All];
			expr = DeleteCases[expr, Cell[_, "WorkflowTitle", ___], All];

			platformList = ToLowerCase @ DeleteDuplicates @ Cases[
				expr,
				StyleBox[s:_String, "WorkflowPlatform", ___] -> s,
				All
			];

			If[Positive @ Length[platformList],
				workflowTabsList = Cell[CellGroupData[Flatten[{
					Cell["", "WorkflowTabsList"],
					Cell[#, "WorkflowTabItem"]& /@ platformList
				}]]]
			];

			(* Formatting note: f9f53f6a: workflows: fix top-of-page handling *)
			expr = expr /. Notebook[{list:__}, rest:___] :> Notebook[
				{
					Cell[CellGroupData[Flatten[
						{
							Cell["", "WorkflowMainHeadingWrapper"],
							anchorBarGridCell,
							workflowTitleCell,
							workflowTabsList
						} /. "WorkflowDockedCell" -> Sequence[]
					]]]
					,
					list
				},
				rest
			]
		];

		dprint["after handling top-of-page items"]; dfile[expr];

		(* ****** IN-PROGRESS remove empty descriptions *)

		expr = expr /. Cell["Description", "WorkflowDescription", ___] :> Unevaluated[Sequence[]];

		dprint["after removing empty descriptions"]; dfile[expr];

		(* ****** UNTESTED wrap WorkflowParenthetical in a cellgroup *)

		expr = expr /.
			{
				first:___,
				c:Longest[Cell[_, "WorkflowParenthetical", ___]..],
				last:___
			} :> {first, Cell[CellGroupData[{c}]], last};

		dprint["after processing WorkflowParenthetical"]; dfile[expr];

		(* ****** TODO standardize the WorkflowPlatform cell to Cell["platform", "WorkflowPlatform"] *)

		(* in ip nbs, WorkflowPlatform is now a cellgroup; there are no instances of WorkflowPlatform not in its own cellgroup.
		nbs without a WorkflowPlatform cell have no platform tabs. *)

		expr = expr /. Module[{
			patt = Cell[
				TextData[{__, StyleBox[p:_String, "WorkflowPlatform", ___], ___}, ___],
				"WorkflowPlatform",
				___
			]
		},
			patt :> Cell[
				If[Positive @ Length @ Cases[expr, patt, All],
					ToLowerCase[p],
					"desktop"
				],
				"WorkflowPlatform"
			]
		];

		dprint["after processing WorkflowPlatform"]; dfile[expr];

		(* ****** TODO nb has WorkflowPlatform but no WorkflowHeader: WorkflowHeaderPlaceholder *)

		(* contains WorkflowPlatform, does not contain a WorkflowHeader => WorkflowPlatform[WorkflowHeaderPlaceholder] *)

		(* any WorkflowPlatform cellgroups that do not contain a WorkflowHeader must be given a WorkflowHeader wrapped around its contents *)

		expr = expr /. Cell[
				CellGroupData[
					{
						a:Cell[_, "WorkflowPlatform", ___],
						rest:__ /; FreeQ[{rest}, Cell[_, "WorkflowHeader", ___]]
					},
					b:___
				],
				c:___
			] :> Cell[
				CellGroupData[
					{
						a,
						Cell[CellGroupData[{
							Cell["", "WorkflowHeaderPlaceholder"],
							rest
						}]]
					},
					b
				],
				c
			];

		dprint["after  WorkflowPlatform: 1; WorkflowHeader: 0; add WorkflowHeaderPlaceholder, STEP 1"]; dfile[expr];

		(* https://jira.wolfram.com/jira/browse/WLDOCS-1902

		and consider this case: a WorkflowPlatform cell is immediately followed by cells
		that are not headers or cellgroups.  it turns out they need to be wrapped in a
		header; otherwise their web elements lose their left margin.

			Cell[CellGroupData[{
		=>        Cell["cloud", "WorkflowPlatform"],
		=>        Cell[ "WorkflowText"]                                    12 lines
		=>        Cell[ "AnnotatedOutput"]                                406 lines
				Cell[CellGroupData[{
						Cell["Programmatically...", "WorkflowHeader", WholeCellGroupOpener -> True],
						Cell[ "WorkflowText"]                              20 lines
						Cell[ "AnnotatedInput"]                            96 lines
						Cell[ "AnnotatedOutput"]                           78 lines
						Cell[ "WorkflowText"]                               5 lines
						etc...
		*)

		expr = expr /.
			{
				a:Cell[_, "WorkflowPlatform", ___],
				b:Cell[_, _String, ___]..,
				c:Cell[CellGroupData[__], ___],
				d:___
			} :> {
				a,
				Cell[CellGroupData[Flatten @ {
					Cell["", "WorkflowHeaderPlaceholder"],
					b
				}]],
				c,
				d
			};

		dprint["after  WorkflowPlatform: 1; WorkflowHeader: 0; add WorkflowHeaderPlaceholder, STEP 2"]; dfile[expr];

		(* ****** TODO no WorkflowPlatform, AND no WorkflowHeader* => *)

		(* and then there are notebooks with no WorkflowPlatform cellgroup at all: there is no tab row at the top, and there is just one page, one container.
		if there is no section-producing element in the content, then the content between header/description and footer needs to be wrapped in one.
		*)

		dprint["entering no WorkflowPlatform, AND no WorkflowHeader*"];

		(* Formatting note: bf3696e6: WLDOCS-1736 - partial fix
			- ignore "Notes" WorkflowHeader cells.  This fixes ExtractImagesFromAWebpage
			(and any similar) from WLDOCS-1736.
		*)
		expr = Module[{cases1, cases2, test1, test2},
			cases1 = Cases[expr, Cell[_, "WorkflowPlatform", ___], All];
			cases2 = Cases[
				expr,
				Cell[TextData[{s:_String /; s =!= "Notes"}], "WorkflowHeader*", ___],
				All
			];
			test1 = (cases1 == {});
			test2 = (cases2 == {});
			dprint["cases1 = "<>ToString[cases1, InputForm]];
			dprint["cases2 = "<>ToString[cases2, InputForm]];
			dprint["test1 = "<>ToString[test1, InputForm]];
			dprint["test2 = "<>ToString[test2, InputForm]];

			If[
				test1 && test2
				,
				dprint["condition is True, attempting to change expr"];

				expr /. Notebook[
					{
						header:Cell[CellGroupData[{Cell[_, "WorkflowMainHeadingWrapper", ___], ___}, ___], ___]
						,
						description:Cell[_, "WorkflowDescription"|"WorkflowText", ___]...
						,
						contents:___
					},
					rest:___
				] :> Notebook[
					{
						header
						,
						Cell[CellGroupData[{Cell["", "WorkflowHeaderPlaceholder"], description, contents}, Open]]
					},
					rest
				]
				,
				dprint["condition is False, returning unchanged expr"];
				expr
			]
		];

		dprint["after WorkflowPlatform: 0; WorkflowHeader: 0; add WorkflowHeaderPlaceholder"]; dfile[expr];

		(* ****** Input -> WorkflowInput as a test *)

		(* Formatting note: 9af05a0d:
			update workflow Input styles to match source docs

			- copy the workflow Input style to a new local style WorkflowInput
			- change those Inputs to WorkflowInputs
		*)
		expr = expr /. {
			StyleBox[a:_, "Input", b:___] -> StyleBox[a, "WorkflowInput", b]
		};

		dprint["after Input -> WorkflowInput"]; dfile[expr];

		(* ****** close Workflow processing *)

	]; (* close entitytype == "Workflow" *)

	(* ***** part 2: entityType == Workflow Guide *)

	If[
		entityType == "Workflow Guide"
		,

		dprint["Export.m: beginning WorkflowGuides..."]; dfile[expr];

		(* ****** Workflow Guide: is this a guide overview? *)

		(* workflow guide overviews have section heads that link to other workflow guides *)
		(* Formatting note: 91128283: https://jira.wolfram.com/jira/browse/WLDOCS-1812 *)
		workflowOverviewQ = Positive @ Count[
			expr,
			c:Cell[_, "WorkflowGuideSection", ___]
				/; (Positive @ Count[c, ButtonBox[_, ___], All]),
			All
		];

		If[workflowOverviewQ,
			dprint[outfile0<>" is a Workflow Guide Overview"],
			dprint[outfile0<>" is not a Workflow Guide Overview"]
		];
		dprint["Workflow Guide: is this a guide overview?  workflowOverviewQ = "<>ToString[workflowOverviewQ, InputForm]];

		(* ****** Workflow Guide: top-of-page material *)

		(* deal with the top-of-page stuff *)
		Module[{workflowtitle, anchorbargrid},
			anchorbargrid = Cases[expr, Cell[_, "AnchorBarGrid", ___], Infinity];
			expr = DeleteCases[expr, Cell[_, "AnchorBarGrid", ___], Infinity];

			(* remove both the title cell and its enclosing cellgroup, replacing the
			latter with a page-containing WorkflowGuidePlatform cellgroup that will
			provide a div and a section to enclose the main content of the page, excluding
			the top-of-page structure and any footers. *)

			workflowtitle = Cases[expr, Cell[_, "WorkflowGuideTitle", ___], Infinity];
			expr = expr /.
				Cell[CellGroupData[{Cell[_, "WorkflowGuideTitle", ___], contents:___}, ___], ___] :>
				Cell[CellGroupData[{Cell["desktop", "WorkflowGuidePlatform"], contents}]];

			expr = expr /. Notebook[list:_List, rest:___] :> Notebook[
				Flatten @ {
					Cell[CellGroupData[
						Flatten @ {
							Cell["", "WorkflowMainHeadingWrapper"],
							anchorbargrid,
							workflowtitle
						},
						Open
					]],
					list
				},
				rest
			]
		];

		dprint["Workflow Guide: top-of-page material, part 1"];
		(* dfile[expr]; *)

		(* Friday, April  6, 2018 at 17:46:43: I don't think I should do this anymore *)
		(* then remove any WorkflowGuideDelimiters that appear immediately after the title and before page content *)
		expr = expr /.
			{
				Cell[CellGroupData[{a:Cell["", "WorkflowMainHeadingWrapper"], b:___}, c:___], d:___],
				Cell["", "WorkflowGuideDelimiter", ___],
				rest:___
			} :> {Cell[CellGroupData[{a, b}, c], d], rest};

		dprint["Workflow Guide: top-of-page material, part 2"];
		dfile[expr];

		(* ****** Workflow Guide: end *)

	]; (* close workflow guide conditional *)

	(* ***** part 3: stuff done to all workflow files *)

	If[StringContainsQ[entityType, "Workflow"],

		(* ****** footers: Related Links and such *)

		dprint["entering workflow footer processing; expr is:"]; dfile[expr];

		(* ******* footer cleanup *)
		(* ******** cleanup 1: remove random junk *)

		(* CreatingDocumentsAndPresentations *)
		expr = expr /.
			Cell[CellGroupData[{
				Cell["Related Guides", "WorkflowGuideRelatedGuidesSection", ___],
				Cell[
					TextData[ButtonBox["\[Ellipsis]", BaseStyle->"Link", ___]],
					"WorkflowGuideEntry",
					___
				]
			}, ___], ___] :> Unevaluated[Sequence[]];

		expr = expr /. Cell[_, "WorkflowFooterDelimiter", ___] :> Unevaluated[Sequence[]];

		(* remove junk leaking in from the ip nb build *)
		(* expr = expr /. *)
		(*   Cell[BoxData[_DocumentationBuild`Make`Private`BuildRelatedFunctions], "Input", ___] *)
		(*   :> Unevaluated[Sequence[]]; *)

		(* then remove any subsequently-empty Related Functions footer sections *)

		(* expr = expr /. *)
		(*   Cell[CellGroupData[{Cell["Related Functions", "FooterHeader", ___], Cell["", "WorkflowFooterDelimiter", ___]...}, ___]] *)
		(*   :> Unevaluated[Sequence[]]; *)

		dprint["workflows: after footer cleanup 1, expr is:"]; dfile[expr];

		(* ******** cleanup 2: remove degenerate footer sections *)
		(* Formatting note: a023ec48: *)

		Map[
			Function[target,
				If[
					0 == Count[expr, Cell[CellGroupData[{Cell[_, target, ___], ___}, ___], ___], All]
				,
					expr = expr /. Cell[_, target, ___] :> Unevaluated[Sequence[]]
				]
			],
			{
				"WorkflowRelatedGuidesSection",
				"WorkflowTechNotesSection",
				"WorkflowRelatedWorkflowsSection",
				"WorkflowRelatedFunctionsSection",
				"WorkflowRelatedLinksSection"
			}
		];

		dprint["workflows: after footer cleanup 2, expr is:"]; dfile[expr];

		(* ******* related functions: leave links unchanged, add correct spacing items from the demo pages *)

		dprint["entering WorkflowRelatedFunctions problem, expr is:"]; dfile[expr, "Format" -> ".nb"];

		(* remove duplicate WorkflowRelatedFunctions cells introduced somewhere in the
		high-res web version of WorkflowFinisher *)

		expr = expr /.
			Cell[links:_, style:"WorkflowRelatedFunctions", rest:___]
			:> Cell[
				links /. {
					" \[FilledVerySmallSquare] " -> Unevaluated[Sequence[
						"\[NonBreakingSpace]",
						StyleBox["", "InlineSeparator"
					]]]
				},
				style,
				rest
			];

		(* Formatting note: 84ae70cd:
			Workflow Related Functions: remove duplicates introduced by the hi-res
			web branch of WorkflowFinisher. *)
		expr = expr /.
			{
				first:_,
				c:Cell[_, "WorkflowRelatedFunctions", ___]..
			} :> Flatten @ {
				first,
				DeleteDuplicates[{c}]
			};

		dprint["leaving WorkflowRelatedFunctions problem, expr is:"]; dfile[expr, "Format" -> ".nb"];

		(* ******* move footers out to the top level of the expression? *)

		(* wherever footer sections may be, move them to the end of the Notebook's main
		list.  then stick the FooterCell at the bottom of the page 

		WLDOCS-1851: *all* Workflow*Section cell styles that are used as footer sections
		need to be listed for posList below.

		In[1]:= m=FileNames["*.nb", "/Users/billw/wri/workspace/Documentation.Notebooks.English.11.3_HighResolution/Documentation/English/System", Infinity];m//Length
		Out[1]= 270

		In[2]:= res={};Module[{e,h},e=Quiet@Get@#;h=Cases[e,s_String/;StringMatchQ[s,"Workflow*Section"],All];AppendTo[res,h]]&/@m;res=DeleteDuplicates@Flatten@res;TableForm[res, TableSpacing->{0,3}]
		Out[2]//TableForm= WorkflowGuideSection [not a footer section, so remove from the list of footer sections below]
						WorkflowGuideRelatedGuidesSection
						WorkflowRelatedWorkflowsSection
						WorkflowRelatedFunctionsSection
						WorkflowTechNotesSection
						WorkflowRelatedLinksSection
						WorkflowNotesSection [not a footer section, so remove from the list of footer sections below]
						WorkflowRelatedGuidesSection
		*)

		expr = Module[{posList, footers},

			expr = expr /. Cell[_, "FooterCell", ___] :> Unevaluated[Sequence[]];

			posList = Position[
				expr,
				Cell[CellGroupData[{
					Cell[
						_,
						s_String /; StringMatchQ[s, {
							"WorkflowGuideRelatedGuidesSection"
							, "WorkflowRelatedWorkflowsSection"
							, "WorkflowRelatedFunctionsSection"
							, "WorkflowTechNotesSection"
							, "WorkflowRelatedLinksSection"
							, "WorkflowRelatedGuidesSection"
						}],
						___
					],
					rest1:___
				}, rest2:___], rest3:___],
				All
			];

			(* Formatting note: 9eed3fa184abe9c359abfb32b6c13a00e62ac77d:
				WLDOCS-1712: add <div id="_footer-offset"></div> for correct spacing above
				permanent site footer

				- when deciding whether to insert the FooterOffset cell, explicitly do not
				touch exprs without "Related *" footers.
			*)
			If[
				Positive @ Length @ posList
			,
				footers = Extract[expr, posList];
				expr = ReplacePart[expr, posList -> Sequence[]];
				expr /. Notebook[first:_List, rest:___] :>
					Notebook[
						Flatten[{{
							first
							, Cell["", "FooterOffset"]
							, footers
							, Cell["", "WorkflowFooterCell"]
						}}]
						,
						rest
					]
			,
				expr
			]
		];

		dprint["after moving footers to the top level of expr, expr is:"]; dfile[expr];

		(* ****** add WorkflowFooterCell if needed *)

		(* at the end of the Notebook's list *)

		dprint["Export.m: add WorkflowFooterCell if needed.  Incoming expr is:"];dfile[expr /. _RasterBox -> rasterBox];

		(* dprint["Export.m transform test: is expr changed by this?"]; *)
		(* exprBefore = expr; *)

		(* Formatting note: 722332747c8fa5b7d70257041233820183e1968d:
			Restore virtual include and footers on some pages; Add WorkflowFooterCell where needed
			https://jira.wolfram.com/jira/browse/WLDOCS-1979
		*)
		If[0 == Count[expr, Cell[_, "WorkflowFooterCell", ___], Infinity],
			dprint["0 == Count[expr, Cell[_, \"WorkflowFooterCell\", ___], Infinity]"];
			expr = expr /.
				Notebook[list:_List, rest:___]
				:> Notebook[
					Append[list, Cell["", "WorkflowFooterCell"]],
					rest
				]
		];

		(* exprAfter = expr; *)
		(* dprint["Export.m transform test: "<>If[exprBefore === exprAfter, "expr is unchanged", "expr is changed"]]; *)
		(* Clear[exprBefore, exprAfter]; *)

		dprint["Export.m: WorkflowFooterCell, outgoing expr is:"];dfile[expr /. _RasterBox -> rasterBox];

		(* ****** normalize "LaTeX" *)

		(* https://jira.wolfram.com/jira/browse/WLDOCS-2737 *)
		(* replace various froufrou forms of the LaTeX logo with the simple string
		"LaTeX", which is universally recognized and accepted by people who use LaTeX *)

		expr = expr /. {
			Cell[BoxData[FormBox[StyleBox["LaTeX", SingleLetterItalics -> False], TraditionalForm]]]
				-> "LaTeX",
			s_String /; StringContainsQ[s, WordBoundary ~~ "Latex" ~~ WordBoundary]
				:> StringReplace[s, "Latex" -> "LaTeX"]
		};

		(* ****** end of workflow processing *)

		dprint["Export.m: leaving workflows & workflowguides..."];
		dfile[expr];
		dprint["======================================================================"];

	]; (* closes StringContainsQ[entityType, "Workflow"] *)

	(* debugOn[]; *)

	(* **** entityType == "Guide"|"Guide Overview"|"Standard Extra Packages" *)

	dprint["Export.m: Guides - entering Guide processing; expr = "]; dfile[expr];

	If[StringMatchQ[entityType, "Guide" | "Guide Overview" | "Standard Extra Packages"],

		(* ***** InlineGuideFunctionListing cells containing GuideText cells have their text cells transformed to <span>s rather than <p>s *)

		(* as of 11.2, only SummaryOfNewFeaturesIn112 contains "GuideText" cells within
		"InlineGuideFunctionListing" cells.  change them to
		"GuideFunctionListingText" for special handling. *)

		(* expr = expr /. a_Cell[_, "InlineGuideFunctionListing", ___] :> ReplaceAll[a, Cell[first:_, "GuideText", rest:___] :> Cell[first, "GuideFunctionListingText", rest]]; *)

		(* Formatting note: 73a92dc5:
			WLDOCS-1553: change "GuideText" cells within "InlineGuideFunctionListing" cells to
			"GuideFunctionListingText" for special handling
		*)
		expr = Module[{posList},
			posList = Position[
				expr,
				Cell[_, "InlineGuideFunctionListing", ___]
			];

			Map[
				Module[{pos, ex},
					pos = #;

					ex = Extract[expr, pos];
					ex = ex /.
						Cell[first:_, "GuideText", rest:___]
						-> Cell[first, "GuideFunctionListingText", rest];

					expr = ReplacePart[expr, pos -> ex];
				]&,
				posList
			];

			expr
		];

		dprint["Export.m: Guides - leaving InlineGuideFunctionListing; expr = "]; dfile[expr];

		(* ***** RecentlyAddedFeatures *)

		If[
			StringMatchQ[uri, "guide/RecentlyAddedFeatures"]
			,
			(* wrap phrases in appropriate structures for css styling *)
			expr = expr /.
				Cell[
					TextData[{StyleBox[s1:_String, ___], ___String, StyleBox[s2:_String, ___]}],
					"GuideFunctionsSection",
					rest:___
				] :> Cell[
					TextData[{
						Cell[s1, "GuideRecentlyAddedUpdates"],
						" ",
						Cell[s2, "GuideRecentlyAddedDate"]
					}],
					"GuideFunctionsSection"
				];

			(* fix the inner instance of nested "GuideFunctionsSection" cells *)
			expr = expr /.
				Cell[first:_, "GuideFunctionsSubsection", rest:___]
				:> Cell[
					first /. Cell[s:_String, "GuideFunctionsSubsection", FontSize -> 15] -> s,
					"GuideFunctionsSubsection",
					rest
				];

			(* remove bitmap magnification as it gives a smaller image (yay) with
			lots of white space at the top (boo).  perhap we can shrink these via css *)
			expr = expr /. {
				Rule[ImageSize, Magnification[0.5]] :> Unevaluated[Sequence[]],
				Rule[Magnification, _] :> Unevaluated[Sequence[]]
			};

			(* remove a mysterious subscript around the new livesteam icon *)
			(* Formatting note: dbbcc7c5:
				fix guide/RecentlyAddedFeatures
				https://jira.wolfram.com/jira/browse/WLDOCS-3023
				the new livestream icon is inside a subscript structure - remove it.
			*)
			expr = expr /.
				Cell[TextData[RowBox[{
					Cell[BoxData[AdjustmentBox[box:_, BoxBaselineShift -> _]]],
					a:___,
					"Livestream",
					b:___
				}]]]
				:> Cell[TextData[RowBox[{
					Cell[BoxData[box]],
					a,
					"Livestream",
					b
				}]]];
		];

		dprint["Export.m: Guides - leaving RecentlyAddedFeatures; expr = "]; dfile[expr];

		(* ***** fix grouping for pages with GuideTOCLink and GuideDelimiterSubsection *)

		(* Formatting note: 57fa3060:
			Add support for the Standard Extra Packages guide page:
				https://bugs.wolfram.com/show?number=362785
		*)
		If[
			entityType == "Standard Extra Packages"
			,

			dprint["entering: fix grouping for pages with GuideTOCLink and GuideDelimiterSubsection; initial expr = "];
			dfile[expr];

			expr = FlattenCellGroups[expr];

			expr = expr /. 
				{
					first:___,
					a:Cell[_, "GuideTitle", ___],
					b:Cell[_, "GuideAbstract", ___],
					rest:___
				} :> {first, Cell[CellGroupData[{a, b}]], rest};

			expr = expr //.
				{
					first:__
					, delim:Cell[_, "GuideDelimiterSubsection", ___]
					, a:Cell[_, "GuideFunctionsSubsection", ___]
					, b:Longest[Cell[_, "GuideTOCLink", ___]..]
					, rest:__
				} :> {first, delim, Cell[CellGroupData[{a, b}]], rest};

			dprint["leaving: fix grouping for pages with GuideTOCLink and GuideDelimiterSubsection; final expr = "];
			dfile[expr];
		];

		(* expr = expr /. { *)
		(*   first:___, *)
		(*   Cell[CellGroupData[{… *)
		(*     a:Cell[_, "GuideTOCLink", ___] *)
		(*     …b:Cell[_, "GuideDelimiterSubsection", ___] *)
		(*   }, Open]], *)
		(*   rest:___ *)
		(* } :> {first, a, b, rest}; *)

		(* ***** handle section heads with inline text following them *)
		(* Formatting note: 98e05382:
			Allow guide sections to appear in GuideText cells
			https://jira.wolfram.com/jira/browse/WLDOCS-2274
		*)

		(* Sometimes people want this all in one line in a guide:
		Section Head - inline text
		It works in-product, so it should work on the web, too.

		2 -> {"/Users/billw/git/unitylink/Documentation/English/Guides/UnityLink.nb"}
		1 -> {"/Users/billw/git/microcontrollerkit/Documentation/English/Guides/MicrocontrollerKit.nb"} *)

		dprint["Guides: entering inline section heads, expr is"]; dfile[expr];

		Module[{patt},
			patt = c:Cell[
				TextData[
					{
						contents:__ /; Positive[Count[
							{contents},
							Cell[_, s:_String /; StringMatchQ[s, "*section*", IgnoreCase -> True], ___],
							Infinity
						]]
					},
					___
				],
				"GuideText",
				___
			];

			If[Positive @ Count[expr, patt, Infinity],
				expr = expr /. patt :> Cell[CellGroupData[{Cell["", "GuideTextSectionWrapper"], c}]]
			]
		];

		dprint["Guides: leaving inline section heads, expr is"]; dfile[expr];

		(* ***** add an "open all"/"close all" button after the first "GuideFunctionsSection" *)

		(* https://jira.wolfram.com/jira/browse/WLDOCS-2398 *)
		(* don't apply this to the guide/RecentlyAddedFeatures snowflake *)
		If[uri =!= "guide/RecentlyAddedFeatures",
			Module[{pos, head},
				pos = Quiet @ Check[
					First @ Position[expr, Cell[_, "GuideFunctionsSection", ___], Infinity],
					{}
				];
				head = Extract[expr, pos];
				head = AppendTo[head, "GuideOpenAll" -> True];
				expr = ReplacePart[expr, pos -> head];
				(* remove per WLDOCS-2398: *)
				If[Positive@Length@pos,
					expr = DeleteCases[
						expr,
						Cell[_, "GuideDelimiterSubsection", ___],
						Infinity
					]
				];
			]
		];

		(* ***** new SymbolIcons for visualization guide pages *)

		(* Formatting note: 449ebdf9:
			Implement new guide icons: https://jira.wolfram.com/jira/browse/WLDOCS-2900 *)
		symbolIconDir = FileNameJoin[{
			$InstallationDirectory, "SystemFiles", "FrontEnd", "SystemResources",
			"Bitmaps", "Documentation", "SymbolIcons"
		}];
		symbolIconFiles = FileNames["*@144dpi.png*", symbolIconDir, Infinity];
		symbolIconDB = Association @ Thread[
			Rule[StringReplace[FileBaseName /@ symbolIconFiles, "@144dpi" -> ""], symbolIconFiles]
		];
		(* symbolIconDB consists of entries like this: *)
		(* "AbsArgPlot" -> "$InstallationDirectory/SystemFiles/FrontEnd/SystemResources/Bitmaps/Documentation/SymbolIcons/A/AbsArgPlot@144dpi.png", *)

		dprint["before SymbolIcon processing, expr = "]; dfile[expr];

		Module[{posList},
			posList = Position[
				expr,
				Cell[
					BoxData[RowBox[{
						DynamicBox[
							FEPrivate`ImportImage[FrontEnd`FileName[{"Documentation", "SymbolIcons", _}, _String]],
							___
						],
						___
					}]],
					___
				]
			];

			Module[{pos, ex, name, imgFile, imgCell},
				pos = #;
				ex = Extract[expr, pos];

				name = First @ Cases[
					ex,
					s_String /; StringContainsQ[s, ".png" ~~ EndOfString],
					Infinity
				];
				imgFile = symbolIconDB[FileBaseName[name]];
				dprint["in posList loop, imgFile = "<>ToString[imgFile, InputForm]];

				imgCell = Cell[BoxData[ToBoxes[Import[imgFile]]], "GuideSymbolIcon"];
				expr = ReplacePart[expr, pos -> imgCell];
			]& /@ posList;
		];

		dprint["after SymbolIcon processing, expr = "]; dfile[expr];

		(* ***** end Guide processing *)
	];

	dprint["Export.m: Guides - leaving Guide processing; expr = "]; dfile[expr];
	(* **** entityType == "Overview" *)

	dprint["Export.m: Overviews - entering Overview processing; expr = "]; dfile[expr];

	(* dprint["Export.m transform test 4: is expr changed by this?"]; *)
	(* exprBefore = expr; *)
	(* Formatting note: 4c9b3aac:
		https://jira.wolfram.com/jira/browse/WLDOCS-1921
		- in tutorial overviews, a TOCDocumentTitle cell should be followed by
		TutorialAbstract, not Text.
	*)
	If[StringMatchQ[entityType, "Overview"],
		expr = expr /.
			{
				a:___,
				b:Cell[_, "TOCDocumentTitle", ___],
				Cell[c1:_, "Text", c2:___],
				d:___
			} :> {a, b, Cell[c1, "TutorialAbstract", c2], d};
	];
	(* exprAfter = expr; *)
	(* dprint["Export.m transform test 4: "<>If[exprBefore === exprAfter, "expr is unchanged", "expr is changed"]]; *)
	(* Clear[exprBefore, exprAfter]; *)

	(* **** entityType == "Entity" *)

	(* Formatting note: 96ee241f:
		Set up usage and abstract for Entity pages

		It seems that Usage and Abstract mean different things in different
		subdirectories of System/ReferencePages.  For entities, changing EntityUsage to
		Usage activates a transform that does the Right Thing.

		Defaults.m: EntityAbstract is the sort of thing that's called Usage elsewhere in
		docs that have no traditional Usage cell.  Anyway, set this up to do the right
		thing.
	*)
	expr = expr /. Cell[a_, "EntityUsage", b___] :> Cell[a, "Usage", b];

	(* **** entityType == "Symbol" *)
	(* ***** standardize the spacing before StyleBox[_, "FullListing"] *)

	dprint["Export.m: entering Symbol FullListing processing; expr = "]; dfile[expr];

	(* https://jira.wolfram.com/jira/browse/WLDOCS-2672

	In ref/Import.html, for example, we have "Listing of Formats »" next to the
	function name.  Import's structure gives nice results on the web; others with
	different whitespace structures result in little or no space between the name and
	the text.

	This appears to be the pattern that produces correct spacing on the web:
		{"Name[Space][MediumSpace][MediumSpace]", ...}

	other patterns found Oct 6 2019:
		{"Name[MediumSpace][MediumSpace]", ...}
		{"Name[MediumSpace][MediumSpace][MediumSpace]", ...}
		{"Name[Space]", ...}
		{"Name", "[MediumSpace][MediumSpace]", ...}
		"Name[Space]", "[MediumSpace][MediumSpace]"

	StringTrim[name] will clear out trailing whitespace from the first string

	instances of the string that contains only \[MediumSpace]:
		StringMatchQ[x, "\[MediumSpace]"..]

	{
		Cell[
			TextData[
				{
					"$PublisherID [MediumSpace][MediumSpace]",
					StyleBox[
						ButtonBox[
							"Request a Publisher ID »",
							BaseStyle -> {"Hyperlink", "FullListing"},
							ButtonData -> {URL["http://datarepository.wolframcloud.com/request-publisher-id"], None}
						],
						"FullListing"
					]
				}
			],
			"ObjectName",
			ExpressionUUID -> "ae9943d6-f0ce-4726-ab0b-290c6f931fb7"
		]
	}
	*)

	If[Positive @ Count[expr, StyleBox[_, "FullListing", ___], Infinity],
		dprint["FullListing: munging ObjectName cell"];
		expr = expr /.
			Cell[
				TextData[
					{
						title:_String,
						spaces:___,
						styleBox:_StyleBox,
						rest1:___
					},
					rest2:___
				],
				"ObjectName",
				rest3:___
			] :> Cell[
				TextData[
					{
						StringTrim[title]<>" \[MediumSpace]\[MediumSpace]",
						styleBox,
						rest1
					},
					rest2
				],
				"ObjectName",
				rest3
			]
	];

	dprint["Export.m: exiting Symbol FullListing processing; expr = "]; dfile[expr];

	(* **** bottom-of-page items *)

	(* https://jira.wolfram.com/jira/browse/WLDOCS-2159: changed structure of
		Related... links at the bottom of each page.

		The plan here is to identify each kind of Related section and then simplify it
		down to the bare head and links wrapped in suitable CellGroupData and Cell.
		Almost like saying RelatedWorkflows[{link1, link2, ...}], where RelatedWorkflows
		would be analogous to a transmog transform defined in Defaults.m (since this
		will apply to nearly every page).
	*)

	(* ***** replace new SeeAlso structure, if present, with the traditional one preserved in its TaggingRules *)

	(* Formatting note: 6bdff7fa:
		Handle new SeeAlso structures correctly

		https://bugs.wolfram.com/show?number=410173

		When a SeeAlso cell is presented in a new dynamic-module format with a
		traditional cell attached as a tagging rule, replace the former with the
		latter.
	*)
	expr = expr /.
		Cell[
			BoxData[GridBox[{{db:_DynamicBox, _DynamicModuleBox}}]],
			"SeeAlsoSection",
			TaggingRules -> {"SeeAlsoGrid" -> gb:_GridBox, ___},
			rest:___
		] -> Cell[
			BoxData[GridBox[{{db, gb}}]],
			"SeeAlsoSection",
			rest
		];

	(* First@Cases[x, Rule[TaggingRules, {Rule["SeeAlsoGrid", t_], ___}] -> t, Infinity] *)

	(* ***** Handling Make.m:AddAttractorIcons *)

	(* The next expressions take footers with attractor icons and convert them to
	old-style footers for use with existing transforms.  Then we can update the
	transforms to provide whatever html markup Sue needs to make the icons appear.
	In the updated transforms, the footer icon image will be chosen based on the
	"*Section" cell style found in the expression.

	for use with updated transforms:
		"*SeeAlsoSection" -> "RelatedFunction.png",
		"*MoreAboutSection" -> "RelatedGuide.png",
		"*TutorialsSection" -> "RelatedTutorial.png",
		"*TechNotesSection" -> "RelatedTechNote.png",
		"*MonographsSection" -> "RelatedMonograph.png",
		"*EntitySection" -> "RelatedEntity.png",
		"*FormatSection" -> "RelatedFormat.png",
		"*InterpreterSection" -> "RelatedInterpreter.png",
		"*WorkflowsSection" -> "RelatedWorkflow.png",
		"*WorkflowGuidesSection" -> "RelatedWorkflowGuide.png" *)


	(* either howto pages don't receive attractor icons, or they're still under development in-product *)
	(* Formatting note: 40160cc6dd8348b7c1be7f9d4df75bce3e9cbbf4:
		- exclude howto pages from the new related-link attractor code.  it seems the
		in-product howto notebooks don't receive the modern link list structures, so
		running them through here just makes a mess.  if in-product howto processing
		is updated to use attractors, remove this exclusion.
	*)
	If[entityType =!= "HowTo",

		dprint["Export.m: initial expr handling AddAttractorIcons:"];
		dfile[expr];

		(* "SeeAlsoSection" *)
		(* Formatting note: a150563f / ca91e18f / 799dcad4 *)
		expr = expr /.
			Cell[
				BoxData[GridBox[{{
					_DynamicBox,
					GridBox[List[_List, {linkListCell:Cell[_, "SeeAlso"]}]]
				}}]],
				"SeeAlsoSection",
				___
			] :> Cell[CellGroupData[
				{Cell["See Also", "SeeAlsoSection"], linkListCell},
				Open
			]];

		(* all others with footer icons *)
		expr = expr /.
			Cell[
				BoxData[GridBox[{{
					_DynamicBox,
					GridBox[{{StyleBox[title:_, style:_String]}, linkList:__List}]
				}}]],
				style:s_String /; StringMatchQ[s, "*Section"],
				___
			] :> Cell[CellGroupData[
				{Cell[Transmogrify`ConvertToString[title], style], linkList},
				Open
			]];

		(* Related Links have a special structure, too *)

		(* Formatting note: d8bd94e9f18b57fff8198130c026b44fd0cbc727:
			Fix guide footer processing and implement Related Links section
			WLDOCS-2159, WLDOCS-2228
		*)
		If[
			Positive @ Count[expr, Cell[_, s_String /; StringContainsQ[s, "RelatedLinksSection"], ___], Infinity]
			,
			dprint["page contains a RelatedLinksSection"];

			Module[{
				icon2linkstyle,
				relatedlinks, newlinks,
				linkPattern1, linkPattern2
			},
				(* this is based on Make.m:TransformRelatedLinksCells *)
				icon2linkstyle = <|
					(* css implementation at /ssh:billw@webdev4.wolfram.com:/www/sites/reference.wolfram.com/htdocs/language-assets/css/global.css *)
					"RelatedLinks-Demonstrations.png"        -> "demonstrations"
					, "RelatedLinks-EIWL.png"                -> "elementary-intro"
					, "RelatedLinks-FastIntro.png"           -> "fast-intro"
					, "RelatedLinks-NKS.png"                 -> "nks"
					, "RelatedLinks-Community.png"           -> "community"
					, "RelatedLinks-DataRepository.png"      -> "data-repo"
					, "RelatedLinks-FunctionRepository.png"  -> "function-repo"
					, "RelatedLinks-Functions.png"           -> "functions"
					, "RelatedLinks-MathWorld.png"           -> "mathworld"
					, "RelatedLinks-NeuralNetRepository.png" -> "nnet-repo"
					, "RelatedLinks-TMJ.png"                 -> "tmj"
					, "RelatedLinks-WolframAlpha.png"        -> "wolframalpha"
					, "RelatedLinks-WolframU.png"            -> "wolframu"
				|>;

				(* according to Make.m we have "RelatedLinks", "WorkflowRelatedLinks", "GuideRelatedLinks" *)
				relatedlinks = Cases[
					expr,
					Cell[_, s:_String /; StringContainsQ[s, "RelatedLinks" ~~ EndOfString], ___],
					Infinity
				];
				dprint["AddAttractorIcons: relatedlinks = "<>ToString[relatedlinks, InputForm]];

				newlinks = Module[{link, icon, style}, Catch[
					link = #;
					dprint["newlinks step 1: link is: "]; dfile[link];
					icon = Quiet @ Check[
						First @ Cases[
							link,
							FrontEnd`FileName[_List, s_String] -> s,
							Infinity
						]
						,
						dprint["no icon, throwing 'link'"];
						Throw[link]
					];
					style = icon2linkstyle[icon];
					dprint["AddAttractorIcons: style = "<>ToString[style, InputForm]];

						(* Formatting note: 6a51659e:
							Fix functions.w.c links in refpages
							https://bugs.wolfram.com/show?number=378457
							https://jira.wolfram.com/jira/browse/WLDOCS-2786
						*)
					linkPattern1 = TextData[{
						Cell[BoxData[DynamicBox[FEPrivate`ImportImage[FrontEnd`FileName[__]]]]],
						Cell[BoxData[TemplateBox[_, s_String /; StringContainsQ[s, "Spacer"]]]],
						text:__
					}];
					linkPattern2 = TextData[
						{
							_ButtonBox,
							ButtonBox[Cell[BoxData[TemplateBox[_, s_String /; StringContainsQ[s, "Spacer"]]]]],
							ButtonBox[text_String]
						},
						___
					];

					Which[
						Positive @ Count[link, linkPattern1, Infinity],
							dprint["RelatedLinks branch 1"];
							link = link /. linkPattern1 :> TextData[{text}];
							(* add cell option "RelatedLinkStyle" *)
							link = link /.
								Cell[BoxData[TemplateBox[a:__, BaseStyle -> {basestyleA:_String /; StringContainsQ[basestyleA, "RelatedLinks"~~EndOfString], basestyleB:___}, b:___]]] :>
								Cell[BoxData[TemplateBox[a   , BaseStyle -> {basestyleA                                                                    , basestyleB    }, b,   "RelatedLinkStyle" -> style]]];
					,
						Positive @ Count[link, linkPattern2, Infinity],
							dprint["RelatedLinks branch 2"];
							link = link /. linkPattern2 :> TextData[{text}];
							(* add cell option "RelatedLinkStyle" *)
							link = link /.
								Cell[TextData[ButtonBox[Cell[TextData[{a:_String}]], BaseStyle -> basestyle:_, ButtonData -> {buttondata_, rest:___}]], "RelatedLinks"] :>
								Cell[TextData[ButtonBox[Cell[TextData[{a        }]], BaseStyle -> basestyle  , ButtonData -> {buttondata , rest    }]], "RelatedLinks", "RelatedLinkStyle" -> style];
					,
						True,
							dprint["RelatedLinks branch 3: leaving link untouched"];
							link
					];
					dprint["newlinks step 2: link is: "]; dfile[link];
					dprint["style = "<>ToString[style, InputForm]];

					If[style == "demonstrations",
						(* Formatting note: 84ae70cd: Fix demonstration link formatting
							https://jira.wolfram.com/jira/browse/WLDOCS-3357 *)
						link = link /.
							{
								first:___,
								a:PatternSequence["(" | " (", _Cell, ")"],
								last:___
							} -> {
								first,
								" ",
								Cell[TextData[{a}], "RelatedLinksParenText"],
								last
							};
					];
					dprint["newlinks step 3: link is: "]; dfile[link];

					link
				]]& /@ relatedlinks;

				dprint["Related Links processing: newlinks = "]; dfile[newlinks];

				(* Formatting note: 07181b9236ab1274e1c5f73bf80aa3781e2f5ba2:
					Fix Related Links in workflows:
						https://jira.wolfram.com/jira/browse/WLDOCS-2379

					Related Links sections in Workflows and WorkflowGuides are formatted as
					they were up through v11.3.  On other pages, they use GridBox formatting
					to handle the new-in-v12 Related Links icons.  brianv says workflow RL
					lists will be updated after v12.0.  Until then, process them the old
					way.

					- temporarily accept old-style Related Link lists for workflow pages.
				*)
				expr = If[StringContainsQ[entityType, "Workflow"],
					expr /.
						Cell[CellGroupData[{Cell[_, s_String /; StringContainsQ[s, "RelatedLinksSection"], ___], ___}, ___], ___]
							:> Cell[CellGroupData[Flatten[{Cell["Related Links", s], newlinks}]]]
					,
					expr /.
						Cell[_, s_String /; StringContainsQ[s, "RelatedLinksSection"], ___]
							:> Cell[CellGroupData[Flatten[{Cell["Related Links", s], newlinks}]]]
				]
			]
		];

		dprint["Export.m: final expr handling AddAttractorIcons:"];
		dfile[expr];
	];

	(* ***** History cell formatting *)

	dprint["Export.m: begin History cell formatting:"]; dfile[expr];

	(* ****** untangle History and Citation cells *)

	(* not all pages with HistorySection also have CitationContainerCell... *)
	(* Formatting note: 7012f918: https://jira.wolfram.com/jira/browse/WLDOCS-3411
		Produce history cells when no citations *)
	expr = expr /.
		{
			first:__,
			c:Cell[_, "HistorySection", ___],
			last:__
		} -> {first, Cell[CellGroupData[{c}]], last};

	(* ...but the rest do *)
	(* Formatting note: 57f7db24 and 3245b31d: https://jira.wolfram.com/jira/browse/WLDOCS-3285
		Web implementation of history summary *)
	expr = expr /.
		{
			first:___,
			Cell[CellGroupData[
				{
					cHistory:Cell[_, "HistorySection"],
					cCitation:Cell[_, "CitationContainerCell"]
				},
				Open
			]],
			rest:___
		} -> {
			first,
			Cell[CellGroupData[{cHistory}]],
			Cell[CellGroupData[{cCitation}]],
			rest
		};

	dprint["Export.m: History cell formatting point 0:"]; dfile[expr];

	(* ****** munge the History cell for better transmogrification *)

	Module[{h, contents, rest, hlist, x, a, b, c},

		(* ******* convert the inner History cells to HistoryItem cells *)

		expr = expr /. Cell[contents:_BoxData, "History", rest:___] -> Cell[contents, "HistoryItem", rest];

		dprint["Export.m: before History processing, initial full expr = "]; dfile[expr];

		(* ******* 1: create 'h': extract the main History cell to 'h' *)

		h = Cases[expr, Cell[_, "History", ___], Infinity];
		dprint["Export.m: History cell formatting point 1: h = "]; dfile[h];

		(* ******* 2: create 'hlist': extract the History cell's main TextData list to 'hlist' *)

		(* but first, some history cells have no TextData list!  i can supply one before processing: *)

		h = h /.
			Cell[
				TextData[Cell[
					BoxData[TemplateBox[
						list:_List,
						"GrayLink",
						BaseStyle -> "History"
					]],
					"HistoryItem"
				]],
				"History"
			] :> Cell[
				TextData[{
					Cell[BoxData[TemplateBox[list, "GrayLink", BaseStyle -> "History"]], "HistoryItem"]
				}],
				"History"
			];

		dprint["Export.m: History cell formatting point 1a: h = "]; dfile[h];

		hlist = Cases[h, Cell[TextData[x:_List], "History"] -> x, Infinity];

		If[Length[hlist] > 1,
			DocsError["Notebook contains more than one \"History\" section."];
		];

		Assert[MatchQ[hlist, {} | {_}]];

		dprint["Export.m: History cell formatting point 2: hlist = "]; dfile[hlist];

		If[Length[hlist] === 1, (

			hlist = First[hlist];

			(* ******* 3: modify 'hlist': unite early unlinked history items in one cell, leaving out HistoryDelimiter cells *)

			(* wrap the unlinked text-only items *)
			hlist = SequenceReplace[
				hlist,
				{a:_String, b:_StyleBox}
					:> Cell[
						TextData[{Cell[a, "HistoryText"], b}],
						"HistoryItem"
					]
			];
			dprint["Export.m: History cell formatting point 3: hlist = "]; dfile[hlist];

			(* ******* 4: add HistoryText cells to the linked entries in 'hlist' *)

			hlist = hlist /.
				Cell[
					BoxData[TemplateBox[
						{
							Cell[TextData[{string:_String, box:_StyleBox}]],
							link:_String
						},
						rest:___
					]],
					"HistoryItem"
				] -> Cell[
					BoxData[TemplateBox[
						{
							Cell[TextData[{Cell[string, "HistoryText"], box}]],
							link
						},
						rest
					]],
					"HistoryItem"
				];

			(* Formatting note: cc13ab71:
				Fix Japanese history section: When built with 12.3.1, Mathematica longnames
				\[NonBreakingSpace] and \[ThinSpace] in Japanese history sections are handled
				incorrectly and are converted to odd binary characters, giving a history cell with
				xml elements showing from the attempt to process the characters.
			*)
			hlist = hlist /.
				s_String :> StringReplace[s, {
					"\[NonBreakingSpace]" -> " ",
					"\[ThinSpace]" -> " "
				}];

			dprint["Export.m: History cell formatting point 4: hlist = "]; dfile[hlist];

			(* ******* 5: assemble the new History cell and replace expr's version *)

			h = h /.
				Cell[TextData[x:_List], "History"]
					-> Cell[TextData[hlist], "History"];
			dprint["Export.m: History cell formatting point 5: h = "]; dfile[h];

			expr = expr /.
				Cell[CellGroupData[{Cell[_, "HistorySection", ___], ___}, ___], ___]
				:>
				Cell[CellGroupData[Flatten@{Cell["", "HistorySection"], h}]];
		)];
	];

	dprint["Export.m: after History cell formatting, expr = :"]; dfile[expr];

	(* ***** Citation cell *)

	dprint["Export.m: beginning citation cell: "]; dfile[expr];

	(* provide better cellgrouping for citation-only pages (with no accompanying history cell): *)
	(* Formatting note: 7825d8ea: *)
	expr = expr /. {
			first:___
			, Cell[CellGroupData[{
				c1:Cell[CellGroupData[{__}, ___], ___]
			, c2:Cell[_, "CitationContainerCell", ___]
			, ___}, ___], ___]
			, rest:___
		} -> {
			first
			, c1
			, Cell[CellGroupData[{c2}]]
			, rest
		};

	(* Formatting note: 1e5d7104:
		Web implementation for citations:
			https://jira.wolfram.com/jira/browse/WLDOCS-3223

		Export.m: harvest citation data from the notebook's tagging rules,
		process it, then replace the notebook's citation cells with simplified
		structures for Transmogrify to handle.
	*)
	citationData = Cases[
		taggingRules["CitationPopupData"],
		Cell[TextData[{contents:__}]] -> {contents},
		All
	];

	citationCells = Module[{
		citationStructure, title, citation
	},
		citationStructure = #;
		title = First @ Cases[citationStructure, StyleBox[s:_String, "CitationFormatsLabel", ___] -> s, Infinity];
		citation = First @ Cases[citationStructure, FormBox[s:_String, TextForm] -> s, Infinity];
		Cell[CellGroupData[{
			Cell[title, "CitationTitle"],
			Cell[citation, "CitationText"]
		}]]
	]& /@ citationData;

	dprint["Export.m: citationCells 1 = "<>ToString[citationCells, InputForm]];

	citationCells = citationCells /.
		Cell[s:_String /; StringContainsQ[s, "note="], "CitationText", rest:___]
		:> Cell[StringReplace[s, ", note=" ~~ __ ~~ "}" -> ", note=[CITATIONTIMESTAMP]}"], "CitationTextJS", rest];

	dprint["Export.m: citationCells 2 = "<>ToString[citationCells, InputForm]];

	expr = ReplaceAll[expr, {
		Cell[CellGroupData[{Cell[_, "CitationContainerCell", ___], ___}, ___], ___]
			-> Cell[CellGroupData[Flatten[{Cell["", "CitationGroup"], citationCells}]]],
		Cell[_, "CitationContainerCell"]
			-> Cell[CellGroupData[Flatten[{Cell["", "CitationGroup"], citationCells}]]]
	}];

	DocsAssert[FreeQ[expr, Cell[_, "CitationContainerCell", ___]]];

	dprint["Export.m: final citation cell: "]; dfile[expr];

	(* ***** SeeAlsoInterpreter and friends *)

	(* the complete in-product lineup as of March 15 2019:

	hits = Cases[expr, Cell[text:_, style:_String /; StringContainsQ[style, StartOfString~~"SeeAlso"~~__], ___] -> Cell[text, style], Infinity];
	hits = DeleteCases[hits, Cell[_, "SeeAlsoSection", ___], Infinity];
	If[Positive@Length@Flatten@hits, Sow[nb, hits]]

	Cell["callback:  ", "SeeAlsoCallback"]
	Cell["Related C Functions:  ", "SeeAlsoC"]
	Cell["Related Classifiers:  ", "SeeAlsoClassifier"]
	Cell["Related Compiled Types:  ", "SeeAlsoCompiledtype"]
	Cell["Related Entity Types:  ", "SeeAlsoEntity"]
	Cell["Related External Evaluation Systems:  ", "SeeAlsoExternalevaluationsystem"]
	Cell["Related Formats:  ", "SeeAlsoFormat"]
	Cell["Related Front End Tokens:  ", "SeeAlsoFrontendobject"]
	Cell["Related Indicator Types:  ", "SeeAlsoIndicator"]
	Cell["Related Interpreter Types:  ", "SeeAlsoInterpreter"]
	Cell["Related Machine Learning Methods:  ", "SeeAlsoMethod"]
	Cell["Related Menu Items:  ", "SeeAlsoMenuitem"]
	Cell["Related Net Decoders:  ", "SeeAlsoNetdecoder"]
	Cell["Related Net Encoders:  ", "SeeAlsoNetencoder"]
	Cell["Related Programs:  ", "SeeAlsoProgram"]
	Cell["Related Service Connections:  ", "SeeAlsoService"]
	Cell["Related System Files:  ", "SeeAlsoFile"] *)

	expr = expr /.
		{
			Cell[
				headtitle:_,
				(* Formatting note: d1434656 and ffafdbdd:
					https://jira.wolfram.com/jira/browse/WLDOCS-2285
					https://jira.wolfram.com/jira/browse/WLDOCS-2137
				*)
				style:
				"SeeAlsoCallback"
				| "SeeAlsoC"
				| "SeeAlsoClassifier"
				| "SeeAlsoCompiledtype"
				| "SeeAlsoEntity"
				| "SeeAlsoExternalevaluationsystem"
				| "SeeAlsoFormat"
				| "SeeAlsoFrontendobject"
				| "SeeAlsoIndicator"
				| "SeeAlsoInterpreter"
				| "SeeAlsoMethod"
				| "SeeAlsoMenuitem"
				| "SeeAlsoNetdecoder"
				| "SeeAlsoNetencoder"
				| "SeeAlsoProgram"
				| "SeeAlsoService"
				| "SeeAlsoFile"
				, headrest:___
			],
			links:__
		} :> Flatten @ {
			Cell[
				headtitle,
				style,
				headrest
			],
			{links} /.
				TemplateBox[first:_, "StringTypeLink", rest:___]
				:> TemplateBox[first, "StringTypeSublink", rest]
		};

	(* **** miscellaneous cleanup *)
	(* ***** WLDOCS-2034: in-product link to a palette should point to a webpage *)

	expr = expr /.
		Rule[ButtonData, "paclet:ref/Chart Elements Palette »"]
		-> Rule[ButtonData, "paclet:workflow/CustomizeCharts"];

	(* ***** WLDOCS-2575: [NonBreakingSpace][RightGuillemet] -> <space>[RightGuillemet] *)

	(* Formatting note: 3f017c74:
		NonBreakingSpace doesn't work out right on the web; substitute a regular
		space character. *)
	expr = expr /.
		TextData[s_String /; StringContainsQ[s, "\[NonBreakingSpace]\[RightGuillemet]"]]
		:> TextData[StringReplace[s, "\[NonBreakingSpace]\[RightGuillemet]" -> " \[RightGuillemet]"]];

	expr = expr /.
		ButtonBox[s_String /; StringContainsQ[s, "\[NonBreakingSpace]\[RightGuillemet]"], rest:___]
		:> ButtonBox[StringReplace[s, "\[NonBreakingSpace]\[RightGuillemet]" -> " \[RightGuillemet]"], rest];

	(* ***** WLDOCS-1399L RightSkeleton and >> to RightGuillemet in ObsolescenceNote *)

	(* Formatting note: a5849aef:
		There are four kinds of ObsolescenceNote - ones that use:
		- \[RightSkeleton] as a final link
		- >> as a final link
		- words as a final link
		- no final link

		All that use >> or \[RightSkeleton] must instead use \[RightGuillemet].
	*)
	expr = expr /.
		Cell[
			first:__,
			"ObsolescenceNote",
			rest:__
		] :> Cell[
			first /. ButtonBox[
				s_String /; StringMatchQ[s, {"\[RightSkeleton]", ">>"}],
				buttonBoxRest:___
			] -> ButtonBox["\[RightGuillemet]", buttonBoxRest],
			"ObsolescenceNote",
			rest
		];

	(* ***** batchcomputationproviders section heads *)

	dprint["Export.m: entering batchcomputationprovider, expr = "];dfile[expr];
	dprint["entityType = "<>ToString[entityType]];

	(* Formatting note: ec5aa603: https://jira.wolfram.com/jira/browse/WLDOCS-3403
		ensure that batchcomputation section heads work the same way as those in ref/service pages
	*)
	If[entityType == "Remote Batch Computation Provider",
		expr = ReplaceAll[
			expr,
			Cell[some___, "Notes", rest___] -> Cell[some, "ServiceNotes", rest]
		];
		expr = ReplaceAll[
			expr,
			Cell[some___, "BatchComputationProviderSection", rest___] -> Cell[some, "ServiceSubsection", rest]
		];
	];

	dprint["Export.m: leaving batchcomputationprovider, expr = "];dfile[expr];

	(* **** top-of-page items *)
	(* ***** clean up any "New in", "Updated in", "ExperimentalBanner" items *)

	(* Formatting note: 8a042766: *)
	(* I've added a step at the beginning of processing to extract this info from
	the expr.  I then retained those structures through all the munging in case
	there are patterns that depend on them.  Now that we're done with them, remove
	them so they're not inadvertently turned into images by Transmogrify *)

	expr = expr /.
		Cell[
			BoxData[TooltipBox[StyleBox[TagBox[GraphicsBox[InsetBox[StyleBox[
				s_String /; StringContainsQ[s, {"NEW IN", "\:306e\:65b0\:6a5f\:80fd", "\:7684\:65b0\:529f\:80fd"}],
			___], ___], ___], ___], ___], ___], ___],
			___
		] :> Unevaluated[Sequence[]];

	expr = expr /.
		Cell[
			BoxData[GraphicsBox[InsetBox[StyleBox[
				s_String /; StringContainsQ[s, {"UPDATED IN", "\:3067\:66f4\:65b0", "\:7684\:66f4\:65b0"}],
			___], ___], ___], ___],
			___
		] :> Unevaluated[Sequence[]];

	(* ExperimentalBanner becomes UnfinishedDraft somewhere up there *)
	expr = expr /. Cell[_, "UnfinishedDraft", ___] :> Unevaluated[Sequence[]];

	(* ***** remove unneeded spacer cells from the ObjectName expression *)

	(* Formatting note: d9cf3998: *)
	expr = expr /.
		{
			a:___,
			b:Cell[_, s1_String /; StringContainsQ[s1, "ObjectName"], ___],
			c:Cell[BoxData[TemplateBox[_, s2_String /; StringContainsQ[s2, "Spacer"], ___], ___], ___],
			d:___
		} :> {a, b, d};

	(* ***** munging for ref/character pages *)
	(* ******* WLDOCS-2000: character name titles have disappeared *)

	(* https://jira.wolfram.com/jira/browse/WLDOCS-2000 *)
	(* Formatting note: 82e5955a *)

	expr = expr /.
		{
			first:___,
			c1:Cell[_, "CharacterName", ___],
			c2:Cell[_, "CharacterImage", ___],
			rest:___
		} :> {
			first,
			Cell[CellGroupData[{c1}, Open]],
			c2,
			rest
		};

	(* ***** special handling for quotation marks in titles *)
	(* author: billw 2018-08-25 *)
	(* ticket: https://jira.wolfram.com/jira/browse/WLDOCS-2023 *)
	(* Formatting note: 0b72dca5 *)

	(* I found this while implementing Entity pages and comparing my html to suef's
	demo page.  Consider the quotation marks in a title:

		- https://reference.wolfram.com/language/ref/classifier/Language
		- https://reference.wolfram.com/language/ref/device/Arduino
		- https://reference.wolfram.com/language/ref/service/ArXiv

	Each quotation mark needs to be transmogrified into its own <span>:

		<h1 class="main-title">&quot;Building&quot;</h1>

	One way to do it: put each one into a distinctively-named cell, then process
	that cell with a transform defined in one of the globally-used transforms -
	probably Defaults.m.

	The list of cell styles is taken from the transform in Functions.m.  Others can
	be added as needed.

	"objectNameQuote" avoids possible name clashes with future OfficialCellStyles.

	Note: this only handles the case of <quote>WordCharacters<quote>.  Are there other patterns in the current docs?  *)

	(*
	expr = expr /. Module[
		{
			styles = {"ObjectName", "ObjectNameSmall", "ObjectNameAlt", "CFunctionName"},
			cellPattern = Cell[s:_String /; StringMatchQ[s, "\""~~WordCharacter..~~"\""], style:Alternatives @@ styles, rest:___],
			first, last
		},

		TextData[cellPattern] :>
			TextData[{
				Cell["", "objectNameQuote"],
				Cell[StringReplace[s, "\"" -> ""], style, rest],
				Cell["", "objectNameQuote"]
			}],

		TextData[{first:___, cellPattern, last:___}] :>
			TextData[{
				first,
				Cell["", "objectNameQuote"],
				Cell[StringReplace[s, "\"" -> ""], style, rest],
				Cell["", "objectNameQuote"],
				last
			}]
	]

	expr = expr /.
		TextData[Cell[a:_String, style:Alternatives @@ styles, b:___]] :>
		TextData[{}]


	TextData[{a, b}]
	*)

	(* ***** move the "Related Interpreter" cell and simplify the resulting EntityAbstract cell *)
	(* Formatting note: d9b9a8a4 *)

	(* Before the Related Interpreter link was introduced in WLDOCS-2144, a
	ref/entity page's abstract text was simply a cell in a list of other cells.
	When Brian implemented the new link he associated it horizontally with the
	adjacent EntityAbstract cell and wrapped the resulting assembly in a GridBox.

	When Sue designed the html markup for the Related Interpreter link, she
	associated the link vertically with the adjacent ObjectNameGrid.  These things
	happen.

	Here, I will:
		- extract the Related Interpreter cell from the new abstract GridBox,
		- reduce the abstract back down to a single cell, and
		- store the Related Interpreter cell as an ObjectName option to be unpacked
		when that cell is transformed *)

	If[Positive @ Count[expr, Cell[_, "EntityRelatedInterpreter", ___], Infinity],

		Module[{riCell},
			(* save the Related Interpreter cell *)
			riCell = Quiet @ Check[
				First @ Flatten @ Cases[
					expr
					, Cell[_, "EntityRelatedInterpreter", ___]
					, Infinity
				],
				{}
			];

			(* remove the Related Interpreter cell and reduce the abstract to its previous simple form *)
			expr = expr /.
				Cell[
					BoxData[GridBox[{{
						abstract:Cell[_, "EntityAbstract", ___],
						_ItemBox,
						Cell[_, "EntityRelatedInterpreter", ___]
					}}]]
					, "EntityAbstract"
					, ___
				] :> abstract;

			(* have the ObjectName cell carry the Related Interpreter cell along to the transform party *)
			expr = expr /.
				Cell[first:_, style:"ObjectName"|"ObjectNameSmall"|"ObjectNameAlt", rest:___]
				-> Cell[first, style, Sequence[rest, "RelatedInterpreterCell" -> riCell]];
		]
	];

	(* ***** DataPageEntityNote and PlatformAvailabilityNote banners *)
	(* Formatting note: 5c7e1050 *)

	(* https://jira.wolfram.com/jira/browse/WLDOCS-2331 - just extract the relevant
	TextData from this thing and send it off to a simple transform to produce some
	html with the proper div class.  Then the associated icon can be handled by css,
	enabling instant global changes when needed. *)

	dprint["expr entering DataPageEntityNote and PlatformAvailabilityNote processing:"]; dfile[expr];

	Module[{pos},
		(* find the large outer DataPageEntityNote cell *)
		pos = Position[
			expr,
			Cell[BoxData[_GridBox], "DataPageEntityNote", ___],
			Infinity
		];
		If[pos =!= {},
			(* replace it with the small inner DataPageEntityNote cell *)
			expr = ReplacePart[
				expr,
				pos -> First[Cases[
					Extract[expr, pos],
					Cell[textdata:_TextData, "DataPageEntityNote", ___] -> Cell[textdata, "DataPageEntityNote"],
					Infinity
				]]
			]
		]
	];

	(* DataPageEntityNote and PlatformAvailabilityNote are similar ideas with
	different m- structures.  this is the same kind of simplification for web
	conversion so fungible assets can be handled via css.  *)

	(* Formatting note: 23378c0f: Update PlatformAvailabilityNote munging *)
	expr = expr /.
		Cell[
			TextData[{Cell[BoxData[_TagBox|GraphicsBox]], contents:___}],
			"PlatformAvailabilityNote",
			___
		] -> Cell[
			TextData[{contents}],
			"PlatformAvailabilityNote"
		];

	dprint["expr leaving DataPageEntityNote and PlatformAvailabilityNote processing:"]; dfile[expr];

	(* ***** ServiceSubtitle and friends (paren text after page title) *)

	(* Formatting note: 912d1406:
		Translate (Data Structure) subtitle
		https://bugs.wolfram.com/show?number=386710
		- ServiceSubtitle is used as a default container for these things
		in-product.  Differentiate them here to introduce a localized phrase.
	*)
	If[entityType == "Data Structure",
		expr = expr /. {
			Cell[_, "ServiceSubtitle"]
			:> Cell[
				Localized["DataStructureSubtitle", language],
				"DataStructureSubtitle"
			]
		}
	];

	(* ***** entityType Interpreter: InterpreterUsage element *)

	dprint["Export.m: entering entityType Interpreter: InterpreterUsage element, expr = "];dfile[expr];

	(* Formatting note: e9b9fbeba422d5963cafdaf6c2dd429e1c8c0d88:
		Fix Related Entities badge on interpreter pages
		https://jira.wolfram.com/jira/browse/WLDOCS-2813
	*)
	If[entityType == "Interpreter",
		interpreterUsage = Cases[
			expr,
			Cell[s:_String, "InterpreterUsage", ___],
			Infinity
		];
		If[Positive @ Length @ interpreterUsage,
			expr = expr /.
				Cell[_BoxData, "InterpreterUsage", ___]
				:> First[interpreterUsage]
		]
	];

	dprint["Export.m: leaving entityType Interpreter: InterpreterUsage element, expr = "];dfile[expr];

	(* ***** remove in-product UPDATED/show changes image *)

	(* Formatting note: 132184d2fa88b6811033de791512e76927df493c:
		Remove in-product UPDATED banner in favor of the textual banner placed by css and the
		iconography div.
		https://jira.wolfram.com/jira/browse/WLDOCS-2911
	*)
	expr = expr /. Cell[BoxData[TagBox[TooltipBox[_TogglerBox, ___], ___], ___], ___] -> Unevaluated[Sequence[]];

	(* ***** root guide page *)
	(* Formatting note: a4372010 *)

	dprint["Export.m: entering final root guide processing; expr = "]; dfile[expr];

	(* If[entityType === "Root Guide", *)
	(*   expr = Module[{actionMenuBoxes, res}, *)
	(*     actionMenuBoxes = Cases[expr, Cell[CellGroupData[{_ActionMenuBox, ___}, ___], ___], Infinity]; *)

	(*     res = Module[{action, contents, list1, list2}, *)
	(*       action = #; *)

	(*       (\* gather what we'll use *\) *)
	(*       contents = Cases[action, _[_, "RootSectionTitle"|"RootItem"|"RootItemDelimiter", ___], Infinity]; *)
	(*       dprint["contents step 1: contents = "]; dfile[contents]; *)

	(*       (\* remove dups *\) *)
	(*       contents = DeleteDuplicates[contents, Part[#1, 2] == "RootSectionTitle" && Part[#2, 2] == "RootSectionTitle"&]; *)
	(*       dprint["contents step 2: contents = "]; dfile[contents]; *)

	(*       (\* clean extraneous stuff out of the RootSectionTitle cell and clean up strings *\) *)
	(*       contents = contents /. StyleBox[a:_, "RootSectionTitle", ___] -> StyleBox[a, "RootSectionTitle"]; *)
	(*       contents[[1, 1]] = StringReplace[contents[[1, 1]], {"\\n" -> " ", "\"" -> ""}]; *)
	(*       dprint["contents step 3: contents = "]; dfile[contents]; *)

	(*       (\* group RootItems into lists and remove delimiters *\) *)
	(*       contents = SplitBy[contents, Part[#, 2] == "RootItem" &]; *)
	(*       dprint["contents step 4: contents = "]; dfile[contents]; *)

	(*       contents = contents /. {{s:_StyleBox}, rest:__} -> {Cell[CellGroupData[{s, rest}]]}; *)
	(*       dprint["contents step 5: contents = "]; dfile[contents]; *)

	(*       (\* put groups into cellgroups for transform convenience *\) *)
	(*       contents = contents /. { *)
	(*         list:{Cell[_, "RootItem"|"RootItemDelimiter", ___], ___} -> Cell[CellGroupData[Flatten[list]]] *)
	(*       }; *)
	(*       dprint["contents step 6: contents = "]; dfile[contents]; *)

	(*       contents = contents //. StyleBox[args:__] -> Cell[args]; *)
	(*       dprint["final contents: "];dfile[contents]; *)

	(*       contents *)
	(*     ]& /@ actionMenuBoxes; *)

	(*     dprint["root guide: res = "];dfile[res]; *)

	(*     Notebook @@ {Flatten[res]} *)
	(*   ] *)
	(* ]; *)

	dprint["Export.m: leaving final root guide processing; expr = "]; dfile[expr];

	(* *** end of the munging - do final ExportWebPage activities and call Transmogrify on the munged expression *)

	(*********************************************************************
	here endeth our last-chance expression munging before running
	Transmogrify
	*********************************************************************)

	(* Formatting note: 463e4429
		Fix monographs: https://jira.wolfram.com/jira/browse/WLDOCS-2221

		- New kludgy flag monographObjectNameGridQ - when a monograph has an
		ObjectNameGrid (some do, some don't), GridBox.m inserts a
		main-heading div for the Grid cell, then Tutorials.m inserts one for
		the monograph title cell.  When processing one of these monographs,
		allow only one of them to insert the div.
	*)
	monographObjectNameGridQ = And[
		entityType == "Monograph",
		Positive@Count[expr, Cell[_, "ObjectNameGrid", ___], Infinity]
	];
	dprint["Export.m: monographObjectNameGridQ = "<>ToString[monographObjectNameGridQ, InputForm]];

	(* Print["Export.m, post-munging point 1"]; *)

    featuredexampledata = If[entityType === "Guide",
		Cases[
			Cases[
				Cases[expr, TaggingRules~(Rule | RuleDelayed)~_],
				"ExamplePageLinks"~(Rule | RuleDelayed)~a_ :> a,
				Infinity
			],
			{_String, _String, _String},
			Infinity
		],
		{}
	];

    BuildLog["   Assemble property list"];

     (* If[entityType === "Guide",
      (* capture Tutorial function list *)
      guideFunctionList = GetGuideFunctionList[expr];
      guideMoreAboutList = GetGuideMoreAboutSectionList[expr];
      guideTutorialsList = GetGuideTutorialsSectionList[expr];
    ];*)

    (* If[entityType === "Tutorial",
      (* capture Tutorial function list *)
      tutorialFunctionList = GetTutorialFunctionList[expr];
      tutorialMoreAboutList = GetTutorialMoreAboutSectionList[expr];
      relatedTutorialsList = GetRelatedTutorialsSectionList[expr];
    ];*)

    (*seeAlsoFunctionList = GetSeeAlsoSectionList[expr];
    moreAboutGuideList = GetMoreAboutSectionList[expr];
    exampleSectionList = {};
    tutorialList = GetTutorialsSectionList[expr];*)

    exampleSectionList = {};

	(* Print["Export.m, post-munging point 2"]; *)

    expr = DeleteCases[expr, Rule["LinkTrails", ___], Infinity];

(* Print["Export.m, post-munging point 3"]; *)

    (* determine relative location of css/js to html page *)
	relativeBase = Replace["WebAssetsPath" /. {opts} /. Options[ExportWebPage], {
		(* this changes depending on page (e.g ref/Sin v. ref/messages/func/tag) *)
		Automatic :> StringRepeat["../", StringCount[uri, "/"]-1],
		value_?StringQ :> value,
		err_ :> DocsError["Invalid value for \"WebAssetsPath\": ``", InputForm[err]]
	}];

	If[completeHTMLQ===True,
		htmlBase = FileNameJoin[{relativeBase, "images"}];
		,
		htmlBase = "";
	];

	htmlTemplateIncludes = "HTMLTemplateIncludes" /. {opts} /. Options[ExportWebPage];

	DocsAssert[
		AssociationQ[htmlTemplateIncludes],
		"bad value for \"HTMLTemplateIncludes\" ExportWebPage option: ``",
		htmlTemplateIncludes
	];

	(* Print["Export.m, post-munging point 4"]; *)
	(* values of these are obtained via GetParameter; dunno why the're called properties here *)
	properties = Join[
		DeleteCases[
			{
				"webmQ" -> webmQ,
				"lwgQ" -> lwgQ,
				"wsmQ" -> wsmQ,
				"packageQ" -> packageQ,
				"monographObjectNameGridQ" -> monographObjectNameGridQ,
				"layoutfile"->layoutfile,
				"title" -> pageTitle,
				"titlemodifier"->pageTitleModifier,
				"contextName" -> contextName,
				"taggingRules" -> taggingRules,
				"buildmode" -> buildmode,
				"url" -> uri,
				"keywords" -> keywordList,
				"flag" -> flag,
				"language" -> language,
				"la" -> la,
				"languageAbbreviations" -> DocumentationBuild`Utils`languageList[[All, 2]],
				"seeAlsoFunctions" -> seeAlsoFunctionList,
				(* "rootCaptions" -> rootCaptions, *)
				"moreAboutGuides" -> moreAboutGuideList,
				"pacletName" -> pacletName,
				"exampleSections" -> exampleSectionList,
				"tutorials" -> tutorialList,
				"urlBaseRemote" -> urlBaseRemote,
				"entityType" -> entityType,
				"specialKeywords" -> specialKeywords,
				"tutorialFunctions" -> tutorialFunctionList,
				"synonyms" -> synonymList,
				"guideFunctions" -> guideFunctionList,
				"tutorialMoreAbout" -> tutorialMoreAboutList,
				"CompleteHTMLQ" -> completeHTMLQ,
				"workflowQ" -> workflowQ,
				"workflowOverviewQ" -> workflowOverviewQ,
				"guideMoreAbout" -> guideMoreAboutList,
				"BuildNumber" -> buildNumber,
				"HistoryData"->({"New", "Modified", "Obsolete", "Excised"} /. history),
				"newin" -> newin,
				"updatedin" -> updatedin,
				"experimentalQ" -> experimentalQ,
				"IsManipulate" -> isManipulate,
				"Cache" -> cache,
				"CacheLocation" -> cachelocation,
				"htmlBase" -> htmlBase,
				"relativeBase" -> relativeBase,
				"HTMLTemplateIncludes" -> htmlTemplateIncludes,
				"summary" -> summaryText,
				"relatedTutorials" -> relatedTutorialsList,
				"guideTutorials" -> guideTutorialsList,
				"featuredexampledata"->featuredexampledata,
				"IsTMJQ"->isTMJQ,
				"IsSolutionsQ"->isSolutionsQ,
        "prerequisite" -> prerequisite
			},
			(Alternatives @@ First /@ params) -> _,
			Infinity
		],
		params
	];

	(* dprint["properties = "<>ToString[properties, InputForm]]; *)

	(* Print["Export.m, post-munging point 5"]; *)
    (* if we just want to see how the conversions _would_ happen,
       just go ahead and do it before notebook creation and manipulation.
       this allows us to do tests like mogtest[sometroublingExpression]
       to debug the XMLTransform/Transmogrify[] *)
    If[ TrueQ[test] || outfile===None,
      Return[If[ test,
        TransmogrifyString[##],
        Transmogrify[##]
      ]&[expr, transform, DefaultParameters->properties, opts]]
    ];

	(* Print["Export.m, post-munging point 6"]; *)
    (* Change WindowSize to a width for nicer line wrapping in gifs *)
    (* if windowsize with is changed here, also change it in the local variables in Transmogrify.m:transmogrifyImage *)
    (* https://jira.wolfram.com/jira/browse/WLDOCS-1023 - 550 is the largest WindowSize that still forces a linebreak in NestList Out[2] *)
    expr = expr /. Rule[WindowSize, {_, _}] :> WindowSize -> {550, 700};

    BuildLog["   Send expr off to Transmogrify"];

    CacheStylesheet[nb];

	(* Print["Export.m, post-munging point 7"]; *)
    (* now call Transmogrify with the appropriate parameters set.
       note that language extension by default goes before the .html
       update 20170108: on our new servers we're now using .html.la, where la is the current language's 2-letter abbreviation *)

    (* if outfile is not already in our desired final format AND we're not processing webm or lwg, whose files must have a .html extension... *)
	(* Formatting note: 90634f12 and 60d6e8ae
		Handle webMathematica links and filenames

		ExportWebPage: test whether we're building a webm file and if so:

		1. use a special url base that provides correct addresses within
			webm's installation in tomcat/apache.

		2. trust that ExportWebPage has been sent the correct filename ending
			in .html and don't append .en to it.  webm's tomcat/apache doesn't
			handle language negotiation out of the box and cannot find
			.html.en files unless the entire filename is given explicitly.

		Detect links to paclet:webMathematica/ and manipulate their targets to
		produce correct paths for webm docs in tomcat/apache.


		Updates for LightweightGridManager docs

		https://jira.wolfram.com/jira/browse/WLDOCS-3229
		https://bugs.wolfram.com/show?number=395458

		- send webmQ and lwgQ off to Defaults.m as parameters retrievable with
		GetParameter so we can have a standard authoritative test for those docs
		throughout DocumentationBuild and avoid the temptation to ad-hockery.
	*)
	If[
		Not[webmQ]
		&& Not[lwgQ]
		&& Not[StringEndsQ[outfile, ".html." ~~ DocumentationBuild`Utils`languageList[[All, 2]]]]
		,
		(* ...then if the file name is not .la.html or is .tm.html, insert a .la after the .html *)
		(* Print["Export.m:outfile 1 = "<>ToString[outfile]]; *)

		If[completeHTMLQ === True,
			langext = "";
			,
			langext = "." <> la;
		];

		(* retain this odd test from the original code *)
		(* Read as:
			If `outfile` doesn't end with '.XX.html', or it ends with '.tm.html', add the
			two letter language extension. E.g. if the file is 'file.tm.html', rename it
			to 'file.tm.en.html'. *)
		outfile = StringReplace[
			outfile,
			RegularExpression["(?(?<!\\.\\w\\w)|(?<=\\.tm))\\.html"] -> langext <> ".html"
		];

		(* Print["Export.m:outfile 2 = "<>ToString[outfile]]; *)
		(* if the file name is .la.html, change it to .html.la *)
		outfile = StringReplace[
			outfile,
			"." ~~ la:DocumentationBuild`Utils`languageList[[All, 2]] ~~ ".html" ~~ EndOfString
				-> ".html."<>la
		]
	];

	dprint["final expression sent to Transmogrify:"]; dfile[expr]; dfile[expr, "Format" -> ".nb"];
	dprint["transform = "<>ToString[transform]];

	dprint["Converting final notebook expression to HTML via Transmogrify..."];
	dprint["'outfile' exiting Transmog = "<>ToString[outfile, InputForm]];

	(* upon successful transmogrification, this return value r is the name of the output file as a string *)
  dprint["Transmogrify called with arguments:"];
  dprint["  outfile = "<>ToString[outfile, InputForm]];
  dprint["  expr with Head = "<>ToString[Head[expr], InputForm]];
  dprint["  transform = "<>ToString[transform, InputForm]];
  dprint["  {properties} = "<>ToString[{properties}, InputForm]];
  dprint["  {opts} = "<>ToString[{Sequence @@ DeleteCases[{opts}, DefaultParameters ~(Rule|RuleDelayed)~ _]}, InputForm]];
	r = PubsEvaluateWithFE @ Transmogrify[
		outfile,
		expr,
		transform,
		DefaultParameters -> properties,
		Sequence @@ DeleteCases[{opts}, DefaultParameters ~(Rule|RuleDelayed)~ _]
	];

	(* Print["Done."]; *)

	(* Print["Export.m, post-munging point 9"]; *)
    BuildLog["** Leaving ExportWebPage **"];

    If[Head[r] =!= String,
		First[r],
		r
	]

,
	(* fail expression for Check *)

	errMessageList = DocumentationBuild`Utils`Private`formatMessageList[$MessageList];
	BuildLog[StringJoin["   MESSAGES: ", errMessageList]];

	If[Head[r] =!= String,
		First[r],
		r
	]
]
]
(* End: ExportWebPage *)

(******************************************************************************)

DocLinkURL[
	(* The paclet / document the link is originating from. *)
	{pacletName_, uri_},
	(* The link destination:
		* another document in the current paclet
		* a document from another paclet
		* a web address ("http://...")
	*)
	destination0_,
	urlBaseRemote_,
	completeHTMLQ_
] := Module[{
	destination = destination0,
	uriDepth,
	isAWebMathematicaLink,
	isALightweightGridLink
},
	DocsAssert[StringQ[pacletName]];
	DocsAssert[StringQ[uri]];
	DocsAssert[StringQ[destination]];
	DocsAssert[StringQ[urlBaseRemote]];
	DocsAssert[BooleanQ[completeHTMLQ]];

	If[StringStartsQ[destination, "paclet:http"],
		(* FIXME:
			This is caused by code in MakeActionMenu which prepends "paclet:" to links
			which don't already start with "paclet:". That is not correct for web links.
			Note: The webMathematica documentation has links which exercise this code.
			      I think any links in the "Related Links" documentation section have this
				  problem.
		*)
		DocsWarning[
			"warning processing ``: fixing up \"paclet:http*\" link destination: ``",
			uri,
			destination
		];
		destination = StringReplace[destination, StartOfString ~~ "paclet:" -> ""];
	];

	If[StringContainsQ[destination, "paclet://"],
		(* Note:
			"paclet://" style links (as opposed to just "paclet:") appear to be a
			historical alternative representation of documentation links. A grep within
			the CVS system documentation folder returns ~a dozen remaining instances of
			links with this style (with possibly more instances in paclet documentation).
			If any of these links do happen to exist, we issue a warning about them and
			transform them into the modern "paclet:" form. We need to do this so that the
			`uriDepth` value computed below is correct. *)
		DocsWarning[
			"warning processing ``: invalid \"paclet://*\" link destination: ``",
			uri,
			destination
		];

		destination = StringReplace[destination, "paclet://" -> "paclet:"];
	];

	If[StringMatchQ[destination, "" | "paclet:"],
		DocsWarning[
			"warning processing ``: link destination is empty: ``",
			uri,
			InputForm[destination]
		];
		Return["", Module];
	];

	(* if we're given an http link, trust it to be correct and not need ".html" added to it *)
	If[StringStartsQ[destination, "http" | "https"],
		Return[destination, Module];
	];

	(*-------*)
	(* Setup *)
	(*-------*)

	(* Formatting note: 90634f12
		Detect links to paclet:webMathematica/ and manipulate their targets to
		produce correct paths for webm docs in tomcat/apache. *)
	isAWebMathematicaLink = TrueQ[StringContainsQ[destination, "paclet:webMathematica/"]];

	(* Formatting note: 91dc5304
		Like webMathematica, LightweightGridManager's documentation is intended
		for viewing in-product in a tomcat installation rather than on the web
		at ref.w.c.

		Define and use isALightweightGridLink similar to isAWebMathematicaLink and use
		it to remove an unneeded part of the lwg paclet uri.
	*)
	isALightweightGridLink = TrueQ[StringContainsQ[destination, "paclet:LightweightGridManager/"]];

	(* Validate that no links into webMathematica or LightweightGridManager exist from
		outside those paclets. If such links exist, the logic below which removes the
		"webMathematica/" or "LightweightGridMananger/" prefix from the link
		destination will make those links point at the wrong place. *)

	If[isAWebMathematicaLink && pacletName =!= "webMathematica",
		DocsError[
			"link to a webMathematica documentation page (``) from outside the"
			<> " webMathematica paclet is not supported: ``",
			InputForm[destination],
			InputForm[pacletName]
		]
	];

	If[isALightweightGridLink && pacletName =!= "LightweightGridManager",
		DocsError[
			"link to a LightweightGridManager documentation page (``) from outside the"
			<> " LightweightGridManager paclet is not supported: ``",
			InputForm[destination],
			InputForm[pacletName]
		]
	];

	(*------------*)
	(* Link logic *)
	(*------------*)

	If[pacletName === "Alpha",
		destination = StringReplace[destination, "paclet:" -> ""];

		destination = StringReplace[
			destination,
			StartOfString ~~ "Alpha/" -> ""
		];

		Return[DocumentationBuild`Utils`DotHTML["/documentation/" <> destination], Module];
	];

	Which[
		completeHTMLQ && StringContainsQ[destination, pacletName],
			(* If we're building standalone (complete) documentation, and
				`link` is a link into a page within the same paclet, use a
				relative link. *)
			uriDepth = Length[StringSplit[uri, "/"]];

			DocumentationBuild`Utils`DotHTML @ StringReplace[
				destination,
				{
					pacletName <> "/" -> "",
					"paclet:" -> StringRepeat["../", uriDepth - 2]
				}
			],

		True,
			If[isAWebMathematicaLink || isALightweightGridLink,
				destination = StringReplace[destination, pacletName <> "/" -> ""];
			];

			DocumentationBuild`Utils`DotHTML[
				DocumentationBuild`Internal`UpdateLinkDestination @ StringReplace[
					destination,
					"paclet:" -> urlBaseRemote
				]
			]
	]
]

(******************************************************************************)
(** ExportNavigatorWebPage **)

Options[ExportNavigatorWebPage] := {
  "BaseURL" -> "http://reference.wolfram.com/language/",
  "ImageBaseURL" -> "http://reference.devel.wolfram.com/mathematicaImages/"
};

ExportNavigatorWebPage[outfile_String, nb_String, opts___?OptionQ] :=
  ExportNavigatorWebPage[outfile, Import[nb], opts];

ExportNavigatorWebPage[outfile_String, nb_Notebook, opts___?OptionQ] :=
  Module[{base, imgbase}, Block[{groupid = 0},
    base = "BaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    imgbase = "ImageBaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    Export[outfile,
      XMLElement["html", {}, {
        XMLElement["head", {}, {
          XMLElement["link", {
            "rel" -> "stylesheet", "type" -> "text/css",
            "href" -> "/css/Navigator.css"
          }, {}],
          XMLElement["script", {
            "type" -> "text/javascript",
            "language" -> "javascript",
            "src" -> "/javascript/jquery.js"
          }, {" "}],
          XMLElement["script", {
            "type" -> "text/javascript",
            "language" -> "javascript",
            "src" -> "/javascript/Navigator.js"
          }, {" "}]
        }],
        XMLElement["body", {}, {
          XMLElement["div", {"id" -> "TabBar"}, {
            XMLElement["table", {}, {XMLElement["tr", {}, {
              XMLElement["td", {}, {
                XMLElement["div",
                  {"id" -> "FunctionNavigatorTab", "class" -> "Tab TabActive"},
                  {
                    XMLElement["a", {
                        "id" -> "FunctionNavigatorLink",
                        "href" -> "javascript:nav_tab(\"FunctionNavigator\")"
                      },
                      {"Function Navigator"}]
              }] }],
              XMLElement["td", {}, {
                XMLElement["div",
                  {"id" -> "VirtualBookTab", "class" -> "Tab"},
                  {
                    XMLElement["a", {
                        "id" -> "VirtualBookLink",
                        "href" -> "javascript:nav_tab(\"VirtualBook\")"
                      },
                      {"Virtual Book"}]
              }] }]
            }] }],
            XMLElement["div", {"id" -> "TabBottom"}, {
              XML`RawXML["&nbsp;"]
            }]
          }],
          XMLElement["div", {"id" -> "FunctionNavigator"},
            Map[ExportNavigatorWebPage[outfile, #, opts]&,
              Cases[nb,
                Cell[CellGroupData[{Cell[_, "Level 1", ___], ___}, ___], ___],
                {2}]
          ]],
          XMLElement["div", {"id" -> "VirtualBook"},
            Map[ExportNavigatorWebPage[outfile, #, opts]&,
              Cases[nb,
                Cell[CellGroupData[{Cell[_, "TutorialLevel 1", ___], ___}, ___], ___],
                {2}]
          ]]
        }]
      }],
    "XML"];
  ]];

ExportNavigatorWebPage[
    outfile_String,
    Cell[CellGroupData[{h_Cell, r___Cell}, ___], ___],
    opts___?OptionQ] :=
  Module[{base, imgbase, bbox, str, tag, class},
    base = "BaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    imgbase = "ImageBaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    groupid++;
    bbox = First@Cases[h, _ButtonBox, Infinity, 1];
    tag = First@Cases[h, (CellTags -> t_) :> t];
    If[StringLength[tag] > 7 && StringTake[tag, 7] == "paclet:",
      tag = StringDrop[tag, 7]];
    If[tag == "guide/NewIn60AlphabeticalListing",
      str = "New In 6.0",
    (* Else *)
      str = Transmogrify`ConvertToString[bbox];
      str = StringReplace[str, {
        RegularExpression["^\\s*\"(.*)\"\\s*$"] :> "$1"
      }];
    ];
    class = StringReplace[h[[2]], {" " -> "", "Tutorial" -> ""}];
    Module[{id = groupid},
      XMLElement["div", {
          "class" -> class,
          "id" -> ToString[id]
        }, Flatten@{
          XMLElement["div", {"class" -> "Head"}, {
            XMLElement["span", {"class" -> "Dingbat"}, {
              XMLElement["a", {"href" ->
                  "javascript:nav_toggle(\"" <> ToString[id] <> "\")"}, {
                XMLElement["img", {"src" -> (imgbase <> "closedGroup.gif")}, {}]
              }]
            }],
            XMLElement["a", {"href" ->
                ("javascript:nav_open(\"" <> base <> tag <> ".html\")") },
              {str}]
          }],
          If[class == "Level1",
            (Export[
              ToFileName[{DirectoryName[outfile], "HTMLFiles"}, ToString[id] <> ".txt"],
              #, "XML"];
            ); {},
            #          
          ]&[
            XMLElement["div", {"class" -> "Content"},
              Map[ExportNavigatorWebPage[outfile, #, opts]&, {r}]]
          ]
      }]
    ]
  ];

ExportNavigatorWebPage[
    outfile_String,
    Cell[_, _String, "Delimiter", ___],
    opts___?OptionQ] :=
  XMLElement["div", {"class" -> "Delimiter"}, {
    XMLElement["hr", {"class" -> "Delimiter"}, {}]
  }];

ExportNavigatorWebPage[
    outfile_String,
    expr:Cell[_, _String, ___],
    opts___?OptionQ] :=
  Module[{base, imgbase,bbox, str, tag, class},
    base = "BaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    imgbase = "ImageBaseURL" /. {opts} /. Options[ExportNavigatorWebPage];
    bbox = First@Cases[expr, _ButtonBox, Infinity, 1];
    str = Transmogrify`ConvertToString[bbox];
    str = StringReplace[str, {
      RegularExpression["^\\s*\"(.*)\"\\s*$"] :> "$1"
    }];
    tag = First@Cases[expr, (CellTags -> t_) :> t];
    If[StringLength[tag] > 7 && StringTake[tag, 7] == "paclet:",
      tag = StringDrop[tag, 7]];
    class = StringReplace[expr[[2]], " " -> ""];
    XMLElement["div", {"class" -> class}, {
      XMLElement["span", {"class" -> "Dingbat"}, {XML`RawXML["&bull; "]}],
      XMLElement["a", {"href" ->
          ("javascript:nav_open(\"" <> base <> tag <> ".html\")") },
        {str}]
    }]
  ];

ExportNavigatorWebPage[c__] := {};


(* * "future landing page", whatever that was *)

GetFutureLandingContent[dir_String] :=
  Module[{manifest, res, outExpr1 = outExpr2 = {}, logstream, increment, count = 0},
  	
    Print["Finding the manifest"];
    Print[dir];
    manifest = FileNames["*.nb", dir, Infinity];
    increment = Floor[N[Length@manifest / 100]];

    Print["Processing the manifest, which involves finding the flag of each notebook"];
     res = Reap[Module[{nb, nbExpr, flag},
      (* poor man's progress meter *)
      count++;
      If[increment > 1 && Mod[count, increment] === 0,
        Print[ToString@count<> " of "<> ToString@Length@manifest<> ": "<>ToString[Round[100*N[count/Length@manifest]]]<>"%"]];
      (* end poor man's progress meter *)
      nb = #;
      nbExpr = Quiet@Get@nb;
      flag = DocumentFlag[nbExpr];
      Which[
        (* test1 *)
        MemberQ[{None, "AwaitingFutureDesignReviewFlag"}, flag],
        (* expr1 *)
        Which[
          Count[nbExpr, Cell[s_String /; StringMatchQ[s, ___~~"7"~~___], "HistoryData", CellTags -> "New"], Infinity] > 0,
          Switch[nb,
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Guides"}]~~__], Sow[nb, "Newin7Guides"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "HowTos"}]~~__], Sow[nb, "Newin7HowTos"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "AppleScript"}]~~__], Sow[nb, "Newin7AppleScript"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Characters"}]~~__], Sow[nb, "Newin7Characters"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "C"}]~~__], Sow[nb, "Newin7C"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Files"}]~~__], Sow[nb, "Newin7Files"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Formats"}]~~__], Sow[nb, "Newin7Formats"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "FrontEndObject"}]~~__], Sow[nb, "Newin7FrontEndObject"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "MenuItems"}]~~__], Sow[nb, "Newin7MenuItems"],
(*            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Messages"}]~~__], Sow[nb, "Newin7Messages"],*)
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Methods"}]~~__], Sow[nb, "Newin7Methods"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Programs"}]~~__], Sow[nb, "Newin7Programs"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Symbols"}]~~__], Sow[nb, "Newin7Symbols"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Tutorials"}]~~__], Sow[nb, "Newin7Tutorials"]],
          Count[nbExpr, Cell[s_String /; StringMatchQ[s, ___~~"7"~~___], "HistoryData", CellTags -> "Modified"], Infinity] > 0,
          Switch[nb,
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Guides"}]~~__], Sow[nb, "Modin7Guides"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "HowTos"}]~~__], Sow[nb, "Modin7HowTos"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "AppleScript"}]~~__], Sow[nb, "Modin7AppleScript"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Characters"}]~~__], Sow[nb, "Modin7Characters"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "C"}]~~__], Sow[nb, "Modin7C"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Files"}]~~__], Sow[nb, "Modin7Files"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Formats"}]~~__], Sow[nb, "Modin7Formats"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "FrontEndObject"}]~~__], Sow[nb, "Modin7FrontEndObject"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "MenuItems"}]~~__], Sow[nb, "Modin7MenuItems"],
(*            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Messages"}]~~__], Sow[nb, "Modin7Messages"],*)
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Methods"}]~~__], Sow[nb, "Modin7Methods"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Programs"}]~~__], Sow[nb, "Modin7Programs"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Symbols"}]~~__], Sow[nb, "Modin7Symbols"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Tutorials"}]~~__], Sow[nb, "Modin7Tutorials"]],
        True, ""];
        Which[
          Count[nbExpr, Cell[s_String /; StringMatchQ[s, ___~~"8"~~___], "HistoryData", CellTags -> "New"], Infinity] > 0,
          Switch[nb,
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Guides"}]~~__], Sow[nb, "Newin8Guides"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "HowTos"}]~~__], Sow[nb, "Newin8HowTos"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "AppleScript"}]~~__], Sow[nb, "Newin8AppleScript"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Characters"}]~~__], Sow[nb, "Newin8Characters"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "C"}]~~__], Sow[nb, "Newin8C"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Files"}]~~__], Sow[nb, "Newin8Files"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Formats"}]~~__], Sow[nb, "Newin8Formats"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "FrontEndObject"}]~~__], Sow[nb, "Newin8FrontEndObject"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "MenuItems"}]~~__], Sow[nb, "Newin8MenuItems"],
            (*s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Messages"}]~~__], Sow[nb, "Newin8Messages"],*)
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Methods"}]~~__], Sow[nb, "Newin8Methods"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Programs"}]~~__], Sow[nb, "Newin8Programs"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Symbols"}]~~__], Sow[nb, "Newin8Symbols"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Tutorials"}]~~__], Sow[nb, "Newin8Tutorials"]],
          Count[nbExpr, Cell[s_String /; StringMatchQ[s, ___~~"8"~~___], "HistoryData", CellTags -> "Modified"], Infinity] > 0,
          Switch[nb,
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Guides"}]~~__], Sow[nb, "Modin8Guides"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "HowTos"}]~~__], Sow[nb, "Modin8HowTos"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "AppleScript"}]~~__], Sow[nb, "Modin8AppleScript"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Characters"}]~~__], Sow[nb, "Modin8Characters"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "C"}]~~__], Sow[nb, "Modin8C"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Files"}]~~__], Sow[nb, "Modin8Files"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Formats"}]~~__], Sow[nb, "Modin8Formats"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "FrontEndObject"}]~~__], Sow[nb, "Modin8FrontEndObject"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "MenuItems"}]~~__], Sow[nb, "Modin8MenuItems"],
(*            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Messages"}]~~__], Sow[nb, "Modin8Messages"],*)
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Methods"}]~~__], Sow[nb, "Modin8Methods"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Programs"}]~~__], Sow[nb, "Modin8Programs"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Symbols"}]~~__], Sow[nb, "Modin8Symbols"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Tutorials"}]~~__], Sow[nb, "Modin8Tutorials"]],
        True, ""]
        ,
        (* test2 *)
        flag === "FutureFlag",
        (* expr2 *)
        Switch[nb,
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Guides"}]~~__], Sow[nb, "FutureGuides"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "HowTos"}]~~__], Sow[nb, "FutureHowTos"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "AppleScript"}]~~__], Sow[nb, "FutureAppleScript"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Characters"}]~~__], Sow[nb, "FutureCharacters"],
                        s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "C"}]~~__], Sow[nb, "FutureC"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Files"}]~~__], Sow[nb, "FutureFiles"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Formats"}]~~__], Sow[nb, "FutureFormats"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "FrontEndObject"}]~~__], Sow[nb, "FutureFrontEndObject"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "MenuItems"}]~~__], Sow[nb, "FutureMenuItems"],
            (* s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Messages"}]~~__], Sow[nb, "FutureMessages"], *)
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Methods"}]~~__], Sow[nb, "FutureMethods"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Programs"}]~~__], Sow[nb, "FuturePrograms"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "ReferencePages", "Symbols"}]~~__], Sow[nb, "FutureSymbols"],
            s_String /; StringMatchQ[s, __~~FileNameJoin[{"System", "Tutorials"}]~~__], Sow[nb, "FutureTutorials"]];
      ];
    ]& /@ manifest, _, Rule][[2]]
  ] ;


landinggrid[rr_, col_] := 
  Block[{ls = rr, mm, griddata}, mm = Ceiling[Length[rr]/col];
   ls = PadRight[rr, col*mm, ""];
   griddata = 
    If[Length[rr] > 0, 
     Transpose[Partition[ls, mm, mm, 1, ""]], {{""}}];
   griddata];
   
landinggridDIV[rr_, col_, machinename_] := 
 Module[{lst}, lst = landinggrid[rr, col]; 
  MapIndexed[
   XMLElement[
     "div", {"class" -> 
       Switch[#2, {1}, "columnLeft", {2}, "columnCenter", {3}, 
        "columnRight"]}, Flatten[#1]]&, 
   Transpose[
 Map[If[# =!= "", 
    {XMLElement[
     "a", {"href" -> 
       "http://"<>machinename<>"/language/" <>StringReplace[
 Take[FileNameSplit[#], {-2}][[1]], {
 	"Guides" -> "guide", 
 	"HowTos" -> "howto", 
 	"AppleScript" -> "ref/applescript", 
 	"Characters" -> "ref/character", 
 	"C" -> "ref/c", 
 	"Files" -> "ref/file", 
 	"Formats" -> "ref/format", 
 	"Devices" -> "ref/device",
 	"Interpreters" -> "ref/interpreter",  
 	"Classifiers" -> "ref/classifier",  
 	"Services" -> "ref/service", 
 	"Indicators" -> "ref/indicator",
 	"Example" -> "example",  
 	"FrontEndObject" -> "ref/frontendobject", 
 	"MenuItems" -> "ref/menuitem",
    "Messages" -> "ref/message", 
(*    "Methods" -> "ref/method", *)
    "Programs" -> "ref/program", 
    "Symbols" -> "ref", 
    "Tutorials" -> "tutorial"
    }]<>"/"<> FileBaseName[#]<>".html"}, {If[StringLength[FileBaseName[#]]>=33, StringTake[FileBaseName[#], 30]<>"...", FileBaseName[#]]}], XML`RawXML["&nbsp;"]},
     ""] &, lst, {2}]]]];

grouplandinggrid[type_, rr_, col_, machinename_] := 
 Module[{}, 
  If[rr =!= {}, XMLElement["div", {"class" -> "ItemizedTextParagraph"}, 
   Join[{XMLElement["div", {}, {XMLElement["b", {}, {type}]}]}, 
    landinggridDIV[rr, col, machinename]]], {}]];

MakeLandingContent[nbdir_String, exportfile_String]:=Module[{layout, filename, headerContent, flagged, pagetypes, content, 
	pageText, theYear, theMonth, theDate, landingPage, ensuredata, newin7, newin8, newin8marketing, modifiedin7, modifiedin8, 
	future, confContent, iFrameLink, iFrame},
   theYear=ToString@Extract[Take[DateList[], {1, 3}], 1];
   theMonth=ToString@Extract[Take[DateList[], {1, 3}], 2];
   theDate=ToString@Extract[Take[DateList[], {1, 3}], 3];
   landingPage = theMonth<>"/"<>theDate<>"/"<>theYear;

   layout = DocsAsset["HTMLTransformLayout", "LandingPage.html"];

   content=GetFutureLandingContent[nbdir];
   flagged = {"Newin7", "Modin7", "Newin8", "Modin8", "Future"};
   pagetypes = {"Guides", "HowTos", "AppleScript", "C", "Characters", 
   "Files", "Formats", "FrontEndObject", "MenuItems", (*"Messages", *)
   "Methods", "Programs", "Symbols", "Tutorials", "Examples"};
   (*this ensures that we get an empty list if the Sow above does find any items*)
   ensuredata= Flatten[Table[flagged[[i]] <> pagetypes[[j]] -> {}, {i, 1, Length[flagged]}, {j, 1, Length[pagetypes]}]];
   iFrameLink ="http://www.devel2.wolfram.com/language/new-in-8/embeddedlisting.html";
   iFrame=XMLElement["IFRAME",{"src"->iFrameLink, "width"->"500px;", "height"->"200px;"}, {"If you can see this your browser does not understand IFRAME. 
   You can link directly to the page here: ",
   XMLElement["a",{"url"->"iFrameLink"}, {iFrameLink}]
   }];
   
    newin7=XMLElement["div",{"id"->"div1","style"->"display:none"}, Flatten[{
   	Map[grouplandinggrid[StringReplace[#<>" Pages", "s "->" "],("Newin7"<>#)/.content/.ensuredata, 3, "reference.wolfram.com"]&, pagetypes]
    }]];
  modifiedin7=XMLElement["div",{"id"->"div2","style"->"display:none"}, Flatten[{
  	Map[grouplandinggrid[StringReplace[#<>" Pages", "s "->" "],("Modin7"<>#)/.content/.ensuredata, 3, "reference.wolfram.com"]&, pagetypes]
  	}]];
   newin8=XMLElement["div",{"id"->"div3","style"->"display:none"}, Flatten[{
   	Map[grouplandinggrid[StringReplace[#<>" Pages", "s "->" "],("Newin8"<>#)/.content/.ensuredata, 3, "reference.devel.wolfram.com"]&, pagetypes]
    }]];
  newin8marketing=XMLElement["div",{"id"->"div6","style"->"display:none"}, 
  	{iFrame}
  ];
  modifiedin8=XMLElement["div",{"id"->"div4","style"->"display:none"}, Flatten[{
  	Map[grouplandinggrid[StringReplace[#<>" Pages", "s "->" "],("Modin8"<>#)/.content/.ensuredata, 3, "reference.devel.wolfram.com"]&, pagetypes]
  	}]];
  future=XMLElement["div",{"id"->"div5","style"->"display:none"}, Flatten[{
  	Map[grouplandinggrid[StringReplace[#<>" Pages", "s "->" "],("Future"<>#)/.content/.ensuredata, 3, "reference.future.wolfram.com"]&, pagetypes]
  	}]];
    
   pageText = Transmogrify`FillInLayout[layout, "Text", 
   	"landingPage"->landingPage,
   	"NEWIN7"->ExportString[newin7,"XML"], 
   	"MODIFIEDIN7"->ExportString[modifiedin7,"XML"],
   	"NEWIN8MARKETING"->ExportString[newin8marketing,"XML"], 
   	"NEWIN8"->ExportString[newin8,"XML"], 
   	"MODIFIEDIN8"->ExportString[modifiedin8,"XML"],
   	"FUTURE"->ExportString[future,"XML"]];
   	
   	confContent="SetEnv STYLE_CustomHeaderFile 0
SetEnv STYLE_InternalPageWidth 100%
SetEnv STYLE_Body '<body style=\"margin:0px;\" onLoad=\"javascript:set_domain()\">'";
   	   headerContent="
   <title>Mathematica Documentation Development</title>
   
   <link rel=\"stylesheet\" type=\"text/css\" href=\"/common/includes/CSS.cgi?b=menustyles;p=css2008/reference.wolfram.com/;v=en\">
   <link rel=\"shortcut icon\" href=\"/common/images2003/favicon.ico\" type=\"image/x-icon\">

   <script src=\"/javascript/search.js\" type=\"text/javascript\"></script>
<script type=\"text/javascript\" src=\"/common/javascript/jquery/plugins/droppy/0.1.2rm/jquery.droppy.js\"></script>
<!--Initializes the header menu-->
<script type='text/javascript'>
  $(function() {
    $('#nav').droppy();
  });
  
function togglediv(thediv) {$(thediv).slideToggle();}
</script>

   <style class=\"text/css\">
ul.indentednote a.graylink:active,ul.indentednote a.graylink:link,ul.indentednote a.graylink:visited {
	font-weight:normal;
	color:gray}
div.GuideSection {
	margin-left:110px;
	font-size:14px;
	font-weight:normal;
	border-top-color: #aa0000;
	width:820px;}
a:link,a:visited {color:#aa0000;}
ul.indentednote a,ul.indentednote a:link,ul.indentednote a:visited {font-weight:bold;color:#aa0000;}
h3 {
	font-weight:normal;
	font-size:26px;
	font-family:Georgia,Arial,sanserif}
.ItemizedTextParagraph a:link,.ItemizedTextParagraph a:visited {color:#666;}
.ItemizedTextParagraph a:hover,.ItemizedTextParagraph a:active {color:#cc0000;}
.ItemizedTextParagraph div {padding-bottom:4px;}
div.columnLeft,div.columnCenter,div.columnRight {
	line-height:.85em;
	width:220;
	float: left;
	padding-rigth: 10px;	
	position: relative; 		
	text-align: justify;
	margin: 0;}
div.columnRight {float:right;}
.ItemizedTextParagraph a {display:block}
div.ItemizedTextParagraph {clear:both}
</style>";

   Export[exportfile<>"-conf", confContent, "Text"];  
   Export[exportfile<>"-head", headerContent, "Text"];   	
   Export[exportfile, pageText,"Text"]
   ];



(* Make Future landing page from Bill's report *)
ExportLandingPage[path_String] := 
  Module[{layout, filename, sevenContent, futureContent, parsedFutureContent,
    landingPageVariables, pageText, theYear, theMonth, theDate, landingPage, headerContent},
   theYear=ToString@Extract[Take[DateList[], {1, 3}], 1];
   theMonth=ToString@Extract[Take[DateList[], {1, 3}], 2];
   theDate=ToString@Extract[Take[DateList[], {1, 3}], 3];
   landingPage = theMonth<>"/"<>theDate<>"/"<>theYear;
   layout = DocsAsset["HTMLTransformLayout", "LandingPage.html"];
   filename = ToFileName[{path}, "mathematica.en.html"];
   sevenContent = Import["https://internal.wolfram.com/wpt/reports/mathematica/gregm/NewModified.txt"];
   futureContent = Import["https://internal.wolfram.com/wpt/reports/mathematica/gregm/Future.txt"];
   parsedFutureContent = StringReplace[futureContent, 
   {
   	"<div class=\"GuideSection\">Quick links to future <i>Mathematica</i> pages:</div>"->"", 
   (*"<div class=\"ItemizedTextParagraph\"><b>"~~Shortest[a__]~~"</b></div>"->a<>"&nbsp;",*)
   "<div class=\"ItemizedTextParagraph\"><b>"~~Shortest[a__]~~"</b></div>"->"<div class=\"NewItemizedTextParagraph\"><b>"<>a<>"</b></Newdiv>&nbsp;",
   "<div class=\"ItemizedTextParagraph\">"->"",
   "</div>"->""
   }];
   
 
   
   (*Print@StringSplit[parsedFutureContent, "<div class=\"NewItemizedTextParagraph\">"->"<div class=\"ItemizedTextParagraph\">"];*)
   (*Print@StringSplit[parsedFutureContent,"<div class=\"NewItemizedTextParagraph\"><b>"~~Shortest[a__]~~"</b></div>"];
   landingPageVariables = {"Body1" -> sevenContent,
     						"Body2" -> futureContent, 
     						"landingPage" ->landingPage};*)
   pageText = Transmogrify`FillInLayout[layout, "Text", landingPageVariables ];
   headerContent="
   <title>Wolfram Research, Inc.</title>
   
   <link rel=\"stylesheet\" type=\"text/css\" href=\"/common/includes/CSS.cgi?b=menustyles;p=css2008/reference.wolfram.com/;v=en\">
   <link rel=\"shortcut icon\" href=\"/common/images2003/favicon.ico\" type=\"image/x-icon\">

   <script src=\"/javascript/search.js\" type=\"text/javascript\"></script>
<script type=\"text/javascript\" src=\"/common/javascript/jquery/plugins/droppy/0.1.2rm/jquery.droppy.js\"></script>
<!--Initializes the header menu-->
<script type='text/javascript'>
  $(function() {
    $('#nav').droppy();
  });
  
function togglediv(thediv) {$(thediv).slideToggle();}
</script>

   <style class=\"text/css\">
ul.indentednote a.graylink:active,ul.indentednote a.graylink:link,ul.indentednote a.graylink:visited {
	font-weight:normal;
	color:gray}
div.GuideSection {
	margin-left:110px;
	font-size:14px;
	font-weight:normal;
	border-top-color: #aa0000;
	width:820px;}
a:link,a:visited {color:#aa0000;}
ul.indentednote a,ul.indentednote a:link,ul.indentednote a:visited {font-weight:bold;color:#aa0000;}
h3 {
	font-weight:normal;
	font-size:26px;
	font-family:Georgia,Arial,sanserif}
.ItemizedTextParagraph a:link,.ItemizedTextParagraph a:visited {color:#666;}
.ItemizedTextParagraph a:hover,.ItemizedTextParagraph a:active {color:#cc0000;}
.ItemizedTextParagraph div {padding-bottom:4px;}
div.columnLeft,div.columnCenter,div.columnRight {
	line-height:.85em;
	width:220;
	float: left;
	padding-rigth: 10px;	
	position: relative; 		
	text-align: justify;
	margin: 0;}
div.columnRight {float:right;}
.ItemizedTextParagraph a {display:block}
div.ItemizedTextParagraph {clear:both}
</style>";

   Export[filename<>"-head", headerContent, "Text"];
   Export[filename, pageText, "Text"]
   
   ];

(****************************************************************
            X M L T r a n s f o r m   F U N C T I O N S

    The following helper functions are used only in the 
    Transmogrify XMLTransforms.
 ***************************************************************)
(**
  htmlBoxesQ
  check to see if *Box's need to be images 
**)

htmlBoxesQ[expr_]:=
	Module[{pos}, 
	Length[Select[
		Position[expr,
			(_FractionBox | _RadicalBox | _SqrtBox | _OverscriptBox | _UnderscriptBox | _SubsuperscriptBox | _GridBox | _SliderBox | _Slider2DBox | _CheckboxBox | _RadioButtonBox | _OpenerBox | _ActionMenuBox | _PopupMenuBox | _ProgressIndicatorBox | _InputFieldBox | _ColorSetterBox ),
			{1, Infinity}],
		Function[{pos},
			FreeQ[
			Map[Head[Extract[expr, #]] &,
				Map[Take[pos, #] &, Range[2, Length[pos]]]
			],
			Cell
		]]
	]]=== 0]



(**
inputQueue[] - get the appropriate list of Inputs up to and including an
inputline containing a "%".

Expects "s" to be the result of CreateInputForm[]. Also, use RenumberCells[] to
make sure each example section starts at In[1] for best results.

For a bit of background, see a wadocs bug in which % and %% appeared in compound
expressions: https://jira.wolfram.com/jira/browse/WADOCUMENTATION-695

--- begin quote ---
The copyable-input mechanism takes special care to save previous inputs, so when
it finds % or %% it can look up the inputs and include them in the code to be
copied. The list of inputs is reset at each sectioning head, and the first input
in a section does not resolve % or %%. As mentioned in the code,

    reset the queue if we've started a new section. we're assuming that the
    first input cell does not contain a "%", as that would just be silly...

The assumptions behind the copyable-input mechanism:

 - an Input cell is counted as one input even if the cell contains a compound
   expression

 - % and %% never appear in compound expressions; rather, such expressions are
   broken up into individual inputs as in
   http://reference.wolfram.com/language/tutorial/PermutationLists.html#451173290

So % and %% in compound expressions in the docs will never be handled correctly.

These assumptions are baked into reference.w.c, the authoring practices behind
it, and the tools that build the pages.

What to try

Could the compound inputs be broken up into separate Input cells?

Could the code be rewritten to pass information around in variables?
--- end quote ---

*)
inputQueue[s_String,label_] := Module[{out,refs},
  Catch[

    (* reset the queue if we've started a new section. we're assuming that the
    first input cell does not contain a "%", as that would just be silly... *)
    If[
      label == None
      || StringContainsQ[label /. None -> "", "In[1]"]
      || Not@ValueQ[$InputCount]
    ,
      ClearAll[$InputQueue];
      $InputQueue[$InputCount=1]=s;
      Throw[s]
    ];
                                                                                               
    $InputQueue[++$InputCount] = s;

    out = If[
      StringFreeQ[s,"%"]
      || StringContainsQ[$InputQueue[1],"%"]
    ,
      s (* Return the string if there's no meaningful "%". *)
    ,
      (* Otherwise, return the specified input requested by "%".  Join all
      the % references in the queue, separated by ";\n" *)
      (
        StringJoin @@ (
          (
            (* if # is Range[5]; for example, don't return Range[5];;\n *)
            # <> If[StringContainsQ[#, ";" ~~ EndOfString], "\n", ";\n"]&
          ) /@ Most[#]
        ) <> Last[#]
      )&[
        (* example findInputRefs return value: {"Range[5];", "%"} *)
        findInputRefs[
          $InputQueue[$InputCount],
          {$InputCount}
        ]
      ]
    ];
    out
  ]
]

inputQueue[f___] := "$Failed";


(*  fun recursion to get all %-referenced inputs.
    i'm sure there's a non-Recursive-MapIndexed way to do this
    but this seemed easiest. :) *)
findInputRefs[s_String, p_] := Module[
  {spl, pos = p[[1]], refs, result}
,
  spl = percentSplit[s];

  (* FIXME: this doesn't handle %1 digitized refs, although
     percentSplit does.  I don't think we use those in our
     documentation though.
  *)
  (* FIXME: this is actually broken when refs contains % refs,
     as it gets them all out of order and doesn't uniqueify
     them in any meaningful way.  Try this:
       $InputQueue[++$InputCount] = "One[1]";
       $InputQueue[++$InputCount] = "Two[%]";
       $InputQueue[++$InputCount] = "Three[%]";
       $InputQueue[++$InputCount] = "Four[%, %%]";
       findInputRefs[$InputQueue[$InputCount], {$InputCount}]
     This bug has existed for a while, but I don't think we
     actually trigger it anywhere.
  *)
  refs = Flatten[
    If[# > 0 && pos =!= #, $InputQueue[pos - #], $InputQueue[pos]] & /@ 
      Reverse[Sort[StringLength /@
        (* v11 ref/Sin, Generalizations & Extensions, %//MatrixForm:
        percentSplit can return a list that contains an empty string, so we need to ensure that we don't do StringTake["", 1] *)
        UnsortedUnion[Cases[spl, pc_ /; (Positive[StringLength[pc]] && StringTake[pc, 1] === "%")]]
    ]] ];
  result = Flatten@Append[MapIndexed[findInputRefs, refs], s];
  result
];

percentSplit[s_String] := percentSplit[Characters[s]];
percentSplit[chars_List] :=
Module[{res = "", c, i = 1, esc = False, str = False},
  While[i <= Length[chars],
    c = chars[[i]];
    Switch[c,
      "\\", (esc = !esc; res = res<>c),
      "\"", (If[esc, esc = False, str = !str]; res = res<>c),
      "%", If[esc || str, res = res<>c, Break[]],
      _, (esc = False; res = res<>c)];
    i++;
  ];
  Flatten[{res,
    If[i > Length[chars], {},
      Module[{dig = "%", c},
        i++;
        While[i < Length[chars],
          c = chars[[i]];
          If[DigitQ[c] || c === "%", dig = dig<>c, Break[]];
          i++];
        {dig,
          If[i > Length[chars], {},
            percentSplit[Drop[chars, i - 1]] ]}
    ]]
  }]
];

inlineModified[Cell[TextData[lst_List], o___]] :=
 Module[{new = lst},
  new = new /. Cell[BoxData["modified"], "InlineFormula"] :> StyleBox["modified", "GuideText", FontWeight->"Plain"];
  new = new /. {" (modified) \[EmptyVerySmallSquare] "} :> {" ", StyleBox["(modified)", "GuideText", FontWeight->"Plain"], " \[EmptyVerySmallSquare] "};
  Cell[TextData[Flatten@new], o]
  ];

inlineSeparator[Cell[TextData[lst_List], o___]] :=
 Module[{new},
  new = lst /. {s_String :> inlineSeparatorString@s};
  Cell[TextData[Flatten@new], o]
  ]
inlineSeparator[c___]:= c;  

inlineSeparatorString[str_String] :=
 Sequence @@ 
  StringReplace[
   str, {RegularExpression[
      "(.*)\\s+\[EmptyVerySmallSquare]\\s+(.*)"] :> {"$1", " ", 
      StyleBox["\[FilledVerySmallSquare]", "InlineSeparator"], " ", 
      "$2"}}]

inlineSeparatorString[c___] := c;


(** ReplaceSpecialCharacters[] 
    this takes care of turning unicode longname \[Character]'s into an
    HTML image representation like <img src="/some/location/Character.gif">
    
    note that simply calling StringSubstitute (=StringReplace[]) is too slow
    for such a large list of rules (currently there are about 652).
    Therefore, thanks to the wonderful magic of oyvind, we employ a hybrid
    use of string replacement + mathematica replacements.
**)

(*  The following code is a one time setup of some pattern variables
    that are actually used in the ReplaceSpecialCharacters[] implementation
    below. *)


(* all single characters we're trying to match are NOT standard printable ASCII. *)

$OneCharPattern = RegularExpression[
  StringPattern`PatternConvert[
    Except[
      CharacterRange[FromCharacterCode[32],FromCharacterCode[126]]
    ]
  ][[1]]
];

$webCharacterReplacements = None;

(* apply rule replacements on string *)
ReplaceSpecialCharacters[ text:(_String|{__String}) ] := Module[{rawCharList, oneCharRules},

	(* Don't Get CharacterReplacements.m if the symbol we want already has a value. *)
	(* load temporary compatibility package until CharacterReplacements is completely fixed *)
	If[$webCharacterReplacements === None,
		$webCharacterReplacements = Get[DocsAsset["Data:WebCharacterReplacements"]];

		DocsAssert[MatchQ[
			$webCharacterReplacements,
			{Repeated[_?StringQ -> (_?StringQ | _XML`RawXML | _XMLElement)]}
		]];
	];

	rawCharList = $webCharacterReplacements;

	ModifyRawCharacterList[list_List]:= Module[{pos1,pos2,newlist},
		pos1=Position[list, "\[OpenCurlyQuote]" -> XML`RawXML["&#39;"], 2];
		pos2=Position[list, "\[CloseCurlyQuote]" -> XML`RawXML["&#39;"], 2];
		newlist=ReplacePart[list, {pos1->"\[OpenCurlyQuote]" -> "'", pos2->"\[CloseCurlyQuote]" -> "'"}];
		newlist
	];

	ModifyRawCharacterListForSolutions[list_List]:= Module[{pos1,pos2,newlist},
		pos1=Position[list, "\[Dash]" -> "-", 2];
		newlist=ReplacePart[list, {pos1->"\[Dash]" -> "\[Dash]"}];
		newlist
	];
		
	rawCharList=Which[GetParameter["IsTMJQ"]===True,
		ModifyRawCharacterList[rawCharList], 
			GetParameter["IsSolutionsQ"]===True,
		ModifyRawCharacterListForSolutions[rawCharList], 
			True, rawCharList];

  oneCharRules = Dispatch @ Prepend[
  	
   rawCharList,
    "\n" -> Which[
        GetParameter["newline2br"], XMLElement["br", {}, {}],
        GetParameter["newline2div"], XML`RawXML[ GetParameter["styleUsageMessageBreak"] ],
        True, "\n"
    ]
  ];

  Sequence @@ StringReplace[ text, { 

    "\t" -> XML`RawXML["&nbsp;&nbsp;&nbsp;&nbsp;"],
    x:$OneCharPattern :> ( x /. oneCharRules )
(*
    x:RegularExpression["[!<>=]=|[-:]>"] :> (XMLElement["img", {"src" -> ("URL" /. #),
        "width" -> ("Width" /. #), "height" -> ("Height" /. #), "alt" -> " "}, {}]&[
            BoxToImage[ToFileName[{"HTMLImages", GetBasenameLa[]},
                GetBasenameLa[] <> "_str_" <> IncrementCounter["SubstringImages"] <> ".gif"],
                StringContent -> Switch[x,
                    "!=", "\[NotEqual]",
                    "<=", "\[LessEqual]",
                    ">=", "\[GreaterEqual]",
                    "==", "\[Equal]",
                    ":>", "\[RuleDelayed]",
                    "->", "\[Rule]",
                    _, x
                ] ]]),
    c:RegularExpression["[^\:0020-\:007E]"] :> Switch[c,
        "\[NonBreakingSpace]",      "\[NonBreakingSpace]",
        "\[AutoOperand]",           "",
        "\[Rule]",                  "->",
        "\[Equal]",                 "=",
        "\[Ellipsis]",              "...", 
        "\[OpenCurlyDoubleQuote]",  "\"", 
        "\[CloseCurlyDoubleQuote]", "\"", 
        "\[OpenCurlyQuote]",        XML`RawXML["&#39;"],
        "\[CloseCurlyQuote]",       XML`RawXML["&#39;"],
        "'",                        XML`RawXML["&#39;"],
        "\[LineSeparator]",         XMLElement["br", {}, {}],
        "\:f35f",                   "",
        "\[IndentingNewLine]",      XMLElement["br", {}, {}],
        "\[ParagraphSeparator]",    XMLElement["p", {}, {}], 
        "\[Hyphen]",                "-", 
        "\[Conjugate]",             "*",
        "\[Ellipsis]",              "...", 
        "\[Colon]",                 ":", 
        "\[VerticalBar]",           "|", 
        "\[Precedes]",              "<", 
        "\[Succeeds]",              ">", 
        "\[Backslash]",             "\\", 
        "\[Dash]",                  "\[Dash]", 
        "\[LongDash]",              "\[LongDash]", 
        "\[InvisibleApplication]",  "", 
        "\[InvisibleSpace]",        "", 
        "\[VeryThinSpace]",         "", 
        "\[ThinSpace]",             " ", 
        "\[MediumSpace]",           " ", 
        "\[ThickSpace]",            " ", 
        "\[NegativeVeryThinSpace]", "", 
        "\[NegativeThinSpace]",     "", 
        "\[NegativeMediumSpace]",   "", 
        "\[NegativeThickSpace]",    "",  
        "\[Null]",                  "", 
        "\[NoBreak]",               "",
        "\t", "\[NonBreakingSpace]\[NonBreakingSpace]\[NonBreakingSpace]\[NonBreakingSpace]", 
        "\n", Which[
        	GetParameter["newline2br"], XMLElement["br", {}, {}],
        	GetParameter["newline2div"], XML`RawXML[ GetParameter["styleUsageMessageBreak"] ],
    	    True, "\n"
        ],
        _, XMLElement["img", {"src" -> ("URL" /. #),
            "width" -> ("Width" /. #), "height" -> ("Height" /. #), "alt" -> " "}, {}]&[
                BoxToImage[ToFileName[{"HTMLImages", GetBasenameLa[]},
                    GetBasenameLa[] <> "_str_" <> IncrementCounter["SubstringImages"] <> ".gif"],
                    StringContent -> c ]]
    ] 
    *)
}]
(*
    oneCharRules
  }] /.
        (LongName[c_] :> (XMLElement["img", {"src" -> ("URL" /. #),
        "width" -> ("Width" /. #), "height" -> ("Height" /. #), "alt" -> " "}, {}]&[
            BoxToImage[ToFileName[{"HTMLImages", GetBasenameLa[]},
                GetBasenameLa[] <> "_str_" <> IncrementCounter["SubstringImages"] <> ".gif"],
                StringContent -> c ]]))
  )
*)
];

RelativeFileNames[pat_String:"*", dir_, lev_:1] :=
    Module[{dirWithTrailingSlash},
        dirWithTrailingSlash = If[StringMatchQ[dir, "*" <> $PathnameSeparator], dir, dir <> $PathnameSeparator];
        StringReplace[#, dirWithTrailingSlash -> ""]& /@ FileNames[pat, dir, lev]
    ];


(* Builds the PacletDB.m file that describes the docs that ship in the layout. This file goes into the layout as a
   pre-cached descriptor that can be read quickly by the PacletManager. The topDir directory should have a
   Documentation/lang/System hierarchy beneatch it, and also a Packages dir beneath it.
   The top-level subdirs within topDir will be the standard set Symbols, Guides, tutorials, etc. 

   The resulting PacletDB.m file is placed into outputDir.
*)
Options[BuildDocPacletIndex] = Options[ExportOnlineNotebooks]

BuildDocPacletIndex[topDir_String, outputDir_String, language_String, OptionsPattern[]]:= 
    Module[{logfile, append, files},
    
        Print["** Entering BuildDocPacletIndex **"];
                
        {logfile, append} = OptionValue[{LogFile, LogAppend}];
        
        (* quick logfile error checking *)
        If[!MatchQ[logfile, _String | None],
            Message[DocumentationExport::logfile, logfile];
            logfile = None;
        ];
        
        (* Create outdir *) 
        If[Transmogrify`Utilities`Private`createDirectory[ToFileName[{outputDir}]] === $Failed,
            Print[
                "Couldn't create output directory " <> ToFileName[{outputDir}],
                CloseLog->True
            ];
            Return[$Failed]
        ];
        
        Print[
            "BuildDocPacletIndex[] Started", 
            "Source: " <> topDir,
            "PacletDB.m file destination: " <> outputDir,
            "M- Version: " <> ToString[Internal`$CreationID] <> " " <> ToString[$Version],
            "PacletManager CreationID: " <> ToString[PacletManager`Information`$CreationID],
            "Installation Directory: " <> ToString[$InstallationDirectory],
            LogAppend->append
        ];
        
        files = FileNames["*.nb", topDir, Infinity];
        Print[ ToString[Length[files]] <> " Notebooks to Process."];
    
        (* Version 2 of the PM debuts a new format for the PacletInfo descriptor file for docs in the layout.
           We create a PacletDB.m file instead of the PacletInfo.db file used in earlier versions.
        *)
        If[StringQ[PacletManager`Package`BuildDocDBFile[outputDir, topDir, language]],
            Print[ "Successfully built PacletDB.m file", CloseLog->True],
        (* else *)
            Print[ "PacletDB.m file FAILED", CloseLog->True];
            $Failed
        ]
    ]


    

GetTutorialsFromGuide[nbexpr_] := 
  Module[{}, Cases[nbexpr, Cell[___, "GuideTutorial", ___], Infinity]];
  
GetTrainingLink[nbexpr_] := 
  Module[{urls, videourl}, urls = Cases[nbexpr, TemplateBox[{Cell[Localized["LearningResourcesTrainingCoursesText", GetParameter["language"]]], a_, ___}, ___]:>a, Infinity];
  videourl = If[Length[urls] > 0, urls[[1]], ""]
  ];
  
GetVideoLink[nbexpr_] := 
  Module[{urls, videourl}, urls = Cases[nbexpr, TemplateBox[{Cell[Localized["LearningResourcesVideosText", GetParameter["language"]]], a_, ___}, ___]:>a, Infinity];
  videourl = If[Length[urls] > 0, urls[[1]], ""]
  ];
  
GetCommunityLink[nbexpr_] := 
  Module[{urls, videourl}, urls = Cases[nbexpr, TemplateBox[{Cell[Localized["LearningResourcesCommunityText", GetParameter["language"]]], a_, ___}, ___]:>a, Infinity];
  videourl = If[Length[urls] > 0, urls[[1]], ""]
  ];  

GetDemonstrationLinkFromGuide[nbexpr_] := 
  Module[{urls, videourl}, urls = Cases[nbexpr, TemplateBox[{Cell[Localized["LearningResourcesDemonstrationsText", GetParameter["language"]]], a_, ___}, ___]:>a, Infinity];
  videourl = If[Length[urls] > 0, urls[[1]], ""]
  ];

GetHowTosFromGuide[nbexpr_] := 
  Module[{}, 
   Cases[nbexpr, Cell[___, "GuideHowTo", ___], Infinity]
   ];



CreateLearningResources[nbexpr_, uri_] := 
  Module[{tutorials, howtos, trainingcourse, video, community, 
    demonstration, includetutorials, includehowtos, 
    includedemonstration, includetrainingcourse, includevideo, 
    includecommunity, leftcellgroup, rightcellgroup}, 
   tutorials = GetTutorialsFromGuide[nbexpr] /. Cell[a___, "GuideTutorial", b___]:>Cell[a, "LearningResourceTutorial", b];
   howtos = GetHowTosFromGuide[nbexpr] /. Cell[a___, "GuideHowTo", b___]:>Cell[a, "LearningResourceHowTo", b];
   trainingcourse = ConvertToString[GetTrainingLink[nbexpr]];
   video = ConvertToString[GetVideoLink[nbexpr]];
   demonstration = ConvertToString[GetDemonstrationLinkFromGuide[nbexpr]];
   community = ConvertToString[GetCommunityLink[nbexpr]];
   includetutorials = If[Length[tutorials] > 0, True, False];
   includehowtos = If[Length[howtos] > 0, True, False];
   includetrainingcourse = 
    If[StringLength[trainingcourse] > 0, True, False];
   includevideo = If[StringLength[video] > 0, True, False];
   includedemonstration = 
    If[StringLength[demonstration] > 0, True, False];
   includecommunity = If[StringLength[community] > 0, True, False];
   leftcellgroup = 
    If[includetutorials || includehowtos, 
     Cell[CellGroupData[
       Join[{Cell["Left", "GuideLearningResourcesLeftSection", 
          CellGroupingRules -> {"SectionGrouping", 31}]}, 
        Flatten[{If[includetutorials, 
           Cell[CellGroupData[
             Join[{Cell[Localized["TutorialsSectionText", GetParameter["language"]], 
                "GuideLearningResourcesTutorialSection"]}, 
              tutorials]]], {}],
          If[includehowtos, 
           Cell[CellGroupData[
             Join[{Cell[Localized["GuideHowTosSectionText", GetParameter["language"]], 
                "GuideLearningResourcesHowToSection"]}, 
              howtos]]], {}]}]]]], {}];
   rightcellgroup = 
    If[includedemonstration || includetrainingcourse || includevideo || includecommunity,
     Cell[CellGroupData[
       Join[{Cell["Right", "GuideLearningResourcesRightSection", 
          CellGroupingRules -> {"SectionGrouping", 31}]}, Flatten[{
          Cell[Localized["LearningResourcesRelatedWebText", GetParameter["language"]], "SmallTitle"],
          
          If[includetrainingcourse, 
           Cell[TextData[
             ButtonBox[Localized["LearningResourcesTrainingCoursesText", GetParameter["language"]], BaseStyle -> "ResourceTraining", 
              ButtonData -> URL[trainingcourse]]], 
            "ResourceTraining"], {}],
          
          If[includevideo, 
           Cell[TextData[
             ButtonBox[Localized["LearningResourcesVideosText", GetParameter["language"]], BaseStyle -> "ResourceVideos", 
              ButtonData -> URL[video]]], "ResourceVideos"], {}],
          
          If[includedemonstration, 
           Cell[TextData[
             ButtonBox[Localized["LearningResourcesDemonstrationsText", GetParameter["language"]], BaseStyle -> "ResourceDemonstrations", 
              ButtonData -> URL[demonstration]]], 
            "ResourceDemonstrations"], {}],
          
          If[includecommunity, 
           Cell[TextData[
             ButtonBox[Localized["LearningResourcesCommunityText", GetParameter["language"]], BaseStyle -> "ResourceCommunity", 
              ButtonData -> URL[community]]], 
            "ResourceCommunity"], {}]}]]]], {}];
   If[includetutorials || includehowtos || includedemonstration || 
     includetrainingcourse || includevideo || includecommunity,
    Flatten[{leftcellgroup,
         rightcellgroup}], {}]
   ];
   
 CreateFeaturedPage[featuredexampledata_List] := 
 Module[{featuredexampledata2}, 
  If[Length[featuredexampledata] > 0,
  	featuredexampledata2 = Partition[Flatten[featuredexampledata], 3];
   Map[Cell[
         CellGroupData[{Cell["/language/example/"<>FileBaseName[#[[3]]] <> ".html", "FeaturedExampleLink", 
            CellGroupingRules -> {"SectionGrouping", 31}], 
           Cell[BoxData[ToBoxes[Uncompress[#[[1]]]]], 
            "FeaturedExampleThumbnailImage"], 
           Cell[#[[2]], "FeaturedExampleText"]}]] &, 
       featuredexampledata2], {}]];
       
AddFirstAndLastToexpandWraps[expr2_] := Module[{pos},
  pos = Position[expr2, "expandWrap"];
  Which[Length[pos] === 1, 
   ReplacePart[expr2, First[pos] -> "expandWrap first last"],
   Length[pos] > 1, 
   ReplacePart[
    expr2, {First[pos] -> "expandWrap first", 
     Last[pos] -> "expandWrap last"}],
   True, expr2]
  ];

deleteNonASCII[str_String] := Module[{}, 
	FromCharacterCode[DeleteCases[DeleteCases[ToCharacterCode[str], chrctr_ /; chrctr > 126], 
		33 | 35 | 36 | 37 | 38 | 40 | 41 | 42 | 44 | 47 | 58 | 60 | 61 | 62 | 63 | 64 | 91 | 93 | 94 | 96 | 123 | 124 | 125 | 126]]];
  
MakeMarketingThumbnailList[expr_, language_] := Module[{tds},
  (*expr is a list of pairs {thumbail,caption}*)
  tds = {
  	XMLElement[
       "div", {},
       {XMLElement[
       "a", {"href" -> 
         Cases[{#[[1]]}, 
           XMLElement["a", {___, Rule["href", b___], ___}, {___}] :> 
            b, Infinity][[1]]}, 
       Flatten[{Flatten[{Flatten[{Cases[{#[[1]]}, 
              w : XMLElement["img", {___}, {___}] :> w, 
              Infinity]}]}]}]]}], 
      XMLElement[
       "div", {}, {XMLElement[
       "a", {"href" -> 
         Cases[{#[[1]]}, 
           XMLElement["a", {___, Rule["href", b___], ___}, {___}] :> 
            b, Infinity][[1]]}, 
       Flatten[{Flatten[
          Cases[{#[[2]]}, 
           XMLElement[
             "a", {___}, {x___}] :> {XMLElement[
              "span", {"class" -> "chevron-after"}, {x}]}, 
           Infinity]]}]]}]} & /@ expr;
  XMLElement["section", {"class" -> "examples"}, 
   Flatten[{XMLElement[
      "h3", {}, {Localized["MarketingRelatedExamples", language]}], 
     XMLElement["div", {"class" -> "table"}, 
      Map[XMLElement["div", {"class" -> "heirs-width-25per"}, #] &, 
       Flatten[Map[Transpose, Partition[tds, 4, 4, 1, {}]], 1]]]}]]];

   
End[]
EndPackage[]
